{/*****************************************************************/
 /  Copyright (c) 1998 IMCA SYSTEMS                                /
 /  All Rights Reserved.
 /  Version
 /  MODULE: Specifical Grid functionality optimized for
 /          RapidTree/FlyTreeView/DBFlyTreeView Rapid Technology modules
 /  Some code derived from Borland (Inprise)
 /  NOTES: Some code Derived from Borland
 /  HISTORY:
 // Moved ColCount, RowCount to Public section
 // Specific InplaceEdit
 // Specific Drawing Cells and Background
 // Specific Drag
 // Specific Multiselection
 // Specific draw full rect
}

unit exgrid;

{$R exgrid.dcr}

interface
{$I DFS.inc}

uses Messages, Windows, SysUtils, Classes, Graphics, Menus, Controls, Forms,
  StdCtrls, ClipBrd {$ifdef DFS_COMPILER_4_UP}, Commctrl{$endif};

const
  MaxCustomExtents = MaxListSize;
  MaxShortInt = High(ShortInt);
  DefColWidth = 75;
type

  EInvalidGridOperation = class(Exception);
  { Internal grid types }
  TGetExtentsFunc = function(Index: Longint): Integer of object;

  TExGridAxisDI = record
    OptimalLineWidth: Integer;
    FixedBounds: Integer;
    GridBounds: Integer;
    GridExtent: Integer;
    LastFullVisibleCell: Longint;
    FullVisBounds: Integer;
    FixedCellCount: Integer;
    FirstGridCell: Integer;
    GridCellCount: Integer;
    GetExtent: TGetExtentsFunc;
  end;

  TExGridDI = record
    Horz, Vert: TExGridAxisDI;
  end;

  TGridState = (gsNormal, gsSelecting, gsRowSizing, gsColSizing,
    gsRowMoving, gsColMoving);
  TGridMovement = gsRowMoving..gsColMoving;

  TCustomExGrid = class;

  TInplaceEdit = class(TCustomEdit)
  private
    FGrid: TCustomExGrid;
    FClickTime: Longint;
    procedure InternalMove(const Loc: TRect; Redraw: Boolean);
    procedure SetGrid(Value: TCustomExGrid);
    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMPaste(var Message); message WM_PASTE;
    procedure WMCut(var Message); message WM_CUT;
    procedure WMClear(var Message); message WM_CLEAR;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure DblClick; override;
    {$ifdef DFS_COMPILER_4_UP}
    function DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
      MousePos: TPoint): Boolean; override;
    {$endif}
    function EditCanModify: Boolean; 
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure BoundsChanged; virtual;
    procedure UpdateContents; virtual;
    procedure WndProc(var Message: TMessage); override;
    property  Grid: TCustomExGrid read FGrid;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Deselect;
    procedure Hide;
    procedure Invalidate; {$ifdef DFS_COMPILER_4_UP}reintroduce;{$endif}
    procedure Move(const Loc: TRect);
    function PosEqual(const Rect: TRect): Boolean;
    procedure SetFocus; {$ifdef DFS_COMPILER_4_UP}reintroduce;{$endif}
    procedure UpdateLoc(const Loc: TRect);
    function Visible: Boolean;
  end;

  TSelectionWay = (swNone, swByMouse, swByKeyBoard);

  TGridOption = (goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine,
    goRangeSelect, goDrawFocusSelected, goRowSizing, goColSizing, goRowMoving,
    goColMoving, goEditing, goTabs, goRowSelect,
    goAlwaysShowEditor, goThumbTracking);

  TGridOptions = set of TGridOption;
  // gdNoUseSelColor - special flag for drawing on another canvas
  TExGridDrawState = set of (gdSelected, gdFocused, gdFixed, gdPrepaint, gdPrepaintIndentZone, gdClicked, gdNoUseSelColor);
  TGridScrollDirection = set of (sdLeft, sdRight, sdUp, sdDown);

  TGridCoord = record
    X: Longint;
    Y: Longint;
  end;

  TGridRect = record
    case Integer of
      0: (Left, Top, Right, Bottom: Longint);
      1: (TopLeft, BottomRight: TGridCoord);
  end;


  TSelectCellEvent = procedure (Sender: TObject; ACol, ARow: Longint;
    var CanSelect: Boolean) of object;
  TDrawCellEvent = procedure (Sender: TObject; aCanvas : TCanvas; ACol, ARow: Longint;
    Rect: TRect; State: TExGridDrawState) of object;
  TSizedEvent = procedure (Index, Value : integer) of Object;
  TSizingEvent = procedure (Index : integer; var Value : integer) of Object;

  TGetEditEvent = procedure (Sender: TObject; ACol, ARow: Longint; var Value: string) of object;
  TSetEditEvent = procedure (Sender: TObject; ACol, ARow: Longint; const Value: string) of object;
  TMovedEvent = procedure (Sender: TObject; FromIndex, ToIndex: Longint) of object;

  {$ifdef DFS_COMPILER_4_UP}
  TScrollBarStyle = (sbsRegular, sbsEncarta, sbsFlat);
  {$endif}

  TCustomExGrid = class(TCustomControl)
  private
    //FInSize : boolean;
    FDragRowColPt : TPoint;
    FSelectionCleared : Boolean;
    {$ifdef DFS_COMPILER_4_UP}
    fFlatSBPresent : boolean;
    //FControlRowExtents. FControlColExtents : boolean;
    FScrollBarsColor : TColor;
    FScrollBarsStyle : TScrollBarStyle;
    {$endif}
    FScrolling, FMouseWheel : boolean;
//    FClickTime: Longint;
    FFixedRowAsButton, FFixedColAsButton : boolean;
    RestRect : TRect;
    RestoreArrow, DragLArrow, DragRArrow, DragImg, DragImgMask : TBitmap;
    DragRect : TRect;
    FDragCoord : TPoint;
    FAnchor, FEndAnchor: TGridCoord;
    FBorderStyle: TBorderStyle;
    FCanEditModify: Boolean;
    FLineColor, FFixedLineColor : TColor;
    FColCount: Longint;
    FColWidths: Pointer;
    FTabStops: Pointer;
    FCurrent: TGridCoord;
    FDefaultColWidth: Integer;
    FDefaultRowHeight: Integer;
    FFixedCols: Integer;
    FFixedRows: Integer;
    FFixedColor: TColor;
    FGridLineWidth: Integer;
    FOptions: TGridOptions;
    FRowCount: Longint;
    FRowHeights: Pointer;
    FScrollBars: TScrollStyle;
    FTopLeft: TGridCoord;
    FSizingIndex: Longint;
    FSizingPos, FSizingOfs: Integer;
    FMoveIndex, FMovePos: Longint;
    FHitTest: TPoint;
    FInplaceEdit: TInplaceEdit;
    FInplaceCol, FInplaceRow: Longint;
    FColOffset: Integer;
    FEditorMode: Boolean;

    FOnColumnWidthChanged : TSizedEvent;
    FOnColumnWidthChanging : TSizingEvent;
    FOnRowHeightChanged : TSizedEvent;
    FOnRowHeightChanging : TSizingEvent;
    FOnAppendRow : TNotifyEvent;
    FOnColumnMoved: TMovedEvent;
    FOnGetEditMask: TGetEditEvent;
    FOnGetEditText: TGetEditEvent;
    FOnRowMoved: TMovedEvent;
    FOnSelectCell: TSelectCellEvent;
    FOnSetEditText: TSetEditEvent;
    FOnDrawCell: TDrawCellEvent;
    FOnTopLeftChanged: TNotifyEvent;
    function IsGridEmpty : boolean;
    //new Move Col method
    procedure PrepareDragImage(aCol,aRow : integer; DrI, DrIM : boolean);
    procedure ClampInView(const Coord: TGridCoord);
    function CalcCoordFromPoint(X, Y: Integer;
      const DrawInfo: TExGridDI): TGridCoord;
    procedure CancelMode;

    procedure ChangeSize(NewColCount, NewRowCount: Longint);

    procedure DrawSizingLine(const DrawInfo: TExGridDI);
    procedure GridRectToScreenRect(GridRect: TGridRect;
      var ScreenRect: TRect; IncludeLine,IncludeHExtent, IncludeVExtent : Boolean);

    procedure Initialize;

    procedure ModifyScrollBar(ScrollBar, ScrollCode, Pos: Cardinal;
      UseRightToLeft: Boolean);

    procedure MoveAdjust(var CellPos: Longint; FromIndex, ToIndex: Longint);
    procedure MoveAnchor(const NewAnchor: TGridCoord; MoveFocus : Boolean);
    Procedure DrawDragImg(aLeft, aTop, aWidth, aHeight:integer; aImageCHandle : {integer} HDC; UseBF : Boolean);  // CEDR 20.09.02 - range
    Procedure DrawInserMark;
    Procedure RestoreInserMark;
    procedure MoveAndScroll(Mouse, CellHit: Longint; var DrawInfo: TExGridDI;
      var Axis: TExGridAxisDI; Scrollbar: Integer; const MousePt: TPoint);
    procedure MoveCurrent(ACol, ARow: Longint; MoveAnchor, Show : Boolean; Shift: TShiftState; SelWay : TSelectionWay);

    procedure ResizeCol(Index: Longint; OldSize, NewSize: Integer);
    procedure ResizeRow(Index: Longint; OldSize, NewSize: Integer);
    procedure SelectionMoved(const OldSel: TGridRect);
    procedure ScrollDataInfo(DX, DY: Integer; var DrawInfo: TExGridDI);
    procedure TopLeftMoved(const OldTopLeft: TGridCoord);
    procedure UpdateScrollPos;
    function GetTabStops(Index: Longint): Boolean;
    function GetVisibleColCount: Integer;
    function GetVisibleRowCount: Integer;
    function IsActiveControl: Boolean;
    procedure ReadColWidths(Reader: TReader);
    procedure ReadRowHeights(Reader: TReader);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetCol(Value: Longint);
    procedure SetColCount(Value: Longint);
    procedure SetDefaultColWidth(Value: Integer);
    procedure SetDefaultRowHeight(Value: Integer);
    procedure SetEditorMode(Value: Boolean);
    procedure SetFixedColor(Value: TColor);
    procedure SetFixedCols(Value: Integer);
    procedure SetFixedRows(Value: Integer);
    procedure SetGridLineWidth(Value: Integer);
    procedure SetLeftCol(Value: Longint);
    procedure SetOptions(Value: TGridOptions);
    procedure SetRow(Value: Longint);
    procedure SetRowCount(Value: Longint);
    procedure SetScrollBars(Value: TScrollStyle);

    procedure SetTabStops(Index: Longint; Value: Boolean);
    procedure SetTopRow(Value: Longint);

    procedure UpdateText;
    procedure WriteColWidths(Writer: TWriter);
    procedure WriteRowHeights(Writer: TWriter);
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure CMCancelMode(var Msg: TMessage); message CM_CANCELMODE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    procedure CMDesignHitTest(var Msg: TCMDesignHitTest); message CM_DESIGNHITTEST;
    procedure CMWantSpecialKey(var Msg: TCMWantSpecialKey); message CM_WANTSPECIALKEY;
    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;
    procedure WMChar(var Msg: TWMChar); message WM_CHAR;
    procedure WMCancelMode(var Msg: TWMCancelMode); message WM_CANCELMODE;
    procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
    procedure WMGetDlgCode(var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMHScroll(var Msg: TWMHScroll); message WM_HSCROLL;
    procedure WMKillFocus(var Msg: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMLButtonDown(var Message: TMessage); message WM_LBUTTONDOWN;
    procedure WMNCHitTest(var Msg: TWMNCHitTest); message WM_NCHITTEST;
    procedure WMSetCursor(var Msg: TWMSetCursor); message WM_SETCURSOR;
    procedure WMSetFocus(var Msg: TWMSetFocus); message WM_SETFOCUS;
    procedure WMSize(var Msg: TWMSize); message WM_SIZE;
    procedure WMTimer(var Msg: TWMTimer); message WM_TIMER;
    procedure WMVScroll(var Msg: TWMVScroll); message WM_VSCROLL;
    procedure SetLineColor(Value : TColor);
    procedure SetFixedLineColor(Value : TColor);
    procedure SetFixedRowAsButton(Value : boolean);
    procedure SetFixedColAsButton(Value : boolean);
    procedure SetSelection(Value: TGridRect);
    function GetSelection: TGridRect;
    procedure CalcDIXY(var DrawInfo: TExGridDI;
      UseWidth, UseHeight: Integer);
    procedure ChangeGridOrientation(RightToLeftOrientation: Boolean);
    {$ifdef DFS_COMPILER_4_UP}
    procedure SetScrollBarsStyle(value : TScrollBarStyle);
    procedure SetScrollbarsColor(value : TColor);
    {$endif}
  protected
    FGridState: TGridState;
    FNotProcessMouse : Boolean;
    FSaveCellExtents: Boolean;
    procedure InternalSetRow(Value: Longint); virtual;
    procedure ClearSelection;
    procedure Click; override;
    procedure DestroyWnd; override;
    procedure CreateWnd; override;
    procedure DblClick; override;
    function GetOptimalKeys : TShiftstate;
    procedure UpdateRowHeight(Index: Longint; Value: Integer);
    procedure MoveEditor(aCol, aRow : integer; aRect : TRect; Redraw : boolean); virtual;

    function CalcMaxTopLeft(const Coord: TGridCoord;
      const DrawInfo: TExGridDI): TGridCoord;
    procedure MoveTopLeft(ALeft, ATop: Longint);

    procedure InvalidateGrid;
    procedure InvalidateNonFixed;
    procedure SetGridState(aState : TGridState);
    function GetGridState : TGridState;
    procedure StopDrag;
    procedure FocusCell(ACol, ARow: Longint; MoveAnchor : Boolean; Shift: TShiftState; SelWay : TSelectionWay);virtual;
    function AllowDesignSizing : boolean; virtual;
    function GetColWidths(Index: Longint): Integer; virtual;//new 11/9/1
    function GetRowHeights(Index: Longint): Integer;virtual;//new 11/9/1
    procedure SetRowHeights(Index: Longint; Value: Integer);virtual;//new 3/4/2
    procedure SetColWidths(Index: Longint; Value: Integer);virtual;//new 3/4/2
    procedure AppendRow; dynamic;
    procedure InvalidateRect(ARect: TGridRect);
    procedure Updatewidths; virtual;
    procedure CalcDI(var DrawInfo: TExGridDI);
    procedure CalcFI(var DrawInfo: TExGridDI);
    procedure CalcSizingState(X, Y: Integer; var State: TGridState;
      var Index: Longint; var SizingPos, SizingOfs: Integer;
      var FixedInfo: TExGridDI); virtual;
    function CreateEditor: TInplaceEdit; virtual;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    function MouseHit(X,Y : integer) : boolean; virtual;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure AdjustSize(Index, Amount: Longint; Rows: Boolean); {$ifdef DFS_COMPILER_4_UP}reintroduce;{$endif} dynamic;

    procedure DoExit; override;

    function CanEditAcceptKey(Key: Char): Boolean; dynamic;
    function CanGridAcceptKey(Key: Word; Shift: TShiftState): Boolean; dynamic;
    function CanEditModify: Boolean; dynamic;
    function CanEditShow: Boolean; virtual;
    procedure UpdateEdit; virtual;
    {$ifdef DFS_COMPILER_4_UP}
    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean; override;
    {$endif}
    procedure HideEdit; virtual;
    function GetEditText(ACol, ARow: Longint): string; dynamic;
    procedure SetEditText(ACol, ARow: Longint; const Value: string); dynamic;
    function GetEditMask(ACol, ARow: Longint): string; dynamic;
    function GetEditLimit: Integer; dynamic;
    function GetGridWidth: Integer;
    function GetGridHeight: Integer;
    procedure HideEditor; virtual;
    procedure ShowEditor; virtual;
    procedure ShowEditorChar(Ch: Char); virtual;
    procedure InvalidateEditor;
    function InSelection(aCol, aRow : integer) : Boolean; virtual;
    procedure ColumnWidthChanging(Index : integer; var Value: integer); dynamic;
    procedure ColumnWidthChanged(Index, Value : integer); dynamic;
    procedure ColCountChanged; virtual;
    procedure RowHeightChanging(Index : integer; var Value: integer); dynamic;
    procedure RowHeightChanged(Index, Value : integer); dynamic;

    procedure ColumnMoved(FromIndex, ToIndex: Longint); dynamic;
    procedure MoveRow(FromIndex, ToIndex: Longint);
    procedure RowMoved(FromIndex, ToIndex: Longint); dynamic;
    procedure DrawFixedFrame(aCanvas : TCanvas; aRect : TRect; aCol, aRow : integer; Down : Boolean); virtual; abstract;
    procedure DrawCell(aCanvas : TCanvas; ACol, ARow: Longint; ARect: TRect;
      AState: TExGridDrawState; Var Continue : boolean); virtual;

    procedure FillNotCellArea(aRect : Trect); virtual;
    function GetNonFixedArea : TRect;//used for drawing backgrounds
    function NeedToShowEditor : boolean; virtual;
    procedure DefineProperties(Filer: TFiler); override;
    procedure Loaded; override;
    procedure MoveColRow(ACol, ARow: Longint; MoveAnchor, MoveFocus, Show: Boolean);
    function SelectCell(var ACol, ARow: Longint; MoveAnchor : Boolean; Shift: TShiftState; SelWay : TSelectionWay): Boolean; virtual;
    procedure SizeChanged(OldColCount, OldRowCount: Longint); dynamic;
    function Sizing(X, Y: Integer): Boolean;
    procedure ScrollData(DX, DY: Integer);

    procedure TopLeftChanged; dynamic;
    procedure TimedScroll(Direction: TGridScrollDirection); dynamic;
    procedure DrawCells(aCanvas : TCanvas; ACol, ARow: Longint; StartX, StartY, StopX, StopY: Integer;
           Color: TColor; IncludeDrawState: TExGridDrawState; aDI : TExGridDI; UseLines : Boolean); virtual;

    procedure Paint; override;
    procedure DrawLines(aCanvas : TCanvas; aCol, aRow : integer; aRect : TRect; VertLine, HorzLine : Boolean; aLinecolor : TColor); virtual;
    procedure GetDrawColor(aRow, aCol : Integer; var aColor, aFontColor : TColor; var aChanged : boolean);virtual;
    procedure ColWidthsChanged; dynamic;
    procedure RowHeightsChanged; dynamic;
    procedure DeleteColumn(ACol: Longint); virtual;
    procedure DeleteRow(ARow: Longint); virtual;
    procedure UpdateDesigner;
    function BeginColumnDrag(var Origin, Destination: Longint;
      const MousePt: TPoint): Boolean; dynamic;
    function BeginRowDrag(var Origin, Destination: Longint;
      const MousePt: TPoint): Boolean; dynamic;
    function CheckColumnDrag(var Origin, Destination: Longint;
      const MousePt: TPoint): Boolean; dynamic;
    function CheckRowDrag(var Origin, Destination: Longint;
      const MousePt: TPoint): Boolean; dynamic;
    function EndColumnDrag(var Origin, Destination: Longint;
      const MousePt: TPoint): Boolean; dynamic;
    function EndRowDrag(var Origin, Destination: Longint;
      const MousePt: TPoint): Boolean; dynamic;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property Color default clWindow;

    property Scrolling : boolean read FScrolling;
    property GridHeight: Integer read GetGridHeight;
    property GridWidth: Integer read GetGridWidth;
    property HitTest: TPoint read FHitTest;
    property InplaceEditor: TInplaceEdit read FInplaceEdit;
    property Options: TGridOptions read FOptions write SetOptions
      default [goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine,
      goRangeSelect];
    property ParentColor default False;
    property TabStops[Index: Longint]: Boolean read GetTabStops write SetTabStops;
    function CanColSizing(X,Y : integer) : Boolean; virtual;
  public
    FSizingAboveFixed : boolean;//for sizing column above colheaders
    Customizing : Boolean; // for sizing at runtime in prop editors
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure UpdateScrollRange;
    procedure DrawRTLText(aCanvas : TCanvas; Data : String; var anRect : TRect; Flags, VFlags : integer);
    //for redirect output
    procedure DrawOn(aCanvas : TCanvas; asControl : Boolean); virtual;
    function BoxRect(ALeft, ATop, ARight, ABottom: Longint): TRect;
    function CellRect(ACol, ARow: Longint): TRect;
    procedure MouseToCell(X, Y: Integer; var ACol, ARow: Longint);
    function MouseCoord(X, Y: Integer): TGridCoord;
    procedure MoveColumn(FromIndex, ToIndex: Longint);
    procedure InvalidateCell(ACol, ARow: Longint); virtual;
    procedure InvalidateCol(ACol: Longint);
    procedure InvalidateRow(ARow: Longint);
    procedure InvalidateFullRow(ARow : Longint);
    function GetDraggedRow : Longint;
    function GetDropTargetRow : Longint;
    procedure SetDropTargetRow(aRow : Longint);
    property ColOffset : integer read FColOffset;
    property VisibleColCount: Integer read GetVisibleColCount;
    property VisibleRowCount: Integer read GetVisibleRowCount;
    property OnColumnWidthChanged : TSizedEvent read FOnColumnWidthChanged write FOnColumnWidthChanged;
    property OnColumnWidthChanging : TSizingEvent read FOnColumnWidthChanging write FOnColumnWidthChanging;
    property OnRowHeightChanged : TSizedEvent read FOnRowHeightChanged write FOnRowHeightChanged;
    property OnRowHeightChanging : TSizingEvent read FOnRowHeightChanging write FOnRowHeightChanging;
        property TopRow: Longint read FTopLeft.Y write SetTopRow;
    property LineColor : TColor read FlineColor write SetLineColor default clSilver;
    property FixedLineColor : TColor read FFixedLineColor write SetFixedLineColor default clBlack;
    property Col: Longint read FCurrent.X write SetCol;
    property ColCount: Longint read FColCount write SetColCount default 1;
    property ColWidths[Index: Longint]: Integer read GetColWidths write SetColWidths;
    property DefaultColWidth: Integer read FDefaultColWidth write SetDefaultColWidth default DefColWidth;
    property DefaultRowHeight: Integer read FDefaultRowHeight write SetDefaultRowHeight default 24;
    property EditorMode: Boolean read FEditorMode write SetEditorMode;
    property FixedColor: TColor read FFixedColor write SetFixedColor default clBtnFace;
    property FixedCols: Integer read FFixedCols write SetFixedCols default 0;
    property FixedRows: Integer read FFixedRows write SetFixedRows default 1;
    property GridLineWidth: Integer read FGridLineWidth write SetGridLineWidth default 1;
    property LeftCol: Longint read FTopLeft.X write SetLeftCol;
    property Row: Longint read FCurrent.Y write SetRow;
    property RowCount: Longint read FRowCount write SetRowCount default 2;
    property RowHeights[Index: Longint]: Integer read GetRowHeights write SetRowHeights;
    property ScrollBars: TScrollStyle read FScrollBars write SetScrollBars default ssBoth;
    {$ifdef DFS_COMPILER_4_UP}
    property ScrollBarsStyle : TScrollBarStyle read FScrollBarsStyle write SetScrollBarsStyle default sbsRegular;
    property ScrollBarsColor : TColor read FScrollBarsColor write SetScrollbarsColor default clWhite;
    {$endif}
    property Selection: TGridRect read GetSelection write SetSelection;
    //specifies to draw fixed cells as buttons
    property FixedRowAsButton : boolean read FFixedRowAsButton write SetFixedRowAsButton default true;
    property FixedColAsButton : boolean read FFixedColAsButton write SetFixedColAsButton default true;
    property OnGetEditMask: TGetEditEvent read FOnGetEditMask write FOnGetEditMask;
    property OnGetEditText: TGetEditEvent read FOnGetEditText write FOnGetEditText;
    property OnRowMoved: TMovedEvent read FOnRowMoved write FOnRowMoved;
    property OnColumnMoved: TMovedEvent read FOnColumnMoved write FOnColumnMoved;
    property OnSelectCell: TSelectCellEvent read FOnSelectCell write FOnSelectCell;
    property OnSetEditText: TSetEditEvent read FOnSetEditText write FOnSetEditText;
    property OnDrawCell: TDrawCellEvent read FOnDrawCell write FOnDrawCell;
    property OnTopLeftChanged: TNotifyEvent read FOnTopLeftChanged write FOnTopLeftChanged;
    property OnAppendRow : TNotifyEvent read FOnAppendRow write FOnAppendRow;
  published
    property TabStop default True;
  end;

  var VertCursor : HCursor;
      HorzCursor : HCursor;

  function GridRect(Left, Top, Right, Bottom : Longint): TGridRect;
  function PointInGridRect(Col, Row: Longint; const Rect: TGridRect): Boolean;




implementation

uses Math, Consts, dialogs, ISConst;


{$IFNDEF DFS_CPPB_5}
//  function AlphaBlend; external 'msimg32.dll' name 'AlphaBlend';
{$endif}

type
  PIntArray = ^TIntArray;
  TIntArray = array[0..MaxCustomExtents] of Integer;

const msimgdll = 'msimg32.dll';

var
  //Win98OrLater, Win2KOrLater : Boolean;
  BF : TBlendFunction;
  AlphaBlend : function (DC: HDC; p2, p3, p4, p5: Integer; DC6: HDC; p7, p8, p9, p10: Integer; p11: TBlendFunction): Bool stdcall = nil;


procedure InitAlphaBlend;
var msimglib : THandle;
   OldError: Longint;
begin
  OldError := SetErrorMode(SEM_NOOPENFILEERRORBOX);
  try
   msimglib := LoadLibrary(msimgdll);
   if (msimglib <> 0)  then
      @Alphablend := GetProcAddress(msimglib, 'AlphaBlend');
  finally
    SetErrorMode(OldError);
  end;
end;

{$ifndef DFS_COMPILER_4_UP}
function Max(A,B: Integer): Integer;
begin
  if A > B then
    Result := A
  else
    Result := B;
end;

function Min(A,B: Integer): Integer;
begin
  if A < B then
    Result := A
  else
    Result := B;
end;

{$endif}
{$ifdef DFS_COMPILER_3_UP}
procedure InvalidOp(const id: string);
begin
  raise EInvalidGridOperation.Create(id);
end;
{$ELSE}
procedure InvalidOp(id: Integer);
begin
  raise EInvalidGridOperation.CreateRes(id);
end;
{$ENDIF}


function GridRect(Left, Top, Right, Bottom : Longint): TGridRect;
begin
   Result.Top:=Top;
   Result.Left:=Left;
   Result.Right:=Right;
   Result.Bottom:=Bottom;
end;

function _GridRect(Coord1, Coord2: TGridCoord): TGridRect;
begin
  with Result do
  begin
    Left := Coord2.X;
    if Coord1.X < Coord2.X then Left := Coord1.X;
    Right := Coord1.X;
    if Coord1.X < Coord2.X then Right := Coord2.X;
    Top := Coord2.Y;
    if Coord1.Y < Coord2.Y then Top := Coord1.Y;
    Bottom := Coord1.Y;
    if Coord1.Y < Coord2.Y then Bottom := Coord2.Y;
  end;
end;

function PointInGridRect(Col, Row: Longint; const Rect: TGridRect): Boolean;
begin
  Result := (Col >= Rect.Left) and (Col <= Rect.Right) and (Row >= Rect.Top)
    and (Row <= Rect.Bottom);
end;

type
  TXorRects = array[0..3] of TRect;

procedure XorRects(const R1, R2: TRect; var XorRects: TXorRects);
var
  Intersect, Union: TRect;

  function PtInRect(X, Y: Integer; const Rect: TRect): Boolean;
  begin
    with Rect do Result := (X >= Left) and (X <= Right) and (Y >= Top) and
      (Y <= Bottom);
  end;

  function Includes(const P1: TPoint; var P2: TPoint): Boolean;
  begin
    with P1 do
    begin
      Result := PtInRect(X, Y, R1) or PtInRect(X, Y, R2);
      if Result then P2 := P1;
    end;
  end;

  function Build(var R: TRect; const P1, P2, P3: TPoint): Boolean;
  begin
    Build := True;
    with R do
      if Includes(P1, TopLeft) then
      begin
        if not Includes(P3, BottomRight) then BottomRight := P2;
      end
      else if Includes(P2, TopLeft) then BottomRight := P3
      else Build := False;
  end;

begin
  FillChar(XorRects, SizeOf(XorRects), 0);
  if not Bool(IntersectRect(Intersect, R1, R2)) then
  begin
    XorRects[0] := R1;
    XorRects[1] := R2;
  end
  else
  begin
    UnionRect(Union, R1, R2);
    if Build(XorRects[0],
      Point(Union.Left, Union.Top),
      Point(Union.Left, Intersect.Top),
      Point(Union.Left, Intersect.Bottom)) then
      XorRects[0].Right := Intersect.Left;
    if Build(XorRects[1],
      Point(Intersect.Left, Union.Top),
      Point(Intersect.Right, Union.Top),
      Point(Union.Right, Union.Top)) then
      XorRects[1].Bottom := Intersect.Top;
    if Build(XorRects[2],
      Point(Union.Right, Intersect.Top),
      Point(Union.Right, Intersect.Bottom),
      Point(Union.Right, Union.Bottom)) then
      XorRects[2].Left := Intersect.Right;
    if Build(XorRects[3],
      Point(Union.Left, Union.Bottom),
      Point(Intersect.Left, Union.Bottom),
      Point(Intersect.Right, Union.Bottom)) then
      XorRects[3].Top := Intersect.Bottom;
  end;
end;

procedure ModifyExtents(var Extents: Pointer; Index, Amount: Longint;
  Default: Integer);
var
  LongSize, OldSize: LongInt;
  NewSize: Integer;
  I: Integer;
begin
  if Amount <> 0 then
  begin
    if not Assigned(Extents) then OldSize := 0
    else OldSize := PIntArray(Extents)^[0];
    if (Index < 0) or (OldSize < Index) then InvalidOp(SIndexOutOfRange);
    LongSize := OldSize + Amount;
    if LongSize < 0 then InvalidOp(STooManyDeleted)
    else if LongSize >= MaxListSize - 1 then InvalidOp(SGridTooLarge);
    NewSize := Cardinal(LongSize);
    if NewSize > 0 then Inc(NewSize);
    ReallocMem(Extents, NewSize * SizeOf(Integer));
    if Assigned(Extents) then
    begin
      I := Index + 1;
      while I < NewSize do
      begin
        PIntArray(Extents)^[I] := Default;
        Inc(I);
      end;
      PIntArray(Extents)^[0] := NewSize-1;
    end;
  end;
end;

procedure UpdateExtents(var Extents: Pointer; NewSize: Longint;
  Default: Integer);
var
  OldSize: Integer;
begin
  OldSize := 0;
  if Assigned(Extents) then OldSize := PIntArray(Extents)^[0];
  ModifyExtents(Extents, OldSize, NewSize - OldSize, Default);
end;

procedure MoveExtent(var Extents: Pointer; FromIndex, ToIndex: Longint);
var
  Extent: Integer;
begin
  if Assigned(Extents) then
  begin
    Extent := PIntArray(Extents)^[FromIndex];
    if FromIndex < ToIndex then
      Move(PIntArray(Extents)^[FromIndex + 1], PIntArray(Extents)^[FromIndex],
        (ToIndex - FromIndex) * SizeOf(Integer))
    else if FromIndex > ToIndex then
      Move(PIntArray(Extents)^[ToIndex], PIntArray(Extents)^[ToIndex + 1],
        (FromIndex - ToIndex) * SizeOf(Integer));
    PIntArray(Extents)^[ToIndex] := Extent;
  end;
end;

function CompareExtents(E1, E2: Pointer): Boolean;
var
  I: Integer;
begin
  Result := False;
  if E1 <> nil then
  begin
    if E2 <> nil then
    begin
      for I := 0 to PIntArray(E1)^[0] do
        if PIntArray(E1)^[I] <> PIntArray(E2)^[I] then Exit;
      Result := True;
    end
  end
  else Result := E2 = nil;//
end;

function LongMulDiv(Mult1, Mult2, Div1: Longint): Longint; stdcall;
  external 'kernel32.dll' name 'MulDiv';

type
  TSelection = record
    StartPos, EndPos: Integer;
  end;

//For replace to TFlyInplaceEdit (finplace.pas)
constructor TInplaceEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ParentCtl3D := False;
  Ctl3D := False;
  TabStop := False;
  //BorderStyle := bsNone;
  ControlStyle := ControlStyle - [csFramed];
  {$ifdef DFS_COMPILER_4_UP}
  DoubleBuffered := False;
  {$endif}
end;

procedure TInplaceEdit.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or ES_MULTILINE;
end;

procedure TInplaceEdit.SetGrid(Value: TCustomExGrid);
begin
  FGrid := Value;
end;

procedure TInplaceEdit.CMShowingChanged(var Message: TMessage);
begin
end;

procedure TInplaceEdit.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  if goTabs in Grid.Options then
    Message.Result := Message.Result or DLGC_WANTTAB;
end;

procedure TInplaceEdit.WMPaste(var Message);
begin
  if not EditCanModify then Exit;
  inherited
end;

procedure TInplaceEdit.WMClear(var Message);
begin
  if not EditCanModify then Exit;
  inherited;
end;

procedure TInplaceEdit.WMCut(var Message);
begin
  if not EditCanModify then Exit;
  inherited;
end;

procedure TInplaceEdit.DblClick;
begin
  Grid.DblClick;
end;
{$ifdef DFS_COMPILER_4_UP}
function TInplaceEdit.DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
  MousePos: TPoint): Boolean;
begin
  Result := Grid.DoMouseWheel(Shift, WheelDelta, MousePos);
end;
{$endif}

function TInplaceEdit.EditCanModify: Boolean;
begin
  Result := Grid.CanEditModify;
end;

procedure TInplaceEdit.KeyDown(var Key: Word; Shift: TShiftState);

  procedure SendToParent;
  begin
    Grid.KeyDown(Key, Shift);
    Key := 0;
  end;

  procedure ParentEvent;
  var
    GridKeyDown: TKeyEvent;
  begin
    GridKeyDown := Grid.OnKeyDown;
    if Assigned(GridKeyDown) then GridKeyDown(Grid, Key, Shift);
  end;

  function ForwardMovement: Boolean;
  begin
    Result := goAlwaysShowEditor in Grid.Options;
  end;

  function Ctrl: Boolean;
  begin
    Result := ssCtrl in Shift;
  end;

  function Selection: TSelection;
  begin
    SendMessage(Handle, EM_GETSEL, Longint(@Result.StartPos), Longint(@Result.EndPos));
  end;

  function RightSide: Boolean;
  begin
    with Selection do
      Result := ((StartPos = 0) or (EndPos = StartPos)) and
        (EndPos = GetTextLen);
   end;

  function LeftSide: Boolean;
  begin
    with Selection do
      Result := (StartPos = 0) and ((EndPos = 0) or (EndPos = GetTextLen));
  end;

begin
  case Key of
    VK_UP, VK_DOWN, VK_PRIOR, VK_NEXT, VK_ESCAPE: SendToParent;
    VK_INSERT:
      if Shift = [] then SendToParent
      else if (Shift = [ssShift]) and not Grid.CanEditModify then Key := 0;
    VK_LEFT: if ForwardMovement and (Ctrl or LeftSide) then SendToParent;
    VK_RIGHT: if ForwardMovement and (Ctrl or RightSide) then SendToParent;
    VK_HOME: if ForwardMovement and (Ctrl or LeftSide) then SendToParent;
    VK_END: if ForwardMovement and (Ctrl or RightSide) then SendToParent;
    VK_F2:
      begin
        ParentEvent;
        if Key = VK_F2 then
        begin
          Deselect;
          Exit;
        end;
      end;
    VK_TAB: if not (ssAlt in Shift) then SendToParent;
  end;
  if (Key = VK_DELETE) and not Grid.CanEditModify then Key := 0;
  if Key <> 0 then
  begin
    ParentEvent;
    inherited KeyDown(Key, Shift);
  end;
end;

procedure TInplaceEdit.KeyPress(var Key: Char);
var
  Selection: TSelection;
begin
  Grid.KeyPress(Key);
  if (Key in [#32..#255]) and not Grid.CanEditAcceptKey(Key) then
  begin
    Key := #0;
    MessageBeep(0);
  end;
  case Key of
    #9, #27: Key := #0;
    #13:
      begin
        SendMessage(Handle, EM_GETSEL, Longint(@Selection.StartPos), Longint(@Selection.EndPos));
        if (Selection.StartPos = 0) and (Selection.EndPos = GetTextLen) then
          Deselect else
          SelectAll;
        Key := #0;
      end;
    ^H, ^V, ^X, #32..#255:
      if not Grid.CanEditModify then Key := #0;
  end;
  if Key <> #0 then inherited KeyPress(Key);
end;

procedure TInplaceEdit.KeyUp(var Key: Word; Shift: TShiftState);
begin
  Grid.KeyUp(Key, Shift);
end;

procedure TInplaceEdit.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    WM_SETFOCUS:
      begin
        if (GetParentForm(Self) = nil) or
        GetParentForm(Self).SetFocusedControl(Grid) then
        Dispatch(Message);
        Exit;
      end;
    WM_LBUTTONDOWN:
      begin
        if UINT(GetMessageTime - FClickTime) < GetDoubleClickTime then
          Message.Msg := WM_LBUTTONDBLCLK;
        FClickTime := 0;
      end;
  end;
  inherited WndProc(Message);
end;

procedure TInplaceEdit.Deselect;
begin
  SendMessage(Handle, EM_SETSEL, $7FFFFFFF, Longint($FFFFFFFF));
end;

procedure TInplaceEdit.Invalidate;
var
  Cur: TRect;
begin
  ValidateRect(Handle, nil);
  InvalidateRect(Handle, nil, True);
  Windows.GetClientRect(Handle, Cur);
  MapWindowPoints(Handle, Grid.Handle, Cur, 2);
  ValidateRect(Grid.Handle, @Cur);
  InvalidateRect(Grid.Handle, @Cur, False);
end;

procedure TInplaceEdit.Hide;
begin
  if HandleAllocated and IsWindowVisible(Handle) then
  begin
    Invalidate;
    SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_HIDEWINDOW or SWP_NOZORDER or
      SWP_NOREDRAW);
    if Focused then
      Windows.SetFocus(Grid.Handle);
  end;
end;

function TInplaceEdit.PosEqual(const Rect: TRect): Boolean;
var
  Cur: TRect;
begin
  GetWindowRect(Handle, Cur);
  MapWindowPoints(HWND_DESKTOP, Grid.Handle, Cur, 2);
  Result := EqualRect(Rect, Cur);
end;

procedure TInplaceEdit.InternalMove(const Loc: TRect; Redraw: Boolean);
begin
  if IsRectEmpty(Loc) then Hide
  else
  begin
    CreateHandle;
    Redraw := Redraw or not IsWindowVisible(Handle);
    Invalidate;
    with Loc do
      SetWindowPos(Handle, HWND_TOP, Left, Top, Right - Left, Bottom - Top,
        SWP_SHOWWINDOW or SWP_NOREDRAW);
    BoundsChanged;
    if Redraw then Invalidate;
    if Grid.Focused then
      Windows.SetFocus(Handle);
  end;
end;

procedure TInplaceEdit.BoundsChanged;
var
  R: TRect;
begin
  R := Rect(2, 2, Width - 2, Height);
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@R));
  SendMessage(Handle, EM_SCROLLCARET, 0, 0);
end;

procedure TInplaceEdit.UpdateLoc(const Loc: TRect);
begin
  InternalMove(Loc, False);
end;

function TInplaceEdit.Visible: Boolean;
begin
  Result := IsWindowVisible(Handle);
end;

procedure TInplaceEdit.Move(const Loc: TRect);
begin
  InternalMove(Loc, True);
end;

procedure TInplaceEdit.SetFocus;
begin
  if IsWindowVisible(Handle) then
    Windows.SetFocus(Handle);
end;

procedure TInplaceEdit.UpdateContents;
begin
  Text := '';
  //EditMask := Grid.GetEditMask(Grid.Col, Grid.Row);
  Text := Grid.GetEditText(Grid.Col, Grid.Row);
  MaxLength := Grid.GetEditLimit;
end;

{ TCustomExGrid }

constructor TCustomExGrid.Create(AOwner: TComponent);
const
  GridStyle = [csCaptureMouse, csOpaque, csClickEvents, csDoubleClicks];
begin
  inherited Create(AOwner);
  if NewStyleControls then
    ControlStyle := GridStyle
  else
    ControlStyle := GridStyle + [csFramed];
  {$ifdef DFS_COMPILER_4_UP}
  fFlatSBPresent := false;//Scrollbars/?
  FScrollBarsColor := clWhite;
  FScrollBarsStyle := sbsRegular;
  {$endif}
  FFixedRowAsButton := true;
  FFixedColAsButton := true;

  FCanEditModify := True;
  FColCount := 1;
  FRowCount := 2;
  FFixedCols := 0;
  FFixedRows := 1;
  FGridLineWidth := 1;
  FlineColor := ClSilver;
  FFixedLineColor := ClBlack;

  FOptions := [goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine,
    goRangeSelect];

  FFixedColor := clBtnFace;
  FScrollBars := ssBoth;
  FBorderStyle := bsSingle;
  FDefaultColWidth := DefColWidth;
  FDefaultRowHeight := 24;
  FSaveCellExtents := True;
  Customizing :=False;
  FEditorMode := False;
  FMovePos := -1;
  Color := clWindow;
  ParentColor := False;
  TabStop := True;
  SetBounds(Left, Top, FColCount * FDefaultColWidth,
    FRowCount * FDefaultRowHeight);
  Initialize;
end;

destructor TCustomExGrid.Destroy;
begin
  FInplaceEdit.Free;
  FreeMem(FColWidths);
  FreeMem(FRowHeights);
  FreeMem(FTabStops);
  inherited Destroy;
end;

procedure TCustomExGrid.AdjustSize(Index, Amount: Longint; Rows: Boolean);
var
  NewCur: TGridCoord;
  OldRows, OldCols: Longint;
  MovementX, MovementY: Longint;
  MoveRect: TGridRect;
  ScrollArea: TRect;
  AbsAmount: Longint;

  function DoSizeAdjust(var Count: Longint; var Extents: Pointer;
    DefaultExtent: Integer; var Current: Longint): Longint;
  var
    I: Integer;
    NewCount: Longint;
  begin
    NewCount := Count + Amount;
    if NewCount < Index then InvalidOp(STooManyDeleted);
    if (Amount < 0) and Assigned(Extents) then
    begin
      Result := 0;
      for I := Index to Index - Amount - 1 do
        Inc(Result, PIntArray(Extents)^[I]);
    end
    else
      Result := Amount * DefaultExtent;
    if Extents <> nil then
      ModifyExtents(Extents, Index, Amount, DefaultExtent);
    Count := NewCount;
    if Current >= Index then
      if (Amount < 0) and (Current < Index - Amount) then Current := Index
      else Inc(Current, Amount);
  end;

begin
  if Amount = 0 then Exit;
  NewCur := FCurrent;
  OldCols := ColCount;
  OldRows := RowCount;
  MoveRect.Left := FixedCols;
  MoveRect.Right := ColCount - 1;
  MoveRect.Top := FixedRows;
  MoveRect.Bottom := RowCount - 1;
  MovementX := 0;
  MovementY := 0;
  AbsAmount := Amount;
  if AbsAmount < 0 then AbsAmount := -AbsAmount;
  if Rows then
  begin
    MovementY := DoSizeAdjust(FRowCount, FRowHeights, DefaultRowHeight, NewCur.Y);
    MoveRect.Top := Index;
    if Index + AbsAmount <= TopRow then MoveRect.Bottom := TopRow - 1;
  end
  else
  begin
    MovementX := DoSizeAdjust(FColCount, FColWidths, DefaultColWidth, NewCur.X);
    MoveRect.Left := Index;
    if Index + AbsAmount <= LeftCol then MoveRect.Right := LeftCol - 1;
  end;
  GridRectToScreenRect(MoveRect, ScrollArea, True, False, False);
  if not IsRectEmpty(ScrollArea) then
  begin
    ScrollWindow(Handle, MovementX, MovementY, @ScrollArea, @ScrollArea);
    UpdateWindow(Handle);
  end;
  SizeChanged(OldCols, OldRows);
  if (NewCur.X <> FCurrent.X) or (NewCur.Y <> FCurrent.Y) then
    MoveCurrent(NewCur.X, NewCur.Y, True,  True, [], swNone);
end;

function TCustomExGrid.BoxRect(ALeft, ATop, ARight, ABottom: Longint): TRect;
var
  GridRect: TGridRect;
begin
  GridRect.Left := ALeft;
  GridRect.Right := ARight;
  GridRect.Top := ATop;
  GridRect.Bottom := ABottom;
  GridRectToScreenRect(GridRect, Result, False, False, False);
end;

procedure TCustomExGrid.DoExit;
begin
  inherited DoExit;
  if not (goAlwaysShowEditor in Options) then Editormode:=false;//HideEditor;
end;

function TCustomExGrid.CellRect(ACol, ARow: Longint): TRect;
begin
  Result := BoxRect(ACol, ARow, ACol, ARow);
end;

function TCustomExGrid.CanEditAcceptKey(Key: Char): Boolean;
begin
  Result := True;
end;

function TCustomExGrid.IsGridEmpty : boolean;
begin
  Result:= (RowCount=0) or (Rowcount=Fixedrows) or (Colcount=0) or (ColCount=Fixedcols);
end;

function TCustomExGrid.CanGridAcceptKey(Key: Word; Shift: TShiftState): Boolean;
begin
  if IsGridEmpty then
     Result:=false //Grid is empty
  else
     Result := True;
end;

function TCustomExGrid.CanEditModify: Boolean;
begin
  Result := FCanEditModify;
end;

function TCustomExGrid.CanEditShow: Boolean;
begin
  Result := ([goRowSelect, goEditing] * Options = [goEditing]) and
    (FEditorMode) and not (csDesigning in ComponentState) and HandleAllocated and
    ((goAlwaysShowEditor in Options) or IsActiveControl);
end;

function TCustomExGrid.IsActiveControl: Boolean;
var
  H: Hwnd;
  //ParentForm: TCustomForm;
begin
  Result := False;
  //modified -501
  {ParentForm := GetParentForm(Self);
  if Assigned(ParentForm) then
  begin
    if (ParentForm.ActiveControl = Self) then
      Result := Focused;//True
  end
  else
  begin}
    H := GetFocus;
    while IsWindow(H) and (Result = False) do
    begin
      if H = WindowHandle then
        Result := True
      else
        H := GetParent(H);
    end;
  {end;}
end;

function TCustomExGrid.GetEditMask(ACol, ARow: Longint): string;
begin
  Result := '';
  if Assigned(FOnGetEditMask) then FOnGetEditMask(Self, ACol, ARow, Result);
end;

function TCustomExGrid.GetEditText(ACol, ARow: Longint): string;
begin
  Result := '';
  if Assigned(FOnGetEditText) then FOnGetEditText(Self, ACol, ARow, Result);
end;

procedure TCustomExGrid.SetEditText(ACol, ARow: Longint; const Value: string);
begin
  if Assigned(FOnSetEditText) then FOnSetEditText(Self, ACol, ARow, Value);
end;

function TCustomExGrid.GetEditLimit: Integer;
begin
  Result := 0;
end;

procedure TCustomExGrid.HideEditor;
begin
  FEditorMode := False;
  HideEdit;
end;

procedure TCustomExGrid.ShowEditor;
begin
  if FeditorMode then HideEditor;
  FEditorMode := True;
  UpdateEdit;
end;

procedure TCustomExGrid.ShowEditorChar(Ch: Char);
begin
  ShowEditor;
  if FInplaceEdit <> nil then
    PostMessage(FInplaceEdit.Handle, WM_CHAR, Word(Ch), 0);
end;

procedure TCustomExGrid.InvalidateEditor;
begin
  FInplaceCol := -1;
  FInplaceRow := -1;
  UpdateEdit;
end;

procedure TCustomExGrid.ReadColWidths(Reader: TReader);
var
  I: Integer;
begin
  with Reader do
  begin
    ReadListBegin;
    for I := 0 to ColCount - 1 do ColWidths[I] := ReadInteger;
    ReadListEnd;
  end;
end;

procedure TCustomExGrid.ReadRowHeights(Reader: TReader);
var
  I: Integer;
begin
  with Reader do
  begin
    ReadListBegin;
    //for I := 0 to RowCount - 1 do RowHeights[I] := ReadInteger;
    //ReadListEnd;
    i :=0;
    while not Reader.EndOfList do
    begin
      RowHeights[I] := ReadInteger;
      inc(i);
    end;
    Reader.ReadListEnd;
  end;
end;

procedure TCustomExGrid.WriteColWidths(Writer: TWriter);
var
  I: Integer;
begin
  with Writer do
  begin
    WriteListBegin;
    for I := 0 to ColCount - 1 do WriteInteger(ColWidths[I]);
    WriteListEnd;
  end;
end;

procedure TCustomExGrid.WriteRowHeights(Writer: TWriter);
var
  I: Integer;
begin
  with Writer do
  begin
    WriteListBegin;
    for I := 0 to RowCount - 1 do WriteInteger(RowHeights[I]);
    WriteListEnd;
  end;
end;

procedure TCustomExGrid.Loaded;
begin
    inherited;
    Initialize;
end;

procedure TCustomExGrid.DefineProperties(Filer: TFiler);

  function DoColWidths: Boolean;
  begin
    if Filer.Ancestor <> nil then
      Result := not CompareExtents(TCustomExGrid(Filer.Ancestor).FColWidths, FColWidths)
    else
      Result := FColWidths <> nil;
  end;

  function DoRowHeights: Boolean;
  begin
    if Filer.Ancestor <> nil then
      Result := not CompareExtents(TCustomExGrid(Filer.Ancestor).FRowHeights, FRowHeights)
    else
      Result := FRowHeights <> nil;
  end;


begin
  inherited DefineProperties(Filer);
  if FSaveCellExtents then
    with Filer do
    begin

      DefineProperty('ColWidths', ReadColWidths, WriteColWidths, DoColWidths);
      DefineProperty('RowHeights', ReadRowHeights, WriteRowHeights, DoRowHeights);
    end;
end;

procedure TCustomExGrid.MoveColumn(FromIndex, ToIndex: Longint);
var
  Rect: TGridRect;
begin
  if FromIndex = ToIndex then Exit;
  if Assigned(FColWidths) then
  begin
    MoveExtent(FColWidths, FromIndex + 1, ToIndex + 1);
    MoveExtent(FTabStops, FromIndex + 1, ToIndex + 1);
  end;
  MoveAdjust(FCurrent.X, FromIndex, ToIndex);
  MoveAdjust(FAnchor.X, FromIndex, ToIndex);
  MoveAdjust(FEndAnchor.X, FromIndex, ToIndex);
  MoveAdjust(FInplaceCol, FromIndex, ToIndex);


  Rect.Top := 0;
  Rect.Bottom := VisibleRowCount+1;
  if FromIndex < ToIndex then
  begin
    Rect.Left := FromIndex;
    Rect.Right := ToIndex;
  end
  else
  begin
    Rect.Left := ToIndex;
    Rect.Right := FromIndex;
  end;
  InvalidateRect(Rect);

  ColumnMoved(FromIndex, ToIndex);
  if Assigned(FColWidths) then
    ColWidthsChanged;
  UpdateEdit;
end;


procedure TCustomExGrid.MoveRow(FromIndex, ToIndex: Longint);
begin
  if Assigned(FRowHeights) then
    MoveExtent(FRowHeights, FromIndex + 1, ToIndex + 1);
  MoveAdjust(FCurrent.Y, FromIndex, ToIndex);
  MoveAdjust(FAnchor.Y, FromIndex, ToIndex);
  MoveAdjust(FInplaceRow, FromIndex, ToIndex);
  RowMoved(FromIndex, ToIndex);
  if Assigned(FRowHeights) then
    RowHeightsChanged;
  UpdateEdit;
end;

procedure TCustomExGrid.RowMoved(FromIndex, ToIndex: Longint);
begin
  if Assigned(FOnRowMoved) then FOnRowMoved(Self, FromIndex, ToIndex);
end;

function TCustomExGrid.MouseCoord(X, Y: Integer): TGridCoord;
var
  DrawInfo: TExGridDI;
begin
  CalcDI(DrawInfo);
  Result := CalcCoordFromPoint(X, Y, DrawInfo);
  if Result.X < 0 then Result.Y := -1
  else if Result.Y < 0 then Result.X := -1;
end;

procedure TCustomExGrid.MoveColRow(ACol, ARow: Longint; MoveAnchor, MoveFocus, Show: Boolean);
begin
  MoveCurrent(ACol, ARow, MoveAnchor, Show, [], swNone);
end;

function TCustomExGrid.SelectCell(var ACol, ARow: Longint; MoveAnchor : Boolean; Shift: TShiftState; SelWay : TSelectionWay): Boolean;
begin
  Result := True;
  if Assigned(FOnSelectCell) then FOnSelectCell(Self, ACol, ARow, Result);
end;

procedure TCustomExGrid.SizeChanged(OldColCount, OldRowCount: Longint);
begin
  UpdateWidths;
end;

function TCustomExGrid.Sizing(X, Y: Integer): Boolean;
var
  DrawInfo: TExGridDI;
  State: TGridState;
  Index: Longint;
  Pos, Ofs: Integer;
begin
  State := FGridState;
  if State = gsNormal then
  begin
    CalcDI(DrawInfo);
    CalcSizingState(X, Y, State, Index, Pos, Ofs, DrawInfo);
  end;
  Result := State <> gsNormal;
end;

procedure TCustomExGrid.TopLeftChanged;
begin
  if FEditorMode and (FInplaceEdit <> nil) then
       MoveEditor(Col,Row, CellRect(Col, Row), False);
  if Assigned(FOnTopLeftChanged) then FOnTopLeftChanged(Self);
end;

procedure TCustomExGrid.GetDrawColor(aRow, aCol : Integer; var aColor, aFontColor : TColor; var aChanged : boolean);
begin
end;

procedure TCustomExGrid.DrawRTLText(aCanvas : TCanvas; Data : String; var anRect : TRect; Flags, VFlags : integer);
var
//   BmpSize : TSize;
   voff : integer;
   NRect : Trect;
{$ifdef DFS_COMPILER_4_UP}
  Hold : integer;
{$endif}

begin
   {$ifdef DFS_COMPILER_4_UP}
   if UseRightToLeftAlignment then
   begin
     AnRect.Left := ClientWidth - AnRect.Left;
     AnRect.Right := ClientWidth - AnRect.Right;
     Hold := AnRect.Left;
     AnRect.Left := AnRect.Right;
     AnRect.Right := Hold;
     ChangeGridOrientation(False);
   end;
   {$endif}

   Case VFlags of
     DT_TOP: DrawText(aCanvas.Handle,PChar(Data), -1, anRect, {$ifdef DFS_COMPILER_4_UP}DrawTextBiDiModeFlags(Flags){$else}Flags{$endif});
     DT_BOTTOM,
     DT_VCENTER:
     begin
       if (Flags and DT_CALCRECT) = 0 then
       begin
          Flags := Flags + DT_CALCRECT;
          NRect := anRect;
          DrawText(aCanvas.Handle,PChar(Data), -1, nRect, {$ifdef DFS_COMPILER_4_UP}DrawTextBiDiModeFlags(Flags){$else}Flags{$endif});
          Flags := Flags - DT_CALCRECT;
           //GetTextExtentPoint32(aCanvas.Handle, Pchar(Data), Length(Data), BmpSize);
           voff := ((anRect.Bottom-anRect.Top) - (nRect.Bottom-nRect.Top));

         if VFlags = DT_VCENTER then
           voff := voff div 2;
         if voff < 0 then voff := 0;
       end
       else
         voff := 0;

       if vOff <> 0 then
         NRect := Rect(anRect.Left, anRect.Top+voff, anRect.Right, anRect.Bottom)
       else
         NRect := anRect;

        DrawText(aCanvas.Handle, PChar(Data), -1, nRect, {$ifdef DFS_COMPILER_4_UP}DrawTextBiDiModeFlags(Flags){$else}Flags{$endif});
        aNrect.Right := Nrect.Right;
        aNrect.Bottom := Nrect.Bottom;
        aNrect.Left := Nrect.Left;
     end;
   end;

   {$ifdef DFS_COMPILER_4_UP}
   if UseRightToLeftAlignment then
      ChangeGridOrientation(True);
   {$endif}
end;

procedure TCustomExGrid.DrawCell(aCanvas : TCanvas; ACol, ARow: Longint; ARect: TRect;
  AState: TExGridDrawState; Var Continue : boolean);
{$ifdef DFS_COMPILER_4_UP}
  var Hold : integer;
{$endif}
begin
  if Assigned(FOnDrawCell) then
  begin
    {$ifdef DFS_COMPILER_4_UP}
    if UseRightToLeftAlignment then
    begin
      ARect.Left := ClientWidth - ARect.Left;
      ARect.Right := ClientWidth - ARect.Right;
      Hold := ARect.Left;
      ARect.Left := ARect.Right;
      ARect.Right := Hold;
      ChangeGridOrientation(False);
    end;
    {$endif}
    FOnDrawCell(Self,aCanvas, ACol, ARow, ARect, AState);
    {$ifdef DFS_COMPILER_4_UP}
    if UseRightToLeftAlignment then ChangeGridOrientation(True);
    {$endif}
  end;
end;

procedure TCustomExGrid.DrawLines(aCanvas : TCanvas; aCol, aRow : integer; aRect : TRect; VertLine, HorzLine : Boolean; aLinecolor : TColor);
 const
    FlatPenStyle = PS_Geometric or PS_Solid or PS_EndCap_Flat or PS_Join_Miter;
 var
    LogBrush: TLOGBRUSH;
begin
     with aRect do
     begin
       //aCanvas.Pen.Width := GridLinewidth;
       with aCanvas do
       if GridLinewidth > 1 then
          begin
          LogBrush.lbStyle := BS_Solid;
          if HorzLine then
           LogBrush.lbColor := aLineColor
          Else
           LogBrush.lbColor := Color;
          LogBrush.lbHatch := 0;
          aCanvas.Pen.Handle := ExtCreatePen(FlatPenStyle, GridLinewidth, LogBrush, 0, nil);
       end
       else
       begin
          if HorzLine then
           aCanvas.Pen.Color := aLineColor
          Else
           aCanvas.Pen.Color := Color;
       end;

         MoveToEx(aCanvas.Handle, Left, Bottom,nil);
         LineTo(aCanvas.Handle,Right+1,Bottom);


       with aCanvas do
       if GridLinewidth > 1 then
          begin
          LogBrush.lbStyle := BS_Solid;
          if HorzLine then
           LogBrush.lbColor := aLineColor
          Else
           LogBrush.lbColor := Color;
          LogBrush.lbHatch := 0;
          aCanvas.Pen.Handle := ExtCreatePen(FlatPenStyle, GridLinewidth, LogBrush, 0, nil);
       end
       else
       begin
          if VertLine then
           aCanvas.Pen.Color := aLineColor
          Else
           aCanvas.Pen.Color := Color;
       end;


//       if VertLine then
//         aCanvas.Pen.Color := aLineColor
//       Else
//         aCanvas.Pen.Color := Color;

       //begin
         {$ifdef DFS_COMPILER_4_UP}
         //it's bug of RighttoLeft Tech -
         if UseRightToLeftAlignment then Inc(Right);
         {$endif}
         MoveToEx(aCanvas.Handle, Right, Top, nil);
         LineTo(aCanvas.Handle, Right,Bottom+1);
       //end;
     end;
end;

procedure TCustomExGrid.Paint;
begin
  DrawOn(Canvas, true);
end;

procedure TCustomExGrid.DrawCells(aCanvas : TCanvas; ACol, ARow: Longint; StartX, StartY, StopX, StopY: Integer;
    Color: TColor; IncludeDrawState: TExGridDrawState; aDI : TExGridDI; UseLines : Boolean);
var
    CurCol, CurRow: Longint;
    Where: TRect;
    aDrawState: TExGridDrawState;
    Focused : Boolean;
    aCont : boolean;
begin
  CurRow := ARow;
  Where.Top := StartY;
  aDrawState := [];
  while (Where.Top < StopY) and (CurRow < RowCount) do
  begin
    CurCol := ACol;
    Where.Left := StartX;
    Where.Bottom := Where.Top + RowHeights[CurRow];
    aCont:=true;
    while (Where.Left < StopX) and aCont do
    begin
      if Curcol < ColCount then
         Where.Right := Where.Left + ColWidths[CurCol]
      else Where.Right := Where.Left+StopX;
      if {(Where.Right > Where.Left) and} RectVisible(aCanvas.Handle, Where) then
      begin
        aDrawState := IncludeDrawState;
        Focused := IsActiveControl;
        if Focused and (CurRow = Row) and (CurCol = Col)  then
          Include(aDrawState, gdFocused);
        if PointInGridRect(CurCol, CurRow, Selection) then
          Include(aDrawState, gdSelected);

          with aCanvas do
             Font.Assign(Self.Font);
          if  (csDesigning in ComponentState) then
          with aCanvas do
          begin
              if (gdSelected in aDrawState) and
                (not (gdFocused in aDrawState) or
                ([goDrawFocusSelected, goRowSelect] * Options <> [])) then
              begin
               if Focused then //new 15/06/2001
                Brush.Color := clHighlight
               else
                Brush.Color := clInactiveCaption;
                Font.Color := clHighlightText;
              end
              else
                Brush.Color := Color;
              FillRect(Where);
            end;

           //drawLines
            if  UseLines then
            begin
              if (gdFixed in aDrawstate) and ((FFixedColAsButton and (CurCol<Fixedcols)) or (FFixedRowAsButton and (CurRow<Fixedrows))) then
              begin
                  //aCanvas.Pen.color := FFixedlineColor;
                  DrawLines(aCanvas, CurCol, CurRow, Where, (goFixedVertline in Options),(goFixedHorzline in Options), FFixedlineColor);
              end
              else
                  //aCanvas.Pen.color := FLineColor;
                  DrawLines(aCanvas, CurCol, CurRow, Where, (goVertLine in Options),(goHorzLine in Options),FLineColor);
              end;
            end;

      DrawCell(ACanvas, CurCol, CurRow, Where, aDrawState, aCont);

      Where.Left := Where.Right + aDI.Horz.OptimalLineWidth;
      Inc(CurCol);
    end;
    Where.Top := Where.Bottom + aDI.Vert.OptimalLineWidth;
    Inc(CurRow);
  end;
end;


procedure TCustomExGrid.DrawOn(aCanvas : TCanvas; asControl : Boolean);
var
  DrawInfo: TExGridDI;
  Sel: TGridRect;
  AFocRect, FocRect: TRect;
  UseLines : boolean;

begin
  {$ifdef DFS_COMPILER_4_UP}
  if UseRightToLeftAlignment then ChangeGridOrientation(True);
  {$endif}
  CalcDI(DrawInfo);
  with DrawInfo do
  begin
    Sel := Selection;
    UseLines := (Horz.OptimalLineWidth > 0) or (Vert.OptimalLineWidth > 0);
    if AsControl then
    begin
     //corner
     DrawCells(aCanvas, 0, 0, 0, 0, Horz.FixedBounds, Vert.FixedBounds, FixedColor, [gdFixed],DrawInfo, UseLines);
     //fixed columns
     DrawCells(aCanvas, LeftCol, 0, Horz.FixedBounds - FColOffset, 0, Horz.GridExtent, Vert.FixedBounds, FixedColor, [gdFixed],DrawInfo, UseLines);
     //fixed rows
     if not IsGridEmpty then //12/7/2
       DrawCells(aCanvas, 0, TopRow, 0, Vert.FixedBounds, Horz.FixedBounds, Vert.GridBounds, FixedColor, [gdFixed], DrawInfo, UseLines);
     //other cells
     if not IsGridEmpty then //04.08.00
       DrawCells(aCanvas, LeftCol, TopRow, Horz.FixedBounds - FColOffset, Vert.FixedBounds, Horz.GridExtent, Vert.GridBounds, Color, [], DrawInfo, UseLines);
    end
    else
    begin //based on rows technology
       // something to do
       //not yet implemented
    end;

    if AsControl then
    if not (csDesigning in ComponentState) and
      (goDrawFocusSelected in Options) and (goRowSelect in Options) and Focused and (not IsGridEmpty)then
    begin
      GridRectToScreenRect(GetSelection, FocRect, False, False, False);
      {$ifdef DFS_COMPILER_4_UP}
      if not UseRightToLeftAlignment then
        aCanvas.DrawFocusRect(FocRect)
      else
      begin
        AFocRect := FocRect;
        AFocRect.Left := FocRect.Right;
        AFocRect.Right := FocRect.Left;
        DrawFocusRect(aCanvas.Handle, AFocRect);
      end;
      {$else}
        aCanvas.DrawFocusRect(FocRect);
      {$endif}
    end;

    if AsControl then
    if Vert.GridBounds < Vert.GridExtent then
    begin
      aCanvas.Brush.Color := Color;
      FillNotCellArea(Rect(0, Vert.GridBounds, Horz.GridExtent, Vert.GridExtent));
    end;
  end;
  {$ifdef DFS_COMPILER_4_UP}
  if UseRightToLeftAlignment then ChangeGridOrientation(False);
  {$endif}

  end;


function TCustomExGrid.NeedToShowEditor : boolean;
begin
   Result:=true;
end;

procedure TCustomExGrid.FillNotCellArea(aRect : TRect);
begin
   Canvas.FillRect(aRect);
end;

function TCustomExGrid.CalcCoordFromPoint(X, Y: Integer;
  const DrawInfo: TExGridDI): TGridCoord;

  function DoCalc(const AxisInfo: TExGridAxisDI; N: Integer): Integer;
  var
    I, Start, Stop: Longint;
    Line: Integer;
  begin
    with AxisInfo do
    begin
      if N < FixedBounds then
      begin
        Start := 0;
        Stop :=  FixedCellCount - 1;
        Line := 0;
      end
      else
      begin
        Start := FirstGridCell;
        Stop := GridCellCount - 1;
        Line := FixedBounds;
      end;
      Result := -1;
      for I := Start to Stop do
      begin
        Inc(Line, GetExtent(I) + OptimalLineWidth);
        if N < Line then
        begin
          Result := I;
          Exit;
        end;
      end;
    end;
  end;

  function DoCalcRightToLeft(const AxisInfo: TExGridAxisDI; N: Integer): Integer;
  var
    I, Start, Stop: Longint;
    Line: Integer;
  begin
    N := ClientWidth - N;
    with AxisInfo do
    begin
      if N < FixedBounds then
      begin
        Start := 0;
        Stop :=  FixedCellCount - 1;
        Line := ClientWidth;
      end
      else
      begin
        Start := FirstGridCell;
        Stop := GridCellCount - 1;
        Line := FixedBounds;
      end;
      Result := -1;
      for I := Start to Stop do
      begin
        Inc(Line, GetExtent(I) + OptimalLineWidth);
        if N < Line then
        begin
          Result := I;
          Exit;
        end;
      end;
    end;
  end;

begin
  {$ifdef DFS_COMPILER_4_UP}
  if not UseRightToLeftAlignment then
    Result.X := DoCalc(DrawInfo.Horz, X)
  else
    Result.X := DoCalcRightToLeft(DrawInfo.Horz, X);
  {$else}
   Result.X := DoCalc(DrawInfo.Horz, X);
  {$endif}
  Result.Y := DoCalc(DrawInfo.Vert, Y);
end;

procedure TCustomExGrid.CalcDI(var DrawInfo: TExGridDI);
begin
  CalcDIXY(DrawInfo, ClientWidth, ClientHeight);
end;

procedure TCustomExGrid.CalcDIXY(var DrawInfo: TExGridDI;
  UseWidth, UseHeight: Integer);

  procedure CalcAxis(var AxisInfo: TExGridAxisDI; UseExtent: Integer);
  var
    I: Integer;
  begin
    with AxisInfo do
    begin
      GridExtent := UseExtent;
      GridBounds := FixedBounds;
      FullVisBounds := FixedBounds;
      LastFullVisibleCell := FirstGridCell;
      for I := FirstGridCell to GridCellCount - 1 do
      begin
        Inc(GridBounds, GetExtent(I) + OptimalLineWidth);
        if GridBounds > GridExtent + OptimalLineWidth then
        begin
          GridBounds := GridExtent;
          Break;
        end;
        LastFullVisibleCell := I;
        FullVisBounds := GridBounds;
      end;
    end;
  end;

begin
  CalcFI(DrawInfo);
  CalcAxis(DrawInfo.Horz, UseWidth);
  CalcAxis(DrawInfo.Vert, UseHeight);
end;

procedure TCustomExGrid.CalcFI(var DrawInfo: TExGridDI);

  procedure CalcFixedAxis(var Axis: TExGridAxisDI; LineOptions: TGridOptions;
    FixedCount, FirstCell, CellCount: Integer; GetExtentFunc: TGetExtentsFunc);
  var
    I: Integer;
  begin
    with Axis do
    begin
      if LineOptions * Options = [] then
        OptimalLineWidth := 0
      else
        OptimalLineWidth := GridLineWidth;

      FixedBounds := 0;
      for I := 0 to FixedCount - 1 do
        Inc(FixedBounds, GetExtentFunc(I) + OptimalLineWidth);

      FixedCellCount := FixedCount;
      FirstGridCell := FirstCell;
      GridCellCount := CellCount;
      GetExtent := GetExtentFunc;
    end;
  end;

begin
  CalcFixedAxis(DrawInfo.Horz, [goFixedVertLine, goVertLine], FixedCols,
    LeftCol, ColCount, GetColWidths);
  CalcFixedAxis(DrawInfo.Vert, [goFixedHorzLine, goHorzLine], FixedRows,
    TopRow, RowCount, GetRowHeights);
end;

{ Calculates the TopLeft that will put the given Coord in view }
function TCustomExGrid.CalcMaxTopLeft(const Coord: TGridCoord;
  const DrawInfo: TExGridDI): TGridCoord;

  function CalcMaxCell(const Axis: TExGridAxisDI; Start: Integer): Integer;
  var
    Line: Integer;
    I, Extent: Longint;
  begin
    Result := Start;
    with Axis do
    begin
      Line := GridExtent + OptimalLineWidth;
      for I := Start downto FixedCellCount do
      begin
        Extent := GetExtent(I);
        if Extent > 0 then
        begin
          Dec(Line, Extent);
          Dec(Line, OptimalLineWidth);
          if Line < FixedBounds then
          begin
            if (Result = Start) and (GetExtent(Start) <= 0) then
              Result := I;
            Break;
          end;
          Result := I;
        end;
      end;
    end;
  end;

begin
  Result.X := CalcMaxCell(DrawInfo.Horz, Coord.X);
  Result.Y := CalcMaxCell(DrawInfo.Vert, Coord.Y);
end;

procedure TCustomExGrid.CalcSizingState(X, Y: Integer; var State: TGridState;
  var Index: Longint; var SizingPos, SizingOfs: Integer;
  var FixedInfo: TExGridDI);

  procedure CalcAxisState(const AxisInfo: TExGridAxisDI; Pos: Integer;
    NewState: TGridState);
  var
    I, Line, Back, Range: Integer;
  begin
    {$ifdef DFS_COMPILER_4_UP}
    if UseRightToLeftAlignment then
      Pos := ClientWidth - Pos;
    {$endif}
    with AxisInfo do
    begin
      Line := FixedBounds;
      Range := OptimalLineWidth;
      Back := 0;
      if Range < 7 then
      begin
        Range := 7;
        Back := (Range - OptimalLineWidth) shr 1;
      end;
      for I := FirstGridCell to GridCellCount - 1 do
      begin
        Inc(Line, GetExtent(I));
        if Line > GridBounds then Break;
        if (Pos >= Line - Back) and (Pos <= Line - Back + Range) then
        begin
          State := NewState;
          SizingPos := Line;
          SizingOfs := Line - Pos;
          Index := I;
          Exit;
        end;
        Inc(Line, OptimalLineWidth);
      end;
      // for fixed sizing
      if (Pos <= FixedBounds)  then
      begin
       Line := 0;
       for I:= 0 to FixedCellCount-1 do
       begin
         Inc(Line, GetExtent(I));
         if Line > FixedBounds + Range then Break;
         if (Pos >= Line - Back) and (Pos <= Line - Back + Range) then
         begin
           State := NewState;
           SizingPos := Line;
           SizingOfs := Line - Pos;
           Index := I;
           Exit;
         end;
         Inc(Line, OptimalLineWidth);
       end;
      end;
      // last column
      if (GridBounds = GridExtent) and (Pos >= GridExtent - Back)
        and (Pos <= GridExtent) then
      begin
        State := NewState;
        SizingPos := GridExtent;
        SizingOfs := GridExtent - Pos;
        Index := LastFullVisibleCell;
      end;
    end;
  end;

  function XOutsideHorzFixedBounds: Boolean;
  begin
    with FixedInfo do
     {$ifdef DFS_COMPILER_4_UP}
      if not UseRightToLeftAlignment then
        Result := X > Horz.FixedBounds
      else
        Result := X < ClientWidth - Horz.FixedBounds;
     {$else}
        Result := X > Horz.FixedBounds;
     {$endif}
  end;

  function XOutsideOrEqualHorzFixedBounds: Boolean;
  begin
    with FixedInfo do
     {$ifdef DFS_COMPILER_4_UP}
      if not UseRightToLeftAlignment then
        Result := X >= Horz.FixedBounds
      else
        Result := X <= ClientWidth - Horz.FixedBounds;
      {$else}
        Result := X >= Horz.FixedBounds;
      {$endif}
  end;


var
  EffectiveOptions: TGridOptions;
begin
  State := gsNormal;
  Index := -1;
  EffectiveOptions := Options;
  if ((csDesigning in ComponentState) and AllowDesignSizing) or Customizing then
    EffectiveOptions := EffectiveOptions + [goColSizing, goRowSizing];
  if [goColSizing, goRowSizing] * EffectiveOptions <> [] then
    with FixedInfo do
    begin
      Vert.GridExtent := ClientHeight;
      Horz.GridExtent := ClientWidth;
      if (goColSizing in EffectiveOptions) then
      begin
        //for effective column resizing
        if not ((not FSizingAboveFixed) and (Y >= Vert.FixedBounds)) 
           or (not (Y >= Vert.GridBounds) or (not CanColSizing(X,Y))) then
        CalcAxisState(Horz, X, gsColSizing);
      end;
      if (Y > Vert.FixedBounds) and (goRowSizing in EffectiveOptions) then
      begin
        if XOutsideOrEqualHorzFixedBounds then Exit;
        CalcAxisState(Vert, Y, gsRowSizing);
      end;
    end;
end;

function TCustomExGrid.CanColSizing(X,Y : integer) : Boolean;
begin
   Result:=true;
end;

procedure TCustomExGrid.ChangeGridOrientation(RightToLeftOrientation: Boolean);
var
  Org: TPoint;
  Ext: TPoint;
begin
  if RightToLeftOrientation then
  begin
    Org := Point(ClientWidth,0);
    Ext := Point(-1,1);
//    SetMapMode(Canvas.Handle, mm_Anisotropic);
//    SetWindowOrgEx(Canvas.Handle, Org.X, Org.Y, nil);
//    SetViewportExtEx(Canvas.Handle, ClientWidth, ClientHeight, nil);
//    SetWindowExtEx(Canvas.Handle, Ext.X*ClientWidth, Ext.Y*ClientHeight, nil);
  end
  else
  begin
    Org := Point(0,0);
    Ext := Point(1,1);
  end;

  SetMapMode(Canvas.Handle, mm_Anisotropic);
  SetWindowOrgEx(Canvas.Handle, Org.X, Org.Y, nil);
  SetViewportExtEx(Canvas.Handle, ClientWidth, ClientHeight, nil);
  SetWindowExtEx(Canvas.Handle, Ext.X*ClientWidth, Ext.Y*ClientHeight, nil);

end;

procedure TCustomExGrid.ChangeSize(NewColCount, NewRowCount: Longint);
var
  OldColCount, OldRowCount: Longint;
  OldDrawInfo: TExGridDI;

  procedure MinRedraw(const OldInfo, NewInfo: TExGridAxisDI; Axis: Integer);
  var
    R: TRect;
    First: Integer;
  begin
    First := Min(OldInfo.LastFullVisibleCell, NewInfo.LastFullVisibleCell);
    // Get the rectangle around the leftmost or topmost cell in the target range.
    R := CellRect(First and not Axis, First and Axis);
    R.Bottom := Height;
    R.Right := Width;
    Windows.InvalidateRect(Handle, @R, False);
  end;

  procedure DoChange;
  var
    Coord: TGridCoord;
    NewDrawInfo: TExGridDI;
  begin
    if FColWidths <> nil then
    begin
      UpdateExtents(FColWidths, ColCount, DefaultColWidth);
      UpdateExtents(FTabStops, ColCount, Integer(True));
    end;
    if FRowHeights <> nil then
      UpdateExtents(FRowHeights, RowCount, DefaultRowHeight);

    Coord := FCurrent;


    if (Row >= RowCount) or (Row < FixedRows) then
    begin
       if RowCount>0 then
         Coord.Y := RowCount - 1 //09.11
       else Coord.Y := 0;
    end;
    if (Col >= ColCount) or (Col < Fixedcols)then
    begin
       if Colcount > 0 then
         Coord.X := ColCount - 1//09.11
       else Coord.X := 0;
    end;

    if (FCurrent.X <> Coord.X) or (FCurrent.Y <> Coord.Y) then
    begin
      //MoveCurrent(Coord.X, Coord.Y, True, True, [], swNone);   // CEDR 06.10.02 [G1]
      FocusCell(Coord.X, Coord.Y, True, [], swNone);   // CEDR 06.10.02 [G1]  to select 1st added node
    end;

    if (LeftCol <> OldDrawInfo.Horz.FirstGridCell) or
      (TopRow <> OldDrawInfo.Vert.FirstGridCell) then
    begin
      InvalidateGrid
    end
    else if HandleAllocated then
    begin
      CalcDI(NewDrawInfo);
      MinRedraw(OldDrawInfo.Horz, NewDrawInfo.Horz, 0);
      MinRedraw(OldDrawInfo.Vert, NewDrawInfo.Vert, -1);
    end;
    UpdateScrollRange;
    SizeChanged(OldColCount, OldRowCount);
  end;

begin
 if HandleAllocated then
    CalcDI(OldDrawInfo);

  OldColCount := FColCount;
  OldRowCount := FRowCount;
  FColCount := NewColCount;
  FRowCount := NewRowCount;

  if FixedCols > NewColCount then FFixedCols := NewColCount - 1;
  if FixedRows > NewRowCount then FFixedRows := NewRowCount - 1;
  try
    DoChange;
  except
    { Could not change size so try to clean up by setting the size back }
    FColCount := OldColCount;
    FRowCount := OldRowCount;
    DoChange;
    InvalidateGrid;
    raise;
  end;
end;

{ Will move TopLeft so that Coord is in view }
procedure TCustomExGrid.ClampInView(const Coord: TGridCoord);
var
  DrawInfo: TExGridDI;
  MaxTopLeft: TGridCoord;
  OldTopLeft: TGridCoord;
begin
  if not HandleAllocated then Exit;
  CalcDI(DrawInfo);
  with DrawInfo, Coord do
  begin
    if (X > Horz.LastFullVisibleCell) or
      (Y > Vert.LastFullVisibleCell) or (X < LeftCol) or (Y < TopRow) then
    begin
      OldTopLeft := FTopLeft;
      MaxTopLeft := CalcMaxTopLeft(Coord, DrawInfo);
      Update;
      if X < LeftCol then FTopLeft.X := X
      else if X > Horz.LastFullVisibleCell then FTopLeft.X := MaxTopLeft.X;
      if Y < TopRow then FTopLeft.Y := Y
      else if Y > Vert.LastFullVisibleCell then FTopLeft.Y := MaxTopLeft.Y;
      TopLeftMoved(OldTopLeft);
    end;
  end;
end;

procedure TCustomExGrid.DrawSizingLine(const DrawInfo: TExGridDI);
var
  OldPen: TPen;
begin
  OldPen := TPen.Create;
  try
    with Canvas, DrawInfo do
    begin
      OldPen.Assign(Pen);
      Pen.Style := psDot;
      Pen.Mode := pmXor;
      Pen.Width := 1;
      try
        if FGridState = gsRowSizing then
        begin
          MoveTo(0, FSizingPos);
          LineTo(Horz.GridBounds, FSizingPos);
        end
        else
        begin
          MoveTo(FSizingPos, 0);
          LineTo(FSizingPos, Vert.GridBounds);
        end;
      finally
        Pen := OldPen;
      end;
    end;
  finally
    OldPen.Free;
  end;
end;


procedure TCustomExGrid.StopDrag;
begin
     KillTimer(Handle, 1);
     FGridState := gsNormal;
end;



procedure TCustomExGrid.SetLineColor(Value : TColor);
begin
    if FLineColor <> Value then
    begin
        FlineColor := Value;
        InvalidateGrid;
    end;
end;

procedure TCustomExGrid.SetFixedLineColor(Value : TColor);
begin
    if FFixedLineColor <> Value then
    begin
        FFixedLineColor := Value;
        InvalidateGrid;
    end;

end;


procedure TCustomExGrid.FocusCell(ACol, ARow: Longint; MoveAnchor : Boolean; Shift: TShiftState; SelWay : TSelectionWay);
begin
  //if not IsGridEmpty then //19/10/01     // CEDR 06.10.02 [G1] rem-ed to allow focus first added node
  //begin
    HideEditor;
    MoveCurrent(ACol, ARow, MoveAnchor, True, Shift, SelWay);
    UpdateEdit;
  //end;
  //Click;
end;

procedure TCustomExGrid.GridRectToScreenRect(GridRect: TGridRect;
  var ScreenRect: TRect; IncludeLine, IncludeHExtent, IncludeVExtent: Boolean);

  function LinePos(const AxisInfo: TExGridAxisDI; Line: Integer): Integer;
  var
    Start, I: Longint;
  begin
    with AxisInfo do
    begin
      Result := 0;
      if Line < FixedCellCount then
        Start := 0
      else
      begin
        if Line >= FirstGridCell then
          Result := FixedBounds;
        Start := FirstGridCell;
      end;
      for I := Start to Line - 1 do
      begin
        Inc(Result, GetExtent(I) + OptimalLineWidth);
        if Result > GridExtent then
        begin
          Result := 0;
          Exit;
        end;
      end;
    end;
  end;

  function CalcAxis(const AxisInfo: TExGridAxisDI;
    GridRectMin, GridRectMax: Integer;
    var ScreenRectMin, ScreenRectMax: Integer; IncludeExtent: Boolean): Boolean;
  begin
    Result := False;
    with AxisInfo do
    begin
      if (GridRectMin >= FixedCellCount) and (GridRectMin < FirstGridCell) then
        if GridRectMax < FirstGridCell then
        begin
          FillChar(ScreenRect, SizeOf(ScreenRect), 0); { erase partial results }
          Exit;
        end
        else
          GridRectMin := FirstGridCell;
      if GridRectMax > LastFullVisibleCell then
      begin
        GridRectMax := LastFullVisibleCell;
        if GridRectMax < GridCellCount - 1 then Inc(GridRectMax);
        if LinePos(AxisInfo, GridRectMax) = 0 then
          Dec(GridRectMax);
      end;

      ScreenRectMin := LinePos(AxisInfo, GridRectMin);
      ScreenRectMax := LinePos(AxisInfo, GridRectMax);
      if ScreenRectMax = 0 then
        ScreenRectMax := ScreenRectMin + GetExtent(GridRectMin)
      else
        Inc(ScreenRectMax, GetExtent(GridRectMax));
      if (ScreenRectMax > GridExtent) or IncludeExtent then
        ScreenRectMax := GridExtent;
      if IncludeLine then Inc(ScreenRectMax, OptimalLineWidth);
    end;
    Result := True;
  end;

var
  DrawInfo: TExGridDI;
  Hold: Integer;
begin
  FillChar(ScreenRect, SizeOf(ScreenRect), 0);
  if (GridRect.Left > GridRect.Right) or (GridRect.Top > GridRect.Bottom) then
    Exit;
  CalcDI(DrawInfo);
  with DrawInfo do
  begin
    if GridRect.Left > Horz.LastFullVisibleCell + 1 then Exit;
    if GridRect.Top > Vert.LastFullVisibleCell + 1 then Exit;

    if CalcAxis(Horz, GridRect.Left, GridRect.Right, ScreenRect.Left,
      ScreenRect.Right, IncludeHExtent) then
    begin
      CalcAxis(Vert, GridRect.Top, GridRect.Bottom, ScreenRect.Top,
        ScreenRect.Bottom, IncludeVExtent);
    end;
  end;
  {$ifdef DFS_COMPILER_4_UP}
  if UseRightToLeftAlignment and (Canvas.CanvasOrientation = coLeftToRight) then
  begin
    Hold := ScreenRect.Left;
    ScreenRect.Left := ClientWidth - ScreenRect.Right;
    ScreenRect.Right := ClientWidth - Hold;
  end;
  {$endif}
end;

procedure TCustomExGrid.Initialize;
begin
  FTopLeft.X := FixedCols;
  FTopLeft.Y := FixedRows;
  FCurrent := FTopLeft;
  FAnchor := FCurrent;
  //newsel
  FEndAnchor := FCurrent;
  if goRowSelect in Options then
  begin
     FAnchor.X := ColCount-1;
     //newSel
     FEndAnchor.X := ColCount-1;
  end;
end;

procedure TCustomExGrid.InvalidateCell(ACol, ARow: Longint);
var
  Rect: TGridRect;
begin
 if not (EditorMode and (aCol = Col) and (aRow = Row)) then
 begin
  if (ACol < 0) or (ARow < 0) or (ACol >= ColCount) or (ARow >= RowCount) then
    exit;
  Rect.Top := ARow;
  Rect.Left := ACol;
  Rect.Bottom := ARow;
  Rect.Right := ACol;
  InvalidateRect(Rect);
 end; 
end;

procedure TCustomExGrid.InvalidateCol(ACol: Longint);
var
  Rect: TGridRect;
begin
  if not HandleAllocated then Exit;
  Rect.Top := 0;
  Rect.Left := ACol;
  Rect.Bottom := VisibleRowCount+1;
  Rect.Right := ACol;
  InvalidateRect(Rect);
end;

procedure TCustomExGrid.InvalidateFullRow(ARow : Longint);
var
  InvalidRect: TRect;
begin
  if not HandleAllocated then Exit;
  GridRectToScreenRect(GridRect(0,aRow,VisibleColCount+1, aRow), InvalidRect, True, True, False);
  Windows.InvalidateRect(Handle, @InvalidRect, true);//true);
end;

procedure TCustomExGrid.InvalidateRow(ARow: Longint);
var
  Rect: TGridRect;
begin
  if not HandleAllocated then Exit;
  Rect.Top := ARow;
  Rect.Left := 0;
  Rect.Bottom := ARow;
  Rect.Right := VisibleColCount+1;
  InvalidateRect(Rect);
end;

procedure TCustomExGrid.SetGridState(aState : TGridState);
begin
   FGridstate := aState;
   fMovePos :=-1;
   Update;
end;

function TCustomExGrid.GetGridState : TGridState;
begin
   Result := FGridstate;
end;

procedure TCustomExGrid.InvalidateGrid;
begin
  Invalidate;
end;

function TCustomExGrid.GetNonFixedArea : TRect;
var
  DrawInfo: TExGridDI;
begin
  CalcDI(DrawInfo);
  with DrawInfo do
    Result := Rect(Horz.FixedBounds, Vert.FixedBounds, Horz.GridExtent, Vert.GridBounds);
end;

procedure TCustomExGrid.InvalidateNonFixed;
var
  InvalidRect: TRect;
begin
 if not HandleAllocated then Exit;
 InvalidRect := GetNonFixedArea;
 Windows.InvalidateRect(Handle, @InvalidRect, true);//true);
end;

procedure TCustomExGrid.InvalidateRect(ARect: TGridRect);
var
  InvalidRect: TRect;
begin
  if not HandleAllocated then Exit;
  GridRectToScreenRect(ARect, InvalidRect, True,False, False);
  Windows.InvalidateRect(Handle, @InvalidRect, True);//true);
end;

procedure TCustomExGrid.ModifyScrollBar(ScrollBar, ScrollCode, Pos: Cardinal;
  UseRightToLeft: Boolean);
var
  NewTopLeft, MaxTopLeft: TGridCoord;
  DrawInfo: TExGridDI;
  RTLFactor: Integer;

  function Min_: Longint;
  begin
    if ScrollBar = SB_HORZ then Result := FixedCols
    else Result := FixedRows;
  end;

  function Max_: Longint;
  begin
    if ScrollBar = SB_HORZ then Result := MaxTopLeft.X
    else Result := MaxTopLeft.Y;
  end;

  function PageUp: Longint;
  var
    MaxTopLeft: TGridCoord;
  begin
    MaxTopLeft := CalcMaxTopLeft(FTopLeft, DrawInfo);
    if ScrollBar = SB_HORZ then
      Result := FTopLeft.X - MaxTopLeft.X else
      Result := FTopLeft.Y - MaxTopLeft.Y;
    if Result < 1 then Result := 1;
  end;

  function PageDown: Longint;
  var
    DrawInfo: TExGridDI;
  begin
    CalcDI(DrawInfo);
    with DrawInfo do
      if ScrollBar = SB_HORZ then
        Result := Horz.LastFullVisibleCell - FTopLeft.X else
        Result := Vert.LastFullVisibleCell - FTopLeft.Y;
    if Result < 1 then Result := 1;
  end;

  function CalcScrollBar(Value, ARTLFactor: Longint): Longint;
  begin
    Result := Value;
    case ScrollCode of
      SB_LINEUP:
        Dec(Result, ARTLFactor);
      SB_LINEDOWN:
        Inc(Result, ARTLFactor);
      SB_PAGEUP:
        Dec(Result, PageUp * ARTLFactor);
      SB_PAGEDOWN:
        Inc(Result, PageDown * ARTLFactor);
      SB_THUMBPOSITION, SB_THUMBTRACK:
        if (goThumbTracking in Options) or (ScrollCode = SB_THUMBPOSITION) then
        begin
          {$ifdef DFS_COMPILER_4_UP}
          if (not UseRightToLeftAlignment) or (ARTLFactor = 1) then
            Result := Min_ + LongMulDiv(Pos, Max_ - Min_, MaxShortInt)
          else
            Result := Max_ - LongMulDiv(Pos, Max_ - Min_, MaxShortInt);
          {$else}
            Result := Min_ + LongMulDiv(Pos, Max_ - Min_, MaxShortInt);
          {$endif}
        end;

      SB_BOTTOM:
        Result := Max_;
      SB_TOP:
        Result := Min_;
    end;
  end;

  procedure ModifyPixelScrollBar(Code, Pos: Cardinal);
  var
    NewOffset: Integer;
    OldOffset: Integer;
    R: TRect;
    GridSpace, ColWidth: Integer;
  begin
    NewOffset := FColOffset;
    ColWidth := ColWidths[DrawInfo.Horz.FirstGridCell];
    GridSpace := ClientWidth - DrawInfo.Horz.FixedBounds;
    case Code of
      SB_LINEUP: Dec(NewOffset, Canvas.TextWidth('0') * RTLFactor);
      SB_LINEDOWN: Inc(NewOffset, Canvas.TextWidth('0') * RTLFactor);
      SB_PAGEUP: Dec(NewOffset, GridSpace * RTLFactor);
      SB_PAGEDOWN: Inc(NewOffset, GridSpace * RTLFactor);
      SB_THUMBPOSITION,
      SB_THUMBTRACK:
        if (goThumbTracking in Options) or (Code = SB_THUMBPOSITION) then
        begin
          {$ifdef DFS_COMPILER_4_UP}
          if not UseRightToLeftAlignment then
            NewOffset := Pos
          else
            NewOffset := Max_ - Integer(Pos);
          {$else}
            NewOffset := Pos;
          {$endif}

        end;
      SB_BOTTOM: NewOffset := 0;
      SB_TOP: NewOffset := ColWidth - GridSpace;
    end;

    // CEDR 03.10.02 -     NewOffset:
    if NewOffset >= ColWidth - GridSpace then
      NewOffset:=ColWidth-GridSpace;
    if NewOffset<0 then
      NewOffset:=0;
{    if NewOffset < 0 then
      NewOffset := 0
    else if NewOffset >= ColWidth - GridSpace then
        NewOffset := ColWidth - GridSpace;}
    if NewOffset <> FColOffset then
    begin
      OldOffset := FColOffset;
      FColOffset := NewOffset;
      ScrollData(OldOffset - NewOffset, 0);
      //FillChar(R, SizeOf(R), 0);
      //R.Bottom := FixedRows+VisibleRowcount+1;
      //InvalidateRect(R);
      //Update;
      R := ClientRect;
      Windows.InvalidateRect(Handle, @R, True);
      UpdateScrollPos;
    end;
  end;

var
  Temp: Longint;
begin
 FScrolling := true;
 try
  {$ifdef DFS_COMPILER_4_UP}
  if (not UseRightToLeftAlignment) or (not UseRightToLeft) then
    RTLFactor := 1
  else
    RTLFactor := -1;
  {$else}
    RTLFactor := 1;
  {$endif}
  //if Visible and CanFocus and TabStop and not (csDesigning in ComponentState) then
  // if not Focused then
  //  SetFocus;
  CalcDI(DrawInfo);

  if (ScrollBar = SB_HORZ) and (ColCount = 1) then
  begin
    ModifyPixelScrollBar(ScrollCode, Pos);
    Exit;
  end;
  MaxTopLeft.X := ColCount - 1;
  MaxTopLeft.Y := RowCount - 1;
  MaxTopLeft := CalcMaxTopLeft(MaxTopLeft, DrawInfo);
  NewTopLeft := FTopLeft;
  if ScrollBar = SB_HORZ then
    repeat
      Temp := NewTopLeft.X;
      NewTopLeft.X := CalcScrollBar(NewTopLeft.X, RTLFactor);
    until (NewTopLeft.X <= FixedCols) or (NewTopLeft.X >= MaxTopLeft.X)
      or (ColWidths[NewTopLeft.X] > 0) or (Temp = NewTopLeft.X)
  else
    repeat
      Temp := NewTopLeft.Y;
      NewTopLeft.Y := CalcScrollBar(NewTopLeft.Y, 1);
    until (NewTopLeft.Y <= FixedRows) or (NewTopLeft.Y >= MaxTopLeft.Y)
      or (RowHeights[NewTopLeft.Y] > 0) or (Temp = NewTopLeft.Y);

  {$ifdef DFS_COMPILER_4_UP}
  NewTopLeft.X := Math.Max(FixedCols, Math.Min(MaxTopLeft.X, NewTopLeft.X));
  NewTopLeft.Y := Math.Max(FixedRows, Math.Min(MaxTopLeft.Y, NewTopLeft.Y));
  {$else}
  NewTopLeft.X := Max(FixedCols, Min(MaxTopLeft.X, NewTopLeft.X));
  NewTopLeft.Y := Max(FixedRows, Min(MaxTopLeft.Y, NewTopLeft.Y));
  {$endif}

  if (NewTopLeft.X <> FTopLeft.X) or (NewTopLeft.Y <> FTopLeft.Y) then
    MoveTopLeft(NewTopLeft.X, NewTopLeft.Y);
 finally
   FScrolling :=false;
 end;
end;

procedure TCustomExGrid.MoveAdjust(var CellPos: Longint; FromIndex, ToIndex: Longint);
var
  Min, Max: Longint;
begin
  if CellPos = FromIndex then CellPos := ToIndex
  else
  begin
    Min := FromIndex;
    Max := ToIndex;
    if FromIndex > ToIndex then
    begin
      Min := ToIndex;
      Max := FromIndex;
    end;
    if (CellPos >= Min) and (CellPos <= Max) then
      if FromIndex > ToIndex then
        Inc(CellPos) else
        Dec(CellPos);
  end;
end;

procedure TCustomExGrid.MoveAnchor(const NewAnchor: TGridCoord; MoveFocus : Boolean);
var
  OldSel: TGridRect;
begin
  if [goRangeSelect, goEditing] * Options = [goRangeSelect] then
  begin
    OldSel := Selection;
    FAnchor := NewAnchor;
    //newsel
    FEndAnchor := NewAnchor;

    if goRowSelect in Options then
    begin
       FAnchor.X := ColCount - 1;
       //newsel
       FEndAnchor.X := ColCount - 1;
    end;
    ClampInView(NewAnchor);
    SelectionMoved(OldSel);
  end
  else
    MoveCurrent(NewAnchor.X, NewAnchor.Y, True,  True, [], swNone);
end;

function TCustomExGrid.InSelection(aCol, aRow : integer) : Boolean;
begin
   Result := PointInGridRect(aCol, ARow, Selection);
end;

procedure TCustomExGrid.MoveCurrent(ACol, ARow: Longint; MoveAnchor, Show : Boolean; Shift: TShiftState; SelWay : TSelectionWay);
var
  OldSel: TGridRect;
  OldCurrent: TGridCoord;
begin
//   Shift -    Fcurrent
  if (ACol < 0) or (ARow < 0) or (ACol > ColCount) or (ARow > RowCount) then
    InvalidOp(SIndexOutOfRange);
  FSelectionCleared := false;
  if SelectCell(ACol, ARow, MoveAnchor, Shift, SelWay) then
  begin

    OldSel := Selection;
    OldCurrent := FCurrent;

    if (not (goAlwaysShowEditor in Options)) or (RowCount = FixedRows)
     then HideEditor;

    FCurrent.X := ACol;
    FCurrent.Y := ARow;

//    if (OldCurrent.X<>FCurrent.X)or(OldCurrent.Y<>FCurrent.Y) then
//      FocusCell(ACol, ARow, MoveAnchor, Shift, SelWay);    // CEDR tmp

    if goRowSelect in Options then
    begin
       FCurrent.X := FixedCols;
    end;
    //    -   selection
    //  ,   
    if MoveAnchor or not (goRangeSelect in Options) then
    begin
      FAnchor := FCurrent;
      if goRowSelect in Options then
         FAnchor.X := ColCount-1;
      //newsel
      FEndAnchor := FAnchor;
      if goRowSelect in Options then
         FEndAnchor.X := Fixedcols;
    end
    else
    begin
      if  (not PointInGridRect(FCurrent.X, Fcurrent.Y, Selection))  then
      begin
       if (FEndAnchor.Y >= FAnchor.Y) then
       begin
        if (FCurrent.Y < FAnchor.Y)  then
           FAnchor.Y := FCurrent.Y
        else if (FCurrent.Y > FEndAnchor.Y)  then
           FEndAnchor.Y := FCurrent.Y
       end
       else if (FEndAnchor.Y < FAnchor.Y) then
       begin
         if (FCurrent.Y < FEndAnchor.Y)  then
           FEndAnchor.Y := FCurrent.Y
         else if (FCurrent.Y > FAnchor.Y)  then
           FAnchor.Y := FCurrent.Y
       end;

       if (FEndAnchor.X >= FAnchor.X) then
       begin
        if (FCurrent.X < FAnchor.X) then
           FAnchor.X := FCurrent.X
        else if (FCurrent.X > FEndAnchor.X) then
           FEndAnchor.X := FCurrent.X;
       end
       else if (FEndAnchor.X < FAnchor.X) then
       begin
         if (FCurrent.X < FEndAnchor.X) then
           FEndAnchor.X := FCurrent.X
         else if (FCurrent.X > FAnchor.X) then
           FAnchor.X := FCurrent.X;
       end;
      end
      else
         FEndAnchor := FCurrent;
    end;
    if Show then
      ClampInView(FCurrent);//);

    SelectionMoved(OldSel);
    with OldCurrent do InvalidateCell(X, Y);
    with FCurrent do InvalidateCell(X, Y);
    if (goAlwaysShowEditor in Options) and not (FixedRows=Rowcount) then
        ShowEditor;
    //Click;
  end;
end;

procedure TCustomExGrid.MoveTopLeft(ALeft, ATop: Longint);
var
  OldTopLeft: TGridCoord;
begin
  if (ALeft = FTopLeft.X) and (ATop = FTopLeft.Y) then Exit;
//  Update;
  Repaint;
  OldTopLeft := FTopLeft;
  FTopLeft.X := ALeft;
  FTopLeft.Y := ATop;
  TopLeftMoved(OldTopLeft);
end;

procedure TCustomExGrid.ResizeCol(Index: Longint; OldSize, NewSize: Integer);
begin
  invalidateGrid;
end;

procedure TCustomExGrid.ResizeRow(Index: Longint; OldSize, NewSize: Integer);
begin
  InvalidateGrid;
end;

procedure TCustomExGrid.SelectionMoved(const OldSel: TGridRect);
var
  OldRect, NewRect: TRect;
  AXorRects: TXorRects;
  I: Integer;
begin
  if not HandleAllocated then Exit;
  GridRectToScreenRect(OldSel, OldRect, True, False, False);
  GridRectToScreenRect(Selection, NewRect, True, False, False);
  XorRects(OldRect, NewRect, AXorRects);
  for I := Low(AXorRects) to High(AXorRects) do
    Windows.InvalidateRect(Handle, @AXorRects[I], True);
end;

procedure TCustomExGrid.ScrollDataInfo(DX, DY: Integer;
  var DrawInfo: TExGridDI);
var
  ScrollArea: TRect;
  TitlesRect : TGridRect;
  ScrollFlags: Integer;
//  IStep,Step : integer;
//  ScrollBmp : TBitmap;

begin
 //SendMessage(Handle, WM_SETREDRAW, Ord(False), 0);
 //try
  with DrawInfo do
  begin
    ScrollFlags := SW_INVALIDATE+SW_ERASE+SW_SMOOTHSCROLL;
    { Scroll the area }
    if DY = 0 then
    begin
      { Scroll both the column titles and data area at the same time }
      {$ifdef DFS_COMPILER_4_UP}
      if not UseRightToLeftAlignment then
        ScrollArea := Rect(Horz.FixedBounds, 0, Horz.GridExtent-1, Vert.GridExtent)
      else
      begin
        ScrollArea := Rect(ClientWidth - Horz.GridExtent + 1, 0, ClientWidth - Horz.FixedBounds, Vert.GridExtent);
        DX := -DX;
      end;
      {$else}
        ScrollArea := Rect(Horz.FixedBounds, 0, Horz.GridExtent-1, Vert.GridExtent);
      {$endif}

      ScrollWindowEx(Handle, DX, 0, @ScrollArea, @ScrollArea, 0, nil, ScrollFlags);

    end
    else if DX = 0 then
    begin
      { Scroll both the row titles and data area at the same time }
      ScrollArea := Rect(0, Vert.FixedBounds, Horz.GridExtent, Vert.GridExtent);
      ScrollWindowEx(Handle, 0, DY, @ScrollArea, @ScrollArea, 0, nil, ScrollFlags);
{      ScrollBmp := TBitmap.Create;
      try
        ScrollBmp.Width := ScrollArea.Right-ScrollArea.Left;
        ScrollBmp.Height := ScrollArea.Bottom-ScrollArea.Top;
        BitBlt(ScrollBmp.Canvas.Handle, 0, 0, ScrollBmp.Width, ScrollBmp.Height, Canvas.Handle, ScrollArea.Left, ScrollArea.Top, SRCCOPY);
        Canvas.Brush.Color := Color;
        Step := Abs(DY);
        IStep :=AnimationInc;
        While IStep < Step do
        begin
           if DY < 0 then
           begin
              BitBlt(Canvas.Handle, ScrollArea.Left, ScrollArea.Top, ScrollBmp.Width, ScrollBmp.Height-iStep, ScrollBmp.Canvas.Handle, 0, IStep, SRCCOPY);
              Canvas.FillRect(Rect(0, (ClientHeight-IStep-AnimationInc), ClientWidth, (ClientHeight-IStep)));
           end
           else
           begin
             BitBlt(Canvas.Handle, ScrollArea.Left, ScrollArea.Top+IStep, ScrollBmp.Width, ScrollBmp.Height-iStep, ScrollBmp.Canvas.Handle, 0, 0, SRCCOPY);
             Canvas.FillRect(Rect(0, IStep+AnimationInc, ClientWidth, IStep));
           end;
           Inc(IStep, AnimationInc);
        end;
        Sleep(1);
      finally
        ScrollBmp.Free;
      end;
}
    end
    else
    begin
      { Scroll titles and data area separately }
      { Column titles }
      ScrollArea := Rect(Horz.FixedBounds, 0, Horz.GridExtent-1, Vert.FixedBounds);
      ScrollWindowEx(Handle, DX, 0, @ScrollArea, @ScrollArea, 0, nil, ScrollFlags);
      { Row titles }
      ScrollArea := Rect(0, Vert.FixedBounds, Horz.FixedBounds, Vert.GridExtent);
      ScrollWindowEx(Handle, 0, DY, @ScrollArea, @ScrollArea, 0, nil, ScrollFlags);
      { Data area }
      ScrollArea := Rect(Horz.FixedBounds, Vert.FixedBounds, Horz.GridExtent,
        Vert.GridExtent);
      ScrollWindowEx(Handle, DX, DY, @ScrollArea, @ScrollArea, 0, nil, ScrollFlags);

      with TitlesRect do
      begin
         Left := LeftCol;
         Top :=  0;
         Right := Colcount - LeftCol;
         Bottom := FixedRows;
      end;
      InvalidateRect(TitlesRect);
    end;
  end;
  if goRowSelect in Options then
    InvalidateRect(Selection);
  //InvalidateGrid;
 //finally
 //  SendMessage(Handle, WM_SETREDRAW, ord(True), 0);
 //  UpdateWindow(Handle);
 //end;
end;

procedure TCustomExGrid.ScrollData(DX, DY: Integer);
var
  DrawInfo: TExGridDI;
begin
  CalcDI(DrawInfo);
  ScrollDataInfo(DX, DY, DrawInfo);
end;

procedure TCustomExGrid.TopLeftMoved(const OldTopLeft: TGridCoord);

  function CalcScroll(const AxisInfo: TExGridAxisDI;
    OldPos, CurrentPos: Integer; var Amount: Longint): Boolean;
  var
    Start, Stop: Longint;
    I: Longint;
  begin
    Result := False;
    with AxisInfo do
    begin
      if OldPos < CurrentPos then
      begin
        Start := OldPos;
        Stop := CurrentPos;
      end
      else
      begin
        Start := CurrentPos;
        Stop := OldPos;
      end;
      Amount := 0;
      for I := Start to Stop - 1 do
      begin
        Inc(Amount, GetExtent(I) + OptimalLineWidth);
        if Amount > (GridBounds - FixedBounds) then
        begin
          { Scroll amount too big, redraw the whole thing }
          InvalidateGrid;
          Exit;
        end;
      end;
      if OldPos < CurrentPos then Amount := -Amount;
    end;
    Result := True;
  end;

var
  DrawInfo: TExGridDI;
  Delta: TGridCoord;
  ToY, ToX : Boolean;
begin
  UpdateScrollPos;
  CalcDI(DrawInfo);
  Invalidate;//
  ToX := CalcScroll(DrawInfo.Horz, OldTopLeft.X, FTopLeft.X, Delta.X);
  ToY := CalcScroll(DrawInfo.Vert, OldTopLeft.Y, FTopLeft.Y, Delta.Y);
  {if CalcScroll(DrawInfo.Horz, OldTopLeft.X, FTopLeft.X, Delta.X) or
    CalcScroll(DrawInfo.Vert, OldTopLeft.Y, FTopLeft.Y, Delta.Y) then
  }
  if ToX or ToY then
    ScrollDataInfo(Delta.X, Delta.Y, DrawInfo);

  TopLeftChanged;
  {ifdef DFS_COMPILER_4_UP}
   //Repaint;
  {ELSE}
   // Invalidate;//new 06/2001//Repaint;//Update;
  {endif}
end;

procedure TCustomExGrid.UpdateScrollPos;
var
  DrawInfo: TExGridDI;
  MaxTopLeft: TGridCoord;
  GridSpace, ColWidth: Integer;

  procedure SetScroll(Code: Word; Value: Integer);
  begin
    {$ifdef DFS_COMPILER_4_UP}
    if UseRightToLeftAlignment and (Code = SB_HORZ) then
      if ColCount <> 1 then Value := MaxShortInt - Value
      else                  Value := (ColWidth - GridSpace) - Value;
    {$endif}
    {$ifdef DFS_COMPILER_4_UP}
    if fFlatSBPresent then
    begin
     if FlatSB_GetScrollPos(Handle, Code) <> Value then //ScrollBars/?
       FlatSB_SetScrollPos(Handle, Code, Value, True);  //ScrollBars/?
    end
    else
    begin
      if GetScrollPos(Handle, Code) <> Value then //ScrollBars/?
        SetScrollPos(Handle, Code, Value, True);  //ScrollBars/?
    end;
    {$else}
     if GetScrollPos(Handle, Code) <> Value then //ScrollBars/?
       SetScrollPos(Handle, Code, Value, True);  //ScrollBars/?
    {$endif}
  end;

begin
  if (not HandleAllocated) or (ScrollBars = ssNone) then Exit;
  CalcDI(DrawInfo);
  MaxTopLeft.X := ColCount - 1;
  MaxTopLeft.Y := RowCount - 1;
  MaxTopLeft := CalcMaxTopLeft(MaxTopLeft, DrawInfo);
  if ScrollBars in [ssHorizontal, ssBoth] then
    if ColCount = 1 then
    begin
      ColWidth := ColWidths[DrawInfo.Horz.FirstGridCell];
      GridSpace := ClientWidth - DrawInfo.Horz.FixedBounds;
      if (FColOffset > 0) and (GridSpace > (ColWidth - FColOffset)) and
        (ColWidth>=GridSpace)
      then
        ModifyScrollbar(SB_HORZ, SB_THUMBPOSITION, ColWidth - GridSpace, True)
      else
        SetScroll(SB_HORZ, FColOffset)
    end
    else
      SetScroll(SB_HORZ, LongMulDiv(FTopLeft.X - FixedCols, MaxShortInt,
        MaxTopLeft.X - FixedCols));
  if ScrollBars in [ssVertical, ssBoth] then
   SetScroll(SB_VERT, LongMulDiv(FTopLeft.Y - FixedRows, MaxShortInt,
      MaxTopLeft.Y - FixedRows));
end;

procedure TCustomExGrid.Updatewidths;
begin
end;

procedure TCustomExGrid.UpdateScrollRange;
var
  MaxTopLeft, OldTopLeft: TGridCoord;
  DrawInfo: TExGridDI;
  OldScrollBars: TScrollStyle;
  Updated: Boolean;

  procedure DoUpdate;
  begin
    if not Updated then
    begin
      {$ifdef DFS_COMPILER_4_UP}
       Update;
      {$else}
       Invalidate;
      {$endif}
      Updated := True;
    end;
  end;

  function ScrollBarVisible(Code: Word): Boolean;
  var
    Min, Max: Integer;
  begin
    Result := False;
    if (ScrollBars = ssBoth) or
      ((Code = SB_HORZ) and (ScrollBars = ssHorizontal)) or
      ((Code = SB_VERT) and (ScrollBars = ssVertical)) then
    begin
     {$ifdef DFS_COMPILER_4_UP}
     if fFlatSBPresent then
       FlatSB_GetScrollRange(Handle, Code, Min, Max)//Scrollbars/?
     else
       GetScrollRange(Handle, Code, Min, Max);//Scrollbars/?

     Result := Min <> Max;
     {$else}
      GetScrollRange(Handle, Code, Min, Max);
      Result := Min <> Max;
     {$endif}
    end;
  end;

  procedure CalcSizeInfo;
  begin
    CalcDIXY(DrawInfo, DrawInfo.Horz.GridExtent, DrawInfo.Vert.GridExtent);
    MaxTopLeft.X := ColCount - 1;
    MaxTopLeft.Y := RowCount - 1;
    MaxTopLeft := CalcMaxTopLeft(MaxTopLeft, DrawInfo);
  end;

  procedure SetAxisRange(var Max, Old, Current: Longint; Code: Word;
    Fixeds: Integer);
  begin
    CalcSizeInfo;
    {$ifdef DFS_COMPILER_4_UP}
    if fFlatSBPresent then
    begin
     if Fixeds < Max then
       FlatSB_SetScrollRange(Handle, Code, 0, MaxShortInt, True)//Scrollbars/?
     else
      FlatSB_SetScrollRange(Handle, Code, 0, 0, True);//Scrollbars/?

      FlatSB_ShowScrollBar(Handle, Code, Fixeds < Max);
    end
    else
    begin
     if Fixeds < Max then
      SetScrollRange(Handle, Code, 0, MaxShortInt, True)//Scrollbars/?
     else
     begin
      SetScrollRange(Handle, Code, 0, 0, True);//Scrollbars/?
     end;
     ShowScrollBar(Handle, Code, Fixeds < Max);
    end;
    {$else}
    if Fixeds < Max then
      SetScrollRange(Handle, Code, 0, MaxShortInt, True)//Scrollbars/?
    else
      SetScrollRange(Handle, Code, 0, 0, True);//Scrollbars/?
      ShowScrollBar(Handle, Code, Fixeds < Max);
    {$endif}

    if Old > Max then
    begin
      DoUpdate;
      Current := Max;
    end;
  end;

  procedure SetHorzRange;
  var
    Range: Integer;
  begin
    if OldScrollBars in [ssHorizontal, ssBoth] then
      if ColCount = 1 then
      begin
        Range := ColWidths[0] - ClientWidth;
        if Range < 0 then Range := 0;
            {$ifdef DFS_COMPILER_4_UP}
            if fFlatSBPresent then
            begin
              FlatSB_SetScrollRange(Handle, SB_HORZ, 0, Range, True);
              FlatSB_ShowScrollBar(Handle, SB_HORZ, Range > 0);
            end
            else
            begin
              if Range > 0 then
                SetScrollRange(Handle, SB_HORZ, 0, Range, True);//Scrollbars/?
              ShowScrollBar(Handle, SB_HORZ, Range > 0);  
            end;
            {$else}
               if Range > 0 then
                 SetScrollRange(Handle, SB_HORZ, 0, Range, True);//Scrollbars/?
               ShowScrollBar(Handle, SB_HORZ, Range > 0);  
            {$endif}
      end
      else
        SetAxisRange(MaxTopLeft.X, OldTopLeft.X, FTopLeft.X, SB_HORZ, FixedCols);
  end;

  procedure SetVertRange;
  begin
    if OldScrollBars in [ssVertical, ssBoth] then
      SetAxisRange(MaxTopLeft.Y, OldTopLeft.Y, FTopLeft.Y, SB_VERT, FixedRows);
  end;

begin
  if (ScrollBars = ssNone) or not HandleAllocated {or not Showing} then Exit;
  with DrawInfo do
  begin
    Horz.GridExtent := ClientWidth;
    Vert.GridExtent := ClientHeight;
    { Ignore scroll bars for initial calculation }
    if ScrollBarVisible(SB_HORZ) then
      Inc(Vert.GridExtent, GetSystemMetrics(SM_CYHSCROLL));
    if ScrollBarVisible(SB_VERT) then
      Inc(Horz.GridExtent, GetSystemMetrics(SM_CXVSCROLL));
  end;
  OldTopLeft := FTopLeft;
  { Temporarily mark us as not having scroll bars to avoid recursion }
  OldScrollBars := FScrollBars;
  FScrollBars := ssNone;
  Updated := False;

  //SendMessage(Handle, WM_SETREDRAW, ord(False), 0);
  try
    { Update scrollbars }
    SetHorzRange;
    DrawInfo.Vert.GridExtent := ClientHeight;
    SetVertRange;
    if DrawInfo.Horz.GridExtent <> ClientWidth then
    begin
      DrawInfo.Horz.GridExtent := ClientWidth;
      SetHorzRange;
    end;
  finally
    FScrollBars := OldScrollBars;
    SendMessage(Handle, WM_NCPAINT , 0, 0); //it for sizeGrip functionality
  end;
  UpdateScrollPos;

  if (FTopLeft.X <> OldTopLeft.X) or (FTopLeft.Y <> OldTopLeft.Y) then
    TopLeftMoved(OldTopLeft);
end;

function TCustomExGrid.CreateEditor: TInplaceEdit;
begin
  Result := TInplaceEdit.Create(Self);
end;

procedure TCustomExGrid.DestroyWnd;
begin
   HideEditor;
   {$ifdef DFS_COMPILER_4_UP}
   if fFlatSBPresent then//Scrollbars/?
   begin
       UnInitializeFlatSB(Handle);
       fFlatSBPresent :=false;
   end;
   {$endif}
   inherited;
end;

procedure TCustomExGrid.CreateWnd;
begin
  inherited;
  {$ifdef DFS_COMPILER_4_UP}
   if (FScrollBars <> ssNone) then//Scrollbars/?
   begin
      try
       if (not UseRightToLeftAlignment) then
       begin
         fFlatSBPresent := InitializeFlatSB(Handle);
       end;
       FlatSB_SetScrollProp(Handle, WSB_PROP_HSTYLE, Ord(FScrollBarsStyle), False);
       FlatSB_SetScrollProp(Handle, WSB_PROP_VSTYLE, Ord(FScrollBarsStyle), False);

       FlatSB_SetScrollProp(Handle, WSB_PROP_VBKGCOLOR, ColorToRGB(FScrollBarsColor), False);
       FlatSB_SetScrollProp(Handle, WSB_PROP_HBKGCOLOR, ColorToRGB(FScrollBarsColor), False);

       //FlatSB_ShowScrollBar(Handle, SB_VERT, FScrollBars in [ssVertical, ssBoth]);
       //FlatSB_ShowScrollBar(Handle, SB_HORZ, FScrollBars in [ssHorizontal, ssBoth]);
      except
       fFlatSBPresent := False;
      end;
   end;
   {$endif}
end;

procedure TCustomExGrid.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or WS_TABSTOP;
    if FScrollBars in [ssVertical, ssBoth] then Style := Style or WS_VSCROLL;
    if FScrollBars in [ssHorizontal, ssBoth] then Style := Style or WS_HSCROLL;

    WindowClass.style := CS_DBLCLKS;
    if FBorderStyle = bsSingle then
      if NewStyleControls and Ctl3D then
      begin
        Style := Style and not WS_BORDER;
        ExStyle := ExStyle or WS_EX_CLIENTEDGE;
      end
      else
        Style := Style or WS_BORDER;
  end;
end;

procedure TCustomExGrid.KeyDown(var Key: Word; Shift: TShiftState);
var
  NewTopLeft, NewCurrent, MaxTopLeft: TGridCoord;
  DrawInfo: TExGridDI;
  PageWidth, PageHeight: Integer;
  RTLFactor: Integer;

  procedure CalcPageExtents;
  begin
    CalcDI(DrawInfo);
    PageWidth := DrawInfo.Horz.LastFullVisibleCell - LeftCol;
    if PageWidth < 1 then PageWidth := 1;
    PageHeight := DrawInfo.Vert.LastFullVisibleCell - TopRow;
    if PageHeight < 1 then PageHeight := 1;
  end;

  procedure Restrict(var Coord: TGridCoord; MinX, MinY, MaxX, MaxY: Longint);
  begin
    with Coord do
    begin
      if X > MaxX then X := MaxX
      else if X < MinX then X := MinX;
      if Y > MaxY then Y := MaxY
      else if Y < MinY then Y := MinY;
    end;
  end;

begin
  inherited KeyDown(Key, Shift);
  if not CanGridAcceptKey(Key, Shift) then exit;//Key := 0;
  {$ifdef DFS_COMPILER_4_UP}
  if UseRightToLeftAlignment then
    RTLFactor := -1
  else
  {$endif}
    RTLFactor := 1;

  NewCurrent := FCurrent;
  NewTopLeft := FTopLeft;
  CalcPageExtents;
  {if ssCtrl in Shift then
    case Key of
      88://cut
        SendMessage(Handle, WM_CUT,0,0);
      86://paste
        SendMessage(Handle, WM_PASTE,0,0);
      67://copy
        SendMessage(Handle, WM_COPY,0,0);
      
      VK_UP:
      begin
        Dec(NewTopLeft.Y);
      end;
      VK_DOWN:
      begin
        Inc(NewTopLeft.Y); //new 4/3/2002
      end;
      VK_LEFT:
        if not (goRowSelect in Options) then
        begin
          Dec(NewCurrent.X, PageWidth * RTLFactor);
          Dec(NewTopLeft.X, PageWidth * RTLFactor);
        end;
      VK_RIGHT:
        if not (goRowSelect in Options) then
        begin
          Inc(NewCurrent.X, PageWidth * RTLFactor);
          Inc(NewTopLeft.X, PageWidth * RTLFactor);
        end;
      VK_PRIOR:
      begin
        NewCurrent.Y := TopRow;
      end;
      VK_NEXT:
      begin
        NewCurrent.Y := DrawInfo.Vert.LastFullVisibleCell;
      end;
      VK_HOME:
        begin
          NewTopLeft.X := FixedCols;
          NewTopLeft.Y := FixedRows;
        end;
      VK_END:
        begin
          //NewTopLeft.X := FixedCols;
          //NewTopLeft.Y := FixedRows;
          ModifyScrollBar(SB_VERT, SB_BOTTOM, 0, True);
          ModifyScrollBar(SB_HORZ, SB_BOTTOM, 0, True);
          Exit;
        end;
    end
  else}//4/3/2002
    case Key of
      VK_UP:
        if (ssCtrl in Shift) and (not (gorangeSelect in Options)) then
          Dec(NewTopLeft.Y) //new 4/3/2002
        else
          Dec(NewCurrent.Y);
      VK_DOWN:
        if (ssCtrl in Shift) and (not (gorangeSelect in Options)) then
          Inc(NewTopLeft.Y) //new 4/3/2002
        else
          Inc(NewCurrent.Y);
      VK_LEFT:
        if goRowSelect in Options then
          Dec(NewCurrent.Y, RTLFactor)
        else
          Dec(NewCurrent.X, RTLFactor);
      VK_RIGHT:
        if goRowSelect in Options then
           Inc(NewCurrent.Y, RTLFactor)
        else
          Inc(NewCurrent.X, RTLFactor);
      VK_NEXT:
        begin
          Inc(NewCurrent.Y, PageHeight);
          Inc(NewTopLeft.Y, PageHeight);
        end;
      VK_PRIOR:
        begin
          Dec(NewCurrent.Y, PageHeight);
          Dec(NewTopLeft.Y, PageHeight);
        end;
      VK_HOME:
        if goRowSelect in Options then
          NewCurrent.Y := FixedRows
        else if Col = FixedCols then
          NewCurrent.Y := FixedRows
        else
          NewCurrent.X := FixedCols;
      VK_END:
        if goRowSelect in Options then
          NewCurrent.Y := RowCount - 1
        else if Col = ColCount-1 then
          NewCurrent.Y := Rowcount-1
        else
          NewCurrent.X := ColCount - 1;
      VK_TAB:
        if not (ssAlt in Shift) then
        repeat
          if ssShift in Shift then
          begin
            Dec(NewCurrent.X);
            if NewCurrent.X < FixedCols then
            begin
              NewCurrent.X := ColCount - 1;
              Dec(NewCurrent.Y);
              if NewCurrent.Y < FixedRows then NewCurrent.Y := RowCount - 1;
            end;
            Shift := [];
          end
          else
          begin
            Inc(NewCurrent.X);
            if NewCurrent.X >= ColCount then
            begin
              NewCurrent.X := FixedCols;
              Inc(NewCurrent.Y);
              if NewCurrent.Y >= RowCount then NewCurrent.Y := FixedRows;
            end;
          end;
        until TabStops[NewCurrent.X] or (NewCurrent.X = FCurrent.X);
      VK_F2: if not Editormode then EditorMode := True;
      0: exit;
    end;
  MaxTopLeft.X := ColCount - 1;
  MaxTopLeft.Y := RowCount - 1;
  MaxTopLeft := CalcMaxTopLeft(MaxTopLeft, DrawInfo);
  Restrict(NewTopLeft, FixedCols, FixedRows, MaxTopLeft.X, MaxTopLeft.Y);
  if (NewTopLeft.X <> LeftCol) or (NewTopLeft.Y <> TopRow) then
    MoveTopLeft(NewTopLeft.X, NewTopLeft.Y);

  Restrict(NewCurrent, FixedCols, FixedRows, ColCount - 1, RowCount - 1);
  if (NewCurrent.X <> Col) or (NewCurrent.Y <> Row) then
    FocusCell(NewCurrent.X, NewCurrent.Y, not ((ssShift in Shift) or (ssCtrl in Shift)), Shift, swByKeyBoard)
  else if (NewCurrent.Y = RowCount-1) then//if last row
    AppendRow;


end;

procedure TCustomExGrid.AppendRow;
begin
   if Assigned(FOnAppendRow) then
      OnAppendrow(Self);
end;

procedure TCustomExGrid.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  if not (goAlwaysShowEditor in Options) and (goEditing in Options) and (Key in [#13,#27]) then
  begin
    if FEditorMode then
      HideEditor
    else
      ShowEditor;
    Key := #0;
  end;
end;

procedure TCustomExGrid.MoveAndScroll(Mouse, CellHit: Longint;
  var DrawInfo: TExGridDI; var Axis: TExGridAxisDI;
  ScrollBar: Integer; const MousePt: TPoint);
  var aCoord : TGridCoord;

  procedure DoScroll;
  begin
    if (Mouse-3 < Axis.FixedBounds) then
      begin
        ModifyScrollbar(ScrollBar, SB_LINEUP, 0, False);
        Update;
        CalcDI(DrawInfo);    // this changes contents of Axis var
        CellHit := Axis.FirstGridCell;
      end
      else if (Mouse+3 >= Axis.FullVisBounds) then
      begin
        ModifyScrollBar(Scrollbar, SB_LINEDOWN, 0, False);
        Update;
        CalcDI(DrawInfo);    // this changes contents of Axis var
        CellHit := Axis.LastFullVisibleCell;
      end
      else if CellHit < 0 then
         CellHit := FMovePos;
  end;
begin
  {$ifdef DFS_COMPILER_4_UP}
  if UseRightToLeftAlignment and (ScrollBar = SB_HORZ) then
    Mouse := ClientWidth - Mouse;
  {$endif}

    //changed to drag selection
    if ((FGridState = gsColMoving))
    and ((MousePt.X<>FDragCoord.X)or(MousePt.Y<>FDragCoord.Y)) then  // CEDR 04.10.02 - dragimage didnt moved when mouse moved vertically
    begin
      if  CheckColumnDrag(FMoveIndex, CellHit, MousePt) then
        FMovePos := CellHit;//mousecoord

      aCoord := MouseCoord(MousePt.X,MousePt.Y);
      if (Dragimg=nil) then // not captured
      begin
         FDragCoord := MousePt;
         PrepareDragImage(aCoord.X, aCoord.Y, true, true);
         SetCapture(Handle);
      end;

      //restore previous display
      RestoreInserMark;
      DrawDragImg(DragRect.Left, DragRect.Top, DragImg.Width, DragImg.Height , DragImgMask.Canvas.Handle, false);

      DoScroll;


      OffsetRect(DragRect, -(FDragCoord.x - MousePt.X), -(FDragCoord.Y - MousePt.Y));
      //Prepare image for restore
      DragImgMask.Canvas.CopyRect(rect(0,0,DragImgMask.Width,DragImgMask.Height), Canvas, DragRect);

      FDragCoord :=MousePt;
      //draw dragged column
      DrawInserMark;
      DrawDragImg(DragRect.Left, DragRect.Top, DragImg.Width, DragImg.Height, DragImg.Canvas.Handle, true);

    end
    else if ((FGridState = gsRowMoving) and CheckRowDrag(FMoveIndex, CellHit, MousePt)) then
    begin
      if (CellHit < Selection.Top) or (CellHit > Selection.Bottom) then
        SetDropTargetRow(CellHit);
      DoScroll;  
    end;
end;

procedure TCustomExGrid.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  CellHit: TGridCoord;
  DrawInfo: TExGridDI;
  Fedited : Boolean;
begin
 if not FNotProcessMouse then
 begin
  //MoveDrawn := False;
  Windows.SetFocus(Handle);//getfocus
  Fedited:=EditorMode;
  //HideEdit;
  Editormode :=false;
  if not (csDesigning in ComponentState) and
    (CanFocus {or (GetParentForm(Self) = nil)}) then
  begin
    //Windows.SetFocus(Handle);//getfocus
    if not IsActiveControl then
    begin
      MouseCapture := False;
      Exit;
    end;
  end;
  if (Button = mbLeft) and (ssDouble in Shift) then
    DblClick
  else

  if (Button = mbLeft) then
  begin
    CalcDI(DrawInfo);
    { Check grid sizing }
     CalcSizingState(X, Y, FGridState, FSizingIndex, FSizingPos, FSizingOfs,
       DrawInfo);
    if FGridState <> gsNormal then
    begin
      {$ifdef DFS_COMPILER_4_UP}
      if UseRightToLeftAlignment then
        FSizingPos := ClientWidth - FSizingPos;
      {$endif}
      DrawSizingLine(DrawInfo);
      Exit;
    end;
    CellHit := CalcCoordFromPoint(X, Y, DrawInfo);
    //check for place of cursor
    if (CellHit.X >= FixedCols) and (CellHit.Y >= FixedRows) then
    begin
       if (goRowMoving in Options) then
       begin
          if not Fedited then FGridState := gsSelecting;
          //Update;
          //SetTimer(Handle, 1, 60, nil);
       end;

       FocusCell(CellHit.X, CellHit.Y, not ((ssShift in Shift) or (ssCtrl in Shift) or InSelection(CellHit.X, CellHit.Y)),Shift, swByMouse );
       FDragCoord :=Point(X,Y);//register mousedown

    end
    //10/3/2
    else if (goRowMoving in Options) and (CellHit.X >= 0) and
      (CellHit.X < FixedCols) and (CellHit.Y >= FixedRows) then
    begin
      FMoveIndex := CellHit.Y;
      FMovePos := FMoveIndex;
      FDragCoord :=Point(X,Y);
      if BeginRowDrag(FMoveIndex, FMovePos, Point(X,Y)) then
      begin
        FGridState := gsRowMoving;
        Update;
        SetTimer(Handle, 1, 60, nil);
      end;
    end
    else if (goColMoving in Options) and (CellHit.Y >= 0) and
      (CellHit.Y < FixedRows) and (CellHit.X >= FixedCols) then
    begin
      FMoveIndex := CellHit.X;
      FMovePos := FMoveIndex;
      if BeginColumnDrag(FMoveIndex, FMovePos, Point(X,Y)) then
      begin//
        FGridState := gsColMoving;
        if (Dragimg=nil) then // not captured
        begin
         FDragCoord := Point(X,Y);
         PrepareDragImage(Cellhit.X, Cellhit.Y, true, true);
         SetCapture(Handle);
        end;
        Update;
        SetTimer(Handle, 1, 60, nil);
        exit;
      end;
    end;

   end;

 end;
 inherited MouseDown(Button, Shift, X, Y);
end;



procedure TCustomExGrid.WMMouseMove(var Message: TWMMouseMove);
//procedure MouseMove(Shift: TShiftState; X, Y: Integer);
var
  DrawInfo: TExGridDI;
  CellHit: TGridCoord;
  X,Y, MinDrag : integer;
begin
  X := Message.XPos;
  Y := Message.YPos;
  CalcDI(DrawInfo);
  case FGridState of
    gsSelecting:
    begin
       CellHit := CalcCoordFromPoint(X, Y, DrawInfo);
       MinDrag := Mouse.DragThreshold;
       if ((goRowMoving in Options) and not (EditorMode)) and
       ((Abs(FDragCoord.X - X) > MinDrag) or (Abs(FDragCoord.Y - Y) > MinDrag)) then
       begin
         FMoveIndex := Selection.Top;//CellHit.Y;//changed for multiselect drag
         FMovePos := FMoveIndex;
         FDragCoord :=Point(X,Y);
         if BeginRowDrag(FMoveIndex, FMovePos, Point(X,Y)) then
         begin
           FGridState := gsRowMoving;
           SetTimer(Handle, 1, 60, nil); //10/3/2
         end;
       end
    end;
    gsColMoving, gsRowMoving: //10/3/2
      begin
        CellHit := CalcCoordFromPoint(X, Y, DrawInfo);
        if //(CellHit.X >= FixedCols) and (CellHit.Y >= FixedRows) and
          (CellHit.X <= DrawInfo.Horz.LastFullVisibleCell+1) and
          (CellHit.Y <= DrawInfo.Vert.LastFullVisibleCell+1) then
          case FGridState of
            gsColMoving:
             //if (CellHit.X >= FixedCols) then
             begin
               MoveAndScroll(X, CellHit.X, DrawInfo, DrawInfo.Horz, SB_HORZ, Point(X,Y));
               CellHit := CalcCoordFromPoint(X, Y, DrawInfo);
             end;
            gsRowMoving:
            begin
             //if (CellHit.Y >= FixedRows) then
              MoveAndScroll(Y, CellHit.Y, DrawInfo, DrawInfo.Vert, SB_VERT, Point(X,Y));
            end;
          end;
      end;

    gsRowSizing, gsColSizing:
      begin
        DrawSizingLine(DrawInfo); // XOR it out
        if FGridState = gsRowSizing then
          FSizingPos := Y + FSizingOfs else
          FSizingPos := X + FSizingOfs;
        DrawSizingLine(DrawInfo); // XOR it back in
      end;
  end;
  inherited;// MouseMove(Shift, X, Y);
end;

procedure TCustomExGrid.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  DrawInfo: TExGridDI;
  NewSize: Integer;
  CellHit: TGridCoord;
  aNewSel : TGridRect;
  i, SelCount : integer;

  function FixedResizeLine(const AxisInfo: TExGridAxisDI): Integer;
  var
    I: Integer;
  begin
    with AxisInfo do
    begin
      Result := 0;
      for I := 0 to FSizingIndex - 1 do
        Inc(Result, GetExtent(I) + OptimalLineWidth);
      Result := FSizingPos - Result;
    end;
  end;

  function ResizeLine(const AxisInfo: TExGridAxisDI): Integer;
  var
    I: Integer;
  begin
    with AxisInfo do
    begin
      Result := FixedBounds;
      //if column on tree
       for I := FirstGridCell to FSizingIndex - 1 do
         Inc(Result, GetExtent(I) + OptimalLineWidth);
       Result := FSizingPos - Result;
    end;
  end;

begin
  try
    case FGridState of
      gsSelecting:
        begin
          MouseMove(Shift, X, Y);
          KillTimer(Handle, 1);
          {if (Shift * [ssShift, ssCtrl] = []) and (Button=mbLeft) then
          begin
            CalcDI(DrawInfo);
            CellHit := CalcCoordFromPoint(X, Y, DrawInfo);//
            FocusCell(CellHit.X, CellHit.Y, True, [], swByMouse );
          end;
          }
          UpdateEdit;
        end;
      gsRowSizing, gsColSizing:
        begin
          CalcDI(DrawInfo);
          DrawSizingLine(DrawInfo);
          {$ifdef DFS_COMPILER_4_UP}
          if UseRightToLeftAlignment then
            FSizingPos := ClientWidth - FSizingPos;
          {$endif}
          if FGridState = gsColSizing then
          begin
            //if non fixed
            if FSizingIndex >= DrawInfo.Horz.FixedCellCount then
               NewSize := ResizeLine(DrawInfo.Horz)
            else
               NewSize := FixedResizeLine(DrawInfo.Horz);

            if NewSize > 1 then
            begin
              ColWidths[FSizingIndex] := NewSize;
              UpdateDesigner;
            end;
          end
          else
          begin
            NewSize := ResizeLine(DrawInfo.Vert);
            if NewSize > 1 then
            begin
              RowHeights[FSizingIndex] := NewSize;
              UpdateDesigner;
            end;
          end;
        end;
      gsColMoving:
        begin
          //if FMoveDrawed then
             //DrawMove;
          //   HideInsertMark;
          KillTimer(Handle, 1);
          if DragImg <> nil then
            DrawDragImg(DragRect.Left, DragRect.Top, DragImg.Width, DragImg.Height , DragImgMask.Canvas.Handle, false);
          aNewSel.Top := 0;
          aNewSel.Bottom := FixedRows-1;
          aNewSel.Left := LeftCol;
          aNewSel.Right := LeftCol+VisibleColcount;
          InvalidateRect(aNewSel);
          if EndColumnDrag(FMoveIndex, FMovePos, Point(X,Y))
            and (FMoveIndex <> FMovePos) and (FmoveIndex<>-1)then
          begin
            MoveColumn(FMoveIndex, FMovePos);
            UpdateDesigner;
          end;
          FMovePos:=-1;
          FMoveIndex:=-1;
          If DragImg<> nil then
             DragImg.Free;
          DragImg := nil;
          If DragImg<> nil then
            DragImgMask.Free;
          DragImgMask := nil;
          if DragRArrow <> nil then
            DragRArrow.Free;
          if DragLArrow <> nil then
            DragLArrow.Free;
          if RestoreArrow <> nil then
            RestoreArrow.Free;
          DragLArrow := nil;
          DragRArrow := nil;
          RestoreArrow := nil;
          RestRect := Rect(0,0,0,0);

          ReleaseCapture;      //update
          UpdateEdit;
        end;
      gsRowMoving:
        if (FMovePos<>-1) then
        begin
          //if FmoveDrawed then
          //  HideInsertMark;
            //DrawMove;
          KillTimer(Handle, 1);
          CalcDI(DrawInfo);
          CellHit := CalcCoordFromPoint(X, Y, DrawInfo);//
          if (FmoveIndex = FMovePos) then
          begin
            FmovePos :=-1;
            FGridState := gsSelecting;
            //new 20/12/2000
            SetTimer(Handle, 1, 60, nil);

            if (goEditing in Options) and NeedToShowEditor then
            begin
            if (CellHit.X = FCurrent.X) and (CellHit.Y = FCurrent.Y) then//
                //place to drag moment - if mouseup - to showeditor/if mouse down - to draggable func
                ShowEditor
             else //
               UpdateEdit;
             end;
            end
            else if EndRowDrag(FMoveIndex, FMovePos, Point(X,Y))
                 and (FMoveIndex <> FMovePos) then
            begin
             SelCount := Selection.Bottom-Selection.Top;
             if FMoveIndex < FMovePos then
             begin
              anewSel.Top:=  FMovePos-SelCount;
              anewSel.Bottom:=anewSel.Top+SelCount;
              aNewSel.Left:=Selection.Left;
              anewSel.Right :=Selection.Right;

              for i:=0 to SelCount do
                 MoveRow(FMoveIndex, FMovePos); //->
              //FMoveDrawed :=false;
              FMoveIndex:=-1;
            end
            else
            begin
               for i:=0 to SelCount do
                 MoveRow(FMoveIndex+i, FMovePos+i);

               anewSel.Top:=FMovePos;
               anewSel.Bottom:=FMovePos+SelCount;
               aNewSel.Left:=Selection.Left;
               anewSel.Right :=Selection.Right;
               //FMoveDrawed :=false;
               FMoveIndex:=-1;
            end;
            //initiate new sel
            SetSelection(anewSel);
            UpdateDesigner;
          end;
          UpdateEdit;
        end;

    else
     begin
      CalcDI(DrawInfo);
      CellHit := CalcCoordFromPoint(X, Y, DrawInfo);//
      if (goEditing in Options) and NeedToShowEditor then//
        if (CellHit.X = FCurrent.X) and (CellHit.Y = FCurrent.Y)
        and MouseHit(X,Y)then//
        //place to drag moment - if mouseup - to showeditor/if mouse down - to draggable func
          ShowEditor//
        else //
          UpdateEdit;
     end
    end;
    inherited MouseUp(Button, Shift, X, Y);
  finally
    FGridState := gsNormal;
    //HideInsertMark;
  end;
end;


function TCustomExGrid.GetDraggedRow : Longint;
begin
    if FGridState = gsRowMoving then
       Result:=FMoveIndex
    else
       Result:= -1;

end;

procedure TCustomExGrid.SetDropTargetRow(aRow : Longint);
begin
   if (FMovePos<>aRow)  then
   begin
    if  (FMovePos<>-1) and (FGridState = gsrowMoving) then
      invalidateRow(FmovePos);
    FmovePos := aRow;
    if (FGridState = gsrowMoving) then
     invalidateRow(FmovePos);
   end;
end;

function TCustomExGrid.GetDropTargetRow : Longint;
begin
//    if FGridState = gsRowMoving then
       Result:=FMovePos
//    else
//       Result:= -1;
end;


procedure TCustomExGrid.Click;
begin
  inherited;
end;


function TCustomExGrid.MouseHit(X,Y : integer) : boolean;
begin
   Result:=true;
end;


Procedure TCustomExGrid.PrepareDragImage(aCol,aRow : integer; DrI, DrIM : boolean);
var CRect : TRect;
begin
 DragRect := CellRect(aCol, aRow);
 CRect := Rect(0, 0, DragRect.Right - DragRect.Left, DragRect.Bottom - DragRect.Top);
 if DrI then
 begin
  if DragImg <> nil then DragImg.Free;
  DragImg := TBitmap.Create;
  DragImg.Height := CRect.Bottom;// + 1;
  DragImg.Width := CRect.Right;// + 1;
  //drawto
  DragImg.Canvas.CopyRect(CRect, Canvas, DragRect);
  DragImg.Canvas.Brush.Color := clBlack;
  DragImg.Canvas.FrameRect(CRect);
  DragImg.IgnorePalette:= true;
 end;

 if DrIM then
 begin
  if DragImgMask <> nil then DragImgMask.Free;
  DragImgMask := TBitmap.Create;
  DragImgMask.Height := CRect.Bottom;// + 1;
  DragImgMask.Width := CRect.Right;// + 1;
  DragImgMask.Canvas.CopyRect(CRect, Canvas, DragRect);
 end;

end;

Procedure TCustomExGrid.RestoreInserMark;
begin
 if (RestoreArrow <> nil) and (FmoveIndex <> FMovePos) then
  if not RestoreArrow.empty then
    Canvas.Draw(RestRect.Left, RestRect.Top, RestoreArrow);
end;

Procedure TCustomExGrid.DrawInserMark;
var Crect : TRect;
    aX, aY : integer;
begin
if FMoveIndex <> FMovePos then
begin
  if DragLArrow = nil then
  begin
    DragLArrow := Tbitmap.Create;
    with DragLArrow do
    begin
      Handle := LoadBitmap(Hinstance, 'DRAGLEFTARROW');
      Transparent := True;
      TransParentColor := Canvas.pixels[0,0];
    end;

  end;
  if DragRArrow = nil then
  begin
    DragRArrow := Tbitmap.Create;
    with DragRArrow do
    begin
      Handle := LoadBitmap(Hinstance, 'DRAGRIGHTARROW');
      Transparent := True;
      TransParentColor := Canvas.pixels[0,0];
    end;
  end;

  if RestoreArrow = nil then
  begin
    RestoreArrow := Tbitmap.Create;
    with RestoreArrow do
    begin
      Width := DragRArrow.Width;
      Height := DragRArrow.Height;
    end;
  end;

  Crect := CellRect(FMovePos, 0);
  if FMoveIndex < FMovePos then
  begin
      aX := CRect.Right-DragLArrow.Width- GridLinewidth;
      aY := ((Crect.Bottom - Crect.Top) - DragLArrow.Height) div 2;
      RestRect := Rect(aX, aY, ax + DragLArrow.Width, aY + DragLArrow.Height);
      //resote bitmap
      RestoreArrow.Canvas.CopyRect(rect(0,0, RestoreArrow.Width, RestoreArrow.Height), Canvas, RestRect);
      Canvas.Draw(aX, aY, DragLArrow);
  end
  else //if FMoveIndex > FMovePos then
  begin
      aX := CRect.Left;
      aY := ((Crect.Bottom - Crect.Top) - DragRArrow.Height) div 2;
      RestRect := Rect(aX, aY, aX + DragLArrow.Width, aY + DragRArrow.Height);
      RestoreArrow.Canvas.CopyRect(rect(0,0, RestoreArrow.Width, RestoreArrow.Height), Canvas, RestRect);
      Canvas.Draw(aX, aY, DragRArrow);
  end;
end;
end;

Procedure TCustomExGrid.DrawDragImg(aLeft, aTop, aWidth, aHeight:integer; aImageCHandle : {integer} HDC; UseBF : Boolean);  // CEDR 20.09.02 - range
 var SDC : HDC;
     Pt : TPoint;
 begin
  SDC := GetDC(0);
  try
   Pt := ClientToScreen(Point(0,0));
   inc(aLeft, Pt.X);
   inc(aTop, Pt.Y);
   if UseBF and (@Alphablend <> nil) then
   begin
    //Alpha blending
     AlphaBlend(SDC{Canvas.Handle}, aLeft, aTop, aWidth, aHeight, aImageCHandle, 0, 0, aWidth, aHeight, BF);
   end
   else
     BitBlt(SDC{Canvas.Handle}, aLeft, aTop, aWidth, aHeight, aImageCHandle, 0, 0, SRCCopy);

 finally
    ReleaseDC(0,SDC);
 end;
end;


function TCustomExGrid.AllowDesignSizing : boolean;
begin
   Result := true;
end;

function TCustomExGrid.GetColWidths(Index: Longint): Integer;
begin
  if (FColWidths = nil) or (Index >= ColCount) then
    Result := DefaultColWidth
  else
    Result := PIntArray(FColWidths)^[Index + 1];
end;

function TCustomExGrid.GetRowHeights(Index: Longint): Integer;
begin
  if (FRowHeights = nil) or (Index >= RowCount) then
      Result := DefaultRowHeight
  else
    Result := PIntArray(FRowHeights)^[Index + 1];
end;

function TCustomExGrid.GetGridWidth: Integer;
var
  DrawInfo: TExGridDI;
begin
  CalcDI(DrawInfo);
  Result := DrawInfo.Horz.GridBounds;
end;

function TCustomExGrid.GetGridHeight: Integer;
var
  DrawInfo: TExGridDI;
begin
  CalcDI(DrawInfo);
  Result := DrawInfo.Vert.GridBounds;
end;

function TCustomExGrid.GetSelection: TGridRect;
begin
//  Result := _GridRect(FCurrent, FAnchor);
  if not FSelectionCleared then
    Result := _GridRect(FendAnchor, FAnchor)
  else Result := GridRect(-1, -1,-1,-1);  
end;

procedure TCustomExGrid.ClearSelection;
begin
   FSelectionCleared := True;
end;

function TCustomExGrid.GetTabStops(Index: Longint): Boolean;
begin
  if FTabStops = nil then Result := True
  else Result := Boolean(PIntArray(FTabStops)^[Index + 1]);
end;

function TCustomExGrid.GetVisibleColCount: Integer;
var
  DrawInfo: TExGridDI;
begin
  CalcDI(DrawInfo);
  Result := DrawInfo.Horz.LastFullVisibleCell - LeftCol + 1;
end;

function TCustomExGrid.GetVisibleRowCount: Integer;
var
  DrawInfo: TExGridDI;
begin
  CalcDI(DrawInfo);
  Result := DrawInfo.Vert.LastFullVisibleCell - TopRow + 1;
end;

procedure TCustomExGrid.SetBorderStyle(Value: TBorderStyle);
begin
  if FBorderStyle <> Value then
  begin
    FBorderStyle := Value;
    RecreateWnd;//createWnd
  end;
end;

procedure TCustomExGrid.SetCol(Value: Longint);
begin
  if Col <> Value then
    FocusCell(Value, Row, True,[], swNone);
end;

procedure TCustomExGrid.SetColCount(Value: Longint);
begin
  if FColCount <> Value then
  begin
    if Value < 1 then Value := 1;
    if not (csLoading in ComponentState) then
      if Value <= FixedCols then
         FixedCols := Value - 1;
    ChangeSize(Value, RowCount);
    if goRowSelect in Options then
    begin
      FAnchor.X := ColCount-1;
      Invalidate;
    end;
  end;
  ColCountChanged;
end;

procedure TCustomExGrid.ColCountChanged;
begin
end;

procedure TCustomExGrid.ColumnWidthChanged(Index, Value : integer);
begin
   If Assigned(FOnColumnWidthchanged) then
      FOnColumnWidthchanged(Index, Value);
end;

procedure TCustomExGrid.ColumnWidthChanging(Index : integer; var Value: integer);
begin
   If Assigned(FOnColumnWidthchanging) then
      FOnColumnWidthchanging(Index, Value);
end;

procedure TCustomExGrid.SetColWidths(Index: Longint; Value: Integer);
begin
  if FColWidths = nil then
    UpdateExtents(FColWidths, ColCount, DefaultColWidth);
  if Index >= ColCount then InvalidOp(SIndexOutOfRange);
  if Value <> PIntArray(FColWidths)^[Index + 1] then
  begin
    ColumnWidthchanging(index, Value);
    ResizeCol(Index, PIntArray(FColWidths)^[Index + 1], Value);
    PIntArray(FColWidths)^[Index + 1] := Value;
    ColWidthsChanged;
    ColumnWidthChanged(Index, Value);
  end;
end;

procedure TCustomExGrid.SetDefaultColWidth(Value: Integer);
begin
  if FColWidths <> nil then UpdateExtents(FColWidths, 0, 0);
  FDefaultColWidth := Value;
  ColWidthsChanged;
  InvalidateGrid;
end;

procedure TCustomExGrid.SetDefaultRowHeight(Value: Integer);
begin
  if FRowHeights <> nil then UpdateExtents(FRowHeights, 0, 0);
  FDefaultRowHeight := Value;
  RowHeightsChanged;
  InvalidateGrid;
end;

procedure TCustomExGrid.SetFixedColor(Value: TColor);
begin
  if FFixedColor <> Value then
  begin
    FFixedColor := Value;
    InvalidateGrid;
  end;
end;

procedure TCustomExGrid.SetFixedCols(Value: Integer);
begin
  if FFixedCols <> Value then
  begin
   //if not (csLoading in Componentstate) then
   //begin
    if Value < 0 then InvalidOp(SIndexOutOfRange);
    if Value >= ColCount then
    begin
       ColCount := Value+1;
    end;
    //InvalidOp(SFixedColTooBig);
   //end;
    FFixedCols := Value;
   //if not (csLoading in Componentstate) then
   //begin
    Initialize;
    InvalidateGrid;
   //end;
  end;
  //Showmessage( 'SetfixedCols:'+Name);
end;

procedure TCustomExGrid.SetFixedRows(Value: Integer);
begin
  if FFixedRows <> Value then
  begin
    if Value < 0 then InvalidOp(SIndexOutOfRange);
    if Value >= RowCount then InvalidOp(SFixedRowTooBig);
    FFixedRows := Value;
    Initialize;
    InvalidateGrid;
  end;
  //Showmessage( 'SetFixedRows:'+Name);
end;

procedure TCustomExGrid.SetEditorMode(Value: Boolean);
begin
  if (not Value) and (FEditorMode) then
  begin
    FEditorMode := False;
    HideEditor
  end
  else if (Value) and (not FEditorMode) then
  begin
    ShowEditor;
    //if FInplaceEdit <> nil then FInplaceEdit.Deselect;
  end;
  //FgridState := gsNormal;
end;

procedure TCustomExGrid.SetGridLineWidth(Value: Integer);
begin
  if FGridLineWidth <> Value then
  begin
    FGridLineWidth := Value;
    InvalidateGrid;
  end;
end;

procedure TCustomExGrid.SetLeftCol(Value: Longint);
begin
  if FTopLeft.X <> Value then MoveTopLeft(Value, TopRow);
end;

procedure TCustomExGrid.SetOptions(Value: TGridOptions);
begin
  if FOptions <> Value then
  begin
    if goRowSelect in Value then
    begin
      Exclude(Value, goAlwaysShowEditor);
      Exclude(Value, goEditing);
    end;
    FOptions := Value;
    if goRowSelect in Value then
       MoveCurrent(Col, Row,  True, False, [], swNone);
    InvalidateGrid;
  end;
end;

procedure TCustomExGrid.InternalSetRow(Value: Longint);//4/26/2
begin
  FCurrent.Y := value;
end;

procedure TCustomExGrid.SetRow(Value: Longint);
begin
  if (Row <> Value) and (Value <>-1) then
     FocusCell(Col, Value, True, [], swNone);
end;

procedure TCustomExGrid.SetRowCount(Value: Longint);
begin
  if FRowCount <> Value then
  begin
    //if Value < 1 then Value := 1;
    //if Value <= FixedRows then FixedRows := Value - 1;
    if Value <= FixedRows then
       Value := FixedRows;
    if TopRow<FixedRows then TopRow :=FixedRows;   
    ChangeSize(ColCount, Value);
  end;
end;

procedure TCustomExGrid.RowHeightChanging(Index : integer; var Value: integer);
begin
   if assigned(FOnRowHeightChanging) then
      FOnRowHeightChanging(Index, Value);
end;

procedure TCustomExGrid.RowHeightChanged(Index, Value : integer); 
begin
   if assigned(FOnRowHeightChanged) then
      FOnRowHeightChanged(Index, Value);
end;

procedure TCustomExGrid.UpdateRowHeight(Index: Longint; Value: Integer);
begin
  if FRowHeights = nil then
    UpdateExtents(FRowHeights, RowCount, DefaultRowHeight);
  if (FRowHeights=nil) or (Index >= RowCount) then exit;//InvalidOp(SIndexOutOfRange);
   PIntArray(FRowHeights)^[Index + 1] := Value+2;
end;

procedure TCustomExGrid.SetRowHeights(Index: Longint; Value: Integer);
begin
 if FSaveCellExtents then
 begin
   if (FRowHeights = nil) then
    UpdateExtents(FRowHeights, RowCount, DefaultRowHeight);

   if Index >= RowCount then exit;//InvalidOp(SIndexOutOfRange);

   if Value <> PIntArray(FRowHeights)^[Index + 1] then
   begin
    RowHeightChanging(Index, Value);
    ResizeRow(Index, PIntArray(FRowHeights)^[Index + 1], Value);
    PIntArray(FRowHeights)^[Index + 1] := Value;
    RowHeightsChanged;
    RowHeightChanged(Index, Value);
   end;
 end;
end;

{$ifdef DFS_COMPILER_4_UP}
procedure TCustomExGrid.SetScrollBarsStyle(value : TScrollBarStyle);
begin
  if (FScrollBarsStyle <> value) then
  begin
     //scrollbars/?
      FScrollBarsStyle := value;
     if fFlatSBPresent then
     begin
      FlatSB_SetScrollProp(Handle, WSB_PROP_HSTYLE, Ord(FScrollBarsStyle), true);
      FlatSB_SetScrollProp(Handle, WSB_PROP_VSTYLE, Ord(FScrollBarsStyle), true);
     end;
  end;
end;

procedure TCustomExGrid.SetScrollbarsColor(value : TColor);
begin
  if (FScrollBarsColor <> value) then
  begin
    FScrollBarsColor := value;
    if fFlatSBPresent then
    begin
      FlatSB_SetScrollProp(Handle, WSB_PROP_VBKGCOLOR, ColorToRGB(FScrollBarsColor), true);
      FlatSB_SetScrollProp(Handle, WSB_PROP_HBKGCOLOR, ColorToRGB(FScrollBarsColor), true);
    end;
  end;
end;
{$endif}

procedure TCustomExGrid.SetScrollBars(Value: TScrollStyle);
begin
  if FScrollBars <> Value then
  begin
    FScrollBars := Value;
    RecreateWnd;
  end;
end;

procedure TCustomExGrid.SetSelection(Value: TGridRect);
var
  OldSel: TGridRect;
begin
  OldSel := Selection;
  FAnchor := Value.TopLeft;
  FCurrent := Value.BottomRight;
  SelectionMoved(OldSel);
end;

procedure TCustomExGrid.SetTabStops(Index: Longint; Value: Boolean);
begin
  if FTabStops = nil then
    UpdateExtents(FTabStops, ColCount, Integer(True));
  if Index >= ColCount then InvalidOp(SIndexOutOfRange);
  PIntArray(FTabStops)^[Index + 1] := Integer(Value);
end;

procedure TCustomExGrid.SetTopRow(Value: Longint);
begin
  if FTopLeft.Y <> Value then MoveTopLeft(LeftCol, Value);
end;

procedure TCustomExGrid.HideEdit;
begin
  FEditorMode := False;
  if FInplaceEdit <> nil then
    try
      UpdateText;
    finally
      FInplaceCol := -1;
      FInplaceRow := -1;
      FInplaceEdit.Hide;
    end;
end;

procedure TCustomExGrid.UpdateEdit;

  procedure UpdateEditor;
  begin
    FInplaceCol := Col;
    FInplaceRow := Row;
    FInplaceEdit.UpdateContents;
    if FInplaceEdit.MaxLength = -1 then FCanEditModify := False
    else FCanEditModify := True;
    FInplaceEdit.SelectAll;
  end;

begin
  if CanEditShow then
  begin
    if FInplaceEdit = nil then
    begin
      FInplaceEdit := CreateEditor;
      FInplaceEdit.SetGrid(Self);
      FInplaceEdit.Parent := Self;
      UpdateEditor;
    end
    else
    begin
      if (Col <> FInplaceCol) or (Row <> FInplaceRow) then
      begin
        HideEdit;
        UpdateEditor;
      end;
    end;
    if CanEditShow then MoveEditor(Col, Row, CellRect(Col, Row), true);
  end;
end;


procedure TCustomExGrid.MoveEditor(aCol, aRow : integer; aRect : TRect; Redraw : Boolean);
begin
 if FInplaceEdit<> nil then
 begin
   if Redraw then
     FInplaceEdit.Move(aRect)
   else
     FInplaceEdit.UpdateLoc(aRect);
 end;
end;

procedure TCustomExGrid.UpdateText;
begin
  if (FInplaceCol <> -1) and (FInplaceRow <> -1) then
    SetEditText(FInplaceCol, FInplaceRow, FInplaceEdit.Text);
end;

procedure TCustomExGrid.WMChar(var Msg: TWMChar);
begin
  if (goEditing in Options) and (Char(Msg.CharCode) in [^H, #32..#255]) then
    ShowEditorChar(Char(Msg.CharCode));
//  else //new 9/3/2
    inherited;
end;

procedure TCustomExGrid.WMCommand(var Message: TWMCommand);
begin
  with Message do
  begin
    if (FInplaceEdit <> nil) and (Ctl = FInplaceEdit.Handle) then
      case NotifyCode of
        EN_CHANGE: UpdateText;
      end;
  end;
end;

procedure TCustomExGrid.WMGetDlgCode(var Msg: TWMGetDlgCode);
begin
  Msg.Result := DLGC_WANTARROWS;
  if goRowSelect in Options then Exit;
  if goTabs in Options then Msg.Result := Msg.Result or DLGC_WANTTAB;
  if goEditing in Options then Msg.Result := Msg.Result or DLGC_WANTCHARS;
end;

procedure TCustomExGrid.WMKillFocus(var Msg: TWMKillFocus);
begin
  inherited;
  InvalidateRect(Selection);
  if (FInplaceEdit <> nil) and (Msg.FocusedWnd <> FInplaceEdit.Handle) then
    HideEdit;
end;

procedure TCustomExGrid.WMLButtonDown(var Message: TMessage);
begin
  inherited;
  if FInplaceEdit <> nil then
     FInplaceEdit.FClickTime := GetMessageTime;
end;

procedure TCustomExGrid.WMNCHitTest(var Msg: TWMNCHitTest);
begin
  DefaultHandler(Msg);
  FHitTest := ScreenToClient(SmallPointToPoint(Msg.Pos));
end;

procedure TCustomExGrid.WMSetCursor(var Msg: TWMSetCursor);
var
  DrawInfo: TExGridDI;
  State: TGridState;
  Index: Longint;
  Pos, Ofs: Integer;
  Cur: HCURSOR;
begin

  Cur :=0;
  with Msg do
  begin
    if (HitTest = HTCLIENT) then
    begin
      if FGridState = gsNormal then
      begin
        CalcDI(DrawInfo);
        CalcSizingState(FHitTest.X, FHitTest.Y, State, Index, Pos, Ofs,
          DrawInfo);
      end
      else
       State := FGridState;


      if State = gsRowSizing then
        Cur := VertCursor//Screen.Cursors[crSizeNS]
      else if State = gsColSizing then
      begin
        Cur := HorzCursor;//Screen.Cursors[crSizeWE]        //getcursor
      end
      else Cur:=0;
    end;
  end;
  if Cur <> 0 then SetCursor(Cur) //setsystemcursor
  else
    inherited;
end;

procedure TCustomExGrid.WMSetFocus(var Msg: TWMSetFocus);
begin
  inherited;
  if (FInplaceEdit = nil) or (Msg.FocusedWnd <> FInplaceEdit.Handle) then
  begin
    InvalidateRect(Selection);
    UpdateEdit;
  end;

end;

procedure TCustomExGrid.WMSize(var Msg: TWMSize);
begin
      inherited;
      ChangeSize(ColCount, RowCount);
//      Updatewidths;
//      UpdateScrollRange;
      {$ifdef DFS_COMPILER_4_UP}
//      if UseRightToLeftAlignment then Invalidate;
      {$endif}

end;

procedure TCustomExGrid.WMVScroll(var Msg: TWMVScroll);
begin
  ModifyScrollBar(SB_VERT, Msg.ScrollCode, Msg.Pos, True);
end;

procedure TCustomExGrid.WMHScroll(var Msg: TWMHScroll);
begin
  ModifyScrollBar(SB_HORZ, Msg.ScrollCode, Msg.Pos, True);
end;

procedure TCustomExGrid.CancelMode;
var
  DrawInfo: TExGridDI;
begin
  try
    case FGridState of
      gsSelecting:
        KillTimer(Handle, 1);
      gsRowSizing, gsColSizing:
        begin
          CalcDI(DrawInfo);
          DrawSizingLine(DrawInfo);
        end;
      gsColMoving, gsRowMoving:
        begin
          //DrawMove;
          //HideInsertMark;
          KillTimer(Handle, 1);
        end;
    end;
  finally
    FGridState := gsNormal;
  end;
end;

procedure TCustomExGrid.WMCancelMode(var Msg: TWMCancelMode);
begin
 {$ifndef ACTIVEX}
  inherited;
 {$endif} 
  CancelMode;
end;

procedure TCustomExGrid.CMCancelMode(var Msg: TMessage);
begin
  if Assigned(FInplaceEdit) and FEditorMode then
     FInplaceEdit.WndProc(Msg);
  inherited;
  CancelMode;
end;

procedure TCustomExGrid.CMFontChanged(var Message: TMessage);
begin
  if FInplaceEdit <> nil then FInplaceEdit.Font := Font;
  inherited;
end;

procedure TCustomExGrid.CMCtl3DChanged(var Message: TMessage);
begin
  inherited;
  RecreateWnd;
end;

procedure TCustomExGrid.CMDesignHitTest(var Msg: TCMDesignHitTest);
begin
  Msg.Result := Longint(BOOL(Sizing(Msg.Pos.X, Msg.Pos.Y)));
end;

procedure TCustomExGrid.CMWantSpecialKey(var Msg: TCMWantSpecialKey);
begin
  inherited;
  if (goEditing in Options) and (Char(Msg.CharCode) = #13) then Msg.Result := 1;
end;

procedure TCustomExGrid.TimedScroll(Direction: TGridScrollDirection);
var
  MaxAnchor, NewAnchor: TGridCoord;
begin
  NewAnchor := FAnchor;
  MaxAnchor.X := ColCount - 1;
  MaxAnchor.Y := RowCount - 1;
  if (sdLeft in Direction) and (FAnchor.X > FixedCols) then Dec(NewAnchor.X);
  if (sdRight in Direction) and (FAnchor.X < MaxAnchor.X) then Inc(NewAnchor.X);
  if (sdUp in Direction) and (FAnchor.Y > FixedRows) then Dec(NewAnchor.Y);
  if (sdDown in Direction) and (FAnchor.Y < MaxAnchor.Y) then Inc(NewAnchor.Y);
  if (FAnchor.X <> NewAnchor.X) or (FAnchor.Y <> NewAnchor.Y) then
    MoveAnchor(NewAnchor, True);
end;

procedure TCustomExGrid.WMTimer(var Msg: TWMTimer);
var
  Point: TPoint;
  DrawInfo: TExGridDI;
  ScrollDirection: TGridScrollDirection;
  CellHit: TGridCoord;
  LeftSide: Integer;
  RightSide: Integer;
begin
 if Msg.TimerID = 1 then
 begin
  if not (FGridState in [gsSelecting, gsRowMoving, gsColMoving]) then
  begin
    KillTimer(Handle, 1);
    Exit;
  end;
  GetCursorPos(Point);
  Point := ScreenToClient(Point);
  CalcDI(DrawInfo);
  ScrollDirection := [];
  with DrawInfo do
  begin
    CellHit := CalcCoordFromPoint(Point.X, Point.Y, DrawInfo);
    case FGridState of
      gsColMoving:
       //if (FDragRowColPt.X <> Point.X)  then
       if (Abs(FDragRowColPt.X - Point.X) > Mouse.DragThreshold) and
       (FDragRowColPt.Y <> CellHit.X) then//or (Abs(FDragRowColPt.Y - Point.Y) > Mouse.DragThreshold) then
       begin
        MoveAndScroll(Point.X, CellHit.X, DrawInfo, Horz, SB_HORZ, Point);
        FDragRowColPt.X := Point.X;
        FDragRowColPt.Y := CellHit.X
        //OutputDebugString(Pchar('X'+ intToStr(FdragRowcolPt.X)+':Y'+ intToStr(FdragRowcolPt.Y)));
       end;
      gsRowMoving:
      begin
        //if (Abs(FDragRowColPt.X - Point.X) > Mouse.DragThreshold) or (Abs(FDragRowColPt.Y - Point.Y) > Mouse.DragThreshold) then
        if (FDragRowColPt.Y <> Point.Y)  then
        begin
          MoveAndScroll(Point.Y, CellHit.Y, DrawInfo, Vert, SB_VERT, Point);
          FDragRowColPt := Point;
        end;
      end;
      gsSelecting:
      begin
        {$ifdef DFS_COMPILER_4_UP}
        if not UseRightToLeftAlignment then
        begin
          if Point.X < Horz.FixedBounds then Include(ScrollDirection, sdLeft)
          else if Point.X > Horz.FullVisBounds then Include(ScrollDirection, sdRight);
        end
        else
        begin
          LeftSide := ClientWidth - Horz.FullVisBounds;
          RightSide := ClientWidth - Horz.FixedBounds;
          if Point.X < LeftSide then Include(ScrollDirection, sdRight)
          else if Point.X > RightSide then Include(ScrollDirection, sdLeft);
        end;
        {$else}
          if Point.X < Horz.FixedBounds then Include(ScrollDirection, sdLeft)
          else if Point.X > Horz.FullVisBounds then Include(ScrollDirection, sdRight);
        {$endif}

        if Point.Y < Vert.FixedBounds then Include(ScrollDirection, sdUp)
        else if Point.Y > Vert.FullVisBounds then Include(ScrollDirection, sdDown);
        if ScrollDirection <> [] then
           TimedScroll(ScrollDirection);
      end;
    end;
  end;
 end;
 inherited;
end;

procedure TCustomExGrid.ColWidthsChanged;
begin
  UpdateScrollRange;
  UpdateEdit;
end;

procedure TCustomExGrid.RowHeightsChanged;
begin
  UpdateScrollRange;
  UpdateEdit;
end;

procedure TCustomExGrid.DeleteColumn(ACol: Longint);
begin
  MoveColumn(ACol, ColCount-1);
  ColCount := ColCount - 1;
end;

procedure TCustomExGrid.DeleteRow(ARow: Longint);
begin
  MoveRow(ARow, RowCount - 1);
  RowCount := RowCount - 1;
end;

procedure TCustomExGrid.SetFixedRowAsButton(Value : boolean);
begin
   if FFixedRowAsButton<> Value then
   begin
      FFixedRowAsButton := Value;
      InvalidateGrid;
   end;

end;

procedure TCustomExGrid.SetFixedColAsButton(Value : boolean);
begin
   if FFixedColAsButton<> Value then
   begin
      FFixedColAsButton := Value;
      InvalidateGrid;
   end;
end;

procedure TCustomExGrid.UpdateDesigner;
var
  ParentForm: {$ifdef DFS_COMPILER_3_UP}TCustomForm{$ELSE}TForm{$ENDIF};
begin
  if (csDesigning in ComponentState) and HandleAllocated and
    not (csUpdating in ComponentState) then
  begin
    ParentForm := GetParentForm(Self);
    if Assigned(ParentForm) and Assigned(ParentForm.Designer) then
      ParentForm.Designer.Modified;
  end;
end;


procedure TCustomExGrid.DblClick;
begin
   inherited;
end;

function TCustomExGrid.GetOptimalKeys : TShiftstate;
begin
  Result:=[];
  if (GetKeyState(VK_SHIFT) < 0) and not FMouseWheel then Include(Result, ssShift);
  if (GetKeyState(VK_CONTROL) < 0) and not FMouseWheel then Include(Result, ssCtrl);

end;

{$ifdef DFS_COMPILER_4_UP}
{function TCustomExGrid.DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
      MousePos: TPoint): Boolean;
begin
   Result := inherited DoMouseWheel(Shift, WheelDelta, MousePos);
end;}

function TCustomExGrid.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean;
begin
  Result := inherited DoMouseWheelDown(Shift, MousePos);
  if not Result then
  begin
    FMouseWheel:=true;
    try
    if (ssShift in Shift) and (Row < RowCount - 1) then //Row := Row + 1;
      FocusCell(Col, Row+1,True, Shift, swByMouse)
    else if (ssCtrl in Shift) then
       ModifyScrollbar(SB_VERT, SB_PAGEDOWN, 0, False)
    else
       ModifyScrollbar(SB_VERT, SB_LINEDOWN, 0, False);
     Result := True;
    finally
      FMouseWheel:=False;
    end;
  end;
end;

function TCustomExGrid.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean;
begin
  Result := inherited DoMouseWheelUp(Shift, MousePos);
  if not Result then
  begin
   FMouseWheel:=true;
    try
     if (ssShift in Shift) and (Row > FixedRows) then //Row := Row - 1;
       FocusCell(Col, Row-1,True, Shift, swByMouse)
     else if (ssCtrl in Shift) then
       ModifyScrollbar(SB_VERT, SB_PAGEUP, 0, False)
     else
       ModifyScrollbar(SB_VERT, SB_LINEUP, 0, False);
     Result := True;
    finally
      FMouseWheel:=false;
    end;
  end;
end;
{$endif}
function TCustomExGrid.CheckColumnDrag(var Origin,
  Destination: Longint; const MousePt: TPoint): Boolean;
begin
  Result := (Destination <> -1) and (Destination < Colcount);
end;

function TCustomExGrid.CheckRowDrag(var Origin,
  Destination: Longint; const MousePt: TPoint): Boolean;
begin
  Result := True;
end;

function TCustomExGrid.BeginColumnDrag(var Origin, Destination: LongInt; const MousePt: TPoint): Boolean;
begin
  Result := True;
end;

function TCustomExGrid.BeginRowDrag(var Origin, Destination: Longint; const MousePt: TPoint): Boolean;
begin
  Result := True;
end;

function TCustomExGrid.EndColumnDrag(var Origin, Destination: Longint; const MousePt: TPoint): Boolean;
begin
  Result := True;
end;

function TCustomExGrid.EndRowDrag(var Origin, Destination: Longint; const MousePt: TPoint): Boolean;
begin
  Result := True;
end;

procedure TCustomExGrid.CMShowingChanged(var Message: TMessage);
begin
  inherited;
  if Showing then UpdateScrollRange;
end;

procedure TCustomExGrid.ColumnMoved(FromIndex, ToIndex: Longint);
begin
  if Assigned(FOnColumnMoved) then FOnColumnMoved(Self, FromIndex, ToIndex);
end;

procedure TCustomExGrid.MouseToCell(X, Y: Integer; var ACol, ARow: Longint);
var
  Coord: TGridCoord;
begin
  Coord := MouseCoord(X, Y);
  ACol := Coord.X;
  ARow := Coord.Y;
end;

initialization
  VertCursor := LoadCursor(0, IDC_SIZENS);
  HorzCursor := LoadCursor(0, IDC_SIZEWE);
//  Win98OrLater :=  (Win32Platform = VER_PLATFORM_WIN32_WINDOWS ) and ((Win32MajorVersion > 4) or
//   ((Win32MajorVersion = 4) and (Win32MinorVersion > 0)));
//  Win2kOrLater := (Win32Platform = VER_PLATFORM_WIN32_NT) and ((Win32MajorVersion > 4));
  InitAlphaBlend;
  with BF do
  begin
    BlendOp := AC_SRC_OVER;
    BlendFlags := 0;
    SourceConstantAlpha:= 160;
    AlphaFormat:=0;
  end;

finalization
end.



