unit Finplace;
{/*****************************************************************/
 /  Copyright (c) 1998 IMCA SYSTEMS                                /
 /  All Rights Reserved.                                           /
 /  Creation Date: 06.08.98                                        /
 /  Version                                                        /
 /  MODULE: Inplace Editor Control - real Plug for Controls        /
 /*****************************************************************/}
{$I DFS.inc}
{$ifdef DFS_COMPILER_4}
 {$ObjExportAll On}
{$endif}

{$Define Registered}
interface
uses Windows, Classes, Forms, CommCtrl, Controls, ExtCtrls,stdctrls,dialogs,
      Graphics, Messages, SysUtils, Menus, ComCtrls,Consts,ExGrid,Mask
      {$IFDEF DFS_DELPHI_6_UP},Variants{$endif};


type
  TSelection = record
    StartPos, EndPos: Integer;
  end;

 
  TISSizedListBox = class(TCustomListBox)
  private
    //sizegrip
    FSizeGrip : Boolean;
    FSizing : Boolean;
    Offset : TPoint;
    aGripBitmap : TBitmap;
    procedure SetSizeGrip(const Value : Boolean);
    function  IsSizeGrip(Pt : TSmallPoint) : Boolean;
    procedure WMNCCalcSize(var Message: TWMNCCalcSize); message WM_NCCALCSIZE;
    procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
    procedure WMNCLMouseDown(var Message: TMessage); message WM_NCLBUTTONDOWN;
    procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure PrepareSizeGrip;
   protected
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure ValidateState;
   public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property SizeGrip : Boolean read FSizeGrip write SetSizeGrip default False;
  end;

  TISCheckListBox = class(TISSizedListBox)
  private
    FAllowGrayed: Boolean;
    FFlat: Boolean;
    FStandardItemHeight: Integer;
    //FCheckWidth, FCheckHeight: Integer;
    FOnClickCheck: TNotifyEvent;
    FSaveStates: TList;
    procedure ResetItemHeight;
    procedure DrawCheck(R: TRect; AState: TCheckBoxState);
    procedure SetChecked(Index: Integer; Checked: Boolean);
    function GetChecked(Index: Integer): Boolean;
    procedure SetState(Index: Integer; AState: TCheckBoxState);
    function GetState(Index: Integer): TCheckBoxState;
    procedure ToggleClickCheck(Index: Integer);
    procedure InvalidateCheck(Index: Integer);
    function CreateWrapper(Index: Integer): TObject;
    function ExtractWrapper(Index: Integer): TObject;
    function GetWrapper(Index: Integer): TObject;
    function HaveWrapper(Index: Integer): Boolean;
    procedure SetFlat(Value: Boolean);
    procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure WMDestroy(var Msg : TWMDestroy);message WM_DESTROY;
  protected
    procedure DrawItem(Index: Integer; Rect: TRect;
      State: TOwnerDrawState); override;
    procedure SetItemData(Index: Integer; AData: LongInt); {$ifdef DFS_DELPHI_4_UP} override;{$endif}
    function GetItemData(Index: Integer): LongInt; {$ifdef DFS_DELPHI_4_UP} override;{$endif}
    procedure ResetContent; {$ifdef DFS_DELPHI_4_UP} override;{$endif}
    procedure DeleteString(Index: Integer); {$ifdef DFS_DELPHI_4_UP} override;{$endif}

    procedure KeyPress(var Key: Char); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;

    procedure ClickCheck; dynamic;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure DestroyWnd; override;
    function GetCheckWidth: Integer;
    
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Checked[Index: Integer]: Boolean read GetChecked write SetChecked;
    property State[Index: Integer]: TCheckBoxState read GetState write SetState;

  published
    property OnClickCheck: TNotifyEvent read FOnClickCheck write FOnClickCheck;
    property Align;
    property AllowGrayed: Boolean read FAllowGrayed write FAllowGrayed default False;
    {$ifdef DFS_DELPHI_4_UP}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property BorderStyle;
    property DragKind;
    property DragMode;
    property ImeMode;
    property ImeName;
    property ParentBiDiMode;
    property OnEndDock;
    property OnEndDrag;
    property OnStartDock;
    property OnStartDrag;
    {$endif}
    property Flat: Boolean read FFlat write SetFlat default True;
    property Color;
    property Columns;

    property Ctl3D;
    property DragCursor;
    property Enabled;
    property Font;
    property IntegralHeight;
    property ItemHeight;
    property Items;

    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Sorted;
    property Style;
    property TabOrder;
    property TabStop;
    property TabWidth;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;

    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;


  { TISPopupListbox }
  TISPopupListbox = class(TISSizedListBox)
  private
    FSearchText: String;
    FSearchTickCount: Longint;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure KeyPress(var Key: Char); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);override;
  public
    {$IFDEF DFS_DELPHI_4_UP}
    function UseRightToLeftAlignment: Boolean; override;
    {$ENDIF}

  end;

  TISPopupCheckListbox = class(TISCheckListBox)
  private
    FSearchText: String;
    FSearchTickCount: Longint;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure KeyPress(var Key: Char); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    function GetStrValue : string;
    procedure SetStrValue(Value : string);
  public
    {$IFDEF DFS_DELPHI_4_UP}
    function UseRightToLeftAlignment: Boolean; override;
    {$ENDIF}

    property StrValue : string read GetStrValue write SetStrValue;
  end;

type
TCustomFlyInplaceEdit=class;
TEditStyle = (etsSimple, etsEllipsis, etsPickList, etsCheckList, etsCheckBox,etsAuto);
TEditOption =(eosAlwaysShowEditor, eosGoTabs);
TeditOptions = set Of TEditOption;
//for future use
TIEDataType = (iedBoolean,iedInteger,iedCurrency,iedPicklist,iedChecklist);
TIEGetEditDataEvent = procedure(Sender: TCustomFlyInplaceEdit;  var AData : String; var Pwd : Char) of object;
TIEValidateDataEvent = procedure(Sender: TCustomFlyInplaceEdit; AData : String; var aCancel : boolean) of object;
TIEGetEditstyleEvent = procedure(Sender: TCustomFlyInplaceEdit; var aEditstyle : TEditStyle) of object;
TIEGetPickListEvent = procedure(Sender: TCustomFlyInplaceEdit; var aPickList : Tstrings) of object;
TIEGetReadOnlyEvent = procedure(Sender: TCustomFlyInplaceEdit; var aReadOnly : boolean) of object;

TIEEditButtonClickEvent = procedure(Sender: TCustomFlyInplaceEdit) of object;
TIEHideEditorEvent = procedure(Sender: TCustomFlyInplaceEdit) of object;
TIEUpdateDataEvent= procedure(Sender: TCustomFlyInplaceEdit) of object;
TIEGetClickKeyEvent=procedure(Sender: TCustomFlyInplaceEdit; var aKey : Word; var aShift : TShiftState) of object;
TIEUpdateEditEvent =procedure(Sender: TCustomFlyInplaceEdit) of object;


 TCustomFlyInplaceEdit = class(TCustomMaskEdit)
  private
    FControl: TWinControl;
    FValidating, FShowed : boolean;
    FEditOptions : TEditOptions;
    FPasswordChar : Char;//new
    FAlignment: TAlignment;
    //callback from FControl events
     FOnGetEditData : TIEGetEditDataEvent;
    FOnValidateData : TIEValidateDataEvent;
    FOnGetEditStyle : TIEGetEditstyleEvent;
    FonGetPickList : TIEGetPickListEvent;
    FonGetReadOnly : TIEGetReadOnlyEvent;

    FonEditButtonClick:TIEEditButtonClickEvent;
    FOnHideEditor : TIEHideEditorEvent;
    FOnUpdateData : TIEUpdateDataEvent;
    FonGetClickKey : TIEGetClickKeyEvent;
    FonUpdateEdit : TIEUpdateEditEvent;
    FonKeyUpParent : TKeyEvent;
    FonKeydownParent : TKeyEvent;
    FonKeyPressParent : TKeyPressEvent;
    FonDblClickParent :  TNotifyEvent;

    FClickTime: Longint;
    FCancelData : boolean;
    FUndoData : string;

    procedure InternalMove(const Loc: TRect; Redraw: Boolean; Scrolling : Boolean);

    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMPaste(var Message); message WM_PASTE;
    procedure WMCut(var Message); message WM_CUT;
    procedure WMClear(var Message); message WM_CLEAR;
    procedure SetPasswordChar(Value: Char);
    procedure SetAlignment(Value : TAlignment);
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure DblClick; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure BoundsChanged; virtual;
    procedure UpdateContents; virtual;
    procedure WndProc(var Message: TMessage); override;
//    procedure CreateWnd; override;
     //virtual engine
    procedure GetEditData( var AData : String; var Pwd : Char);
    procedure ValidateData(AData : String; var aCancel : boolean);
    procedure GetEditstyle(var aEditstyle : TEditStyle);
    procedure GetPickList(var aPickList : Tstrings);
    procedure GetReadOnly(var aReadOnly : boolean);

    procedure EditButtonClick;
    procedure HideEditor;
    procedure UpdateData;
    procedure UpdateEdit;//update location and data of editor
    procedure GetClickKey(var aKey : Word; var aShift : TShiftState);
    procedure KeyUpParent(var Key: Word; Shift: TShiftState);
    procedure KeydownParent(var Key: Word; Shift: TShiftState);
    procedure KeyPressParent(var Key: Char);
    procedure DblClickParent;

  public
    {$IFDEF DFS_DELPHI_4_UP}
    function UseRightToLeftAlignment: Boolean; override;
    {$ENDIF}
    procedure Hide(Scrolling : boolean);

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Deselect;
    procedure Invalidate;{$ifdef DFS_COMPILER_4_UP} reintroduce; {$endif}
    procedure Move(const Loc: TRect);
    function PosEqual(const Rect: TRect): Boolean;
    procedure SetFocus; override;
    procedure UpdateLoc(const Loc: TRect; Scrolling : Boolean);
    function Visible: Boolean;
    property UndoData : string read FUndoData write FUndoData;
    property Validating : Boolean read FValidating write FValidating;
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    property EditOptions : TEditOptions read FeditOptions write FeditOptions;
    property CancelData : boolean read FcancelData write FcancelData;//cancelling update data
    property ClickTime : Longint read FClickTime write FClickTime;
    property  Control: TWinControl read FControl write Fcontrol;
    property  OnGetEditData : TIEGetEditDataEvent read FOnGetEditData write FOnGetEditData;
    property  OnValidateData : TIEValidateDataEvent read FOnValidateData write FOnValidateData;
    property  OnGetEditStyle : TIEGetEditstyleEvent read FOnGetEditStyle write FOnGetEditStyle;
    property  OnGetPickList : TIEGetPickListEvent read FonGetPickList write FonGetPickList;
    property  OnGetReadOnly : TIEGetReadOnlyEvent read FonGetReadOnly write FonGetReadOnly;

    property  OnEditButtonClick : TIEEditButtonClickEvent read FonEditButtonClick write FonEditButtonClick;
    property  OnHideEditor : TIEHideEditorEvent read FonHideEditor write FonHideEditor;
    property  OnUpdateData : TIEUpdateDataEvent read FonUpdateData write FonUpdateData;
    property  OnGetClickKey : TIEGetClickKeyEvent read FonGetClickKey write FonGetClickKey;
    property  OnUpdateEdit : TIEUpdateEditEvent read FonUpdateEdit write FonUpdateEdit;
    property  OnKeyUpParent : TKeyEvent read FonKeyUpParent write FonKeyUpParent;
    property  OnKeydownParent : TKeyEvent read FonKeydownParent write FonKeydownParent;
    property  OnKeyPressParent : TKeyPressEvent read FonKeyPressParent write FonKeyPressParent;
    property  OnDblClickParent :  TNotifyEvent read FonDblClickParent write FonDblClickParent;
    property EditMask;
    property Font;//passwordchar
    Property Passwordchar : Char read FPasswordChar write SetPasswordChar default #0;
  end;

// TFlyInplaceEdit


  TFlyInplaceEdit = class(TCustomFlyInplaceEdit)
  private
    FButtonWidth: Integer;
    FSizeGrip : Boolean;
    FState : TCheckBoxState;
    FPickList: TCustomListbox;//TISPopupListbox;
    FActiveList: TWinControl;
    FEditStyle: TEditStyle;
    FListVisible: Boolean;
    FTracking: Boolean;
    FPressed: Boolean;
    procedure ListMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SetEditStyle(Value: TEditStyle);
    procedure PrepareActiveList;
    procedure StopTracking;
    procedure TrackButton(X,Y: Integer);

    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CancelMode;
    procedure WMCancelMode(var Message: TMessage); message WM_CancelMode;
    procedure WMKillFocus(var Message: TMessage); message WM_KillFocus;

    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message wm_LButtonDblClk;
    procedure WMPaint(var Message: TWMPaint); message wm_Paint;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SetCursor;
    function OverButton(const P: TPoint): Boolean;
    function ButtonRect: TRect;
    function CheckRect : TRect;

  protected
    procedure BoundsChanged; override;
    procedure CloseUp(Accept: Boolean);
    procedure DoDropDownKeys(var Key: Word; Shift: TShiftState);

    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure PaintWindow(DC: HDC); override;
    procedure WndProc(var Message: TMessage); override;

    property  ActiveList: TWinControl read FActiveList write FActiveList;
//    procedure CreateWnd; override;
    property  PickList: TCustomListbox read FPickList;
  public
    procedure CancelMode(Sender : Tobject);
    procedure UpdateContents; override;
    procedure DropDown;
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
    property  EditStyle: TEditStyle read FEditStyle write SetEditStyle;
    property DropDownSizable : Boolean read FSizeGrip write FSizeGrip; 
  end;

implementation
uses IsConst;

{ TISSizedListBox }
constructor TISSizedListBox.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   aGripBitmap := nil;
end;


destructor TISSizedListBox.Destroy;
begin
  if aGripBitmap<> nil then
    aGripBitmap.Free;
  inherited;
end;

procedure TISSizedListBox.SetSizeGrip(const Value: Boolean);
begin
  if FSizeGrip <> Value then
  begin
     FSizeGrip := Value;
     ValidateState;
  end;
end;

procedure TISSizedListBox.CMColorChanged(var Message: TMessage);
begin
   inherited;
   ValidateState;
end;


procedure TISSizedListBox.ValidateState;
begin
  if FSizeGrip then
     PrepareSizeGrip
  else
  begin
     aGripBitmap.Free;
     aGripBitmap :=nil;
  end;
 {$ifdef DFS_COMPILER_4_UP}
  Perform(CM_BORDERCHANGED, 0, 0);
 {$endif}
end;

procedure TISSizedListBox.WMNCCalcSize(var Message: TWMNCCalcSize);
begin
  inherited;
  if FSizeGrip then
    Dec(Message.CalcSize_Params^.rgrc[0].Bottom, GetSystemMetrics(SM_CYVSCROLL));
end;



procedure TISSizedListBox.WMNCPaint(var Message: TMessage);
var
  MaskDC, DC: HDC;
  sR, RC, RW, SaveRW: TRect;
  EdgeSize: Integer;
  WinStyle: Longint;
  Save: THandle;

begin
  { Get window DC that is clipped to the non-client area }
  if FSizeGrip then
  begin
    DC := GetWindowDC(Handle);
    try
      Windows.GetClientRect(Handle, RC);
      GetWindowRect(Handle, RW);
      MapWindowPoints(0, Handle, RW, 2);
      OffsetRect(RC, -RW.Left, -RW.Top);
      ExcludeClipRect(DC, RC.Left, RC.Top, RC.Right, RC.Bottom);

      SaveRW := RW;
      EdgeSize := GetSystemMetrics(SM_CYVSCROLL);
      Inc(RC.Bottom, EdgeSize);
      RW := RC;
      with RW do
        begin
          WinStyle := GetWindowLong(Handle, GWL_STYLE);
          if (WinStyle and WS_VSCROLL) <> 0 then
            Inc(Right, GetSystemMetrics(SM_CYVSCROLL));
          if (WinStyle and WS_HSCROLL) <> 0 then
            Inc(Bottom, GetSystemMetrics(SM_CXHSCROLL));
        end;
      with RW do
      {$ifdef DFS_COMPILER_4_UP}
      if UseRightToLeftAlignment then
        sr := Rect(Left, Bottom-EdgeSize, Left+EdgeSize, Bottom)
      else
      {$endif}
        sr := Rect(Right-EdgeSize, Bottom-EdgeSize, Right, Bottom);

      IntersectClipRect(DC, RW.Left, RW.Top, RW.Right, RW.Bottom);
      RW := SaveRW;
      { Erase parts not drawn }
      OffsetRect(RW, -RW.Left, -RW.Top);
      Windows.FillRect(DC, RW, Brush.Handle);
      Save := 0;
      MaskDC := 0;
      try
        MaskDC := CreateCompatibleDC(0);
        Save := SelectObject(MaskDC, aGripBitmap.MaskHandle);
        TransparentStretchBlt(DC, SR.Left, Sr.Top, EdgeSize,
        EdgeSize, aGripBitmap.Canvas.Handle, 0, 0, EdgeSize,
        EdgeSize, MaskDC, 0, 0);
      finally
        if Save <> 0 then SelectObject(MaskDC, Save);
        if MaskDC <> 0 then DeleteDC(MaskDC);
      end;
    finally
      ReleaseDC(Handle, DC);
    end;
  end;
  inherited;
end;

procedure TISSizedListBox.PrepareSizeGrip;
var  EdgeSize, Flags : integer;
     aBmp : TBitmap;
begin
    EdgeSize := GetSystemMetrics(SM_CYVSCROLL);
    aGripBitmap := TBitmap.Create;
    aGripBitmap.Width := EdgeSize;
    aGripBitmap.Height := EdgeSize;

    {$ifdef DFS_COMPILER_4_UP}
    if UseRighttoLeftAlignment then //
      Flags := DFCS_SCROLLSIZEGRIPRIGHT
    else
    {$endif}
     Flags := DFCS_SCROLLSIZEGRIP;//+DFCS_TRANSPARENT

    if (not Ctl3d) and (BordersTyle= bsNone) then
           Flags := Flags + DFCS_FLAT;

    DrawFrameControl(aGripBitmap.Canvas.Handle, Rect(0,0, Edgesize, EdgeSize),DFC_SCROLL,Flags);
    {$ifdef DFS_COMPILER_4_UP}
    if UseRightToLeftAlignment then
    begin
      aBmp := TBitmap.Create;
      try
         aBmp.Width := EdgeSize;
         aBmp.Height := EdgeSize;
         abmp.Canvas.StretchDraw(Rect(0, 0, EdgeSize, EdgeSize),aGripBitmap);
         aGripBitmap.Assign(aBmp);
      finally
         aBmp.Free;
      end;
    end;
    {$endif}
end;

function TISSizedListBox.IsSizeGrip(Pt : TSmallPoint) : Boolean;
var RW : TRect;
    EdgeSize : integer;
begin
  GetWindowRect(Handle, RW);
  Dec(Pt.X, RW.Left);
  Dec(Pt.Y, RW.Top);
  EdgeSize := GetSystemMetrics(SM_CYVSCROLL);
  {$ifdef DFS_COMPILER_4_UP}
  if UseRightToLeftAlignment then
   Result := (Pt.X>0) and (Pt.X < EdgeSize) and
     ((Pt.Y)>(RW.Bottom-RW.Top)-EdgeSize) and ((Pt.Y)<(RW.Bottom-RW.Top))
  else
  {$endif}
   Result := ((Pt.X)>(RW.Right-RW.Left)-EdgeSize) and ((Pt.X)<(RW.Right-RW.Left)) and
     ((Pt.Y)>(RW.Bottom-RW.Top)-EdgeSize) and ((Pt.Y)<(RW.Bottom-RW.Top));
end;

procedure TISSizedListBox.WMNCHitTest(var Message: TWMNCHitTest); //message WM_NCHITTEST;
begin
  inherited;
  if FSizeGrip then
  with Message do
  begin
     if IsSizeGrip(Pos) then
        Result := HTGROWBOX;
  end;
end;

procedure TISSizedListBox.WMSetCursor(var Message: TWMSetCursor);
begin
  if (not (csDesigning in ComponentState)) and FSizeGrip then
  begin
  with Message do
   case Smallint(HitTest) of
     HTGROWBOX:
       {$ifdef DFS_COMPILER_4_UP}
       if UseRightToLeftAlignment then
        Windows.SetCursor(Screen.Cursors[crSizeNESW])
       else
       {$endif}
        Windows.SetCursor(Screen.Cursors[crSizeNWSE]);
     else inherited;
   end;
   end
   else inherited;
end;

procedure TISSizedListBox.WMNCLMouseDown(var Message: TMessage);
var aPt : TSmallPoint;
    Sizingstart : TPoint;
begin
  inherited;
   With Message do
   if FSizeGrip then
   begin
      aPt := TSmallPoint(Lparam);
      if IsSizeGrip(aPt) then
      begin
        FSizing :=True;
        SizingStart.X := aPt.X;
        SizingStart.Y := aPt.Y;
        SizingStart := ScreenToClient(SizingStart);
        {$ifdef DFS_COMPILER_4_UP}
        if UseRightToLeftAlignment then
          Offset.X := SizingStart.X
        else
        {$endif}
          Offset.X := Width -SizingStart.X;
        Offset.Y := Height -SizingStart.Y;
        MouseCapture := True;
      end;
   end;
end;

procedure TISSizedListBox.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
   if FSizing then
   begin
     {$ifdef DFS_COMPILER_4_UP}
     if UseRightToLeftAlignment then
       SetWindowPos(Handle, HWND_TOP, Left+(X-Offset.X), Top, Width-(X-Offset.X), Y+Offset.Y,
        SWP_NOACTIVATE)
     else
     {$endif}
        SetWindowPos(Handle, HWND_TOP, Left, Top, X+Offset.X, Y+Offset.Y,
        SWP_NOACTIVATE);

   end
   else
     Inherited;
end;

procedure TISSizedListBox.WMLButtonUp(var Message: TWMLButtonUp);
begin
   if FSizing then
   begin
      Fsizing := False;
      MouseCapture := False;
   end
   else
     inherited;
end;


procedure TISPopupListbox.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or  WS_BORDER;
    ExStyle := WS_EX_TOOLWINDOW or WS_EX_TOPMOST;
    {$IFDEF DFS_DELPHI_4_UP}
    AddBiDiModeExStyle(ExStyle);
    {$endif}
    WindowClass.Style := CS_SAVEBITS;
  end;
end;

{$IFDEF DFS_DELPHI_4_UP}
function TISPopupListbox.UseRightToLeftAlignment: Boolean;
begin
  if Parent <> nil then
   Result := Parent.UseRightToLeftAlignment
  else Result := inherited UseRightToLeftAlignment;
end;
{$ENDIF}

procedure TISPopupListbox.CreateWnd;
begin
  inherited CreateWnd;
  Windows.SetParent(Handle, 0);
  CallWindowProc(DefWndProc, Handle, wm_SetFocus, 0, 0);
end;

procedure TISPopupListbox.Keypress(var Key: Char);
var
  TickCount: Integer;
begin
  case Key of
    #8, #27: FSearchText := '';
    #32..#255:
      begin
        TickCount := GetTickCount;
        if TickCount - FSearchTickCount > 2000 then FSearchText := '';
        FSearchTickCount := TickCount;
        if Length(FSearchText) < 32 then FSearchText := FSearchText + Key;
        SendMessage(Handle, LB_SelectString, WORD(-1), Longint(PChar(FSearchText)));
        Key := #0;
      end;
  end;
  inherited Keypress(Key);
end;

procedure TISPopupListbox.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  //TFlyInplaceEdit(Owner).CloseUp((X >= 0) and (Y >= 0) and
  //    (X < Width) and (Y < Height)); //ptinrect
end;


procedure TISPopupListbox.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
var
  Index: Integer;
begin
  Index := ItemAtPos(Point(X,Y),True);
  if Index <> -1 then itemindex:=index;
    inherited;
end;

//-----------------------------------------------------------------------------

type
  TISCheckListBoxDataWrapper = class
  private
    FData: LongInt;
    FState: TCheckBoxState;
    procedure SetChecked(Check: Boolean);
    function GetChecked: Boolean;
  public
    class function GetDefaultState: TCheckBoxState;
    property Checked: Boolean read GetChecked write SetChecked;
    property State: TCheckBoxState read FState write FState;
  end;

var
  FCheckWidth, FCheckHeight: Integer;

procedure GetCheckSize;
begin
 FCheckWidth := GetSystemMetrics(SM_CXMENUCHECK);
 FCheckHeight := GetSystemMetrics(SM_CYMENUCHECK);
end;

{ TISCheckListBoxDataWrapper }

procedure TISCheckListBoxDataWrapper.SetChecked(Check: Boolean);
begin
  if Check then
     FState := cbChecked
  else
     FState := cbUnchecked;
end;

function TISCheckListBoxDataWrapper.GetChecked: Boolean;
begin
  Result := FState = cbChecked;
end;

class function TISCheckListBoxDataWrapper.GetDefaultState: TCheckBoxState;
begin
  Result := cbUnchecked;
end;

{ TISCheckListBox }

constructor TISCheckListBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFlat := True;
end;

destructor TISCheckListBox.Destroy;
VAR i : integer;
begin
  FSaveStates.Free;
  FSaveStates:=nil;
  for i := 0 to Items.Count -1 do
    ExtractWrapper(i).Free;

  inherited destroy;
end;

procedure TISCheckListBox.CreateWnd;
begin
  inherited CreateWnd;
  if FSaveStates <> nil then
  begin
    FSaveStates.Free;
    FSaveStates := nil;
  end;
  ResetItemHeight;
end;

procedure TISCheckListBox.DestroyWnd;
var
  I: Integer;
begin
  if Items.Count > 0 then
  begin
    FSaveStates := TList.Create;
    for I := 0 to Items.Count -1 do
      FSaveStates.Add(TObject(State[I]));
  end;
  inherited DestroyWnd;
end;

procedure TISCheckListBox.CreateParams(var Params: TCreateParams);
begin
  inherited;
  with Params do
  begin
    if Style and (LBS_OWNERDRAWFIXED or LBS_OWNERDRAWVARIABLE) = 0 then
      Style := Style or LBS_OWNERDRAWFIXED;
    Style := Style or  WS_BORDER;
  end;
end;

function TISCheckListBox.GetCheckWidth: Integer;
begin
  Result := FCheckWidth + 2;
end;

procedure TISCheckListBox.CMFontChanged(var Message: TMessage);
begin
  inherited;
  ResetItemHeight;
end;

procedure TISCheckListBox.ResetItemHeight;
begin
  if HandleAllocated and (Style = lbStandard) then
  begin
    Canvas.Font := Font;
    FStandardItemHeight := Canvas.TextHeight('Wg');
    Perform(LB_SETITEMHEIGHT, 0, FStandardItemHeight);
  end;
end;

procedure TISCheckListBox.DrawItem(Index: Integer; Rect: TRect;
  State: TOwnerDrawState);
var
  R: TRect;
  SaveEvent: TDrawItemEvent;
  ACheckWidth: Integer;
begin
  ACheckWidth := GetCheckWidth;
  if Index < Items.Count then
  begin
    R := Rect;
    {$ifdef DFS_DELPHI_4_UP}
    if not UseRightToLeftAlignment then
    begin
      R.Right := Rect.Left;
      R.Left := R.Right - ACheckWidth;
    end
    else
    begin
      R.Left := Rect.Right;
      R.Right := R.Left + ACheckWidth;
    end;
    {$else}
      R.Right := Rect.Left;
      R.Left := R.Right - ACheckWidth;
      offsetRect(R,-r.left,0);
    {$endif}
    DrawCheck(R, GetState(Index));
  end;

  if (Style = lbStandard) and Assigned(OnDrawItem) then
  begin
    { Force lbStandard list to ignore OnDrawItem event. }
    SaveEvent := OnDrawItem;
    OnDrawItem := nil;
    try
      inherited;
    finally
      OnDrawItem := SaveEvent;
    end;
  end
  else
    inherited;
end;

procedure TISCheckListBox.CNDrawItem(var Message: TWMDrawItem);
begin
  with Message.DrawItemStruct^ do
     {$ifdef DFS_DELPHI_4_UP}
    if not UseRightToLeftAlignment then
      rcItem.Left := rcItem.Left + GetCheckWidth
    else
      rcItem.Right := rcItem.Right - GetCheckWidth;
    {$else}
      rcItem.Left := rcItem.Left + GetCheckWidth;
    {$endif}
  inherited;
end;

procedure TISCheckListBox.DrawCheck(R: TRect; AState: TCheckBoxState);
var
  DrawState: Integer;
  DrawRect: TRect;
  OldBrushColor: TColor;
  OldBrushStyle: TBrushStyle;
  OldPenColor: TColor;
  Rgn, SaveRgn: HRgn;
begin
  SaveRgn := 0;
  DrawRect.Left := R.Left + (R.Right - R.Left - FCheckWidth) div 2;
  DrawRect.Top := R.Top + (R.Bottom - R.Top - FCheckWidth) div 2;
  DrawRect.Right := DrawRect.Left + FCheckWidth;
  DrawRect.Bottom := DrawRect.Top + FCheckHeight;
  case AState of
    cbChecked:
      DrawState := DFCS_BUTTONCHECK or DFCS_CHECKED;
    cbUnchecked:
      DrawState := DFCS_BUTTONCHECK;
    else // cbGrayed
      DrawState := DFCS_BUTTON3STATE or DFCS_CHECKED;
  end;
  with Canvas do
  begin
    if Flat then
    begin
      { Remember current clipping region }
      SaveRgn := CreateRectRgn(0,0,0,0);
      GetClipRgn(Handle, SaveRgn);
      { Clip 3d-style checkbox to prevent flicker }
      with DrawRect do
        Rgn := CreateRectRgn(Left + 2, Top + 2, Right - 2, Bottom - 2);
      SelectClipRgn(Handle, Rgn);
      DeleteObject(Rgn);
    end;
    DrawFrameControl(Handle, DrawRect, DFC_BUTTON, DrawState);
    if Flat then
    begin
      SelectClipRgn(Handle, SaveRgn);
      DeleteObject(SaveRgn);
      { Draw flat rectangle in-place of clipped 3d checkbox above }
      OldBrushStyle := Brush.Style;
      OldBrushColor := Brush.Color;
      OldPenColor := Pen.Color;
      Brush.Style := bsClear;
      Pen.Color := clBtnShadow;
      with DrawRect do
        Rectangle(Left + 1, Top + 1, Right - 1, Bottom - 1);
      Brush.Style := OldBrushStyle;
      Brush.Color := OldBrushColor;
      Pen.Color := OldPenColor;
    end;
  end;
end;

procedure TISCheckListBox.SetChecked(Index: Integer; Checked: Boolean);
begin
  if Checked <> GetChecked(Index) then
  begin
    TISCheckListBoxDataWrapper(GetWrapper(Index)).SetChecked(Checked);
    InvalidateCheck(Index);
  end;
end;

procedure TISCheckListBox.SetState(Index: Integer; AState: TCheckBoxState);
begin
  if AState <> GetState(Index) then
  begin
    TISCheckListBoxDataWrapper(GetWrapper(Index)).State := AState;
    InvalidateCheck(Index);
  end;
end;
    
procedure TISCheckListBox.InvalidateCheck(Index: Integer);
var
  R: TRect;
begin
  R := ItemRect(Index);
  {$ifdef DFS_DELPHI_4_UP}
  if not UseRightToLeftAlignment then
    R.Right := R.Left + GetCheckWidth
  else
  {$endif}
    R.Left := R.Right - GetCheckWidth;
  InvalidateRect(Handle, @R, not (csOpaque in ControlStyle));
  UpdateWindow(Handle);
end;
    
function TISCheckListBox.GetChecked(Index: Integer): Boolean;
begin
  if HaveWrapper(Index) then
    Result := TISCheckListBoxDataWrapper(GetWrapper(Index)).GetChecked
  else
    Result := False;
end;
    
function TISCheckListBox.GetState(Index: Integer): TCheckBoxState;
begin
  if HaveWrapper(Index) then
    Result := TISCheckListBoxDataWrapper(GetWrapper(Index)).State
  else
    Result := TISCheckListBoxDataWrapper.GetDefaultState;
end;
    
procedure TISCheckListBox.KeyPress(var Key: Char);
begin
  inherited;
  if (Key = ' ') then ToggleClickCheck(ItemIndex);
end;
    
procedure TISCheckListBox.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
var
  Index: Integer;
begin
  inherited;
  Index := ItemAtPos(Point(X,Y),True);
  if Index <> -1 then
    {$ifdef DFS_DELPHI_4_UP}
    if not UseRightToLeftAlignment then
    begin
      if X - ItemRect(Index).Left < GetCheckWidth then
        ToggleClickCheck(Index)
    end
    else
    begin
      Dec(X, ItemRect(Index).Right - GetCheckWidth);
      if (X > 0) and (X < GetCheckWidth) then
        ToggleClickCheck(Index)
    end;
    {$ELSE}
     if X - ItemRect(Index).Left < GetCheckWidth then
        ToggleClickCheck(Index)
    {$endif}

end;
    
procedure TISCheckListBox.ToggleClickCheck;
var
  State: TCheckBoxState;
begin
  if (Index >= 0) and (Index < Items.Count) then
  begin
    State := Self.State[Index];
    case State of
      cbUnchecked:
        if AllowGrayed then State := cbGrayed else State := cbChecked;
      cbChecked: State := cbUnchecked;
      cbGrayed: State := cbChecked;
    end;
    Self.State[Index] := State;
    ClickCheck;
    invalidate;
  end;
end;

procedure TISCheckListBox.ClickCheck;
begin
  if Assigned(FOnClickCheck) then FOnClickCheck(Self);
end;


function TISCheckListBox.GetItemData(Index: Integer): LongInt;
begin
  Result := 0;
  if HaveWrapper(Index) then
    Result := TISCheckListBoxDataWrapper(GetWrapper(Index)).FData;
end;


function TISCheckListBox.GetWrapper(Index: Integer): TObject;
begin
  Result := ExtractWrapper(Index);
  if Result = nil then
    Result := CreateWrapper(Index);
end;

function TISCheckListBox.ExtractWrapper(Index: Integer): TObject;
begin
  Result := TISCheckListBoxDataWrapper(
  {$ifdef DFS_DELPHI_4_UP}inherited GetItemData(Index)
  {$else}
     SendMessage(Handle, LB_GETITEMDATA, Index, 0)
   {$endif}  );
  if LB_ERR = Integer(Result) then
    {$ifdef DFS_DELPHI_4_UP}
    raise EListError.CreateFmt(SListIndexError, [Index]);
    {$else}
    raise EListError.CreateFmt('List index out of bounds (%d)', [Index]);
    {$endif}
  if (Result <> nil) and (not (Result is TISCheckListBoxDataWrapper)) then
    Result := nil;
end;

function TISCheckListBox.CreateWrapper(Index: Integer): TObject;
begin
  Result := TISCheckListBoxDataWrapper.Create;
  {$ifdef DFS_DELPHI_4_UP}
  inherited SetItemData(Index, LongInt(Result));
  {$else}
   SendMessage(Handle, LB_SETITEMDATA, Index, LongInt(Result));
  {$endif}
end;

function TISCheckListBox.HaveWrapper(Index: Integer): Boolean;
begin
  Result := ExtractWrapper(Index) <> nil;
end;


procedure TISCheckListBox.SetItemData(Index: Integer; AData: LongInt);
var
  Wrapper: TISCheckListBoxDataWrapper;
begin
  Wrapper := TISCheckListBoxDataWrapper(GetWrapper(Index));
  Wrapper.FData := AData;
  if FSaveStates <> nil then
    if FSaveStates.Count > 0 then
    begin
     Wrapper.FState := TCheckBoxState(FSaveStates[0]);
     FSaveStates.Delete(0);
    end;
end;

procedure TISCheckListBox.ResetContent;
var
  I: Integer;
begin
  for I := 0 to Items.Count - 1 do
    if HaveWrapper(I) then
      GetWrapper(I).Free;
  {$ifdef DFS_DELPHI_4_UP} inherited;{$endif}
end;



procedure TISCheckListBox.DeleteString(Index: Integer);
begin
  if HaveWrapper(Index) then
    GetWrapper(Index).Free;

  {$ifdef DFS_DELPHI_4_UP} inherited;{$endif}
end;

procedure TISCheckListBox.SetFlat(Value: Boolean);
begin
  if Value <> FFlat then
  begin
    FFlat := Value;
    Invalidate;
  end;
end;

procedure TISCheckListBox.WMDestroy(var Msg: TWMDestroy);
begin
  inherited;
end;

{TCustomFlyInplaceEdit}
constructor TCustomFlyInplaceEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ParentCtl3D := false;
  Ctl3D := False;
  TabStop := False;
  ParentFont :=true;
  FControl := TWincontrol(Owner);
  Parent:=TWincontrol(Owner);
  ControlStyle := ControlStyle-[csFramed];
  CancelData:=true;
  {$IFDEF DFS_DELPHI_4_UP}
  DoubleBuffered := False;
  {$endif}
end;

destructor TCustomFlyInplaceEdit.Destroy;
begin
  // Parent:=Fcontrol;
   inherited;
end;

procedure TCustomFlyInplaceEdit.SetPasswordChar(Value: Char);
begin
  if FPasswordChar <> Value then
  begin
    FPasswordChar := Value;
    RecreateWnd;
    if HandleAllocated then
    begin
      SendMessage(Handle, EM_SETPASSWORDCHAR, Ord(FPasswordChar), 0);
      SetTextBuf(PChar(Text));
    end;
  end;

end;

procedure TCustomFlyInplaceEdit.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  if PasswordChar <> #0 then
      Params.Style := Params.Style or ES_PASSWORD
  else Params.Style := Params.Style or ES_MULTILINE;

  Case Ord(FAlignment) of
    ord(taLeftJustify): Params.Style := Params.Style or ES_LEFT;
    ord(taCenter): Params.Style := Params.Style or ES_CENTER;
    ord(taRightJustify): Params.Style := Params.Style or ES_RIGHT;
  end;
  Params.Style := Params.Style and not WS_BORDER;
  if NewStyleControls and Ctl3D and (csFramed in ControlStyle) then
    begin
      Params.ExStyle := Params.ExStyle or WS_EX_CLIENTEDGE;
  end;
  //Params.Style := Params.Style or WS_BORDER;
end;



procedure TCustomFlyInplaceEdit.CMShowingChanged(var Message: TMessage);
begin
  { Ignore showing using the Visible property }
end;

procedure TCustomFlyInplaceEdit.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  if eosGoTabs in FEditOptions then
    Message.Result := Message.Result or DLGC_WANTTAB;
end;

procedure TCustomFlyInplaceEdit.WMPaste(var Message);
var txt : string;
    aCancel : Boolean;
begin
  if Readonly then Exit;
  Txt := Text;
//  Clipboard.Open;
//  txt := Clipboard.AsText;
//  Clipboard.Close;
//  Canceldata := true;
//  ValidateData(Txt, aCancel);
//  Canceldata := false;
//  if not Cancel then
    inherited;

  Canceldata := true;
  aCancel := true;//still typing
  ValidateData(Text, aCancel);
  Canceldata := false;
   if aCancel then
     Text := Txt;

end;

procedure TCustomFlyInplaceEdit.WMClear(var Message);
begin
  if Readonly then Exit;
  inherited;
end;

procedure TCustomFlyInplaceEdit.WMCut(var Message);
var txt : string;
    aCancel : Boolean;
begin
  if Readonly then Exit;
  Txt := Text;

  inherited;

  Canceldata := true;
  aCancel := true;//still typing
  ValidateData(Text, aCancel);
  Canceldata := false;
   if aCancel then
     Text := Txt;

end;

procedure TCustomFlyInplaceEdit.DblClick;
begin
  //FControl.OnDblClick(Self);
   DblClickParent;
end;

procedure TCustomFlyInplaceEdit.SetAlignment(Value: TAlignment);
begin
  if FAlignment <> Value then
  begin
    FAlignment := Value;
    RecreateWnd;
  end;
end;

procedure TCustomFlyInplaceEdit.KeyDown(var Key: Word; Shift: TShiftState);
  procedure SendToParent;
  //var KeyData : longint;
  begin
     //FControl.KeyDown(Key, Shift);
     KeyDownParent(Key,Shift);
     {KeyData:=0;
     if ssAlt in shift then KeyData:=KeyData+$20000000;
     if ssctrl in shift then KeyData:=KeyData+VK_CONTROL;
     if ssShift in shift then KeyData:=KeyData+VK_SHIFT;
     Postmessage(Fcontrol.Handle,WM_KEYDOWN,Key,KeyData);}
  end;

  procedure ParentEvent;
  begin
    KeyDownParent(Key,Shift);
  end;

  function ForwardMovement: Boolean;
  begin
    Result := eosAlwaysShowEditor in FEditOptions;
  end;

  function Ctrl: Boolean;
  begin
    Result := ssCtrl in Shift;
  end;

  function Selection: TSelection;
  begin
    SendMessage(Handle, EM_GETSEL, Longint(@Result.StartPos), Longint(@Result.EndPos));
  end;

  function RightSide: Boolean;
  begin
    with Selection do
      //Result := ((StartPos = 0) or (EndPos = StartPos)) and
      //  (EndPos = GetTextLen);
      Result := (EndPos = GetTextLen) and (EndPos = StartPos);//new 29/3/2
   end;

  function LeftSide: Boolean;
  begin
    with Selection do
      //Result := (StartPos = 0) and ((EndPos = 0) or (EndPos = GetTextLen));
      Result := ((StartPos = 0) and (EndPos = 0));//new 29/3/2
  end;

begin
  //Parent := Fcontrol;
  case Key of

    VK_ADD, VK_SUBTRACT, VK_RETURN,VK_UP, VK_DOWN, VK_PRIOR, VK_NEXT, VK_ESCAPE: SendToParent;
    VK_INSERT:
      if Shift = [] then
      begin
       SendToParent;
      end
      else if (Shift = [ssShift]) and ReadOnly then Key := 0;
    VK_LEFT: if ForwardMovement and (Ctrl or LeftSide) then SendToParent;
    VK_RIGHT: if ForwardMovement and (Ctrl or RightSide) then SendToParent;
    VK_HOME: if ForwardMovement and (Ctrl or LeftSide) then SendToParent;
    VK_END: if ForwardMovement and (Ctrl or RightSide) then SendToParent;
    VK_F2:
      begin
        ParentEvent;
        if Key = VK_F2 then
        begin
          Deselect;
          Exit;
        end;
      end;
    VK_TAB: if not (ssAlt in Shift) then SendToParent;
  end;
  if (Key = VK_DELETE) and ReadOnly then Key := 0;
  if Key <> 0 then
  begin
    //ParentEvent;
    inherited KeyDown(Key, Shift);
  end;
end;

procedure TCustomFlyInplaceEdit.KeyPress(var Key: Char);
var
  Selection: TSelection;
//  aCancel : Boolean;
//  RText, OldText : string;
begin
  //Parent := Fcontrol;
  if Key=#27 then FcancelData:=true;
  //to Parent
  KeyPressParent(Key);

  if (Key in [#32..#255]) and ReadOnly then
  begin
    Key := #0;
    MessageBeep(0);
  end;
  case Key of
    #9, #27: Key := #0;
    #13:
      begin
        SendMessage(Handle, EM_GETSEL, Longint(@Selection.StartPos), Longint(@Selection.EndPos));
        if (Selection.StartPos = 0) and (Selection.EndPos = GetTextLen) then
          Deselect else
          SelectAll;
        Key := #0;
      end;
    ^H, ^V, ^X, #32..#255:
      if ReadOnly then Key := #0;
  end;

{  if (Key in [#32..#255]) then
  begin
    FundoData := Text;
    RText := FundoData + Key;
    OldText := RText;
    aCancel := True;
    Canceldata := true;
    ValidateData(RText, aCancel);
    Canceldata := false;
    if aCancel then
    begin
      if (OldText <> RText) then//if text not modified
      begin
        Text := RText;
        FUndoData := '';
      end
      else
        Text := FUndoData;
      Key := #0;
    end;
  end;
}
  if (Key <> #0) then
    inherited KeyPress(Key);

end;

procedure TCustomFlyInplaceEdit.KeyUp(var Key: Word; Shift: TShiftState);
 begin
    KeyUpParent(Key,Shift);
end;

procedure TCustomFlyInplaceEdit.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    WM_SETFOCUS:
      begin
        //if (GetParentForm(Self) = nil) or GetParentForm(Self).SetFocusedControl(FControl) then
        Dispatch(Message);
        Exit;
      end;
    WM_LBUTTONDOWN:
       begin
        if UINT(GetMessageTime - FClickTime) < GetDoubleClickTime then
          Message.Msg := WM_LBUTTONDBLCLK;
        FClickTime := 0;
      end;
  end;
  inherited WndProc(Message);
end;

procedure TCustomFlyInplaceEdit.Deselect;
begin
  SendMessage(Handle, EM_SETSEL, $7FFFFFFF, Longint($FFFFFFFF));
end;

procedure TCustomFlyInplaceEdit.Invalidate;
var
  Cur: TRect;
begin
  //Parent := Fcontrol;
  ValidateRect(Handle, nil);
  InvalidateRect(Handle, nil, True);
  Windows.GetClientRect(Handle, Cur);
  MapWindowPoints(Handle, FControl.Handle, Cur, 2);
  ValidateRect(FControl.Handle, @Cur);
  InvalidateRect(FControl.Handle, @Cur, False);
end;

{$IFDEF DFS_DELPHI_4_UP}
function TCustomFlyInplaceEdit.UseRightToLeftAlignment: Boolean;
begin
  if Parent <> nil then
   Result := Parent.UseRightToLeftAlignment
  else
     Result := inherited UseRightToLeftAlignment;
end;
{$ENDIF}

procedure TCustomFlyInplaceEdit.Hide(Scrolling : boolean);
begin
  if HandleAllocated {and IsWindowVisible(Handle)} then
  begin
    Invalidate;
    SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_HIDEWINDOW or SWP_NOZORDER or
      SWP_NOREDRAW);
  end;
  if Focused and not Scrolling then
    Windows.SetFocus(FControl.Handle);
end;

function TCustomFlyInplaceEdit.PosEqual(const Rect: TRect): Boolean;
var
  Cur: TRect;
begin
  GetWindowRect(Handle, Cur);
  MapWindowPoints(HWND_DESKTOP, Fcontrol.Handle, Cur, 2);
  Result := EqualRect(Rect, Cur);
end;

procedure TCustomFlyInplaceEdit.InternalMove(const Loc: TRect; Redraw: Boolean; Scrolling : Boolean);
begin
  if IsRectEmpty(Loc) then
    Hide(Scrolling)
  else
  begin
    CreateHandle;
    Redraw := Redraw or not IsWindowVisible(Handle);
    Invalidate;
    with Loc do
      SetWindowPos(Handle, HWND_TOP, Left, Top, Right - Left, Bottom - Top,
        SWP_SHOWWINDOW or SWP_NOREDRAW);
    BoundsChanged;
    if Redraw then Invalidate;
    if FControl.Focused and not Scrolling then
      Windows.SetFocus(Handle);
  end;
end;

procedure TCustomFlyInplaceEdit.BoundsChanged;
var
  R: TRect;
begin
  R := Rect(2, 2, Width - 2, Height);
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@R));
  SendMessage(Handle, EM_SCROLLCARET, 0, 0);
end;

procedure TCustomFlyInplaceEdit.UpdateLoc(const Loc: TRect; Scrolling : Boolean);
begin
  InternalMove(Loc, False, Scrolling);
end;

function TCustomFlyInplaceEdit.Visible: Boolean;
begin
  //Parent := Fcontrol;
  Result := IsWindowVisible(Handle);
end;

procedure TCustomFlyInplaceEdit.Move(const Loc: TRect);
begin
  InternalMove(Loc, True, false);
end;

procedure TCustomFlyInplaceEdit.SetFocus;
begin
  //Parent := Fcontrol;
  if IsWindowVisible(Handle) then
    Windows.SetFocus(Handle);
end;

//procedure TCustomFlyInplaceEdit.CreateWnd;
//begin
//  inherited CreateWnd;
//end;

procedure TCustomFlyInplaceEdit.UpdateContents;
var aText : string;
    apwdChar : Char;
begin
  aText := '';
  apwdChar:=#0;
  EditMask:='';
  GetEditData(atext,apwdChar);//virtual function, Fcontrol must be connected to this events
  Text:=aText;
  Passwordchar := apwdChar;
  invalidate;
end;

//virtual engine
procedure TCustomFlyInplaceEdit.GetEditData( var AData : String; var Pwd : Char);
begin
   if Assigned(FonGetEditData) then FonGetEditData(Self,adata,Pwd);
end;
procedure TCustomFlyInplaceEdit.ValidateData(AData : String; var aCancel : boolean);
begin
   if Assigned(FonValidateData) then FonValidateData(Self,AData,aCancel);
end;

procedure TCustomFlyInplaceEdit.GetEditstyle(var aEditstyle : TEditStyle);
begin
   if Assigned(FonGetEditStyle) then FonGetEditStyle(Self,aEditstyle);
end;
procedure TCustomFlyInplaceEdit.GetPickList(var aPickList : Tstrings);
begin
   if Assigned(fonGetPickList) then fonGetPickList(Self,aPickList);
end;
procedure TCustomFlyInplaceEdit.GetReadOnly(var aReadOnly : boolean);
begin
   if Assigned(FonGetReadOnly) then FonGetReadOnly(Self,areadOnly);
end;

procedure TCustomFlyInplaceEdit.EditButtonClick;
begin
   FShowed :=true;
   if Assigned(FonEditButtonClick) then FonEditButtonClick(Self);
   UpdateData;
   FShowed :=false;
end;

procedure TCustomFlyInplaceEdit.HideEditor;
begin
   if Assigned(FonHideEditor) then FonHideEditor(Self);
end;

procedure TCustomFlyInplaceEdit.UpdateData;
begin
   if Assigned(FonUpdateData) then FonUpdateData(Self);
end;
procedure TCustomFlyInplaceEdit.GetClickKey(var aKey : Word; var aShift : TShiftState);
begin
   if Assigned(FonGetClickKey) then FonGetClickKey(Self,aKey,aShift);
end;

procedure TCustomFlyInplaceEdit.UpdateEdit;
begin
   if Assigned(FonUpdateEdit) then FonUpdateEdit(Self);
end;

procedure TCustomFlyInplaceEdit.KeyUpParent(var Key: Word; Shift: TShiftState);
begin
   if Assigned(FonKeyUpParent) then FonKeyUpParent(Self,Key,Shift);
end;
procedure TCustomFlyInplaceEdit.KeyDownParent(var Key: Word; Shift: TShiftState);
begin
   if Assigned(FonKeyDownParent) then FonKeyDownParent(Self,Key,Shift);
end;
procedure TCustomFlyInplaceEdit.KeyPressParent(var Key: Char);
begin
   if Assigned(FonKeyPressParent) then FonKeyPressParent(Self,Key);
end;
procedure TCustomFlyInplaceEdit.DblClickParent;
begin
   if Assigned(FonDblClickParent) then FonDblClickParent(Self);
end;


{TFlyInplaceEdit}
constructor TFlyInplaceEdit.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  FActiveList := nil;
  FButtonWidth := GetSystemMetrics(SM_CXVSCROLL);
  FEditStyle := etsSimple;
end;

destructor TFlyInplaceEdit.Destroy;
begin

    //if FPickList <> nil then
    //begin
    //   Fpicklist.Free;
    //end;
    CloseUp(false);
    Parent:=Fcontrol;//not remove
    if FActiveList<>nil then
       FActiveList.Free;
   inherited;
end;

procedure TFlyInplaceEdit.BoundsChanged;
var
  R: TRect;
begin
  //Parent := Fcontrol;
  SetRect(R, 2, 2, Width - 2, Height);
  if FEditStyle <> etsSimple then
   {$IFDEF DFS_DELPHI_4_UP}
    if not Parent.UseRightToLeftAlignment then
      Dec(R.Right, FButtonWidth)
    else
      Inc(R.Left, FButtonWidth - 2);
   {$ELSE}
   Dec(R.Right, FButtonWidth);
   {$endif}
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@R));
  SendMessage(Handle, EM_SCROLLCARET, 0, 0);
  {$IFDEF DFS_DELPHI_4_UP}
  if SysLocale.FarEast then
    SetImeCompositionWindow(Font, R.Left, R.Top);
  {$endif}
end;

procedure TFlyInplaceEdit.CloseUp(Accept: Boolean);
var
  ListValue : string;
  aCancel : boolean;
begin
    Parent := Fcontrol;
    if GetCapture <> 0 then
       SendMessage(GetCapture, WM_CANCELMODE, 0, 0);

      if HandleAllocated then
      if (FActiveList<>nil) and (FPicklist=FActiveList) and (EditStyle = etsPickList)
      and FPicklist.HandleAllocated then
      begin
      if (FPickList.ItemIndex <> -1) and (FPickList.ItemIndex < FPickList.items.Count) then
        ListValue := FPickList.Items[FPicklist.ItemIndex];
      end
      else if (FActiveList<>nil) and (FPicklist=FActiveList) and (EditStyle = etsCheckList) and
      FPicklist.HandleAllocated then
        ListValue := TISPopupCheckListbox(FPickList).strValue
      else
        Listvalue:=Text;

    if FListVisible then
    begin
       SetWindowPos(FActiveList.Handle, 0, 0, 0, 0, 0, SWP_NOZORDER or
      SWP_NOMOVE or SWP_NOSIZE or SWP_NOACTIVATE or SWP_HIDEWINDOW);
      FListVisible := False;
      {if FActiveList<>nil then
      begin
        if EditStyle = etsPickList then
          FActiveList.Free
        else
        if EditStyle = etsCheckList then
           TISPopupCheckListbox(FActiveList).Free;

      FActiveList:=nil;
     end;}
    end;

    if Accept then
        if (not VarIsNull(ListValue)) and not ReadOnly then
        begin
          FCancelData:=true;
          //Tree.AValidateNodeData(Tree.Selected,ListValue);
          FValidating:=true;
          try
            ValidateData(ListValue, aCancel);
          finally
            FValidating := False;
          end;
          if HandleAllocated then
             Text:=Listvalue;
        end;
  //end;

end;

procedure TFlyInplaceEdit.DoDropDownKeys(var Key: Word; Shift: TShiftState);
var ListValue : string;
begin
  //Parent := Fcontrol;
  inherited;
  case Key of
    VK_UP, VK_DOWN:
      if ssAlt in Shift then
      begin
        if FListVisible then
           CloseUp(True)
        else
           DropDown;
        Key := 0;
      end
      else if FListVisible then
      begin
         {SendMessage(FActiveList.Handle, WM_LBUTTONDOWN, 0, Integer(MousePos));}
           ListValue:='';
           if EditStyle = etsPickList then
           begin
                if FPickList.ItemIndex <> -1 then
                   ListValue := FPickList.Items[FPicklist.ItemIndex];

           end
           else if EditStyle = etsCheckList then
           begin
                ListValue := TISPopupCheckListbox(FPickList).StrValue;
           end;
           Text := ListValue;
      end;


    VK_RETURN, VK_ESCAPE:
      if FListVisible and not (ssAlt in Shift) then
      begin
        CloseUp(Key = VK_RETURN);
        //HideEditor;
        Key := 0;
      end;
  end;
end;

procedure TFlyInplaceEdit.DropDown;
var
  P: TPoint;
  aitemheight,I,J,Y: Integer;
  //aNode:TTreeNode;
  aStrings : TStringList;
  apwdChar : Char;
  aText : string;
  //amaxlen : integer;
begin
  //Parent := Fcontrol;
  PrepareActiveList;
  if not FListVisible and Assigned(FActiveList) then
  begin
    FActiveList.Width := Width;
    //aNode:=Tree.Selected;
    if EditStyle = etsPickList then
    begin
        TISPopupListbox(FPickList).Color := Color;
        TISPopupListbox(FPickList).Font := Font;
        TISPopupListbox(FPickList).SizeGrip := DropdownSizable;
        if DropDownsizable then
        begin
           TISPopupListbox(FPickList).IntegralHeight := False;
        end;
    end
    else if EditStyle = etsCheckList then
    begin
       TISPopupCheckListbox(FPickList).Color := Color;
       TISPopupCheckListbox(FPickList).Font := Font;
       TISPopupCheckListbox(FPickList).SizeGrip := DropdownSizable;
       if DropDownsizable then
       begin
           TISPopupCheckListbox(FPickList).IntegralHeight := False;
       end;
    end;

      aStrings:=TStringList.create;
      try
        //Tree.aGetPicklist(aNode,Tstrings(astrings));
        GetPickList(Tstrings(astrings));
        fPickList.items.assign(Tstrings(astrings));
      finally
        if aStrings<> nil then aStrings.Free;
      end;
      aItemHeight:=0;
      if EditStyle = etsPickList then
          aitemheight:=TISPopupListbox(FPickList).ItemHeight
      else if EditStyle = etsCheckList then
          aitemheight:=TISPopupCheckListbox(FPickList).ItemHeight;

      if FPickList.Items.Count >= 8 then
        FPickList.Height := 8 * aItemHeight + 4
      else
        FPickList.Height := FPickList.Items.Count * aItemHeight + 4;

      if DropdownSizable then
        FPickList.Height := FPickList.Height+ GetSystemMetrics(SM_CYVSCROLL);


        GetEditData(atext,apwdChar);
        FPickList.ItemIndex := FPickList.Items.IndexOf(atext);

        if EditStyle = etsPickList then
          FPickList.ItemIndex := FPickList.Items.IndexOf(atext)
        else if EditStyle = etsCheckList then
          TISPopupCheckListbox(FPickList).StrValue:=atext;

      J := FPickList.ClientWidth;
      for I := 0 to FPickList.Items.Count - 1 do
      begin
        Y := FPickList.Canvas.TextWidth(FPickList.Items[I]);
        if Y > J then J := Y;
      end;
      FPickList.ClientWidth := J+8;
{    end;}
    P := Parent.ClientToScreen(Point(Left, Top));
    Y := P.Y + Height;
    if Y + FActiveList.Height > GetSystemMetrics(SM_CYSCREEN) then Y := P.Y - FActiveList.Height;
    SetWindowPos(FActiveList.Handle, HWND_TOP, P.X, Y, 0, 0,
      SWP_NOSIZE or SWP_NOACTIVATE or SWP_SHOWWINDOW);
    FListVisible := True;
    Invalidate;
    Windows.SetFocus(Handle);
  end;
end;

type
  TWinControlCracker = class(TWinControl) end;

//procedure TFlyInplaceEdit.CreateWnd;
//begin
//  inherited CreateWnd;
//end;

procedure TFlyInplaceEdit.KeyDown(var Key: Word; Shift: TShiftState);
  var mykey : word;
      MyShift: TShiftState;
procedure KillMessage(Wnd: HWnd; Msg: Integer);
  // Delete the requested message from the queue, but throw back
  // any WM_QUIT msgs that PeekMessage may also return
 var
  M: TMsg;

 begin
  M.Message := 0;
  if PeekMessage(M, Wnd, Msg, Msg, pm_Remove) and (M.Message = WM_QUIT) then
    PostQuitMessage(M.wparam);
 end;

begin
  //Parent := Fcontrol;
  GetClickKey(MyKey,MyShift);
  if (EditStyle = etsEllipsis) and (Key = MyKey) and (shift=MyShift) then
  begin
    EditButtonClick;
    //Parent := Fcontrol;
    CloseUp(false);
    KillMessage(Handle, WM_CHAR);
  end
  else
    inherited KeyDown(Key, Shift);
end;

procedure TFlyInplaceEdit.ListMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 if (Button = mbLeft)  then //(EditStyle <> etsCheckList) then
    CloseUp(PtInRect(FActiveList.ClientRect, Point(X, Y)));
end;

procedure TFlyInplaceEdit.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if (Button = mbLeft) and (FEditStyle <> etsSimple) and
    OverButton(Point(X,Y)) then
  begin
    if FListVisible then
    begin
      //if  (EditStyle <> etsCheckList) then
      CloseUp(true);
    end
    else
    begin
      MouseCapture := True;
      FTracking := True;
      TrackButton(X, Y);
      //if Assigned(FActiveList) then
      DropDown;
    end;
  end;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TFlyInplaceEdit.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  ListPos: TPoint;
  MousePos: TSmallPoint;
begin
  //Parent := Fcontrol;
  if FTracking then
  begin
    TrackButton(X, Y);
    if FListVisible then
    begin
      ListPos := FActiveList.ScreenToClient(ClientToScreen(Point(X, Y)));
      if PtInRect(FActiveList.ClientRect, ListPos) then
      begin
        StopTracking;
        MousePos := PointToSmallPoint(ListPos);
        SendMessage(FActiveList.Handle, WM_LBUTTONDOWN, 0, Integer(MousePos));
        Exit;
      end;
    end;
  end;
  inherited MouseMove(Shift, X, Y);
end;

procedure TFlyInplaceEdit.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  WasPressed: Boolean;
begin
  //Parent := Fcontrol;
  WasPressed := FPressed;
  StopTracking;
  if (Button = mbLeft) and WasPressed then
  begin
   if  (FEditStyle = etsEllipsis) then
      EditButtonClick
   else if ((Feditstyle = etsChecklist) or (Feditstyle = etsPickList) or FlistVisible)
   and OverButton(Point(X,Y)) then
    begin
      //MouseCapture := True;
      //FTracking := true;
      //TrackButton(X, Y);
      StopTracking;
    end
    else
      CloseUp(false);
  end;
  //Parent := Fcontrol;

  inherited MouseUp(Button, Shift, X, Y);
end;

procedure TFlyInplaceEdit.PaintWindow(DC: HDC);
var
  R: TRect;
  DrawState, Flags: Integer;
  W, X, Y: Integer;
begin
  //Parent := Fcontrol;
  if FEditStyle <> etsSimple then
  begin
    R := ButtonRect;
    //if Width >=r.right then Width:=r.left;
    Flags := 0;
    if FEditStyle in [etsPickList,etsCheckList] then
    begin
      //if FActiveList = nil then
      //  Flags := DFCS_INACTIVE
      //else
      if FPressed then
        Flags := DFCS_FLAT or DFCS_PUSHED;
      DrawFrameControl(DC, R, DFC_SCROLL, Flags or DFCS_SCROLLCOMBOBOX);
    end
    else if FEditStyle = etsEllipsis then
    begin
      if FPressed then Flags := BF_FLAT;
      DrawEdge(DC, R, EDGE_RAISED, BF_RECT or BF_MIDDLE or Flags);
      X := R.Left + ((R.Right - R.Left) shr 1) - 1 + Ord(FPressed);
      Y := R.Top + ((R.Bottom - R.Top) shr 1) - 1 + Ord(FPressed);
      W := FButtonWidth shr 3;
      if W = 0 then W := 1;
      PatBlt(DC, X, Y, W, W, BLACKNESS);
      PatBlt(DC, X - (W * 2), Y, W, W, BLACKNESS);
      PatBlt(DC, X + (W * 2), Y, W, W, BLACKNESS);
    end
    else if FEditStyle = etscheckBox then
    begin
       R:=CheckRect;
       case FState of
         cbChecked:
           DrawState := DFCS_BUTTONCHECK or DFCS_CHECKED;
         cbUnchecked:
           DrawState := DFCS_BUTTONCHECK;
         else // cbGrayed
           DrawState := DFCS_BUTTON3STATE or DFCS_CHECKED;
       end;
       DrawFrameControl(Handle, R, DFC_BUTTON, DrawState);
    end;

    ExcludeClipRect(DC, R.Left, R.Top, R.Right, R.Bottom);
  end;
  inherited PaintWindow(DC);
end;

procedure TFlyInplaceEdit.PrepareActiveList;
begin
  if FActiveList<> nil then FActiveList.Free;
  case FEditStyle of
      etsCheckList:
      begin
        FPickList := TISPopupCheckListbox.Create(Self);
        TISPopupCheckListbox(FPickList).Flat:=False;
        FPickList.Visible := False;
        FPickList.Parent := Self;
        TISPopupCheckListbox(FPickList).OnMouseUp := ListMouseUp;
        TISPopupCheckListbox(FPickList).AllowGrayed:=false;
        FActiveList := FPickList;
      end;
    etsPickList:
      begin
        FPickList := TISPopupListbox.Create(Self);
        FPickList.Visible := False;
        FPickList.Parent := Self;
        TISPopupListbox(FPickList).OnMouseUp := ListMouseUp;
        FActiveList := FPickList;
      end;
  else  { cbsNone, cbsEllipsis, or read only field }
    FActiveList := nil;
  end;
end;

procedure TFlyInplaceEdit.SetEditStyle(Value: TEditStyle);
 var aReadOnly : boolean;
begin
  FEditStyle := Value;
  aReadOnly:=false;
  GetReadOnly(aReadOnly);
  Self.ReadOnly := aReadOnly;
  Repaint;
end;

procedure TFlyInplaceEdit.StopTracking;
begin
  if FTracking then
  begin
    TrackButton(-1, -1);
    FTracking := False;
    MouseCapture := False;
  end;
end;

procedure TFlyInplaceEdit.TrackButton(X,Y: Integer);
var
  NewState: Boolean;
  R: TRect;
begin
  //Parent := Fcontrol;
  R := ButtonRect;
  NewState := PtInRect(R, Point(X, Y));
  if FPressed <> NewState then
  begin
    FPressed := NewState;
    InvalidateRect(Handle, @R, False);
  end;
end;

procedure TFlyInplaceEdit.UpdateContents;
var
  agetStyle,NewStyle: TEditStyle;
begin
  //Parent := Fcontrol;
  ReadOnly:=false;
  NewStyle := etsSimple;
  agetStyle:= etsSimple;
  //Tree.aGetEditStyle(Tree.Selected,agetStyle);
  GetEditStyle(aGetStyle);
  case agetStyle of
   etsEllipsis,
   etscheckList,
   etsPickList,
   etsCheckBox: NewStyle := aGetStyle;
  end;
  EditStyle := NewStyle;
  inherited UpdateContents;
end;

procedure TFlyInplaceEdit.CancelMode(Sender : Tobject);
begin
  //Parent := Fcontrol;
  if Editstyle=etsSimple then
     closeup(true)
  else
  if (Sender <> Self) and (Sender <> FActiveList)  then
   if FActiveList=nil then
    CloseUp(false);
end;

procedure TFlyInplaceEdit.CMCancelMode(var Message: TCMCancelMode);
begin
  CancelMode(Message.Sender);
end;

procedure TFlyInplaceEdit.WMCancelMode(var Message: TMessage);
begin
  StopTracking;
  inherited;
end;

procedure TFlyInplaceEdit.WMKillFocus(var Message: TMessage);
begin
  inherited;
  {$IFDEF DFS_DELPHI_4_UP}
  //modified -501
  {if not SysLocale.FarEast then inherited
  else
  begin
    ImeName := Screen.DefaultIme;
    ImeMode := imDontCare;
    inherited;
    if HWND(Message.WParam) <> Fcontrol.Handle then
    begin
      ActivateKeyboardLayout(Screen.DefaultKbLayout, KLF_ACTIVATE);
    end;
  end;}
  {$endif}
if not FValidating then
begin
  if Modified then
  begin
//     UpdateData;
//     UpdateEdit;
     CloseUp(True);
  end
  else
  begin
   CloseUp(False);
   //HideEditor;
  end;
  HideEditor;
 end;
end;

function TFlyInplaceEdit.CheckRect : TRect;
var VOffset, HOffset : integer;
begin
  HOffset:=(Width-FcheckWidth) div 2;
  Voffset:=(Height-FCheckheight) div 2;
  Result := Rect(Hoffset, VOffset, Hoffset+FcheckWidth, VOffset+FCheckHeight);
end;

function TFlyInplaceEdit.ButtonRect: TRect;
begin
 {$IFDEF DFS_DELPHI_4_UP}
 if UseRightToLeftAlignment then
   Result := Rect(0, 0, FButtonWidth, Height)
 else
 {$ENDIF}
    Result := Rect(Width - FButtonWidth, 0, Width, Height);//Result := Rect(0, 0, FButtonWidth, Height);
end;

function TFlyInplaceEdit.OverButton(const P: TPoint): Boolean;
begin
  Result := PtInRect(ButtonRect, P);
end;

procedure TFlyInplaceEdit.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
  with Message do
  if (FEditStyle <> etsSimple) and OverButton(Point(XPos, YPos)) then
    Exit;
  inherited;
end;

procedure TFlyInplaceEdit.WMPaint(var Message: TWMPaint);
begin
  PaintHandler(Message);
end;

procedure TFlyInplaceEdit.WMSetCursor(var Message: TWMSetCursor);
var
  P: TPoint;
begin
  GetCursorPos(P);
  P := ScreenToClient(P);
  if (FEditStyle <> etsSimple) and OverButton(P) then
    Windows.SetCursor(LoadCursor(0, idc_Arrow))
  else
    inherited;
end;

procedure TFlyInplaceEdit.WndProc(var Message: TMessage);
function KeyDataToShiftState(KeyData: Longint): TShiftState;
const
  AltMask = $20000000;
begin
  Result := [];
  if GetKeyState(VK_SHIFT) < 0 then Include(Result, ssShift);
  if GetKeyState(VK_CONTROL) < 0 then Include(Result, ssCtrl);
  if KeyData and AltMask <> 0 then Include(Result, ssAlt);
end;
begin
  case Message.Msg of
    wm_KeyDown, wm_SysKeyDown, wm_Char:
      if EditStyle in [etsPickList,etsCheckList] then
      with TWMKey(Message) do
      begin
        DoDropDownKeys(CharCode, KeyDataToShiftState(KeyData));
        if (CharCode <> 0) and FListVisible then
        begin
          with TMessage(Message) do
            SendMessage(FActiveList.Handle, Msg, WParam, LParam);
            DoDropDownKeys(CharCode, KeyDataToShiftState(KeyData));
            Exit;
        end;
      end
  end;
  inherited WndProc(Message);
end;


//-------TISPopupCheckListbox-------------------------------------------
procedure TISPopupCheckListbox.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or WS_BORDER;
    ExStyle := WS_EX_TOOLWINDOW;
    WindowClass.Style := CS_SAVEBITS;
  end;
end;

procedure TISPopupCheckListbox.CreateWnd;
begin
  inherited CreateWnd;
  Windows.SetParent(Handle, 0);
  CallWindowProc(DefWndProc, Handle, wm_SetFocus, 0, 0);
end;

procedure TISPopupCheckListbox.SetStrValue(Value : string);
var
    i,j: integer;
    aTestList : TStringList;
begin
    if value='' then exit;
    aTestList:=TStringList.Create;
    try
     aTestList.CommaText:=Value;
     for i:=0 to aTestList.Count-1 do
     begin
      for j:=0 to items.count-1 do
       if items[j]=aTestList[i] then
       begin
         Checked[j]:=true;
         break;
       end;

     end;
    finally
      aTestList.Free;
    end;
end;

function TISPopupCheckListbox.GetStrValue : string;
var
    atestList : Tstringlist;
    i : integer;
begin
     Result:='';
     atestList :=Tstringlist.Create;
     try
        for i:=0 to Items.Count-1 do
        if Checked[i] then
        begin
           atestList.Add(items[i]);
        end;
        Result :=atestList.Commatext;
     finally
       atestList.Free
     end;
end;

{$IFDEF DFS_DELPHI_4_UP}
function TISPopupCheckListbox.UseRightToLeftAlignment: Boolean;
begin
  if Parent <> nil then
   Result := Parent.UseRightToLeftAlignment
  else Result := inherited UseRightToLeftAlignment;
end;
{$ENDIF}

procedure TISPopupCheckListbox.Keypress(var Key: Char);
var
  TickCount: Integer;
begin
 if (Key <> ' ') then
  case Key of
    #8, #27: FSearchText := '';
    #32..#255:
      begin
        TickCount := GetTickCount;
        if TickCount - FSearchTickCount > 2000 then FSearchText := '';
        FSearchTickCount := TickCount;
        if Length(FSearchText) < 32 then FSearchText := FSearchText + Key;
        SendMessage(Handle, LB_SelectString, WORD(-1), Longint(PChar(FSearchText)));
        Key := #0;
      end;
  end;
  inherited Keypress(Key);
  TFlyInplaceEdit(Owner).text:=strValue;
end;

procedure TISPopupCheckListbox.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
  var FOnMouseUp_f : TMouseEvent;
  //Index : integer;
begin
  FOnMouseUp_f :=nil;
  //Index := ItemAtPos(Point(X,Y),True);
      if (X > 0) and (X < GetCheckWidth) then
      begin
         FOnMouseUp_f :=OnMouseUp;
         OnMouseUp :=nil;
         TFlyInplaceEdit(Owner).text:=strValue;
      end;
  inherited MouseUp(Button, Shift, X, Y);

  if Assigned(FOnMouseUp_f) then
      OnMouseUp:=FOnMouseUp_f;
end;

initialization
   GetCheckSize;
finalization

end.

