unit ISCalendar;
// popup  
{$R *.res}
{$I DFS.INC}
{$WARNINGS OFF}
interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, ExGrid, Math;

const
  InitRepeat = 400;
  RepeatPause     = 100;

type

  TISDayOfWeek = (doMonday, doTuesday, doWednesday, doThursday,
    doFriday, doSaturday, doSunday, doLocaleDefault);
  TTypeOfCell = ( dtHeading, dtWeekNo, dtTrailingday, dtDay, dtMonth );

  // Options for Calendar-Control
  TISCalendarOption  =
  (coAllowViewChange,
   coPerformTrailingDay,
   coShowTrailingDays,
   coAcceptHoliday,
   coShowFrame,
   coShowWeekNumbers,
   coShowNavButtons,
   coShowTitle,
   coShowHorzLine,
   coShowVertLine);

  TISCalendarOptions = set of TISCalendarOption;

  TISCalendarView = (cvMonthView, cvYearView);

  TDrawCalendarCellEvent = procedure (Sender: TObject; AText: string;
         Col, Row: Longint; Rect: TRect; State: TExGridDrawState;
         AType: TTypeOfCell) of object;

  TCellDisplayStyle = (dsFlat, dsRaised, dsSunken);

  TISCalendarColors = class;

  TISCalendar = class(TCustomExGrid)
  private
    FTimer : integer;
    FSizing, FInit, LeftBEnabled, RightBEnabled, FRightPressed, FLeftPressed : boolean;
    FTitleText : string;
    FFrameWidth, FNonClientHeight, FStartOfWeek : integer;
    FDefaultDOW : boolean;
    FSelectedStyle, FFixedStyle, FTodayStyle, FCalendarStyle : TCellDisplayStyle;
    FDate : TDateTime;
    FOptions: TISCalendarOptions;
    FViewStyle: TISCalendarView;
    FCalendarColors: TISCalendarColors;
    FOnViewChange: TNotifyEvent;
    FOnChange: TNotifyEvent;
    FUpdating: Boolean;
    FMonthOffset: Integer;
    FReadOnly: Boolean;
    FSkipMouseUp : Boolean;
    FUseCurrentDate : Boolean;
    FAutosize: boolean;

    //
    procedure CalcAutosizeBounds(var aWidth, aHeight : integer);
    procedure SetCalendarColors(Value: TISCalendarColors);
    procedure SetCalendarView(Value: TISCalendarView);
    procedure SetOptions(Value: TISCalendarOptions);
    procedure SetTitleText;
    procedure SetColor(Index: Integer; Value: TColor);
    function GetOptions: TISCalendarOptions;
    function GetDateElement(Index: Integer): Integer;
    function GetCalendarView: TISCalendarView;
    procedure SetCalendarDate(Value: TDateTime);
    function GetCalendarDate: TDateTime;
    procedure SetStartOfWeek(Value: TISDayOfWeek);
    function GetStartOfWeek : TISDayOfWeek;
    procedure DoChange(Sender: TObject);
    procedure DoViewChange(Sender: TObject);
    procedure SetDateElement(Index: Integer; Value: Integer);
    function GetDayNum(ACol, ARow: Integer): Integer;
    function GetCalendarWeek: Integer;
    function CalcCalendarWeek(ADate: TDateTime): Integer;
    procedure SetCellDimensions;
    procedure CalendarUpdate(DayOnly: Boolean);
    procedure ChangeMonth(Delta: Integer);
    function GetCellText(ACol, ARow: Integer): string;
    procedure SetFrameWidth(const Value: integer);

    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure WMNCHittest (var Message: TMessage); message WM_NCHITTEST;
    procedure WMNCCalcSize (var Message: TWMNCCalcSize); message WM_NCCALCSIZE;
    procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
    procedure WMTimer(var Msg: TWMTimer); message WM_TIMER;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure SetAutoSize(const Value: boolean);
    procedure SetDisplaysStyle(const Index: Integer;
      const Value: TCellDisplayStyle);
  protected
    procedure DrawFixedFrame(aCanvas : TCanvas; aRect : TRect; aCol, aRow : integer; Down : Boolean); override;
    function AllowDesignSizing : boolean; override;
    procedure CreateHandle; override;
    procedure Change; dynamic;
    procedure Click; override;
//    procedure DblClick; override;
    procedure Loaded; override;
    procedure DrawCell(aCanvas : TCanvas; ACol, ARow: Longint; ARect: TRect;
      AState: TExGridDrawState; Var Continue : boolean); override;
    function SelectCell(var ACol, ARow: Longint; MoveAnchor : Boolean;  Shift: TShiftState; SelWay : TSelectionWay): Boolean; override;
    procedure FocusCell(ACol, ARow: Longint; MoveAnchor: Boolean;  Shift: TShiftState; SelWay : TSelectionWay); override;
    function DaysThisMonth: Integer; virtual;
    function DaysPerMonth(AYear, AMonth: Integer): Integer; virtual;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
         X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton;Shift: TShiftState; X, Y: Integer); override;
    procedure DoDrawCell(Sender: TObject; AText: string;
          ACol, ARow: Longint; ARect: TRect; AState: TExGridDrawState;
          AType: TTypeOfCell); dynamic;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure AdjustAutoSize;
    procedure NextMonth;
    procedure NextYear;
    procedure PrevMonth;
    procedure PrevYear;

    procedure UpdateCalendar; virtual;
    function IsLeapYear(AYear: Integer): Boolean;
    function IsTrailingDay(ACol, ARow: Integer): Boolean;
    function IsCalendarWeek(ACol, ARow: Integer): Boolean;
    function GetCellType(ACol, ARow: Integer): TTypeOfCell;
    function IsToday(ACol, ARow: Integer): Boolean;
    function IsWeekend(ACol, ARow: Integer): Boolean;
    function GetDateFromCell(ACol, ARow: Integer): TDateTime;
    property CellText[ACol, ARow: Integer]: string read GetCellText;
    property WeekNumber: Integer read GetCalendarWeek;
  published
    property DoubleBuffered default true;
    property Date: TDateTime read GetCalendarDate write SetCalendarDate;
    property Align;
    property BorderStyle;
    property FrameWidth : integer read FFrameWidth write SetFrameWidth default 0;
    property Ctl3D;
    property Enabled;
    property Font;
    property ParentFont;
    property ShowHint;
    property StartOfWeek: TISDayOfWeek read GetStartOfWeek write SetStartOfWeek default doLocaleDefault;
    property Day: Integer index 3 read GetDateElement write SetDateElement stored False;
    property Month: Integer index 2 read GetDateElement write SetDateElement stored False;
    property Year: Integer index 1 read GetDateElement write SetDateElement stored False;
    property Colors: TISCalendarColors read FCalendarColors write SetCalendarColors;
    property ViewStyle: TISCalendarView read GetCalendarView write SetCalendarView
                default cvMonthView;
    property Options: TISCalendarOptions read GetOptions write SetOptions
      default [coShowTrailingDays,  coAcceptHoliday,
               coPerformTrailingDay, coAllowViewChange,
               {coShowCellFrames, coShowFrame,} coShowNavButtons,  coShowTitle];
    property OnViewChange: TNotifyEvent read FOnViewChange write FOnViewChange;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property FixedStyle : TCellDisplayStyle index 0 read FFixedStyle write SetDisplaysStyle default dsFlat;
    property TodayStyle : TCellDisplayStyle  index 1 read FTodayStyle write SetDisplaysStyle default dsFlat;
    property CalendarStyle : TCellDisplayStyle  index 2 read FCalendarStyle write SetDisplaysStyle default dsFlat;
    property SelectedStyle : TCellDisplayStyle  index 3 read FSelectedStyle write SetDisplaysStyle default dsFlat;
    property AutoSize : boolean read FAutosize write SetAutosize default true;
    Property OnKeyDown;
    property OnKeyUp;
    property OnKeyPress;
    property OnMouseDown;
    property OnMouseUp;
    property OnMouseMove;
    property OnDblClick;
    property OnClick;
  end;

  TISCalendarColors = class(TPersistent)
  private
    Owner: TISCalendar;
    FBackColor: TColor;
    FTextColor: TColor;
    FTitleBackColor: TColor;
    FTitleTextColor: TColor;
    FMonthBackColor: TColor;
    FTrailingTextColor: TColor;
    FTodaysColor: TColor;
    FTodaysBorderColor: TColor;
    FTodaysTextColor: TColor;
    FSelectedBorderColor: TColor;
    FSelectedColor: TColor;
    FSelectedTextColor: TColor;
    FFrameColor, FLineColor : TColor;
    procedure SetColor(Index: Integer; Value: TColor);
    procedure SetAllColors;
  public
    constructor Create(AOwner: TISCalendar);
    procedure Assign(Source: TPersistent); override;
  published
    property BackColor: TColor index 0 read FBackColor
                        write SetColor default clWindow;
    property TextColor: TColor index 1 read FTextColor
                        write SetColor default clWindowText;
    property TitleBackColor: TColor index 2 read FTitleBackColor
                        write SetColor default clBtnFace;
    property TitleTextColor: TColor index 3 read FTitleTextColor
                        write SetColor default clWindowText;
    property MonthBackColor: TColor index 4 read FMonthBackColor
                        write SetColor default clWhite;
    property TrailingTextColor: TColor index 5 read FTrailingTextColor
                        write SetColor default clGrayText;
    property TodaysColor: TColor index 6 read FTodaysColor write SetColor
                        default clInfobk;
    property TodaysBorderColor: TColor  index 7 read FTodaysBorderColor
                        write SetColor default clWindowFrame;
    property TodaysTextColor: TColor index 8 read FTodaysTextColor
                        write SetColor default clActiveCaption;   
    property SelectedBorderColor: TColor  index 9 read FSelectedBorderColor
                        write SetColor default clGray;
    property SelectedColor: TColor  index 10 read FSelectedColor
                        write SetColor default clGray;
    property SelectedTextColor: TColor  index 11 read FSelectedTextColor
                        write SetColor default clHighlightText;
    property LineColor: TColor  index 12 read FLineColor
                        write SetColor default clSilver;
    property FrameColor: TColor  index 13 read FFrameColor
                        write FFrameColor default clWindowFrame;
  end;



implementation

{$IFNDEF Registered}
uses FTAbout;
{$ENDIF}

{ TISCalendar }
constructor TISCalendar.Create( AOwner: TComponent );
begin
  inherited Create(AOwner);

  FFixedStyle := dsFlat;
  FTodayStyle := dsFlat;
  FCalendarStyle := dsFlat;
  DoubleBuffered := true;

  FAutoSize := true;
  FFramewidth := 0;
  FViewStyle := cvMonthView;
  FDate := SysUtils.Date;
  FCalendarColors := TISCalendarColors.Create(Self);
  FCalendarColors.SetAllColors;
  LeftBEnabled := true;
  RightBEnabled := true;
  FTimer := -1;
  ScrollBars := ssNone;
  GridLineWidth := 1;
  FOptions := [coShowTrailingDays,  coAcceptHoliday,
               coPerformTrailingDay, coAllowViewChange,
                coShowNavButtons,  coShowTitle];

  StartOfWeek := doLocaleDefault;
  DoChange(Self);
  AdjustAutoSize;

  inherited Options := inherited Options  + [goDrawFocusSelected] - [goColSizing, goColMoving, goRowSizing, goRowMoving, goRangeSelect, goVertLine, goHorzLine, goFixedVertLine, goFixedHorzLine];

end;

procedure TISCalendar.Loaded;
begin
  inherited;
  UpdateCalendar;
end;

destructor TISCalendar.Destroy;
begin
   FCalendarColors.Free;
   inherited;
end;

procedure TISCalendar.WMNCHittest (var Message: TMessage);
begin
 With Message do
   Result := HTCLIENT;
end;

procedure TISCalendar.WMTimer(var Msg: TWMTimer);
begin
   if FInit then
   begin
      FInit :=False;
      Killtimer(Handle, 1);
      if FRightPressed or FLeftPressed then
        FTimer := SetTimer(Handle, 1, RepeatPause, nil);
   end;

   if Not FRightPressed and not FLeftPressed then
   begin
     Killtimer(Handle, 1);
     FTimer := -1;
   end
   else if FLeftPressed then //activecontrol
   begin
    if FViewStyle = cvMonthView then
      PrevMonth
    else
      PrevYear;
   end
   else if FRightPressed then
   begin
    if FViewStyle = cvMonthView then
      NextMonth
    else
      NextYear;
   end;
end;

procedure TISCalendar.CMFontChanged(var Message: TMessage);
begin
 inherited;
 if HandleAllocated then
 begin
   if FAutoSize then
     AdjustAutoSize;
   Perform(WM_NCPAINT,0,0);
 end;
end;



procedure TISCalendar.WMNCPaint(var Message: TMessage);
var
  DC: HDC;
  RC, BR, RW, SaveRW: TRect;
begin
  { Get window DC that is clipped to the non-client area }
    DC := GetWindowDC(Handle);
    try
      Windows.GetClientRect(Handle, RC);
      GetWindowRect(Handle, RW);
      MapWindowPoints(0, Handle, RW, 2);
      OffsetRect(RC, -RW.Left, -RW.Top);
      OffsetRect(RW, -RW.Left, -RW.Top);
      if BorderStyle <> bsNone then
      begin
         if Ctl3d then
           inflateRect(RW, -2,-2)
         else inflateRect(RW, -1,-1);
      end;
      ExcludeClipRect(DC, RC.Left, RC.Top, RC.Right, RC.Bottom);


       FillRect(DC, RW, CreateSolidBrush(ColorToRGB(Colors.BackColor)));//GetSysColorBrush(COLOR_BTNFACE));//
       if (coShowframe in Foptions) then
       begin
        BR := RC;
        if (coShowTitle in FOptions) then
          Dec(BR.Top, FNonClientHeight);
        if FFrameWidth > 0 then
          inflateRect(BR, 1,1);
        FrameRect(DC, BR, CreateSolidBrush(ColorToRGB(Colors.FrameColor)));
       end;

      if (coShowTitle in FOptions) then
      begin
       RW := Rect(RC.Left, Rc.Top - FNonClientHeight, RC.Right, rc.Top);
       FillRect(DC, RW, CreateSolidBrush(ColorToRGB(Colors.TitleBackColor)));
       DrawEdge(DC, RW, BDR_RAISEDINNER,BF_RECT);

       SelectObject(DC, Font.Handle);
       SetTextColor(DC, ColorToRGB(Colors.TitleTextColor));
       SetBkColor(DC,ColortoRGB(Colors.TitleBackColor));
       SaveRW := RW;
       InflateRect(RW,-1, -1);

       if ([coShowNavButtons] * FOptions)<> [] then
       begin
        BR := Rect(RW.Left, RW.Top, Rw.Left+GetSystemMetrics(SM_CYVSCROLL), RW.Bottom);
        DrawFrameControl(DC, BR ,DFC_SCROLL, DFCS_SCROLLLEFT+DFCS_FLAT);
        if FLeftPressed then
          FrameRect(DC, BR, GetSysColorBrush(COLOR_HIGHLIGHT))
        else
          FrameRect(DC, BR, GetSysColorBrush(COLOR_BTNFACE));
       end;
       BR := Rect(Rw.Left+GetSystemMetrics(SM_CYVSCROLL), RW.Top, RW.Right-+GetSystemMetrics(SM_CYVSCROLL), RW.Bottom);
       DrawText(DC, PChar(FTitleText), Length(FTitleText), BR,DT_CENTER+DT_SINGLELINE+DT_VCENTER);

       if ([coShowNavButtons] * FOptions)<> [] then
       begin
        BR := Rect(RW.Right-GetSystemMetrics(SM_CYVSCROLL), RW.Top, Rw.Right, RW.Bottom);
        DrawFrameControl(DC, BR ,DFC_SCROLL, DFCS_SCROLLRIGHT+DFCS_FLAT);
        if FRightPressed then
          FrameRect(DC, BR, GetSysColorBrush(COLOR_HIGHLIGHT))
        else
          FrameRect(DC, BR, GetSysColorBrush(COLOR_BTNFACE));
       end;
      end;
     finally
      ReleaseDC(Handle, DC);
    end;

 inherited;
end;

procedure TISCalendar.WMNCCalcSize(var Message: TWMNCCalcSize);
var FHeight, ArrowHeight : integer;
begin
  inherited;
  if (coShowFrame in FOptions) then
    inflaterect(Message.CalcSize_Params^.rgrc[0], -FFrameWidth, -FFrameWidth);

  if (coShowTitle in FOptions) then
  begin
   ArrowHeight := GetSystemMetrics(SM_CYVSCROLL)+4;
   FHeight :=  Canvas.TextHeight('Wg')+4;
   FNonClientHeight := Max(FHeight, ArrowHeight);
   Inc(Message.CalcSize_Params^.rgrc[0].Top, FNonClientHeight);
  end;
end;


procedure TISCalendar.DoDrawCell(Sender: TObject; AText: string;
          ACol, ARow: Longint; ARect: TRect; AState: TExGridDrawState;
          AType: TTypeOfCell);
var
  TheRect: TRect;
  ABorderColor, AFontColor, ABackColor{, AFrameColor} : TColor;
  aStyle :  TCellDisplayStyle;// = (dsFlat, dsRaised, dsSunken);
begin
  TheRect := ARect;
  aStyle := FCalendarStyle;
  ABackColor := FCalendarColors.BackColor;
//  AFrameColor := ABackColor;
  ABorderColor := ABackColor;
  AFontColor := FCalendarColors.TextColor;

  if (AType = dtTrailingday) then
    AFontColor := FCalendarColors.TrailingTextColor
  else if (AType = dtDay) then
    AFontColor := FCalendarColors.TextColor;

  // Colors for the Fixed row and col
  if (gdFixed in AState) then
  begin
    ABackColor := FCalendarColors.TitleBackColor;
//    AFrameColor := ABackColor;
    ABorderColor := FCalendarColors.TitleBackColor;
    AFontColor := FCalendarColors.TitleTextColor;
    aStyle := FFixedStyle;
  end;

  if IsToday(ACol, ARow) then
  begin
    ABackColor := FCalendarColors.TodaysColor;
    AFontColor := FCalendarColors.TodaysTextColor;
    ABorderColor := FCalendarColors.TodaysBorderColor;
    aStyle := FTodayStyle;
  end
  else if (gdSelected in AState) then
   begin
    ABackColor := FCalendarColors.SelectedColor;
    AFontColor := FCalendarColors.SelectedTextColor;
    ABorderColor := FCalendarColors.SelectedBorderColor;
    aStyle := FSelectedStyle;
  end;

  // setting font and do drawing
  Canvas.Font := Font;
  with TheRect, Canvas do
  begin
    // print the contents of cell
    Font.Color := AFontColor;
    Brush.Color := ABackColor;
    TextRect(TheRect, Left + (Right - Left - TextWidth(AText)) div 2,
      Top + (Bottom - Top - TextHeight(AText)) div 2, AText);

{    Brush.Style := bsSolid;
    if not (gdFixed in AState) then
    begin
      Brush.Color := AFrameColor;
      FrameRect(TheRect);
    end;
}
   case aStyle of
     dsRaised:
       DrawEdge(Canvas.Handle, TheRect, BDR_RAISEDINNER, BF_RECT);
     dsSunken:
       DrawEdge(Canvas.Handle, TheRect, BDR_SUNKENINNER, BF_RECT);
     dsFlat:
      begin
       InflateRect(TheRect, -1, -1);
       Brush.Color := ABorderColor;
       FrameRect(TheRect);
      end;
    end;
  end;
end;


procedure TISCalendar.SetCalendarColors(Value: TISCalendarColors);
begin
  if FCalendarColors <> Value then FCalendarColors.Assign(Value);
end;

procedure TISCalendar.SetOptions(Value: TISCalendarOptions);
var
  ToRecreateWnd : boolean;
begin
  ToRecreateWnd := false;
  if FOptions <> Value then
  begin

    if coShowHorzLine in Value then
      inherited Options := inherited Options + [goHorzLine, goFixedHorzLine]
    else
      inherited Options := inherited Options - [goHorzLine, goFixedHorzLine];

    if coShowVertLine in Value then
      inherited Options := inherited Options + [goVertLine, goFixedVertLine]
    else
      inherited Options := inherited Options - [goVertLine, goFixedVertLine];


    if ([coShowFrame] * Value) <> ([coShowFrame] * FOptions) then //(coShowFrame in Value) then
       ToRecreateWnd := True;

    if ([coShowNavButtons] * Value) <> ([coShowNavButtons] * FOptions) then
        ToRecreateWnd := True;

    if ([coShowTitle] * Value) <> ([coShowTitle] * FOptions) then
      ToRecreateWnd := True;


    FOptions := Value;
    if ToRecreateWnd then
      RecreateWnd
    else Invalidate;

    SetCellDimensions;
    Updatecalendar;
  end;
end;


function TISCalendar.GetOptions: TISCalendarOptions;
begin
  Result := FOptions;
end;

function TISCalendar.GetCalendarView: TISCalendarView;
begin
  Result := FViewStyle;
end;

procedure TISCalendar.DoViewChange(Sender: TObject);
begin
  SetTitleText;
  if Assigned(FOnViewChange) then FOnViewChange(Self);
end;

procedure TISCalendar.SetCalendarView(Value: TISCalendarView);
begin
  If FViewStyle <> Value then
  begin
    FViewStyle := Value;
    SetCellDimensions;
    UpdateCalendar;
    DoViewChange(Self);
  end;
end;

procedure TISCalendar.SetColor(Index: Integer; Value: TColor);
begin
  case Index of
    0:
    begin
      Color := Value;
      if HandleAllocated then
        Perform(WM_NCPAINT,0,0);
    end;
    2: FixedColor := Value;
    3: if HandleAllocated then
        Perform(WM_NCPAINT,0,0);
    12: LineColor := Value;
  end;
end;


procedure TISCalendar.PrevMonth;
begin
  ChangeMonth(-1);
end;

procedure TISCalendar.NextMonth;
begin
  ChangeMonth(1);
end;

procedure TISCalendar.NextYear;
begin
  if IsLeapYear(Year) and (Month = 2) and (Day = 29) then Day := 28;
  Year := Year + 1;
end;

procedure TISCalendar.PrevYear;
begin
  if IsLeapYear(Year) and (Month = 2) and (Day = 29) then Day := 28;
  Year := Year - 1;
end;


procedure TISCalendar.DoChange(Sender: TObject);
begin
  SetTitleText;
  if Assigned(FOnChange) then FOnChange(Self);
end;


procedure TISCalendar.SetTitleText;
var
  TheText: string;
begin
  if FViewStyle = cvMonthView then
    TheText := LongMonthNames[Month] + ' ' + IntToStr(Year)
  else
    TheText := IntToStr(Year);

    FTitleText := TheText;
    if HandleAllocated then
      Perform(WM_NCPAINT, 0,0);
end;

procedure TISCalendar.SetCalendarDate(Value: TDateTime);
var
  AYear, AMonth, ADay, VYear, VMonth, VDay: Word;
begin
  if Value <> FDate then begin
    DecodeDate(FDate, AYear, AMonth, ADay);
    DecodeDate(Value, VYear, VMonth, VDay);
    FDate := Value;
    CalendarUpdate((AYear = VYear) and (AMonth = VMonth));
  end;
end;

function TISCalendar.GetCalendarDate: TDateTime;
begin
  Result := FDate;
end;

procedure TISCalendar.SetStartOfWeek(Value: TISDayOfWeek);
var
  DOWFlag: Integer;
  A: array[0..1] of char;
begin
    if Value = doLocaleDefault then
    begin
      GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_IFIRSTDAYOFWEEK, A, SizeOf(A));
      DOWFlag := Ord(A[0]) - Ord('0');
      FDefaultDOW := true;
    end
    else
    begin
      DOWFlag := Ord(Value);
      FDefaultDOW := false;
    end;

    if FStartOfWeek <> dowFlag then
    begin
     FStartOfWeek := DowFlag;
     UpdateCalendar;
    end;

end;

function TISCalendar.GetStartOfWeek : TISDayOfWeek;
begin
 if FDefaultDOW then
    Result := doLocaleDefault
 else
    Result := TISDayOfWeek(FStartOfWeek);
end;

function TISCalendar.AllowDesignSizing : boolean;
begin
   Result := false;
end;

procedure TISCalendar.CreateHandle;
begin
  inherited;
  if FAutosize then
     AdjustAutosize;
end;

procedure TISCalendar.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TISCalendar.Click;
begin
  inherited Click;
end;



procedure TISCalendar.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if Shift = [] then
  begin
      case Key of
        VK_PRIOR:
        begin
          if FViewStyle = cvMonthView then
             PrevMonth//ChangeMonth(-1)
          else
             PrevYear;
          exit;
        end;
        VK_NEXT:
        begin
          if FViewStyle = cvMonthView then
            NextMonth//ChangeMonth(1);
          else
            NextYear;
          exit;
        end;
        VK_LEFT, VK_SUBTRACT:
          if FViewStyle = cvMonthView then
          begin
            if (Day > 1) then
              Day := Day - 1
            else
              FDate := FDate - 1;
            Exit;
          end
          else if Key = VK_SUBTRACT then
             PrevYear;

        VK_RIGHT, VK_ADD:
          if FViewStyle = cvMonthView then
          begin
            if (Day < DaysThisMonth) then
              Day := Day + 1
            else
              FDate := FDate + 1;
            Exit;
          end
          else if Key = VK_ADD then
             NextYear;
      end;
  end;
  inherited KeyDown(Key, Shift);
end;

{
procedure TISCalendar.DblClick;
begin
  inherited DblClick;
  if (coAllowViewChange in FOptions) and not FSkipMouseUp then
  begin
    FSkipMouseUp := True;         // avoid jumping arround
    if (FViewStyle = cvYearView) then
      ViewStyle := cvMonthView
    else
      ViewStyle := cvYearView;
    FSkipMouseUp := false;
  end;
end;
}

procedure TISCalendar.MouseDown(Button: TMouseButton; Shift: TShiftState;
X, Y: Integer);
var R,BR : TRect;
begin
 //if (ssDouble in Shift) then FSkipMouseUp := true;
 if (Y < 0) then //title
 begin
   R := Rect(0,0,Width, Height);
   if (coShowframe in FOptions) then
   begin
     InflateRect(R, -Framewidth, -FrameWidth);
     inc(Y, FrameWidth);
     inc(X, FrameWidth);
   end;
   if (coShowtitle in FOptions) then
   begin
     inc(R.Bottom, FNonClientHeight);
     inc(Y, FNonClientHeight);
   end
   else R.Bottom := R.Top;

   if LeftBEnabled then
   begin
     BR := Rect(R.Left, R.Top, R.Left+GetSystemMetrics(SM_CYVSCROLL), R.Bottom);
     if PtInrect(BR, Point(X, Y)) then
     begin
       MouseCapture := True;
       FLeftPressed := True;
       Perform(WM_NCPAINT,0,0);

       if FViewStyle = cvMonthView then
         PrevMonth
       else
         PrevYear;
       FInit := True;
       FTimer := SetTimer(Handle, 1, InitRepeat, nil);
     end;
   end;
   if RightBEnabled then
   begin
     BR := Rect(R.Right-GetSystemMetrics(SM_CYVSCROLL), R.Top, R.Right, R.Bottom);
     if PtInrect(BR, Point(X, Y)) then
     begin
       MouseCapture := True;
       FRightPressed := True;
       Perform(WM_NCPAINT,0,0);

       if FViewStyle = cvMonthView then
         NextMonth
       else
         NextYear;

       FInit := True;
       FTimer := SetTimer(Handle, 1, InitRepeat, nil);
     end;
   end;
   if (ssDouble in Shift)and not (FRightPressed or FLeftPressed) then
   begin
     if (coAllowViewChange in FOptions) and not FSkipMouseUp then
     begin
       FSkipMouseUp := True;         // avoid jumping arround
       if (FViewStyle = cvYearView) then
         ViewStyle := cvMonthView
       else
         ViewStyle := cvYearView;
       FSkipMouseUp := false;
     end;
   end;

 end
 else
  inherited;

 FSkipMouseUp := false;
end;

procedure TISCalendar.MouseUp(Button: TMouseButton;
                  Shift: TShiftState; X, Y: Integer);
var
  ACoord : TGridCoord;
begin
  inherited MouseUp(Button, Shift, X, Y);
  if (Y < 0 ) then //title
  begin
    FLeftPressed := False;
    FRightPressed := False;
    Perform(WM_NCPAINT,0,0);
  end
  else if (Button = mbLeft) and (coShowTrailingDays in FOptions) then
  begin
    ACoord := MouseCoord(X, Y);
    if IsTrailingDay(ACoord.X, ACoord.Y) and (ACoord.Y > 0) and
        (FSkipMouseUp = False) and (coPerformTrailingDay in FOptions) then
    begin
      FDate := GetDateFromCell(ACoord.X, ACoord.Y);
      UpdateCalendar;
//      Change;
    end;
  end;
  FSkipMouseUp := False;
end;


procedure TISCalendar.WMSize(var Message: TWMSize);
begin
  SetCellDimensions;
  if AutoSize then
  begin
    Message.Width := Width;
    Message.Height := Height;
  end;
  inherited;
end;

type TCrackWincontrol = class(TWincontrol);

procedure TISCalendar.CalcAutosizeBounds(var aWidth, aHeight : integer);
var Max,  i,j : integer;
    Txtsize : TSize;
    Txt : string;
    DC : HDC;
    sc : boolean;
begin
 sc := false;
 if HandleAllocated then
   DC := Canvas.Handle
 else
 begin
   DC := GetDC(0);
   sc := true;
 end;
 try
  Canvas.Font := Font;
  Max :=0;
  for i:= 0 to ColCount-1 do
  begin
   for j:=0 to Rowcount -1 do
   begin
      Txt := GetCellText(i, j);
      GetTextExtentPoint32(DC, Pchar(Txt), Length(Txt), TxtSize);
      //Cur := Canvas.TextWidth(GetCellText(i, j));
      if TxtSize.CX > Max then Max := TxtSize.CX;
   end;
  end;

  //Cur := Canvas.TextHeight('Yg');//row height
  GetTextExtentPoint32(DC, Pchar('Yg'), Length('Yg'), TxtSize);
  if Txtsize.CY > Max then Max := Txtsize.CY;

  Txtsize.CX := Max;
  Txtsize.CY := Max;
  if (coShowVertLine in FOptions) then
    inc(Txtsize.CX, GridLinewidth);
  Inc(Txtsize.CX, 4);

  if (coShowHorzLine in FOptions) then
    inc(Txtsize.CY, GridLinewidth);
  Inc(Txtsize.CY, 4);

  aWidth := ((Txtsize.CX  + 4) * Colcount);
  aHeight := ((Txtsize.CY  + 4) * Rowcount);

 finally
   if sc then
    ReleaseDC(0, DC);
 end;
end;

procedure TISCalendar.AdjustAutoSize;
 var HAdd, WAdd : integer;
begin

  CalcAutosizeBounds(Wadd, HAdd);
  if (coShowFrame in FOptions) then
  begin
     Inc(HAdd, 2*FFrameWidth);
     Inc(WAdd, 2*FFrameWidth);
  end;
  if (coShowTitle in FOptions) then
    inc(Hadd, FNonClientheight);

  If BorderStyle <> bsNone then
  begin
    if Ctl3d then
    begin
      inc(Wadd, 4);
      inc(Hadd, 4);
    end
    else
    begin
      inc(Wadd, 2);
      inc(Hadd, 2);
    end;

  end;
  FSizing := true;
  SetBounds(Left, Top, WAdd, HAdd);
  FSizing := false;

end;


procedure TISCalendar.DrawCell(aCanvas : TCanvas; ACol, ARow: Longint; ARect: TRect;
      AState: TExGridDrawState; Var Continue : boolean);//(ACol, ARow: Longint; ARect: TRect; AState: TExGridDrawState);
var
  AText : string;
begin
  AText := CellText[ACol, ARow];

  DoDrawCell(Self, AText, ACol, ARow, ARect, AState, GetCellType(ACol, ARow));
end;

procedure TISCalendar.FocusCell(ACol, ARow: Longint; MoveAnchor: Boolean;  Shift: TShiftState; SelWay : TSelectionWay);
var
  TheCellText: string;
  AMonth : Integer;
begin
  inherited;
  if not FUpdating then
  begin
    TheCellText := CellText[Col, Row];
    if (FViewStyle = cvMonthView) then
    begin
      if TheCellText <> '' then
      begin
        Day := StrToInt(TheCellText);
      end;
     end
     else
     begin
       AMonth := (Row * 4) + Col + 1;
       if Day > DaysPerMonth(Year, AMonth) then
         Day := DaysPerMonth(Year, AMonth);
       Month := AMonth;
    end;
  end;
end;

function TISCalendar.SelectCell(var ACol, ARow: Longint; MoveAnchor : Boolean;  Shift: TShiftState; SelWay : TSelectionWay): Boolean;//(ACol, ARow: Longint): Boolean;
begin
  if ((not FUpdating) and FReadOnly) or (CellText[ACol, ARow] = '')
       or (IsTrailingDay(ACol, ARow)) then
    Result := False
  else
    Result := inherited SelectCell(ACol, ARow, Moveanchor, Shift, SelWay);
end;

function TISCalendar.CalcCalendarWeek(ADate: TDateTime): Integer;
var
  AYear, AMonth, ADay: Word;
  AWeekDay : Word;
  ANumDays : Word;
  AFirstDayOfYear : TDateTime;
begin
  try
    DecodeDate(ADate, AYear, AMonth, ADay);
    AFirstDayOfYear := EncodeDate(AYear, 1, 1);
    AWeekDay := DayOfWeek(AFirstDayOfYear);
    ANumDays := Trunc(Int(ADate) - AFirstDayOfYear) +
                (7 - DayOfWeek(ADate - 1)) +
                (7 * Ord(AWeekDay in[2..5]));
    Result := ANumDays div 7;
    if Result = 0 then
    begin
      if(DayOfWeek(EncodeDate(AYear - 1, 1, 1)) > 5) or
        (DayOfWeek(EncodeDate(AYear - 1, 12, 31)) < 5) then
        Result := 52
      else
        Result := 53;
    end
    else if Result = 53 then
      if (AWeekDay > 5) or (DayOfWeek(EncodeDate(AYear, 12, 31)) < 5) then
      begin
        Result := 1;
      end;
  except
    Result := 0;
  end;
end;

function TISCalendar.IsLeapYear(AYear: Integer): Boolean;
begin
  Result := (AYear mod 4 = 0) and ((AYear mod 100 <> 0) or (AYear mod 400 = 0));
end;

function TISCalendar.DaysPerMonth(AYear, AMonth: Integer): Integer;
const
  DaysInMonth: array[1..12] of Integer = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
begin
  Result := DaysInMonth[AMonth];
  // recognize leapyear
  if (AMonth = 2) and IsLeapYear(AYear) then Inc(Result);
end;

function TISCalendar.DaysThisMonth: Integer;
begin
  Result := DaysPerMonth(Year, Month);
end;

procedure TISCalendar.ChangeMonth(Delta: Integer);
var
  AYear, AMonth, ADay: Word;
  NewDate: TDateTime;
  CurDay: Integer;
begin
  // Given a value of 1 or -1, moves to Next or Prev month accordingly
  DecodeDate(FDate, AYear, AMonth, ADay);
  CurDay := ADay;
  if Delta > 0 then
    ADay := DaysPerMonth(AYear, AMonth)
  else
    ADay := 1;
  NewDate := EncodeDate(AYear, AMonth, ADay);
  NewDate := NewDate + Delta;
  DecodeDate(NewDate, AYear, AMonth, ADay);
  if DaysPerMonth(AYear, AMonth) > CurDay then
    ADay := CurDay
  else
    ADay := DaysPerMonth(AYear, AMonth);

  Self.Date := EncodeDate(AYear, AMonth, ADay);
end;

function TISCalendar.GetCalendarWeek: Integer;
begin
  Result := CalcCalendarWeek(Self.FDate);
end;

function TISCalendar.IsTrailingDay(ACol, ARow: Integer): Boolean;
begin
  Result := GetCellType(ACol, ARow) = dtTrailingday;
end;

function TISCalendar.IsToday(ACol, ARow: Integer): Boolean;
begin
  Result := False;
 if (ACol < ColCount-1) and (ARow < RowCount-1) then
  if (GetCellType(ACol, ARow) = dtDay) then
    Result := (GetDateFromCell(ACol, ARow) = SysUtils.Date);
end;

function TISCalendar.IsCalendarWeek(ACol, ARow: Integer): Boolean;
begin
  Result := GetCellType(ACol, ARow) = dtWeekNo;
end;

function TISCalendar.IsWeekend(ACol, ARow: Integer): Boolean;
begin
  Result := False;
end;


function TISCalendar.GetDateElement(Index: Integer): Integer;
var
  AYear, AMonth, ADay: Word;
begin
  DecodeDate(FDate, AYear, AMonth, ADay);
  case Index of
    1: Result := AYear;
    2: Result := AMonth;
    3: Result := ADay;
    else Result := -1;
  end;
end;

function TISCalendar.GetCellText(ACol, ARow: Integer): string;
var
  CellType : TTypeOfCell;
  AYear, AMonth, ADay: Word;
begin
  Result := '';                                 // assume empty string

  CellType := GetCellType(ACol, ARow);
  case CellType of

  dtHeading:
    begin
      if FViewStyle = cvMonthView then
      begin
        if (coShowWeekNumbers in FOptions) then
        begin
          if ACol > 0 then //copystr
            Result := Copy(ShortDayNames[(Integer(FStartOfWeek+1) + ACol - 1) mod 7 + 1],1,1);
        end
        else
          Result := Copy(ShortDayNames[(Integer(FStartOfWeek+1) + ACol) mod 7 + 1],1,1);
      end;
    end;

  dtMonth:
    begin    //formats monthnames
      Result := ShortMonthNames[ACol+(ARow * 4) + 1] + '.';
    end;

  dtWeekNo:
    Result := IntToStr(CalcCalendarWeek(GetDateFromCell(1, ARow)));

  dtTrailingday, dtDay:
    begin
      DecodeDate(GetDateFromCell(ACol, ARow), AYear, AMonth, ADay);
      if (CellType = dtDay) or (coShowTrailingDays in FOptions) then
        Result := IntToStr(ADay);
    end;

  else
  end;
end;


function TISCalendar.GetCellType(ACol, ARow: Integer): TTypeOfCell;
var
  DayNum : Integer;
begin
  if (ARow = 0) and (FViewStyle <> cvYearView) then
    Result := dtHeading
  else
  begin
    if FViewStyle = cvYearView then
      Result := dtMonth
    else
    begin
      if (ACol = 0) and (coShowWeekNumbers in FOptions) then
        Result := dtWeekNo
      else
      begin
        DayNum := GetDayNum(ACol, ARow);
        if (DayNum < 1) or (DayNum > DaysThisMonth) then
          Result := dtTrailingday
        else
          Result := dtDay;
      end;
    end;
  end;
end;

function TISCalendar.GetDayNum(ACol, ARow: Integer): Integer;
var
  TheCol : Integer;
begin
  if (coShowWeekNumbers in FOptions) then
    TheCol := ACol - 1
  else
    TheCol := ACol;

  if FMonthOffset = 1 then
    Result := FMonthOffset + TheCol + (ARow - 2) * 7
  else
    Result := FMonthOffset + TheCol + (ARow - 1) * 7;
end;

function TISCalendar.GetDateFromCell(ACol, ARow: Integer): TDateTime;
var
  DayNum : Integer;
  CellType: TTypeOfCell;
  AYear, AMonth, ADay: Word;
begin
  CellType := GetCellType(ACol, ARow);
  DecodeDate(FDate, AYear, AMonth, ADay);
  DayNum := GetDayNum(ACol, ARow);

  if (CellType = dtDay) then
     ADay := DayNum;

  if (CellType = dtTrailingDay) then begin
    if DayNum > 0 then begin
      ADay := DayNum - DaysPerMonth(AYear, AMonth);
      Inc(AMonth);
      if AMonth > 12 then begin
        AMonth := 1;
        Inc(AYear);
      end;
    end
    else begin
      Dec(AMonth);
      if AMonth < 1 then begin
        Dec(AYear);
        AMonth := 12;
      end;
      ADay := DaysPerMonth(AYear, AMonth) + DayNum;
    end;
  end;

  Result := EncodeDate(AYear, AMonth, ADay);
end;


procedure TISCalendar.SetDateElement(Index: Integer; Value: Integer);
var
  AYear, AMonth, ADay: Word;
begin
  if Value > 0 then
  begin
    DecodeDate(FDate, AYear, AMonth, ADay);
    case Index of
      1: if AYear <> Value then
                   AYear := Value else Exit;
      2: if (Value <= 12) and (Value <> AMonth) then
                   AMonth := Value else Exit;
      3: if (Value <= DaysThisMonth) and (Value <> ADay) then
                   ADay := Value else Exit;
      else Exit;
    end;
    FDate := EncodeDate(AYear, AMonth, ADay);
    FUseCurrentDate := False;
    CalendarUpdate((Index = 3) or ((Index = 2) and (FViewStyle = cvYearView)));
    Change;
  end;
end;


procedure TISCalendar.UpdateCalendar;
begin
  SetCellDimensions;
  CalendarUpdate(False);
end;

procedure TISCalendar.CalendarUpdate(DayOnly: Boolean);
var
  AYear, AMonth, ADay: Word;
  ACol, ARow : Integer;
  FirstDate: TDateTime;
begin
  FUpdating := True;
  try
    DecodeDate(FDate, AYear, AMonth, ADay);
    FirstDate := EncodeDate(AYear, AMonth, 1);

    FMonthOffset := 2 - ((DayOfWeek(FirstDate)
                          - Integer(FStartOfWeek) + 7) mod 7);
    if FMonthOffset = 2 then FMonthOffset := -5;

    if FViewStyle = cvMonthView then
    begin
      ACol := (ADay - FMonthOffset) mod 7;
      ARow := (ADay - FMonthOffset) div 7 + 1;
      if FMonthOffset = 1 then
        Inc(ARow);
      if (coShowWeekNumbers in FOptions) then
        Inc(ACol);
      MoveColRow(ACol, ARow, True, true, True);
    end
    else
      MoveColRow(((AMonth - 1) mod 4), ((AMonth - 1) div 4) , True, True, True);

//      if (coShowtitle in FOptions) and HandleAllocated then
//         Perform(WM_NCPAINT,0,0);
      Invalidate;

    if not DayOnly then
       SetTitleText;
  finally
    FUpdating := False;
  end;
end;

procedure TISCalendar.SetCellDimensions;
begin
  if (FViewStyle = cvMonthView) then
  begin
    RowCount := 7;
    if (coShowWeekNumbers in FOptions) then
    begin
      ColCount := 8;
      FixedCols := 1;
    end
    else
    begin
      ColCount := 7;
      FixedCols := 0;
    end;
  end
  else
  begin
    FixedCols := 0;
    FixedRows := 0;
    ColCount := 4;
    RowCount := 3;
  end;
 if HandleAllocated then
 begin
  if AutoSize then
     AdjustAutoSize;
  DefaultRowHeight := ClientHeight div RowCount;
  DefaultColWidth := ClientWidth div Colcount;
 end;
end;

procedure TISCalendar.SetFrameWidth(const Value: integer);
begin
  FFrameWidth := Value;
  if coShowFrame in FOptions then
    RecreateWnd;
end;

//------------------------------
//
//TISCalendarColors
//
//------------------------------
constructor TISCalendarColors.Create(AOwner: TISCalendar);
begin
  Owner              := AOwner;
  FBackColor         := clWindow;
  FTitleBackColor    := clBtnFace;
  FTitleTextColor    := clWindowText;
  FMonthBackColor    := clWhite;
  FTrailingTextColor := clGrayText;
  FTodaysColor       := clInfobk;
  FTodaysBorderColor := clWindowFrame;
  FSelectedBorderColor := clHighlight;
  FTodaysTextColor   := clActiveCaption;
  FSelectedTextColor := clHighlightText;
  FLineColor := clSilver;
  FFrameColor := clWindowFrame;
  FSelectedColor     := clHighlight;//clGray;
  FTextColor         := clWindowText;//clWhite;
end;

procedure TISCalendarColors.Assign(Source: TPersistent);
const
   SAssignError = '%s can not be assigned to %s';
var
  SourceName: string;
begin
  if Source = nil then SourceName := 'nil'
  else SourceName := Source.ClassName;
  if (Source = nil) or not (Source is TISCalendarColors) then
    raise EConvertError.CreateFmt(SAssignError, [SourceName, ClassName]);
  FBackColor         := TISCalendarColors(Source).BackColor;
  FTextColor         := TISCalendarColors(Source).TextColor;
  FTitleBackColor    := TISCalendarColors(Source).TitleBackColor;
  FTitleTextColor    := TISCalendarColors(Source).TitleTextColor;
  FMonthBackColor    := TISCalendarColors(Source).MonthBackColor;
  FTrailingTextColor := TISCalendarColors(Source).TrailingTextColor;
  FTodaysColor       := TISCalendarColors(Source).TodaysColor;
  FTodaysBorderColor := TISCalendarColors(Source).TodaysBorderColor;
  FTodaysTextColor   := TISCalendarColors(Source).TodaysTextColor;
  FSelectedBorderColor := TISCalendarColors(Source).SelectedBorderColor;
  FSelectedColor     := TISCalendarColors(Source).SelectedColor;
  FSelectedTextColor := TISCalendarColors(Source).SelectedTextColor;
  FLineColor := TISCalendarColors(Source).LineColor;
  FFrameColor := TISCalendarColors(Source).FFrameColor;
  Owner.Invalidate;
end;

procedure TISCalendarColors.SetColor(Index: Integer; Value: TColor);
begin

  case Index of
    0: FBackColor         := Value;
    1: FTextColor         := Value;
    2: FTitleBackColor    := Value;
    3:
       FTitleTextColor    := Value;
    4: FMonthBackColor    := Value;
    5: FTrailingTextColor := Value;
    6: FTodaysColor       := Value;
    7: FTodaysBorderColor := Value;
    8: FTodaysTextColor   := Value;
    9: FSelectedBorderColor := Value;
   10: FSelectedColor     := Value;
   11: FSelectedTextColor := Value;
   12 : FLineColor := Value;
   13 : FFrameColor := Value;
  end;

  if Owner <> nil then
  begin
    Owner.SetColor(Index, Value);
    Owner.Invalidate;
  end;


end;

procedure TISCalendarColors.SetAllColors;
begin
  SetColor(0, FBackColor);
  SetColor(1, FTextColor);
  SetColor(2, FTitleBackColor);
  SetColor(3, FTitleTextColor);
  SetColor(4, FMonthBackColor);
  SetColor(5, FTrailingTextColor);
  SetColor(6, FTodaysColor);
  SetColor(7, FTodaysBorderColor);
  SetColor(8, FTodaysTextColor);
  SetColor(9, FSelectedBorderColor);
  SetColor(10, FSelectedColor);
  SetColor(11, FSelectedTextColor);
  SetColor(12, FLineColor);
  SetColor(13, FFrameColor);
  Owner.Invalidate;
end;

procedure TISCalendar.SetAutoSize(const Value: boolean);
begin
 FAutoSize := Value;
 if FAutoSize then
   AdjustAutoSize; //handlecreated
end;

procedure TISCalendar.DrawFixedFrame(aCanvas : TCanvas; aRect : TRect; aCol, aRow : integer; Down : Boolean);
begin
end;

procedure TISCalendar.SetDisplaysStyle(const Index: Integer;
  const Value: TCellDisplayStyle);
begin
  Case Index of
     0: FFixedStyle := Value;
     1: FTodayStyle := Value;
     2: FCalendarStyle := Value;
     3: FSelectedStyle := Value;
  end;
  Invalidate;
end;


initialization
   //register classes for property editors
   {$IFNDEF Registered}
   {$IFNDEF ACTIVEX}
      ShowURToolTip('', false);
   {$ENDIF}
   {$ENDIF}

finalization
end.



