unit ISColled;
{-----------------------------------------------------------------
   Copyright (c) 1999 IMCA SYSTEMS
   All Rights Reserved.
   Creation Date: 1.11.99//
   Version :
   Module: Collection Editor module
-----------------------------------------------------------------}
{$I DFS.inc}
interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  StdCtrls, Menus, ExtCtrls, ComCtrls,  Buttons, exgrid, RapTree,
  Flytree, PropTree;



type
  //TColOption = (coAdd, coDelete, coMove);
  //TColOptions = set of TColOption;
  TCollectionEditorClass = class of TCollEditor;
  TModifiedProc = procedure of object;

  TCollEditor = class(TForm)
    ToolBar1: TPanel;
    Add: TSpeedButton;
    Delete: TSpeedButton;
    MoveUp: TSpeedButton;
    MoveDown: TSpeedButton;
    Panel3: TPanel;
    ListView1: TListView;
    PopupMenu1: TPopupMenu;
    AddMnu: TMenuItem;
    DeleteMnu: TMenuItem;
    MoveUpMnu: TMenuItem;
    MoveDownMnu: TMenuItem;
    SelAll: TMenuItem;
    N2: TMenuItem;
    PropTree1: TPropertiesTree;
    procedure AddClick(Sender: TObject);
    procedure DeleteClick(Sender: TObject);
    procedure MoveUpClick(Sender: TObject);
    procedure MoveDownClick(Sender: TObject);
    procedure ListView1DragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure ListView1DragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure FormCreate(Sender: TObject);
    procedure SelAllClick(Sender: TObject);
    procedure ListView1SelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure ListView1Change(Sender: TObject; Item: TListItem;
      Change: TItemChange);
  private
    FCollectionPropertyName: string;
    FStateLock: Integer;
    FCollectionClassName: string;
    FSelectionError: Boolean;
    FCollection : TCollection;
    FComponent : TComponent;
    DesignerModified :  TModifiedProc;
    procedure SetCollectionPropertyName(const Value: string);
    procedure UpdateButtons;
  protected
    //procedure Activated; override;
    function  CanAdd(Index: Integer): Boolean; virtual;
    procedure LockState;
    procedure UnlockState;
    property StateLock: Integer read FStateLock;
    procedure SelectAll(DoUpdate: Boolean);
    procedure SelectNone(DoUpdate: Boolean);

  public
    { Public declarations }
    function GetItemName(Index, ItemIndex: Integer): string;
    //procedure GetSelection;
    procedure SetSelection;
    procedure UpdateListbox;
    property CollectionPropertyName: string read FCollectionPropertyName
      write SetCollectionPropertyName;
  end;

var
  CollEditor: TCollEditor;

  function ShowCollectionEditorClass(ModifiedProc :  TModifiedProc; AComponent: TComponent;
  ACollection: TCollection; const PropertyName: string): TCollEditor;

implementation

type
  TAccessCollection = class(TCollection);
  TPersistentCracker = class(TPersistent);


{$R *.DFM}

function ShowCollectionEditorClass(ModifiedProc :  TModifiedProc; AComponent: TComponent;
  ACollection: TCollection; const PropertyName: string): TCollEditor;
begin
  Result := TColleditor.Create(Application);
  with Result do
  try
    FCollection := ACollection;
    DesignerModified := ModifiedProc;
    FCollectionClassName := ACollection.ClassName;
    FComponent := AComponent;
    CollectionPropertyName := PropertyName;
    UpdateListbox;
    Show;
  except
    Free;
  end;
end;

procedure ShowCollectionEditor(ModifiedProc :  TModifiedProc; AComponent: TComponent;
  ACollection: TCollection; const PropertyName: string);
begin
  ShowCollectionEditorClass(ModifiedProc, AComponent, ACollection, PropertyName);
end;


function TCollEditor.GetItemName(Index, ItemIndex: Integer): string;
begin
  with TAccessCollection(FCollection) do
  {$IFDEF DFS_COMPILER_3_UP}
    if GetAttrCount < 1 then
      Result := Format('%d - %s',[ItemIndex, FCollection.Items[ItemIndex].DisplayName])
    else Result := GetItemAttr(Index, ItemIndex);
  {$ELSE}
   Result := Format('%d - Item',[ItemIndex])
  {$ENDIF}
end;



procedure TCollEditor.SetCollectionPropertyName(const Value: string);
begin
  if Value <> FCollectionPropertyName then
  begin
    FCollectionPropertyName := Value;
    Caption := Format('Edit ', [FComponent.Name, '.', Value]);
  end;
end;

procedure TCollEditor.SetSelection;
var
  I: Integer;
  List: TList;
begin
  if FStateLock>0 then exit;
  try
    if ListView1.SelCount > 0 then
    begin
      List := TList.Create;
      try
        for I := 0 to ListView1.Items.Count - 1 do
          if ListView1.Items[I].Selected then
          begin
            List.Add(FCollection.Items[I]);
          end;
          PropTree1.SelectComponents(List);
      finally
        List.Free;
      end;
    end
  except
    FSelectionError := True;
    Application.HandleException(ExceptObject);
    Close;
  end;
end;
procedure TCollEditor.LockState;
begin
  Inc(FStateLock);
end;

procedure TCollEditor.UnlockState;
begin
  Dec(FStateLock);
end;

procedure TCollEditor.UpdateListbox;
var
  I, J: Integer;

  procedure UpdateColumns;
  var
    I: Integer;
  begin
    if (FCollection <> nil) and
      (((TAccessCollection(FCollection).GetAttrCount > 0) and
      (ListView1.Columns.Count <> TAccessCollection(FCollection).GetAttrCount)) or
      ((ListView1.Columns.Count = 0) and
      (TAccessCollection(FCollection).GetAttrCount < 1))) then
    begin
      ListView1.HandleNeeded;
      with TAccessCollection(FCollection) do
      begin
        if GetAttrCount >= 1 then
          for I := 0 to GetAttrCount - 1 do
            with ListView1.Columns.Add do
            begin
              Caption := GetAttr(I);
              Width := -2;
            end
        else
          with ListView1.Columns.Add do
            Width := -1;
        if GetAttrCount >= 1 then
          ListView1.ShowColumnHeaders := True
        //else
        //  ListView1.Column[0].Width := ListView1.ClientWidth;
      end;
      //UpdateSizes;
    end;
  end;

  procedure FetchItems(List: TStrings);
  var
    I, J: Integer;
    SubList: TStringList;
  begin
    if FCollection <> nil then
      for I := 0 to FCollection.Count - 1 do
        if CanAdd(I) then
        begin
          SubList := TStringList.Create;
          for J := 1 to TAccessCollection(FCollection).GetAttrCount - 1 do
            SubList.Add(GetItemName(J, I));
          List.AddObject(GetItemName(0, I), SubList);
        end;

  end;

  function ItemsEqual(ListItems: TListItems; Items: TStrings): Boolean;
  var
    I, J: Integer;
  begin
    Result := False;
    if ListItems.Count <> Items.Count then Exit;
    for I := 0 to ListItems.Count - 1 do
    begin
      if ListItems[I].Caption = Items[I] then
      begin
        for J := 0 to ListItems[I].SubItems.Count - 1 do
          if ListItems[I].SubItems[J] <> TStrings(Items.Objects[I])[J] then
            Exit;
      end
      else
        Exit;
    end;
    Result := True;
  end;

var
  TmpItems: TStringList;
begin
  if FCollection = nil then Exit;
  LockState;
  try
    TmpItems := TStringList.Create;
    FetchItems(TmpItems);
    try
      if (TmpItems.Count = 0) or not ItemsEqual(ListView1.Items, TmpItems) then
      begin
        ListView1.Items.BeginUpdate;
        try
          UpdateColumns;
          ListView1.Items.Clear;
          for I := 0 to TmpItems.Count - 1 do
            with ListView1.Items.Add do
            begin
              Caption := TmpItems[I];
              for J := 0 to TStrings(TmpItems.Objects[I]).Count - 1 do
                SubItems.Add(TStrings(TmpItems.Objects[I])[J]);
            end;
        finally
          ListView1.Items.EndUpdate;
        end;
      end;
    finally
      for I := 0 to TmpItems.Count - 1 do
        TStrings(TmpItems.Objects[I]).Free;
      TmpItems.Free;
    end;
  finally
    UnlockState;
  end;
end;


procedure TCollEditor.AddClick(Sender: TObject);
var
  Item: TListItem;
  //PrevCount: Integer;
  //aItem : TCollectionItem;
begin
  //SelectNone(False);
  FCollection.BeginUpdate;
  try
    //PrevCount := FCollection.Count+1;
    //aItem :=
    FCollection.Add;
    { Take into account collections that free items }
    //if aItem=nil{PrevCount <> FCollection.Count} then
    UpdateListBox;
    //else
    //ListView1.Selected := ListView1.Items.Add;
  finally
    FCollection.EndUpdate;
  end;
  //SetSelection;
  if Assigned(DesignerModified) then
     DesignerModified;
  { Focus last added item }
  Lockstate;
  Item := ListView1.Items[ListView1.Items.Count-1];
  Item.Focused := True;
  Item.MakeVisible(False);
  UnlockState;
  SetSelection;
  UpdateButtons;

end;

procedure TCollEditor.DeleteClick(Sender: TObject);
var
  I, J: Integer;
begin
  FCollection.BeginUpdate;
  try
    //Designer.SetSelections(nil);
    if ListView1.Selected <> nil then
      J := ListView1.Selected.Index
    else J := -1;
    if ListView1.SelCount = FCollection.Count then
      FCollection.Clear
    else if ListView1.SelCount > 0 then
      for I := ListView1.Items.Count - 1 downto 0 do
        if ListView1.Items[I].Selected then
          FCollection.Items[I].Free;
  finally
    FCollection.EndUpdate;
  end;
  UpdateListbox;
  if J >= ListView1.Items.Count then
    J := ListView1.Items.Count - 1;
  if (J > -1) and (J < ListView1.Items.Count) then
    ListView1.Selected := ListView1.Items[J];
  //SetSelection;
  if Assigned(DesignerModified) then
     DesignerModified;
  UpdateButtons;

end;

procedure TCollEditor.MoveUpClick(Sender: TObject);
var
  I, InsPos: Integer;
begin
  if (ListView1.SelCount = 0) or
    (ListView1.SelCount = FCollection.Count) then Exit;

  InsPos := 0;
  while not ListView1.Items[InsPos].Selected do
    Inc(InsPos);
  if InsPos > 0 then Dec(InsPos);

  FCollection.BeginUpdate;
  try
     for I := 0 to ListView1.Items.Count - 1 do
       if ListView1.Items[I].Selected then
       begin
         FCollection.Items[I].Index := InsPos;
         Inc(InsPos);
       end;
  finally
    FCollection.EndUpdate;
  end;
  //GetSelection;
  if Assigned(DesignerModified) then
     DesignerModified;
  UpdateButtons;

end;

procedure TCollEditor.MoveDownClick(Sender: TObject);
var
  I, InsPos: Integer;
begin
  if (ListView1.SelCount = 0) or
    (ListView1.SelCount = FCollection.Count) then Exit;

  InsPos := ListView1.Items.Count - 1;
  while not ListView1.Items[InsPos].Selected do
    Dec(InsPos);
  if InsPos < (ListView1.Items.Count -1) then Inc(InsPos);

  FCollection.BeginUpdate;
  try
     for I := ListView1.Items.Count - 1 downto 0 do
       if ListView1.Items[I].Selected then
       begin
         FCollection.Items[I].Index := InsPos;
         Dec(InsPos);
       end;
  finally
    FCollection.EndUpdate;
  end;
  //GetSelection;
  if Assigned(DesignerModified) then
     DesignerModified;
end;

procedure TCollEditor.ListView1DragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
var
  Item: TListItem;
begin
  Item := ListView1.GetItemAt(X, Y);
  Accept := (Item <> nil) and (Source = ListView1) and
    (not Item.Selected);
end;

procedure TCollEditor.ListView1DragDrop(Sender, Source: TObject; X,
  Y: Integer);
var
  Item: TListItem;
  I, J, InsPos: Integer;
  L: TList;
begin
  Item := ListView1.GetItemAt(X, Y);
  if Item <> nil then
    InsPos := Item.Index
  else Exit;
  L := TList.Create;
  try
    for I := 0 to ListView1.Items.Count - 1 do
      if ListView1.Items[I].Selected then
        L.Add(FCollection.Items[I]);

    FCollection.BeginUpdate;
    try
      for I := 0 to L.Count - 1 do
      with TCollectionItem(L[I]) do
      begin
        J := Index;
        Index := InsPos;
        if (J > InsPos) and (InsPos < Collection.Count) then
          Inc(InsPos);
      end;
    finally
      FCollection.EndUpdate;
    end;
  finally
    L.Free;
  end;
  //GetSelection;
  if Assigned(DesignerModified) then
     DesignerModified;
end;

procedure TCollEditor.UpdateButtons;
begin
    MoveUp.Enabled:= (ListView1.items.count>1) and (ListView1.Selected<>nil);
    MoveDown.Enabled:= (ListView1.items.count>1) and (ListView1.Selected<>nil);
    Delete.Enabled:= (ListView1.Selected<>nil);
    MoveUpMnu.Enabled:= (ListView1.items.count>1) and (ListView1.Selected<>nil);
    MoveDownMnu.Enabled:= (ListView1.items.count>1) and (ListView1.Selected<>nil);
    DeleteMnu.Enabled:= (ListView1.Selected<>nil);
    SelAll.Enabled:= ListView1.items.count>0;
end;

procedure TCollEditor.FormCreate(Sender: TObject);
begin

  UpdateButtons;
end;

procedure TCollEditor.SelAllClick(Sender: TObject);
begin
//
  SelectAll(True);
end;


procedure TCollEditor.SelectAll(DoUpdate: Boolean);
var
  I: Integer;
begin
  LockState;
  ListView1.Items.BeginUpdate;
  try
    for I := 0 to Listview1.Items.Count-1 do
      Listview1.Items[I].Selected := True;
  finally
    ListView1.Items.EndUpdate;
    UnlockState;
    if DoUpdate then SetSelection;
  end;
end;

procedure TCollEditor.SelectNone(DoUpdate: Boolean);
var
  I: Integer;
begin
  LockState;
  ListView1.Items.BeginUpdate;
  try
    for I := 0 to Listview1.Items.Count-1 do
      Listview1.Items[I].Selected := False;
  finally
    ListView1.Items.EndUpdate;
    UnlockState;
    if DoUpdate then SetSelection;
  end;
end;

function TCollEditor.CanAdd(Index: Integer): Boolean;
begin
  Result := True;
end;


procedure TCollEditor.ListView1SelectItem(Sender: TObject; Item: TListItem;
  Selected: Boolean);
begin
  //SelectNone(true);
  SetSelection;
  UpdateButtons;
end;

procedure TCollEditor.ListView1Change(Sender: TObject; Item: TListItem;
  Change: TItemChange);
begin
  SetSelection;
  UpdateButtons;
end;

end.
