{----------------------------------------------------------------
 /  Copyright (c) 2000 IMCA SYSTEMS
 /  All Rights Reserved.
 /  Creation Date: 21.05.2000
 /  Version
 /  MODULE: HTMLDrawing function
 // Needed : Align, FontName, Size
 /  NOTES:                                                     /
 /  HISTORY:                                                   /

  Supports tags:
    <b>,<B> - bold face
    <i>,<I> - italic face
    <u>,<U> - underline
    <s>, <S> - strikeout face
    <color=...>, <COLOR = ...> - color of font, where ... - is
    <Font face = "Times" size="24" color="clRed">WWW</font>2222
    color in delphi (clblack, $00FFFF00) or standard HTML (#FF00FF) format
    Also can be used close-tags for these tags
    Smart Wordwrap supports
    Supports Alignment for entire text
 02.10.2000 -
    // corrected bug with draw extra-width text with center and right alignment
----------------------------------------------------------------
Expected : TextRect calculation
-----------------------------------------------------------------}
unit ISHTMLDRaw;

interface
{$I DFS.inc}
uses Windows, Classes, Graphics, SysUtils;

function DrawHTML(var R: TRect; aCanvas: TCanvas; const Text: string; WordWrap, CalcRect : boolean; Alignment, VAlignment : integer) : string;
function StripTags(const Text : string) : string;
implementation

type

  TFontAttrib = record
    Name : string;
    Color : TColor;
    Height : integer;
  end;
  PFontAttrib  = ^TFontAttrib;

  TBmpLine = record
    Bmp : TBitmap;
    EL : integer;
  end;
  PBmpLine  = ^TBmpLine;
var
  BBrush : TBrush;
  BFont: TFont;
  WBrush : TBrush;
  Wfont : TFont;
  FontList : TList;
  LineList : TList;
  BGList : TList;
  STags : Boolean;

procedure PopFont;
begin
  if FontList.Count >0 then
  begin
     WFont.Name := PFontAttrib(FontList[0])^.Name;
     WFont.Height := PFontAttrib(FontList[0])^.Height;
     WFont.Color := PFontAttrib(FontList[0])^.Color;
     Dispose(PFontAttrib(FontList[0]));
     FontList.Delete(0);
  end;
end;

procedure AddToLines(aBmp : TBitmap; aEL : integer);
var pbl : PBmpLine;
begin
  new(pbl);
  pbl^.Bmp := aBmp;
  pbl^.EL := aEL;
  LineList.Add(pbl);
end;


procedure PopBg;
begin
 if BGList.count > 0 then
 begin
  WBrush.Color := TColor(BGList[0]);
  BGList.Delete(0);
 end;
end;

procedure PushBG;
begin
  BgList.Add(Pointer(WBrush.Color))
end;

procedure PushFont;
var Attrib : PFontAttrib;
begin
  new(attrib);
  Attrib^.Name := WFont.Name;
  Attrib^.Height := Wfont.Height;
  Attrib^.Color := Wfont.Color;
  FontList.Insert(0,Attrib);
end;

procedure ClearLineList;
var i : integer;
begin
  for i:=0 to LineList.Count-1 do
  begin
    PBmpLine(LineList[i])^.Bmp.Free;
    Dispose(PBmpLine(LineList[i]));
  end;
  LineList.Clear;
end;

procedure ClearFontList;
var i : integer;
begin
  for i:=0 to FontList.Count-1 do
    Dispose(PFontAttrib(FontList[i]));
  FontList.Clear;
end;

function StripTags(const Text : string) : string;
var R : TRect;
begin
  STags := True;
  Result :=DrawHTML(R, nil, Text, false, true, 0,0);
  STags := False;
end;

function DrawHTML(var R: TRect; aCanvas: TCanvas; const Text: string; WordWrap, CalcRect : boolean; Alignment, VAlignment : integer) : string;
var
 p, aSubStr: PChar;
 aOutStr : String;
 voff, Curwidth, ax, x, y, hc: Integer;
 Code: string;
 OldColor, DefBrushColor, DefFontColor : TColor;
 NeedLine, Parsed : boolean;
 MainBitmap, aBitmap, tmpBitmap : TBitmap; //aBitmap -     
 TxtMetric: TTEXTMETRIC;

 function GetStr : String;
 var
     aStr, PreStr : String;
     LeftMargin : integer;
     LastSpace : PChar;
     RCanvas : TCanvas;
     TmpRight : integer;
 begin
   aStr :='';
   NeedLine :=false;
   LeftMargin := X;
   LastSpace := P;
   RCanvas:=aCanvas;
   if not STags then
   begin
    if not CalcRect then
      RCanvas := aBitmap.Canvas;

     GetTextMetrics(RCanvas.Handle, TxtMetric);
   end;
   while ((p^ <> '<') or not Parsed)  do
   begin
    Parsed :=true;
    if (p^=#13) then
    begin
      inc(p)
    end
    else if (p^=#10) or (p^ = #0) then
    begin
       if p^<>#0 then
         inc(p);
       NeedLine :=True;
       PreStr := PreStr+aStr;
       aStr := '';
       break;
    end
    else if (p^=' ' ) then
    begin //add word to Result
      PreStr := PreStr + aStr + ' ';
      inc(P);
      LastSpace := P;
      //     -    
      //      .    rect
      //   
      if not STags then
        LeftMargin := X + RCanvas.TextWidth(PreStr)+TxtMetric.tmOverhang;
      aStr := '';
    end
    else
    begin
     if not STags then
      TmpRight := X+ RCanvas.Textwidth(PreStr+aStr+ p^)+TxtMetric.tmOverhang
     else TmpRight := 0;

      if ((TmpRight <= (R.Right-R.Left)) or not Wordwrap)
      or (LeftMargin = 0 ){<- if only one word in text} then
      begin
       aStr := aStr + p^;
       inc(p);
      end
      else
      begin
         NeedLine:=True;
         P := LastSpace; //come back to last phrase
         aStr := '';
         break;
      end;

    end;
   end;
   Result := PreStr;
   if aStr<>'' then
     Result := Result + aStr;
 end;

 function GetClrValue(astr : PChar) : TColor;
 var IntClr : integer;
 begin
   Result := clNone;
   while (aStr^ <> '=') and (aStr^<>#0) do
     inc(aStr);
   if (aStr^<>#0) then inc(aStr);
   while (aStr^ = ' ') and (aStr^<>#0) do
     inc(aStr);

   if aStr^<>#0 then
   try
    if (aStr^ in ['C', 'c', '$']) then //Delphi color constants - begining cl or HexDisplayPrefix
     Result := StringToColor(string(aStr))
    else if (aStr^='#') then // standard HTML coding
    begin
      inc(aStr);
      IntClr := StrToInt('$'+string(aStr));
      Result := TColor(
      RGB( LoByte(HiWord(intclr)), Hibyte(Loword(Intclr)), Lobyte(Loword(intclr)))
      );//StringToColor('$'+string(aStr));//rgbtocolor
    end
    else
      Result := DefFontColor;
   except
     Result := DefFontColor;
   end;
 end;

 function GetSectionVal(section, aStr : string) : string;
 var fpos : integer;
 begin
    Result := '';
    fpos := Pos(section, aStr); //strlpos strcomp
    if fPos > 0 then
    begin
     while aStr[fpos] <> '=' do inc(fpos);
     inc(fPos);//skip '='

     while aStr[fpos] = ' ' do inc(fpos); //skip spaces

     if (aStr[fpos] = '"') then inc(fpos);
     while (fpos <= Length(aStr)) do
     begin
       if (aStr[fpos] = ' ') or (aStr[fpos] = '"') then
         break
       else
         Result := Result+aStr[fpos];

       inc(fpos);
     end;
    end;
 end;

 procedure ParseFont(FontStr : string);
 var aFace, aSize, aColor : string;
    Intclr : integer;
 begin
    //----fetch face
    aFace := GetSectionVal(' FACE', FontStr);
    aSize := GetSectionVal(' SIZE', FontStr);
    aColor := GetSectionVal(' COLOR', FontStr);
    if (aFace <> '') or (aSize <> '') or (aColor <> '' )then
    begin
      PushFont;//(aFont);
      if aFace <> '' then
      try
        WFont.Name := aFace;
        WFont.Charset := DEFAULT_CHARSET;
      except
      end;
      if aSize <> '' then
        WFont.Height := StrToIntDef(aSize,0);
      if aColor <> '' then
      try
        if aColor[1]='#' then
        begin
         aColor[1] := '$';
         IntClr := StrToInt(aColor);
         WFont.Color :=  TColor(
         RGB( LoByte(HiWord(intclr)), Hibyte(Loword(Intclr)), Lobyte(Loword(intclr)))
         );
        end
        else if (aColor[1] in ['C', 'c', '$']) then
          WFont.Color := StringToColor(AColor);
      except
      end;
    end;
 end;

 procedure ParseTag;
 var
     OldPtr : PChar;
 begin
   code := '';
   OldPtr := P;
   inc(p);
   Parsed :=false;
//   if not CalcRect then
//     RCanvas := aBitmap.Canvas
//   else
//     RCanvas := aCanvas;

   while (p^ <> '>') and (p^ <> #0) do
   begin
    if not (p^ in [#10, #13]) then
     code := code + UpperCase(p^);
     inc(p);
   end;
   //with RCanvas.Font do
    if Pos('FONT ', Code) > 0 then //fonts!!!
    begin
       ParseFont(code);
       Parsed:=True;
    end
    else if code = 'B' then
    begin
      WFont.Style := WFont.Style + [fsBold];
      Parsed:=true;
    end
    else if code = 'I' then
    begin
      WFont.Style := WFont.Style + [fsItalic];
      Parsed:=true;
    end
    else if code = 'U' then
    begin
      WFont.Style := WFont.Style + [fsUnderline];
      Parsed:=true;
    end
    else if code = 'S' then
    begin
      WFont.Style := WFont.Style + [fsStrikeout];
      Parsed:=true;
    end
    else if code = '/B' then
    begin
      WFont.Style := WFont.Style - [fsBold];
      Parsed:=true;
    end
    else if code = '/I' then
    begin
      WFont.Style := WFont.Style - [fsItalic];
      Parsed:=true;
    end
    else if code = '/U' then
    begin
      WFont.Style := WFont.Style - [fsUnderline];
      Parsed:=true;
    end
    else if code = '/S' then
    begin
      WFont.Style := WFont.Style - [fsStrikeout];
      Parsed:=true;
    end
    else if code = '/FONT' then
    begin
       PopFont;
       Parsed :=true;
    end
    else //bgcolor support
    begin
      aSubstr:=StrPos(Pchar(code), Pchar('/BGCOLOR'));
      if aSubStr<>nil then
      begin
         PopBG;//.Color := DefBrushcolor;
         Parsed:=true;
      end
      else
      begin
        aSubStr := StrPos(Pchar(code), Pchar('BGCOLOR'));
        if (aSubStr <> nil) then
        begin
           PushBG;//(RCanvas.Brush.Color);
           WBrush.Color := GetClrValue(aSubStr);
           Parsed:=true;
        end
        else
        begin
          aSubStr := StrPos(Pchar(code), Pchar('/COLOR'));
          if aSubStr<>nil then
          begin
            PopFont;//(RCanvas.Font);
            //Color := DefFontColor;
            Parsed:=true;
          end
          else
          begin
            aSubStr := StrPos(Pchar(code), Pchar('COLOR'));
            if (aSubStr <> nil) then
            begin
             Pushfont;//(Rcanvas.Font);
             WFont.Color := GetClrValue(aSubStr);
             Parsed:=true;
            end;
          end;
        end;
      end;
    end;
    if not Parsed then
       p:=OldPtr
    else
      if p^>#0 then Inc(p);
  end;

var DrawWidth, MaxHeight, MaxAscent, i : integer;
    BmpSize : TSize;
    aRect : TRect;

//    DC: HDC;
//Begin of HTMLDRAW
begin
 Result:='';
 if Text = '' then exit;
 ClearFontList;
 BGList.Clear;
 if not STags then
  With ACanvas do
  begin
     BBrush.Assign(Brush);
     BFont.Assign(Font);
     WBrush.Assign(Brush);
     WFont.Assign(Font);
  end;
 p := PChar(text);
 x := 0;//r.Left;
 y := 0;//r.Top;
 aX := 0;//r.Left;
 hc := 0;
 aOutStr := '';
 Parsed := true;
 MainBitmap := nil;
 DefBrushColor := clWhite;
// Cbitmap := nil;
if not STags then
 with aCanvas do
 begin
   DefFontColor := Font.Color;
   DefBrushColor := Brush.Color;
   hc := TextHeight('Ag');
 end;

 if (not (CalcRect or STags)) then
 begin
  // 
  MainBitmap := TBitmap.Create;
  with MainBitmap do //prepare main bitmap
  begin
    Width := Abs(R.Right - R.Left);
    Height := (R.Bottom - R.Top);
    Canvas.Brush.Color := DefBrushColor;
    Canvas.FillRect(Rect(0,0, Width, Height));
  end;

  ABitmap := TBitmap.Create;
  {$ifdef DFS_COMPILER_4_UP}
  aBitmap.Canvas.TextFlags := aCanvas.TextFlags;
  {$endif}

  with ABitmap do
  begin
   Width := 0;//(r.Right-R.Left);
   Height := hc;//R.Bottom-r.Top;
   with ABitmap.Canvas do
   begin
    Font.Assign(aCanvas.Font);
    Brush.Color := DefBrushColor;
   end;

   Canvas.Fillrect(Rect(0, 0, Width, Height));
  end;
 end;
 {$ifdef DFS_COMPILER_4_UP}
  if not STags then
  if ACanvas.CanvasOrientation = coRightToLeft then
     if Alignment = 0 then //Left
        Alignment := 1
     else if Alignment =1 then
        Alignment :=0;
  {$endif}
 try

 if p <> nil then
   while (p^ <> #0) or (NeedLine) do
   begin
     if (p^ = '<') and Parsed then
       ParseTag;
     //else
     //begin
       aOutStr := GetStr;
       if STags then
       begin
         Result := Result + aOutStr;
         aOutStr := '';
       end;
       if aOutStr <> '' then
       begin
         if not CalcRect then
         begin
          TmpBitmap := TBitmap.Create;
          with Tmpbitmap do
          begin
            {$ifdef DFS_COMPILER_4_UP}
            TmpBitmap.Canvas.TextFlags := aCanvas.TextFlags;
            {$endif}
            Canvas.Brush.Color := WBrush.color;//aBitmap.Canvas.Brush.Color;//DefBrushColor;
            Canvas.Font.Assign (WFont);//(aBitmap.Canvas.Font);
            GetTextMetrics(Canvas.Handle, TxtMetric);
            GetTextExtentPoint32(Canvas.Handle, Pchar(aOutStr), Length(aOutStr), BmpSize);
            inc(BmpSize.CX, TxtMetric.tmOverhang);
            Height := BmpSize.CY;//Canvas.TextHeight(aOutStr);//necessary after assigning font!
            Width := BmpSize.CX;//Canvas.Textwidth(aOutStr);
            Curwidth := BmpSize.CX;
            aRect := Rect(0,0, Width, Height);
            Canvas.Fillrect(aRect);
            //   ->>
            if Height <> hc then hc := Height;
            Canvas.TextOut(0,0, aOutStr);
            AddToLines(TmpBitmap, TxtMetric.tmAscent);
          end;
         end
         else
         begin
           GetTextMetrics(aCanvas.Handle, TxtMetric);
           GetTextExtentPoint32(aCanvas.Handle, Pchar(aOutStr), Length(aOutStr), BmpSize);
           Curwidth := BmpSize.CX+TxtMetric.tmOverhang;//aCanvas.Textwidth(aOutStr);
           if BmpSize.CY > hc then hc := BmpSize.CY;
           //MaxHeight := Hc;  // never used !
         end;

         if CalcRect then
           if aX < X + Curwidth  then
              aX:= x + Curwidth;//calculate max width

         inc(x, CurWidth);
         aOutStr := ''; //clear string
       end;

       if NeedLine then
       begin
         MaxHeight := 0;
         if not CalcRect then
         begin
          try
           Drawwidth := 0;
           MaxAscent :=0;
           for i:=0 to LineList.Count-1 do
           begin
              inc(DrawWidth, PBmpLine(LineList[i])^.Bmp.Width);
              if PBmpLine(LineList[i])^.Bmp.Height > MaxHeight then
                 MaxHeight := PBmpLine(LineList[i])^.Bmp.Height;
              if PBmpLine(LineList[i])^.EL > MaxAscent then
                 MaxAscent := PBmpLine(LineList[i])^.EL;
           end;
           aBitmap.Width := DrawWidth;
           aBitmap.Height := MaxHeight;
           X :=0;
           for i:=0 to LineList.Count-1 do
           begin //(MaxAscent-PBmpLine(LineList[i])^.EL)
             aBitmap.Canvas.Draw(X,(MaxAscent-PBmpLine(LineList[i])^.EL), PBmpLine(LineList[i])^.Bmp);
             inc(X, PBmpLine(LineList[i])^.Bmp.Width);
           end;
           ClearLineList;

           if MainBitmap<>nil then   //new 06.10.02
           with MainBitmap.Canvas do //new 08/03/02//new 18/06/01
           //with CBitmap.Canvas do
           if (R.Top + Y) < R.Bottom then
           Case Alignment of
            0://left alignment
            with R, aBitmap do
            begin
              {$ifdef DFS_COMPILER_4_UP}
              if ACanvas.CanvasOrientation = coRightToLeft then
                CopyRect(Rect(0, Y, MainBitmap.Width-DrawWidth, MainBitmap.Height), aBitmap.Canvas, Rect(0, 0, DrawWidth, MainBitmap.Height-Y))
                //CopyRect(Rect(Right, Top+Y, Right-DrawWidth, Bottom), aBitmap.Canvas, Rect(0, 0, DrawWidth, Bottom-(Top+Y)))
              else
              {$endif}
               //CopyRect(Rect(left, Top+Y, Left+DrawWidth, Bottom), aBitmap.Canvas, Rect(0, 0, DrawWidth, Bottom-(Top+Y)));
               CopyRect(Rect(0, Y, DrawWidth, MainBitmap.Height), aBitmap.Canvas, Rect(0, 0, DrawWidth, MainBitmap.Height-Y));
            end;
            1://right
              with R, aBitmap do
              {$ifdef DFS_COMPILER_4_UP}
              if ACanvas.CanvasOrientation = coRightToLeft then
                //CopyRect(Rect(left, Top+Y, Left+DrawWidth, Bottom), aBitmap.Canvas, Rect(aBitmap.Width ,0, aBitmap.Width-DrawWidth, Bottom-(Top+Y)))
                CopyRect(Rect(0, Y, DrawWidth, MainBitmap.Height), aBitmap.Canvas, Rect(aBitmap.Width ,0, aBitmap.Width-DrawWidth, MainBitmap.Height-Y))
              else
              {$endif}
               CopyRect(Rect(MainBitmap.Width-DrawWidth, Y, MainBitmap.Width, MainBitmap.Height), aBitmap.Canvas, Rect(aBitmap.Width-DrawWidth,0,aBitmap.Width, MainBitmap.Height-Y));
            2:
            with R, aBitmap do
            {$ifdef DFS_COMPILER_4_UP}
              if ACanvas.CanvasOrientation = coRightToLeft then
               CopyRect(Rect(0, Y, MainBitmap.Width, MainBitmap.Height), aBitmap.Canvas, Rect(MainBitmap.Width+(aBitmap.Width-MainBitmap.Width) div 2,0, (aBitmap.Width-(R.Right-R.Left)) div 2, MainBitmap.Height-Y))
              else
            {$endif}
              CopyRect(Rect(0, Y, MainBitmap.Width, MainBitmap.Height), aBitmap.Canvas, Rect((aBitmap.Width-(R.Right-R.Left)) div 2,0, MainBitmap.Width +(aBitmap.Width-MainBitmap.Width) div 2, MainBitmap.Height-Y));
           end;
          finally
            with aBitmap.Canvas do //clear bitmap
            begin
             OldColor := Brush.Color;
             Brush.Color := DefBrushColor;
             FillRect(Rect(0, 0, aBitmap.Width, aBitmap.Height));
             Brush.Color := OldColor;
            end;
          end;
         end;
         inc(y, MaxHeight);
         x := 0;//r.left;
         NeedLine := False;
       end;
   end;  // while

   //final output
   if not CalcRect then
   with aCanvas,MainBitmap do
   Case VAlignment of
      DT_TOP:
        CopyRect(Rect(R.Left, R.Top, R.Right, R.Top+y), Canvas, Rect(0,0, Width, y));
      DT_BOTTOM, DT_VCENTER:
      begin
        if VAlignment = DT_VCENTER then
          voff := ((R.Bottom-R.Top) - y) div 2
        else
          voff := ((R.Bottom-R.Top) - y);

        if voff > 0 then
          CopyRect(Rect(R.Left, R.Top+vOff, R.Right, R.Top+vOff+y), Canvas, Rect(0,0, Width, y))
        //else if voff < 0 then
        //  CopyRect(R, Canvas, Rect(0, -voff, Width, -voff + y))
        else CopyRect(Rect(R.Left, R.Top, R.Right, R.Top+y), Canvas, Rect(0,0, Width, y));
      end;
   end;

 finally
    if CalcRect then
    begin
      R.Bottom := R.Top+y;
      R.Right := R.Left+aX;
    end;
    //restore
    if Not STags then
    With ACanvas do
    begin
     Brush.Assign(BBrush);
     Font.Assign(BFont);
    end;

    if not CalcRect then
    begin
     aBitmap.Free;
     //Mainbitmap.Dormant;
     //Mainbitmap.FreeImage;
     MainBitmap.Free;
    end;

    ClearLineList;
    ClearFontList;
    STags := False;
 end;
end;


initialization
  BBrush := TBrush.Create;
  BFont:= TFont.Create;
  WBrush := TBrush.Create;
  WFont:= TFont.Create;
  FontList := TList.Create;
  LineList := TList.Create;
  BGList := TList.Create;
finalization
  ClearFontList;
  ClearLineList;
  FontList.Free;
  LineList.Free;
  BGList.Free;

  BBrush.Free;
  BFont.Free;
  WBrush.Free;
  WFont.Free;

end.