unit ISHTMLText;

interface
{$I DFS.inc}
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;
{$R ISHTMLText}

type
  TISVerticalAlign = (vaTop, vaBottom, vaCenter);


  TISHTMLText = class(TCustomControl)
  private
    FBorderStyle: TBorderStyle;
    FAutosize, FInSize, FWordwrap: Boolean;
    FAlignment : TAlignment;
    FVertAlign: TISVerticalAlign;
    FStrings: TStrings;
    FMargin: integer;
    procedure SetStrings(const Value: TStrings);
    procedure SetVertAlign(const Value: TISVerticalAlign);
    procedure SetWordWrap(const Value: Boolean);
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    procedure CMBorderChanged(var Message: TMessage); message CM_BORDERCHANGED;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure SetAlignment(const Value: TAlignment);
    procedure SetAutosize(const Value: boolean);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetMargin(const Value: integer);
    { Private declarations }
  protected
    procedure Paint; override;
    procedure Changed;
    procedure CalculateSize(var aWidth, aHeight : integer);
  public
    procedure CreateParams(var Params: TCreateParams); override;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Canvas;
    procedure AdjustAutoSize;
  published
    { Published declarations }
    property Margin : integer read FMargin write SetMargin default 4;
    property VerticalAlignment : TISVerticalAlign read FVertAlign write SetVertAlign default vaCenter;
    property Alignment : TAlignment read FAlignment write SetAlignment default taLeftJustify;
    property Lines : TStrings read FStrings write SetStrings;
    property WordWrap : Boolean read Fwordwrap write SetWordWrap default false;
    property AutoSize : boolean read FAutosize write SetAutosize default true;
    property Align;
    property Anchors;
    property BiDiMode;
    property BorderWidth;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsNone;
    property Color;
    property Constraints;
    property Ctl3D;
    property UseDockManager default True;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnDockDrop;
    property OnDockOver;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnKeyDown;
    property OnKeyUp;
    property OnKeyPress;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

procedure Register;

implementation

uses ISHTMLDraw;

procedure Register;
begin
  RegisterComponents('9Rays.Net', [TISHTMLText]);
end;

{ TISHTMLText }

constructor TISHTMLText.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csReplicatable, csFramed];
  Width := 105;
  Height := 105;
  FAlignment := taLeftJustify;
  FVertAlign := vaCenter;
  FStrings := TStringList.Create;
  FMargin := 4;
end;

destructor TISHTMLText.Destroy;
begin
 inherited;
 FStrings.Free;
end;

procedure TISHTMLText.SetBorderStyle(Value: TBorderStyle);
begin
  if FBorderStyle <> Value then
  begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TISHTMLText.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of DWORD = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or BorderStyles[FBorderStyle];
    if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
    //WindowClass.style := WindowClass.style and not (CS_HREDRAW or CS_VREDRAW);
  end;
end;

procedure TISHTMLText.AdjustAutoSize;
var aWidth, aHeight : integer;
begin
 if FAutosize and not FInSize then
 begin
   CalculateSize(aWidth, aHeight);
   FInSize := true;
   SetBounds(Left, Top, aWidth, aHeight);
   FInSize := False;
 end;
end;

procedure TISHTMLText.Changed;
begin
  if FAutoSize then
    AdjustAutoSize;
  Invalidate;
end;

procedure TISHTMLText.CalculateSize(var aWidth, aHeight : integer);
const
 VAlignFlags : array [TISVerticalAlign] of Integer =  ( DT_TOP, DT_BOTTOM, DT_VCENTER);
var aRect : TRect;
begin
 if not (csLoading in ComponentState) and AutoSize then
 begin
  aRect := Rect(0,0, Width, Height);
  inflateRect(aRect, -4, -1);
  DrawHTML(aRect, Canvas, FStrings.Text, FWordWrap, True, DT_LEFT, VAlignFlags[FVertAlign]);
  //SetBounds(Left, Top, aRect.Right, aRect.Bottom);
  aWidth := aRect.Right;
  aHeight := aRect.Bottom;
  inc(aWidth, FMargin);
  inc(aHeight, FMargin);
 end;
end;

procedure TISHTMLText.Paint;
const
  VAlignFlags : array [TISVerticalAlign] of Integer =  ( DT_TOP, DT_BOTTOM, DT_VCENTER);
  AlignFlags : array [TAlignment] of Integer= ( DT_LEFT, DT_CENTER, DT_RIGHT);
var aRect : TRect;
begin
  with Canvas do
  begin
    Font := Self.Font;
    Brush.Color := Color;
//    inherited;
    aRect := ClientRect;
    FillRect(aRect);
    InflateRect(aRect, -FMargin, -FMargin);
//    inflateRect(aRect, -4, -1);
    DrawHTML(aRect, Canvas, FStrings.Text, FWordWrap, false, AlignFlags[FAlignment], VAlignFlags[FVertAlign]);
  end;
end;

procedure TISHTMLText.SetStrings(const Value: TStrings);
begin
  FStrings.Assign(Value);
  Changed;
end;

procedure TISHTMLText.SetVertAlign(const Value: TISVerticalAlign);
begin
 if FVertAlign <> Value then
 begin
   FVertAlign := Value;
   Changed;
 end;
end;

procedure TISHTMLText.SetWordWrap(const Value: Boolean);
begin
 if FWordWrap <> Value then
 begin
    FWordwrap := Value;
    Changed;
 end;
end;

procedure TISHTMLText.CMBorderChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TISHTMLText.CMCtl3DChanged(var Message: TMessage);
begin
  inherited;
  RecreateWnd;
end;

procedure TISHTMLText.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Changed;
end;

procedure TISHTMLText.WMSize(var Message: TWMSize);
begin
  if FAutoSize then
  begin
    AdjustAutosize;
    Message.Width := Width;
    Message.Height := Height;
  end;
  //Changed;
  inherited;
end;

procedure TISHTMLText.SetAlignment(const Value: TAlignment);
begin
 if FAlignment <> Value then
 begin
    FAlignment := Value;
    Changed;
 end;
end;

procedure TISHTMLText.SetAutoSize(const Value: boolean);
begin
 if FAutosize <> Value then
 begin
  FAutoSize := Value;
  Changed;
 end;
end;

procedure TISHTMLText.SetMargin(const Value: integer);
begin
 if FMargin <> Value then
 begin
    FMargin := Value;
    Invalidate;
 end;
end;

end.
