unit ISPictEdit;

interface
{$I dfs.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, ComCtrls, rtProps,Isconst,
  {$IFDEF DFS_DELPHI_3_UP} extdlgs, {$ENDIF} clipbrd, Buttons;

type
  TPictureEditForm = class(TForm)
    Panel1: TPanel;
    Image: TImage;
    Panel3: TPanel;
    OpenButton: TSpeedButton;
    SaveButton: TSpeedButton;
    CopyButton: TSpeedButton;
    PasteButton: TSpeedButton;
    ClearButton: TSpeedButton;
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    procedure ToolButtonLoadClick(Sender: TObject);
    procedure ToolButtonSaveClick(Sender: TObject);
    procedure ToolButtonCopyClick(Sender: TObject);
    procedure ToolButtonPasteClick(Sender: TObject);
    procedure ClearButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
  public
  end;

var
  PictureEditForm: TPictureEditForm;

type
  TGraphicEdit = class(TRTPropertyEditor)
    private
      form   : TPictureEditForm;

    public
      function  GetAttributes : TRTPropertyAttributes; override;
      function  GetValue : string; override;
      procedure Edit; override;
      procedure GetImage; virtual; abstract;
      procedure SetImage; virtual; abstract;
  end;

  TPictureEdit = class(TGraphicEdit)
    procedure GetImage; override;
    procedure SetImage; override;
  end;

  TBitmapEdit = class(TGraphicEdit)
    procedure GetImage; override;
    procedure SetImage; override;
  end;

  TIconEdit = class(TGraphicEdit)
    procedure GetImage; override;
    procedure SetImage; override;
    function  GetValue : string; override;
  end;

implementation

function  TGraphicEdit.GetAttributes : TRTPropertyAttributes;
begin
  result := [rtDialog];
end;

{------------------------------------------------------------------}

function  TGraphicEdit.GetValue : string;
begin
  if GetOrdValue = 0 then
    result := '(None)'
  else
    result := '(TBitmap)';
end;

{------------------------------------------------------------------}

procedure TGraphicEdit.Edit;
var
  comp : TPersistent;
begin
  form := TPictureEditForm.Create(nil);
  with form do
    try
      comp := GetPersistent(0);
      if Comp is TComponent then
        Caption := TComponent(Comp).Name + '.' + GetName
      else
        Caption := GetName;

      GetImage;
      if ShowModal = idOk then
        SetImage;
    finally
      Free;
    end;
end;

{------------------------------------------------------------------}

procedure TPictureEdit.GetImage;
begin
  form.Image.Picture.Assign(TPicture(GetOrdValue).Bitmap);
end;

{------------------------------------------------------------------}

procedure TPictureEdit.SetImage;
begin
  TPicture(GetOrdValue).Assign(form.Image.Picture);
  Modified;
end;

{------------------------------------------------------------------}

procedure TBitmapEdit.GetImage;
begin
  form.Image.Picture.Assign(TBitmap(GetOrdValue));
end;

{------------------------------------------------------------------}

procedure TBitmapEdit.SetImage;
begin
  TBitmap(GetOrdValue).Assign(form.Image.Picture.Bitmap);
   Modified;
end;

{------------------------------------------------------------------}

procedure TIconEdit.GetImage;
begin
  form.Image.Picture.Icon.Assign(TIcon(GetOrdValue));
  Modified;
end;

{------------------------------------------------------------------}

procedure TIconEdit.SetImage;
begin
  TIcon(GetOrdValue).Assign(form.Image.Picture.Icon);
  Modified;
end;

{------------------------------------------------------------------}

function TIconEdit.GetValue : string;
begin
  result := '(TIcon)';
end;

{------------------------------------------------------------------}
{$R *.DFM}

{$IFNDEF DFS_DELPHI_3_UP}
Type
  TOpenPictureDialog=TOpenDialog;
  TSavePictureDialog=TSaveDialog;
{$ENDIF}

procedure TPictureEditForm.ToolButtonLoadClick(Sender: TObject);
begin
  with TOpenPictureDialog.Create(self) do
    try

      Filter := tePictFilter;

      if Execute then
        Image.Picture.LoadFromFile(FileName);
    finally
      Free;
    end;
end;

{------------------------------------------------------------------}

procedure TPictureEditForm.ToolButtonSaveClick(Sender: TObject);
begin
  with TSavePictureDialog.Create(self) do
    try

      Filter := tePictFilter;

      if Execute  then
        Image.Picture.SaveToFile(FileName);
    finally
      Free;
    end;
end;

{------------------------------------------------------------------}

procedure TPictureEditForm.ToolButtonCopyClick(Sender: TObject);
begin
  Clipboard.Assign(Image.Picture);
end;

{------------------------------------------------------------------}

type
  TMClipboard = class (TClipboard)
  end;

procedure TPictureEditForm.ToolButtonPasteClick(Sender: TObject);
begin
  TMClipboard(Clipboard).AssignTo(Image.Picture);
end;

{------------------------------------------------------------------}

procedure TPictureEditForm.ClearButtonClick(Sender: TObject);
begin
  Image.Picture.Graphic:=nil;
end;

{------------------------------------------------------------------}

procedure TPictureEditForm.FormCreate(Sender: TObject);
begin
end;

{------------------------------------------------------------------}

end.
