unit IsLists;
{*****************************************************************
  Copyright (c) 2000 IMCA SYSTEMS
  All Rights Reserved.
  MODULE: Some lists classes
*****************************************************************}
interface
{$I DFS.inc}
uses Windows, SysUtils;
const
 { Maximum list size }
  MaxListSize = High(Integer) div 16;

 type
  PPointerList = ^TPointerList;
  TPointerList = array[0..MaxListSize - 1] of Pointer;
  TISListSortCompare = function (Item1, Item2: Pointer): Integer;

  EListError = class(Exception);

  TISList = class(TObject)
  private
    FList: PPointerList;
    FCount: Integer;
    FCapacity: Integer;
  protected
    function Get(Index: Integer): Pointer;
    procedure Grow; virtual;
    procedure Put(Index: Integer; Item: Pointer);
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetCount(NewCount: Integer);
  public
    destructor Destroy; override;
    function Add(Item: Pointer): Integer;
    class procedure Error(const Msg: string; Data: Integer); virtual;
    procedure Clear; dynamic;
    function Delete(Index: Integer) : boolean;
    function Exchange(Index1, Index2: Integer) : boolean;
    function Expand: TISList;
    function First: Pointer;
    function IndexOf(Item: Pointer): Integer;
    function Insert(Index: Integer; Item: Pointer) : boolean;
    function Last: Pointer;
    function Move(CurIndex, NewIndex: Integer) : boolean;
    function Remove(Item: Pointer): Integer;
    procedure Pack;
    procedure Sort(Compare: TISListSortCompare);
    property Capacity: Integer read FCapacity write SetCapacity;
    property Count: Integer read FCount write SetCount;
    property Items[Index: Integer]: Pointer read Get write Put; default;
    property List: PPointerList read FList;
  end;
  {$ifdef DFS_COMPILER_3_UP}
  TISThreadList = class
  private
    FList: TIsList;
    FLock: TRTLCriticalSection;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(Item: Pointer);
    procedure Clear;
    function  LockList: TIsList;
    procedure Remove(Item: Pointer);
    procedure UnlockList;
  end;

    IInterfaceList = interface
  ['{285DEA8A-B865-11D1-AAA7-00C04FB17A72}']
    function Get(Index: Integer): IUnknown;
    function GetCapacity: Integer;
    function GetCount: Integer;
    procedure Put(Index: Integer; Item: IUnknown);
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetCount(NewCount: Integer);

    procedure Clear;
    procedure Delete(Index: Integer);
    procedure Exchange(Index1, Index2: Integer);
    function First: IUnknown;
    function IndexOf(Item: IUnknown): Integer;
    function Add(Item: IUnknown): Integer;
    procedure Insert(Index: Integer; Item: IUnknown);
    function Last: IUnknown;
    function Remove(Item: IUnknown): Integer;
    procedure Lock;
    procedure Unlock;
    property Capacity: Integer read GetCapacity write SetCapacity;
    property Count: Integer read GetCount write SetCount;
    property Items[Index: Integer]: IUnknown read Get write Put; default;
  end;

  TInterfaceList = class(TInterfacedObject, IInterfaceList)
  private
    FList: TISThreadList;
  protected
    { IInterfaceList }
    function Get(Index: Integer): IUnknown;
    function GetCapacity: Integer;
    function GetCount: Integer;
    procedure Put(Index: Integer; Item: IUnknown);
    procedure SetCapacity(NewCapacity: Integer);
    procedure SetCount(NewCount: Integer);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure Delete(Index: Integer);
    procedure Exchange(Index1, Index2: Integer);
    function Expand: TInterfaceList;
    function First: IUnknown;
    function IndexOf(Item: IUnknown): Integer;
    function Add(Item: IUnknown): Integer;
    procedure Insert(Index: Integer; Item: IUnknown);
    function Last: IUnknown;
    function Remove(Item: IUnknown): Integer;
    procedure Lock;
    procedure Unlock;
    property Capacity: Integer read GetCapacity write SetCapacity;
    property Count: Integer read GetCount write SetCount;
    property Items[Index: Integer]: IUnknown read Get write Put; default;
  end;
  {$endif}
  procedure MoveBytes( const Source; var Dest; count : Integer);

implementation
uses IsConst, Consts;

procedure MoveBytes( const Source; var Dest; count : Integer);
asm
{     ->EAX     Pointer to source       }
{       EDX     Pointer to destination  }
{       ECX     Count                   }

        PUSH    ESI
        PUSH    EDI

        MOV     ESI,EAX
        MOV     EDI,EDX

        MOV     EAX,ECX

        CMP     EDI,ESI
        JA      @@down
        JE      @@exit

        SAR     ECX,2           { copy count DIV 4 dwords       }
        JS      @@exit

        REP     MOVSD

        MOV     ECX,EAX
        AND     ECX,03H
        REP     MOVSB           { copy count MOD 4 bytes        }
        JMP     @@exit

@@down:
        LEA     ESI,[ESI+ECX-4] { point ESI to last dword of source     }
        LEA     EDI,[EDI+ECX-4] { point EDI to last dword of dest       }

        SAR     ECX,2           { copy count DIV 4 dwords       }
        JS      @@exit
        STD
        REP     MOVSD

        MOV     ECX,EAX
        AND     ECX,03H         { copy count MOD 4 bytes        }
        ADD     ESI,4-1         { point to last byte of rest    }
        ADD     EDI,4-1
        REP     MOVSB
        CLD
@@exit:
        POP     EDI
        POP     ESI
end;
{TISLIST Class}
destructor TISList.Destroy;
begin
  Clear;
end;

class procedure TISList.Error(const Msg: string; Data: Integer);

  function ReturnAddr: Pointer;
  asm
          MOV     EAX,[EBP+4]
  end;

begin
  raise EListError.CreateFmt(Msg, [Data]) at ReturnAddr;
end;

function TISList.Add(Item: Pointer): Integer;
begin
  Result := FCount;
  if Result = FCapacity then Grow;
  FList^[Result] := Item;
  Inc(FCount);
end;

procedure TISList.Clear;
begin
  SetCount(0);
  SetCapacity(0);
end;

function TISList.Delete(Index: Integer) : boolean;
begin
  if not ((Index < 0) or (Index >= FCount)) then
  begin
   Dec(FCount);
   if Index < FCount then
    MoveBytes(FList^[Index + 1], FList^[Index],
      (FCount - Index) * SizeOf(Pointer));
    Result:=true;
  end
  else Result:=false;
end;


function TISList.Exchange(Index1, Index2: Integer) : boolean;
var
  Item: Pointer;
begin
  if not (((Index1 < 0) or (Index1 >= FCount))
     or ((Index2 < 0) or (Index2 >= FCount))) then
  begin
   Item := FList^[Index1];
   FList^[Index1] := FList^[Index2];
   FList^[Index2] := Item;
   Result :=true;
  end
  else Result:=false;
end;

function TISList.Expand: TISList;
begin
  if FCount = FCapacity then Grow;
  Result := Self;
end;

function TISList.First: Pointer;
begin
  Result := Get(0);
end;

function TISList.Get(Index: Integer): Pointer;
begin
  if not ((Index < 0) or (Index >= FCount)) then
    Result := FList[Index]
  else
    Result := nil;
end;

procedure TISList.Grow;
var
  Delta: Integer;
begin
  if FCapacity > 64 then
    Delta := FCapacity div 4
  else //if FCapacity > 8 then
    Delta := 16;
  //else
  //    Delta := 4;}
  SetCapacity(FCapacity + Delta);
end;

function TISList.IndexOf(Item: Pointer): Integer;
begin
  Result := 0;
  while (Result < FCount) and (FList^[Result] <> Item) do Inc(Result);
  if Result = FCount then Result := -1;
end;

function TISList.Insert(Index: Integer; Item: Pointer) : boolean;
begin
 if not ((Index < 0) or (Index > FCount)) then
 begin
   if FCount = FCapacity then Grow;
   if Index < FCount then
    MoveBytes(FList^[Index], FList^[Index + 1],
      (FCount - Index) * SizeOf(Pointer));
   FList^[Index] := Item;
   Inc(FCount);
   Result := true;
 end
 else Result:=false;
end;

function TISList.Last: Pointer;
begin
  Result := Get(FCount - 1);
end;

function TISList.Move(CurIndex, NewIndex: Integer) : boolean;
var
  Item: Pointer;
begin
  Result:=false;
  if CurIndex <> NewIndex then
  begin
    Item := Get(CurIndex);
    if Item <> nil then
    begin
     Result := Delete(CurIndex);
     if Result then
       Insert(NewIndex, Item);
    end;
  end;
end;

procedure TISList.Put(Index: Integer; Item: Pointer);
begin
  if not ((Index < 0) or (Index >= FCount)) then
  begin
   FList^[Index] := Item;
  end;
end;

function TISList.Remove(Item: Pointer): Integer;
begin
  Result := IndexOf(Item);
  if Result <> -1 then Delete(Result);
end;

procedure TISList.Pack;
var
  I: Integer;
begin
  for I := FCount - 1 downto 0 do if Items[I] = nil then Delete(I);
end;

procedure TISList.SetCapacity(NewCapacity: Integer);
begin
  if not ((NewCapacity < FCount) or (NewCapacity > MaxListSize)) then
   if NewCapacity <> FCapacity then
   begin
    ReallocMem(FList, NewCapacity * SizeOf(Pointer));
    FCapacity := NewCapacity;
   end;
end;

procedure TISList.SetCount(NewCount: Integer);
begin
  if not ((NewCount < 0) or (NewCount > MaxListSize)) then
  begin
    if NewCount > FCapacity then SetCapacity(NewCount);
    if NewCount > FCount then
      FillChar(FList^[FCount], (NewCount - FCount) * SizeOf(Pointer), 0);
    FCount := NewCount;
  end;
end;

procedure QuickSort(SorTISList: PPointerList; L, R: Integer;
  SCompare: TISListSortCompare);
var
  I, J: Integer;
  P, T: Pointer;
begin
  repeat
    I := L;
    J := R;
    P := SorTISList^[(L + R) shr 1];
    repeat
      while SCompare(SorTISList^[I], P) < 0 do Inc(I);
      while SCompare(SorTISList^[J], P) > 0 do Dec(J);
      if I <= J then
      begin
        T := SorTISList^[I];
        SorTISList^[I] := SorTISList^[J];
        SorTISList^[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(SorTISList, L, J, SCompare);
    L := I;
  until I >= R;
end;

procedure TISList.Sort(Compare: TISListSortCompare);
begin
  if (FList <> nil) and (Count > 0) then
    QuickSort(FList, 0, Count - 1, Compare);
end;


{ TISThreadList }
{$ifdef DFS_COMPILER_3_UP}
constructor TISThreadList.Create;
begin
  inherited Create;
  InitializeCriticalSection(FLock);
  FList := TIsList.Create;
end;

destructor TISThreadList.Destroy;
begin
  LockList;    // Make sure nobody else is inside the list.
  try
    FList.Free;
    inherited Destroy;
  finally
    UnlockList;
    DeleteCriticalSection(FLock);
  end;
end;

procedure TISThreadList.Add(Item: Pointer);
begin
  LockList;
  try
    if FList.IndexOf(Item) = -1 then
      FList.Add(Item);
  finally
    UnlockList;
  end;
end;

procedure TISThreadList.Clear;
begin
  LockList;
  try
    FList.Clear;
  finally
    UnlockList;
  end;
end;

function  TISThreadList.LockList: TIsList;
begin
  EnterCriticalSection(FLock);
  Result := FList;
end;

procedure TISThreadList.Remove(Item: Pointer);
begin
  LockList;
  try
    FList.Remove(Item);
  finally
    UnlockList;
  end;
end;

procedure TISThreadList.UnlockList;
begin
  LeaveCriticalSection(FLock);
end;


constructor TInterfaceList.Create;
begin
  inherited Create;
  FList := TISThreadList.Create;
end;

destructor TInterfaceList.Destroy;
begin
  Clear;
  FList.Free;
  inherited Destroy;
end;

procedure TInterfaceList.Clear;
var
  I: Integer;
begin
  if FList <> nil then
  begin
    with FList.LockList do
    try
      for I := 0 to Count - 1 do
        IUnknown(List[I]) := nil;
      Clear;
    finally
      Self.FList.UnlockList;
    end;
  end;
end;

procedure TInterfaceList.Delete(Index: Integer);
begin
  with FList.LockList do
  try
    Self.Put(Index, nil);
    Delete(Index);
  finally
    Self.FList.UnlockList;
  end;
end;

function TInterfaceList.Expand: TInterfaceList;
begin
  with FList.LockList do
  try
    Expand;
    Result := Self;
  finally
    Self.FList.Unlocklist;
  end;
end;

function TInterfaceList.First: IUnknown;
begin
  Result := Get(0);
end;

function TInterfaceList.Get(Index: Integer): IUnknown;
begin
  with FList.LockList do
  try
    Result := IUnknown(List[Index]);
  finally
    Self.FList.UnlockList;
  end;
end;

function TInterfaceList.GetCapacity: Integer;
begin
  with FList.LockList do
  try
    Result := Capacity;
  finally
    Self.FList.UnlockList;
  end;
end;

function TInterfaceList.GetCount: Integer;
begin
  with FList.LockList do
  try
    Result := Count;
  finally
    Self.FList.UnlockList;
  end;
end;

function TInterfaceList.IndexOf(Item: IUnknown): Integer;
begin
  with FList.LockList do
  try
    Result := IndexOf(Pointer(Item));
  finally
    Self.FList.UnlockList;
  end;
end;

function TInterfaceList.Add(Item: IUnknown): Integer;
begin
  with FList.LockList do
  try
    Result := Add(nil);
    IUnknown(List[Result]) := Item;
  finally
    Self.FList.UnlockList;
  end;
end;

procedure TInterfaceList.Insert(Index: Integer; Item: IUnknown);
begin
  with FList.LockList do
  try
    Insert(Index, nil);
    IUnknown(List[Index]) := Item;
  finally
    Self.FList.UnlockList;
  end;
end;

function TInterfaceList.Last: IUnknown;
begin
  with FList.LockList do
  try
    Result := Self.Get(Count - 1);
  finally
    Self.FList.UnlockList;
  end;
end;

procedure TInterfaceList.Put(Index: Integer; Item: IUnknown);
begin
  with FList.LockList do
  try
    if (Index < 0) or (Index >= Count) then Error(SListIndexError, Index);
    IUnknown(List[Index]) := Item;
  finally
    Self.FList.UnlockList;
  end;
end;

function TInterfaceList.Remove(Item: IUnknown): Integer;
begin
  with FList.LockList do
  try
    Result := IndexOf(Pointer(Item));
    if Result > -1 then
    begin
      IUnknown(List[Result]) := nil;
      Delete(Result);
    end;
  finally
    Self.FList.UnlockList;
  end;
end;

procedure TInterfaceList.SetCapacity(NewCapacity: Integer);
begin
  with FList.LockList do
  try
    Capacity := NewCapacity;
  finally
    Self.FList.UnlockList;
  end;
end;

procedure TInterfaceList.SetCount(NewCount: Integer);
begin
  with FList.LockList do
  try
    Count := NewCount;
  finally
    Self.FList.UnlockList;
  end;
end;

procedure TInterfaceList.Exchange(Index1, Index2: Integer);
begin
  with FList.LockList do
  try
    Exchange(Index1, Index2);
  finally
    Self.FList.UnlockList;
  end;
end;

procedure TInterfaceList.Lock;
begin
  FList.LockList;
end;

procedure TInterfaceList.Unlock;
begin
  FList.UnlockList;
end;
{$endif}

end.



