unit ishtmprn;
{*********************************************************}
//  Copyright (c) 2000 IMCA SYSTEMS
//  All Rights Reserved.
//  Version
//  MODULE: Printer Object with Preview abilities
{*********************************************************}
interface

uses
  Windows, Graphics, Printers, IsLists;

type
  TISPrinter = class(TObject)
  private
    { Private declarations }
    FPageNumber : Integer;
    FCanvases : TISList;
    FMetafiles : TISList;
    FDonePrinting : Boolean;

    function GetPageCount : Integer;
    procedure SetCurrentPage(Index : Integer);
    function GetCanvas(Index : Integer): TMetafileCanvas;
    function GetMetafile(Index : Integer): TMetafile;
    function GetPagewidth : integer;
    function GetPageheight : integer;
    function GetTopMargin : integer;
    function GetLeftMargin : integer;

  protected
    { Protected declarations }
  public
    { Public declarations }
    XPixelPerInch : Integer;
    YPixelPerInch : Integer;

    XOff : Integer;
    YOff : Integer;
    Title : String;
    DC: HDC;
    function PageSize : TPoint;
    procedure PrintPage(const PageNum : Integer);
    procedure PrintIt;
    procedure Preview;
    procedure SaveToFile(const TheFile : String; Page : integer);
    function XInch(const aSingle : Single): Integer;
    function YInch(const aSingle : Single): Integer;

    procedure Begindoc;
    procedure EndDoc;
    procedure Abort;//!
    function NewPage : Integer;
    constructor Create;
    destructor Destroy; override;

    property PageCount : Integer  read GetPageCount;
    property CurrentPage : Integer read FPageNumber write
      SetCurrentPage;
    property Canvases[Index : Integer] : TMetafileCanvas read GetCanvas;
    property Metafiles[Index : Integer] : TMetafile read GetMetafile;
    property Pagewidth : integer read GetPagewidth;
    property PageHeight : integer read GetPageheight;
    property TopMargin : integer read GetTopMargin;
    property LeftMargin : integer read GetLeftMargin;
  end;

implementation

uses consts, ishtmprv;{preview form}

function TISPrinter.GetPageCount : Integer;
begin
     Result := FCanvases.Count;
end;

procedure TISPrinter.SetCurrentPage(Index : Integer);
begin
 if (Index <= PageCount) AND (Index > 0) then
    FPageNumber := Index;
end;

function TISPrinter.GetCanvas(Index : Integer): TMetafileCanvas;
begin
 if (Index > 0) AND (Index <= PageCount) then
   Result := TMetafileCanvas(FCanvases[Index - 1])
 else
   Result := nil;
end;

function TISPrinter.GetMetafile(Index : Integer): TMetafile;
begin
 if (Index > 0) AND (Index <= PageCount) {AND FDonePrinting} then
  Result := TMetafile(FMetafiles[Index - 1])
 else
  Result := nil;
end;

function TISPrinter.GetPagewidth : integer;
begin
  result:=Printer.Pagewidth;
end;

function TISPrinter.GetLeftMargin : integer;
begin
  Result := GetDeviceCaps(Printer.Handle, PHYSICALOFFSETX);
end;

function TISPrinter.GetTopMargin : integer;
begin
  Result := GetDeviceCaps(Printer.Handle, PHYSICALOFFSETY);
end;

function TISPrinter.GetPageheight : integer;
begin
  result:=Printer.Pageheight;
end;

function TISPrinter.PageSize : TPoint;
begin
   Escape(Printer.Handle, GETPHYSPAGESIZE, 0, nil, @Result);
end;


procedure TISPrinter.PrintPage(const PageNum : Integer);
var aRect : TRect;
begin
if not FDonePrinting then Abort;
if (PageNum > 0) AND (PageNum <= PageCount) then
begin
  Printer.Title := Title;
  Printer.BeginDoc;
  try //setrect
   SetRect(aRect,0,0,Printer.PageWidth, Printer.PageHeight);
   Printer.Canvas.StretchDraw(aRect, Metafiles[PageNum]);
  finally
    Printer.EndDoc;
  end;
end;
end;

procedure TISPrinter.PrintIt;
var
 i : Integer;
 aRect : TRect;
begin
 if not FDonePrinting then Abort;
 if PageCount > 0 then begin
   Printer.Title := Title;
  if not Printer.Printing then
    Printer.BeginDoc;
  i := 1;
  try
  SetRect(aRect,0,0,Printer.PageWidth, Printer.PageHeight);
  Printer.Canvas.StretchDraw(aRect, Metafiles[i]);
  for i := 2 to PageCount do
  begin
    Printer.NewPage;
    Printer.Canvas.StretchDraw(aRect, Metafiles[i]);
  end;
  finally
    Printer.EndDoc;
  end;
  end;
end;

procedure TISPrinter.Preview;
var FHTMLPreviewForm : THTMLPreviewForm;
begin
 if not FDonePrinting then Abort;
 FHTMLPreviewForm := THTMLPreviewForm.Create(nil);
 try
   FHTMLPreviewForm.SeTISPrinter(Self);
   FHTMLPreviewForm.Caption := Title;
   FHTMLPreviewForm.ShowModal;
 finally
   FHTMLPreviewForm.Free;
 end;
end;

procedure TISPrinter.SaveToFile(const TheFile : String; Page : integer);
begin
 if FDonePrinting then
    Metafiles[Page].SaveToFile(Thefile);
end;


function TISPrinter.XInch(const aSingle : Single): Integer;
begin
  Result := Trunc(XPixelPerInch * aSingle);
end;

function TISPrinter.YInch(const aSingle : Single): Integer;
begin
  Result := Trunc(YPixelPerInch * aSingle);
end;

type
  TPrinterDevice = class
    Driver, Device, Port: String;
    //constructor Create(ADriver, ADevice, APort: PChar);
    //function IsEqual(ADriver, ADevice, APort: PChar): Boolean;
  end;

procedure TISPrinter.BeginDoc;
var
 i : Integer;
 p : TPoint;
 aDevMode : PDeviceMode;
begin
  aDevMode:=nil;
 for i := FMetafiles.Count - 1 downto 0 do
 begin
    TMetafile(FMetafiles[i]).Free;
 end;
 FCanvases.Clear;
 FMetafiles.Clear;
 FPageNumber := 0;
 FDonePrinting := False;
 Escape(Printer.Handle, GETPRINTINGOFFSET, 0, nil, @p);
 XOff := p.X;
 YOff := p.Y;
 XPixelPerInch := GetDeviceCaps(Printer.Handle, LOGPIXELSX);
 YPixelPerInch := GetDeviceCaps(Printer.Handle, LOGPIXELSY);
 //create Device context
 with TPrinterDevice(Printer.Printers.Objects[Printer.PrinterIndex]) do
 begin
   DC := CreateDC(PChar(Driver), PChar(Device), PChar(Port), aDevMode);
   if DC = 0 then
      raise EPrinter.Create(SInvalidPrinter);
 end;
 NewPage;
end;

function TISPrinter.NewPage : Integer;
begin
  Result:=0;
 if not FDonePrinting then
 begin
  Result := FMetafiles.Add(TMetafile.Create) + 1;
  if Printer.Orientation=poPortrait then
  begin//exchange values
    TMetafile(FMetafiles[Result - 1]).MMWidth := GetDeviceCaps(DC, HORZSIZE)*100;
    TMetafile(FMetafiles[Result - 1]).MMHeight := GetDeviceCaps(DC, VERTSIZE)*100;
    TMetafile(FMetafiles[Result - 1]).Transparent:=true;
  end
  else
  begin
    TMetafile(FMetafiles[Result - 1]).MMHeight := GetDeviceCaps(DC, HORZSIZE)*100;
    TMetafile(FMetafiles[Result - 1]).MMwidth := GetDeviceCaps(DC, VERTSIZE)*100;

  end;

  FCanvases.Add(TMetafileCanvas.Create(TMetafile(FMetafiles[Result - 1]),
     DC));


  FPageNumber := Result;
  with Canvases[Result] do
  begin
    Brush.style := bsClear;
    Font.Name := 'Arial';
    Font.Size := 10;
    Refresh;
  end;
  end;
end;

procedure TISPrinter.EndDoc;
var
  i : integer;
begin
 if not FDonePrinting then
 begin
  for i := 1 to PageCount do
  begin
     Canvases[i].Free;
  end;
  //delete device context
  //DeleteDC(DC);
  //DC := 0;
  FDonePrinting := True;
 end;
end;

procedure TISPrinter.Abort;
var i : integer;
begin
 for i := FMetafiles.Count - 1 downto 0 do
 begin
  TMetafile(FMetafiles[i]).Free;
 end;
 FCanvases.Free;
 FMetafiles.Free;
 FPageNumber:=-1;
 if Printer.Printing then
   Printer.Abort;
end;


constructor TISPrinter.Create;
begin
  inherited Create;
  FCanvases := TISList.Create;
  FMetafiles := TISList.Create;
  FPageNumber := 0;
  FDonePrinting := False;
end;

destructor TISPrinter.Destroy;
var i : Integer;
begin
  for i := FMetafiles.Count - 1 downto 0 do
  begin
    TMetafile(FMetafiles[i]).Free;
  end;
  FCanvases.Free;
  FMetafiles.Free;
  if Printer.Printing then
     Printer.Abort;

  inherited Destroy;
end;



end.
