unit PropTree;

{****************************************************************
   Copyright (c) 1999 IMCA SYSTEMS
   All Rights Reserved.
   Creation Date: 06.08.98
   Version 1.01b
   MODULE: TPropTree - Runtime Object Inspector based on CustomFlyTree
   1. Expected Additional Property Editors
   2. Different styles of Drawing
   3. Grouping properties
*****************************************************************}

{$I DFS.inc}
{$R *.RES}
{$M+}
interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, ExtCtrls,Controls, Forms, Dialogs,
  exgrid, finplace, RapTree, Flytree, RtProps, TypInfo {$IFDEF DFS_DELPHI_5_UP},Menus,Contnrs{$ENDIF};

type
  TPropertiesTree = class(TCustomFlyTreeView)
  private
    { Private declarations }
    FStateLock : integer;
    CompList : TPersistentList; //container for properties
    FOnModified : TNotifyEvent;
    CurrentListToAdd : TList;
    aPopupListToAdd : TStringList;
    FPropTypes : TTypeKinds;
    FProptypesChanged : Boolean;
    FMarkRow : TPanelBevel;
    FEXPF: String;
    procedure SetPropTypes(aTypes : TTypeKinds);
    procedure FitToWidth;
    procedure _GetPropEditor(PropEdit : TRTPropertyEditor);
    procedure GetEditorsList(Components: TPersistentList; Props: TTypeKinds;  Proc: TGetPropEditProc);
    //colsizing uitlity
    procedure WMSetCursor(var Msg: TWMSetCursor); message WM_SETCURSOR;
    //utility procs for fill props and subprops to any depth
    procedure FillList(const S : String);
    procedure InitColumns;
    procedure SetMarkRow(aMark : TPanelBevel);
    procedure SetEXPF(const Value: String);
    procedure AssignFilter;
  protected
    { Protected declarations }
    procedure RefreshColumnsmap;override;
    procedure InsertRows(atRow, aCount : longint); override;
    procedure DeleteRows(atRow, aCount : longint); override;
    //drawing proc
    procedure Paint;override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
         X, Y: Integer); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Loaded; override;
    procedure ColWidthsChanged; override;
    //function SelectCell(var ACol, ARow: Longint): Boolean; override;
    procedure PropertyModified(Sender : TObject);
    function SelectCell(var ACol, ARow: Longint; MoveAnchor : Boolean;  Shift: TShiftState; SelWay : TSelectionWay): Boolean; override;
    procedure aGetEditText(Node: TFlyNode; index : integer; Var aData: string; var aPwdChar : boolean; Var MaxLength : integer); override;
    procedure AGetEditStyle(Node: TFlyNode; index : integer; Var aEditStyle : TEditStyle); override;
    procedure DrawDataNode(aCanvas : TCanvas; aNode : TFlyNode; aRect : TRect; aCol, aRow : Longint; AState: TExGridDrawState; Var Continue : boolean);override;
    procedure AValidateNodeData(Node: TFlyNode; index : integer; var aData : string; var Cancel : Boolean); override;
    procedure AEditButtonClick(Node: TFlyNode; index : integer); override;
    procedure AIsReadOnlyNode(Node: TFlyNode; index : integer; var aReadOnly : Boolean); override;
    procedure AGetPicklist(Node: TFlyNode; index : integer; var aPickList : Tstrings); override;
    procedure MoveEditor(Editor : TFlyInplaceEdit; aRect : TRect); override;
    procedure DefineProperties(Filer: TFiler); override;
    function CanColSizing(X,Y : integer) : Boolean;override;
    procedure Lockstate;
    procedure Unlockstate;
    function GetCount : integer;
    function GetModified : Boolean;
    procedure AddPropertiesFromList(aList : Tlist; Node : TFlyNode); virtual;
    procedure AddSubproperties(aNode : TFlyNode; aPropEditor : TRTPropertyEditor);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent);override;
    destructor Destroy;override;
    function CreateTo(aOwner : TComponent) : TCustomFlyTreeView; override;
    procedure InvalidateCell(ACol, ARow: Longint); override;
    procedure SelectComponents(Components : TList);//select components for PropertiesTree
    procedure SelectComponent(Component : TPersistent);//select components for PropertiesTree
    procedure Deselect;//clear tree and freeing property editors
    //new
    function Add(Item: TPersistent): Integer;
    property Count: Integer read GetCount;
    property Modified : boolean read GetModified;
    property ExcludePropertiesFilter : String read FEXPF write SetEXPF;
  published
    property OutlineColor;
    property OutlineStyle;
    property ButtonSize;
    {$ifdef DFS_COMPILER_4_UP}
    property ScrollBarsStyle;
    property ScrollBarsColor;
    {$endif}
    property FitToHeight;
    property FitColumnToClientWidth;
    //copy/cut/paste management
    property OnPasteSelection;
    property OnCopySelection;
    property OnCutSelection;

    property OnModified  : TNotifyEvent read FOnModified write FOnModified;
    property MarkRow : TPanelBevel read FMarkRow write SetMarkRow default bvLowered;// Type of mark current row
    property PropertyTypes : TTypeKinds read FPropTypes write SetPropTypes default tkProperties;// Set of Property types for fetch
    property OnMouseEnter;
    property OnMouseLeave;
    property Flat;
    property OnGetNodeData;
    property OnColumnClick;
    property OnValidateNodeData;
    property OnEditButtonClick;
    property OnGetNodeAttributes;
    property SelectedTextColor;
    property SelectedBackgroundColor;
    property Align;
    property Borderstyle;
    property Ctl3D;
    property Color;
    property ParentCtl3D;
    {$ifdef DFS_COMPILER_4_UP}
    property BorderWidth;
    {$endif}
    property AutoExpand;
    property HotTrack;
    property ToolTips;
    property TooltipColor;
    property ToolTipPause;
    {$ifdef DFS_COMPILER_4_UP}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
    property OnEndDock;
    property OnStartDock;
    property DoubleBuffered;
    {$endif}
    property ShowHint;
    property ParentShowHint;
    property DragMode;
    property Enabled;
    property Font;
    property Indent;
    property ParentColor default False;
    property ParentFont;
    property PopupMenu;
    property SortType;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnChange;
    property OnChanging;
    property OnClick;
    property OnCollapsing;
    property OnCollapsed;
    property OnCompare;
    property OnDblClick;
    property OnDeletion;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    //grid
    property GridLineWidth;
    property Options;
    property DefaultRowHeight;
    property DefaultColWidth;
    property ScrollBars;
    //tree
    property HideSelection;
    property Showlines default false;
    property ShowRoot;
    property ShowButtons;
  end;


  TCharSet= Set of char;

  var FPropsRegistered : boolean;


procedure RegisterProps;

implementation

uses IsConst, Consts, FlyEdit, ISStrEd, ISPictEdit
{$IFNDEF Registered}
,FTAbout
{$ENDIF};


Function RemoveChars(Const Cs : TCharSet; Const S : String) : String;
Var
  i:Integer;
begin
  Result := S;
  for i := Length(Result) downto 1 do
    if Result[I] in Cs then
      Delete(Result, I, 1);
end;

function FindComponent(Const FullPath : String):TComponent;
Var
  i:Integer;
  FormName, CompName : String;
begin
  Result:=nil;
  FormName := ChangeFileExt(FullPath,'');
  CompName := RemoveChars(['.'], ExtractFileExt(FullPath));

  for i:=0 to {$IFDEF DFS_DELPHI_3_UP}Screen.CustomFormCount{$ELSE}Screen.FormCount{$ENDIF} - 1 do
    if CompareText(FormName, Screen.{$IFDEF DFS_DELPHI_3_UP}CustomForms{$ELSE}Forms{$ENDIF}[i].Name) = 0 then
    begin
      Result:= Screen.{$IFDEF DFS_DELPHI_3_UP}CustomForms{$ELSE}Forms{$ENDIF}[i].FindComponent(CompName);
      exit;
    end;

  With Screen do
  for i := 0 to DataModuleCount - 1 do
   if CompareText(DataModules[i].Name,FormName)=0 then
   begin
     result := DataModules[i].FindComponent(CompName);
     exit;
   end;

end;


//-----------------------------------------------------------------------------
constructor TPropertiesTree.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FitColumnToClientWidth := true;
   FixedCols:=0;
   ColCount:=2;
   Fixedrows:=0;
   RowCount:=0;
   ShowLines := false;
   FPropTypes := tkProperties;
   FSizingAboveFixed:=true;
   AutoExpand:= true;
   CompList := TPersistentList.Create;
   Deselect;
end;

//-----------------------------------------------------------------------------
destructor TPropertiesTree.Destroy;
begin
   Lockstate;
   Deselect;
   CompList.Free;
   CompList := nil;
   inherited Destroy;
end;

function TPropertiesTree.CreateTo(aOwner : TComponent) : TCustomFlyTreeView;
begin
   Result := TPropertiesTree.Create(aOwner);
end;

function TPropertiesTree.CanColSizing(X,Y : integer) : Boolean;
begin
   Result:=(Items.Count>0);
end;

//-----------------------------------------------------------------------------
type
  TPersistentCrack = class(TPersistent)
  end;

//-----------------------------------------------------------------------------
function TPropertiesTree.GetCount : Integer;
begin
   Result := CompList.Count;
end;
//-----------------------------------------------------------------------------
function TPropertiesTree.GetModified : Boolean;
begin
   Result := CompList.Changed;
end;

//-----------------------------------------------------------------------------
function TPropertiesTree.Add(Item: TPersistent): Integer;
var aList : TList;
    i : integer;
begin
  aList := TList.Create;
  try
   for i:=0 to CompList.Count-1 do
     aList.Add(CompList[i]);
   Result := aList.Add(Item);
   SelectComponents(aList);
  finally
    aList.Free;
  end;
end;

//-----------------------------------------------------------------------------
procedure TPropertiesTree._GetPropEditor(PropEdit : TRTPropertyEditor);
begin
  if CurrentListtoAdd <> nil then
  begin
     CurrentListToAdd.Add(PropEdit);
     if Propedit<> nil then
        Propedit.OnModified := PropertyModified;
  end;
end;

procedure TPropertiesTree.DefineProperties(Filer: TFiler);
begin
   //Columns.Clear;//prevent for load/write columns
   FWriteColumns := false;
   FWriteMap := false;
   inherited DefineProperties(Filer);
end;

procedure TPropertiesTree.GetEditorsList(Components: TPersistentList; Props: TTypeKinds;  Proc: TGetPropEditProc);
begin
   GetPersistentProperties(Components, Props, Proc);
end;

procedure TPropertiesTree.AddSubproperties(aNode : TFlyNode; aPropEditor : TRTPropertyEditor);
var aList : TList;
begin
   aList := TList.Create;
   try
     //point to callback procedure for addint to needed list
     CurrentListToAdd :=aList;
     aPropEditor.GetProperties(_GetPropEditor);
     CurrentListToAdd :=nil;
     AddPropertiesFromList(aList, aNode);
   finally
     aList.Free;
   end;
end;

procedure TPropertiesTree.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if Selected<>nil then
  begin
    if (Selected.HasChildren) then
    begin
      if Selected.Expanded and (Key=VK_SUBTRACT) then
         Selected.Expanded:=false
      else if not (Selected.Expanded) and (Key=VK_ADD) then
         Selected.Expanded:=true
      else if Key = VK_RETURN then
         Selected.Expanded := not Selected.Expanded;
     end;
     inherited;
  end;

end;



procedure TPropertiesTree.MouseDown(Button: TMouseButton; Shift: TShiftState;
X, Y: Integer);
begin

    if Getcursor = Screen.Cursors[crHSplit] then
       SetGridstate(gsColsizing)
    else
      SetGridstate(gsNormal);
    inherited
end;

procedure TPropertiesTree.WMSetCursor(var Msg: TWMSetCursor);
begin
 if Items.Count = 0 then exit;
 inherited;
 {if gocolsizing in Options then
   with Msg do
   begin
    if (GetGridstate = gsNormal) and (items.Count > 0) and
    ((GetHittest.X>=ColWidths[0]-3) and (GetHittest.X <= ColWidths[0]+3))
    then
    begin
       SetCursor(Screen.Cursors[crHSplit]);
    end
    else
    begin
      SetGridstate(gsNormal);
    end;
   end;
}   
end;


procedure TPropertiesTree.AddPropertiesFromList(aList : Tlist; Node : TFlyNode);
var
  PropName, PrevPropName : String;
  i : integer;
  aNode : TFlyNode;
begin
  PropName :='';
  PrevPropName :='';
   for i:=0 to aList.Count-1 do
   begin
    try
     PropName := TRTPropertyEditor(aList[i]).GetName;
     if (PropName <> PrevPropName) then
     begin
       aNode:=Items.AddChildObject(Node, PropName, aList[i]);
       if TRTPropertyEditor(aList[i]).GetAttributes * [rtSubProperties] <> [] then
       begin
         AddSubproperties(aNode,TRTPropertyEditor(aList[i]));
       end;
     end;
    except
       ShowMessage('Error on '+PropName);
    end;
     PrevPropName := PropName;
  end;
end;


procedure TPropertiesTree.RefreshColumnsmap;
begin
   if not (csloading in ComponentState) then
   begin
      ColCount :=2;
      inherited;
   end;
end;

procedure TPropertiesTree.InsertRows(atRow, aCount : longint);
begin
   inherited;
   FitTowidth;
end;

procedure TPropertiesTree.DeleteRows(atRow, aCount : longint);
begin
   inherited;
   FitTowidth;
end;

procedure TPropertiesTree.SelectComponent(Component : TPersistent);
var aList : TList;
begin
  aList := TList.Create;
  try
    aList.Add(Component);
    SelectComponents(AList);
  finally
    aList.Free;
  end;
end;

//-----------------------------------------------------------------------------
procedure TPropertiesTree.SelectComponents(Components : TList);
var    i : integer;
    OldCurrentProp : string;
    var aList : TList;
    procedure ToCurrentProp;
    var aNode : TFlyNode;
        NoName,Succ : Boolean;
    begin
      Succ:=false;
      NoName:=false;
      aNode:=items.GetFirstNode;
      if OldCurrentProp='' then
        OldCurrentProp:='.Name';
      while aNode<>nil do
      begin
         NoName:= (aNode.GetPath('.')='.Name');
         if aNode.GetPath('.') = OldCurrentProp then
         begin
             aNode.Selected:=true;
             Succ:=true;
             Break;
         end
         else
             aNode:=aNode.GetNext;
      end;
      if (not Succ) and (NoName) then//for get Name Property
      begin
         OldCurrentProp:='';
         ToCurrentProp;
      end
      else if (not Succ) and (items.GetFirstNode<>nil) then//if name property absent
      begin
          Selected:=items.GetFirstNode;//items.GetFirstNode.Selected:=true;
      end;
    end;
begin

   if Selected<>nil then
     OldCurrentProp := Selected.GetPath('.')
  else OldCurrentProp:='';

  //componentstate
  if (Components = nil) or (Components.Count = 0) then
  begin
    Deselect;
    exit;
  end
  else //if CompList list notr itself
  begin
   if (not CompList.EqualsWithList(Components)) or FPropTypesChanged then
   begin
     items.BeginUpdate;
     Deselect;
     CompList.Clear;
     RegisterProps;
     aList := TList.Create;
     try
       //Add Components to CompList
      try
        for i:=0 to Components.Count-1 do
          CompList.Add(TComponent(Components[i]));

         CurrentListToAdd := aList;//pointer for callback proc
         GetEditorsList(CompList, FPropTypes, _GetPropEditor);
      finally
        CurrentListToAdd := nil;
      end;
      //Fill Tree
      AddPropertiesFromList(aList,nil);
    finally
      aList.Free;
      AssignFilter;
      items.endUpdate;
      FitToWidth;
      if items.Count>0 then
      begin
        if ColCount>0 then Col:=1;
        EditorMode:=false;
        ToCurrentProp;//select usual or old current prop
        if Focused then
         EditorMode:=true;
      end;
    end;
   end;
  end;
end;

procedure TPropertiesTree.SetPropTypes(aTypes : TTypeKinds);
var aList : TList;
    i : integer;
begin
   if aTypes <> FPropTypes then
   begin
     aList := TList.Create;
     try
      FPropTypes :=aTypes;
      FPropTypesChanged := True;

      for i:=0 to CompList.Count-1 do
         aList.Add(CompList[i]);

      SelectComponents(aList);
     finally
       FPropTypesChanged := False;
       aList.Free;
     end;
   end;
end;

procedure TPropertiesTree.FitToWidth;
var i, aEffectWidth : integer;
begin
   //fit to width
  if FstateLock>0 then exit;
  LockState;
  try
   aEffectWidth :=0;
   if Columns= nil then exit;
   if Columns.Count<>2 then
   begin
    Defaultcolwidth:=Width div 2;
    //ColCount:=2;
    Columns.BeginUpdate;
    Columns.Clear;
    for i:=0 to 1 do
    begin
      With TTreeColumn.Create(Columns) do
      begin
       case i of
         0:Caption :='Properties';
         1:Caption :='Values';
       end;
       Width:=Defaultcolwidth;
      end;
    end;
    Columns.endUpdate;
    UpdateColumns;
   end;
  //DefaultColWidth := aWidth;
   if handleAllocated then
   begin
    for i:=0 to Columns.Count-1 do
      inc(aEffectWidth, Colwidths[i]);
    aEffectWidth := ColWidths[1] - (aEffectWidth-ClientWidth)-2;
    if aEffectWidth<> ColWidths[1] then
      ColWidths[1] := aEffectWidth;
   end;
  finally
      UnlockState;
  end;
end;

procedure TPropertiesTree.PropertyModified(Sender : TObject);
begin
   if Assigned(FOnModified) then
        FOnModified(Self);
end;

function TPropertiesTree.SelectCell(var ACol, ARow: Longint; MoveAnchor : Boolean;  Shift: TShiftState; SelWay : TSelectionWay): Boolean;
begin
   Result:=inherited SelectCell(ACol, ARow, MoveAnchor, Shift, SelWay);
   if Colcount>0 then
   begin
    if aCol = 0 then aCol := 1;
   end;

end;

procedure TPropertiesTree.FillList(const S : String);
begin
  if aPopupListToAdd <> nil then
    aPopupListToAdd.Add(S);
end;


procedure TPropertiesTree.Paint;
var aRect : TRect;
begin
 if FStateLock<>0 then exit;

 inherited;//tPanel
 if Items.Count>0 then
 begin
   aRect := BoxRect(Fixedcols, Row, VisibleColCount, Row);
   if FmarkRow <> bvNone then
   with canvas do
   begin
     //if goHorzline in Options then
     //  inflateRect(arect,0,-GridLineWidth);
     if FMarkRow=bvRaised then
        DrawEdge(Handle, aRect, BDR_RAISEDINNER, BF_RECT+BF_ADJUST)
     else
        DrawEdge(Handle, aRect, BDR_SUNKENINNER, BF_RECT+BF_ADJUST);
   end;
 end;
end;

procedure TPropertiesTree.InvalidateCell(ACol, ARow: Longint);
begin
  if FStateLock<>0 then exit;
  if FMarkRow<>bvNone then
   InvalidateRow(aRow)
  else
    inherited;
end;

procedure TPropertiesTree.MoveEditor(Editor : TFlyInplaceEdit; aRect : TRect);
begin
  if Editor<>nil then
  if Editor.EditStyle=etsCheckBox then
   Editor.Hide(false)
  else
  begin
    inflaterect(aRect,-1, -1);
    Editor.Move(ARect);
  end;
end;

procedure TPropertiesTree.SetMarkRow(aMark : TPanelBevel);
begin
   if aMark <> FMarkRow then
   begin
        FMarkRow :=aMark;
        Invalidate;
   end;
end;
procedure TPropertiesTree.AGetPicklist(Node: TFlyNode; index : integer; var aPickList : Tstrings);
begin
   if FStateLock<>0 then exit;
   if (Node.Data<>nil) then
   begin
      aPicklist.Clear;
      aPopupListToAdd:=TStringList(aPicklist);
      try
      TRTPropertyEditor(Node.Data).Getvalues(FillList);
      finally
       aPopupListToAdd :=nil;
      end;
      if (rtSortList in TRTPropertyEditor(Node.Data).GetAttributes) then
      TStringList(aPicklist).Sort;
   end;

end;

procedure TPropertiesTree.AIsReadOnlyNode(Node: TFlyNode; index : integer; var aReadOnly : Boolean);
begin
  if FStateLock<>0 then exit;
  if (Node.Data<>nil) and (index=1) then
  begin
     aReadOnly:= (rtReadOnly in TRTPropertyEditor(Node.Data).GetAttributes);
  end;

end;

procedure TPropertiesTree.AEditButtonClick(Node: TFlyNode; index : integer);
begin
    inherited;
    if Node.Data<>nil then
       TRTPropertyEditor(Node.Data).Edit;

end;


procedure TPropertiesTree.AValidateNodeData(Node: TFlyNode; index : integer; var aData : string; var Cancel : Boolean);
begin
  if FStateLock<>0 then exit;
  if Node.Data<>nil then
  begin
     if (TRTPropertyEditor(Node.Data).GetValue<>aData) {and
     not (paDialog in TRTPropertyEditor(Node.Data).GetAttributes)} then
     begin
       try
        try
         TRTPropertyEditor(Node.Data).Value := aData;
        except
           Cancel:=true;
           {$IFDEF DFS_COMPILER_3_UP}
           raise EPropertyError.Create(SInvalidPropertyValue);
           {$ELSE}
           raise EPropertyError.Create(LoadStr(SInvalidPropertyValue));
           {$ENDIF}
        end;
        if (Node.Level>0) then
        begin
          if Node.Parent<>nil then
           InvalidateRow(Node.Parent.GetRow);
        end;
       except
          Cancel:=true;
       end;
     end;
  end;

  inherited;
end;

procedure TPropertiesTree.AGetEditStyle(Node: TFlyNode; index : integer; Var aEditStyle : TEditStyle);
var aPropAttrs : TRTPropertyAttributes;
begin
   if FStateLock<>0 then exit;
   if Node<>nil then
   if (Node.Data<>nil) then
   begin
     aEditStyle:=etsSimple;
     aPropAttrs := TRTPropertyEditor(Node.Data).GetAttributes;
     if (rtValueList in aPropAttrs) or (rtSubProperties in aPropAttrs) then
     begin
       //if TRTPropertyEditor(Node.Data).GetPropType{$ifdef DFS_COMPILER_3_UP}^{$endif}.Kind in [tkSet,tkEnumeration] then
       if not (rtReadOnly in aPropAttrs) then
       begin
        if (TRTPropertyEditor(Node.Data) is TBooleanProperty) or
        (TRTPropertyEditor(Node.Data) is TSetElementProperty) then
        begin
         aEditStyle := etsCheckBox
        end
        else if TRTPropertyEditor(Node.Data) is TSetProperty then
         aEditStyle := etsCheckList
       else //if not (TRTPropertyEditor(Node.Data) is TMethodProperty) then
          aEditStyle := etsPickList
       end
       else if (TRTPropertyEditor(Node.Data) is TFontProperty) or
       (TRTPropertyEditor(Node.Data) is TClassProperty) then
         aEditStyle := etsEllipsis;
     end
     else if rtDialog in aPropAttrs then
     begin
        aEditStyle := etsEllipsis
     end
     else if TRTPropertyEditor(Node.Data).GetPropType{$ifdef DFS_COMPILER_3_UP}^{$endif}.Kind=tkEnumeration then
     begin
        aEditStyle := etsCheckBox;
     end;
   end
   else
      inherited;

end;

procedure TPropertiesTree.DrawDataNode(aCanvas : TCanvas; aNode : TFlyNode; aRect : TRect; aCol, aRow : Longint; AState: TExGridDrawState; Var Continue : boolean);
var
 aData : string;
 PwdChar : boolean;
 YOffset, aHeight, TextLength : integer;
 aColor{, BackColor} : TColor;
 BRect : TRect;
begin
 if TRTPropertyEditor(aNode.Data) Is TColorProperty then
 begin
    aGetEditText(aNode, aCol, aData, PwdChar, TextLength);
    with aCanvas do
      begin
         if (not (gdNoUseSelColor in AState)) and ((aNode.Selected or aNode.DropTarget  or (gdClicked in aState))  and (PointInGridRect(aCol, aRow, Selection)))
         and ((not HideSelection) or (GetFocus = Self.Handle)) then
         begin
            Font.Color := SelectedTextColor;
            Brush.Color:= SelectedBackgroundColor;
            //FillRect(aRect);
         end
         else
         begin
            Font.Color :=Self.Font.Color;
            Brush.Color:= Self.Color;
         end;
    end;
    DrawBackground(aCanvas, ARect, aCol, aRow, aState);
    inflaterect(aRect,-1, -1);
    //Draw
    aColor := StringToColor(aData);

    if (aRect.Bottom - Arect.Top) > Indent then
    begin
       aHeight := Indent;
       YOffset := ((aRect.Bottom-aRect.Top)-aHeight) Div 2;
    end
    else
    begin
       aHeight := aRect.Bottom - ARect.Top;
       YOffset := 0;
    end;
    BRect := Rect(aRect.Left, Arect.top + Yoffset, aRect.Left+aHeight, aRect.Top+Yoffset+aHeight);
    Inc(aRect.Left, aHeight+2);
    DrawText(aCanvas.Handle, PChar(aData), Length(aData),aRect,DT_END_ELLIPSIS+DT_SINGLELINE+DT_VCENTER{$ifdef DFS_COMPILER_4_UP} or DrawTextBiDiModeFlags(0){$endif});
    with Brect, aCanvas do
    begin
      if aColor =clBlack then
         Pen.Color := ClWhite
      else
         Pen.Color := ClBlack;
      Brush.Color := aColor;
      Rectangle(Left, Top, Right, Bottom);
    end;
 end
 else inherited;
end;

procedure TPropertiesTree.aGetEditText(Node: TFlyNode; index : integer; Var aData: string; var aPwdChar : boolean; Var MaxLength : integer);
begin
  if FStateLock<>0 then exit;
  if Node<>nil then
  begin
   if Index = StructureCol then
      aData :=Node.Text
   else
   begin
     try
      if Node.Data<>nil then
         aData := TRTPropertyEditor(Node.Data).GetValue;
     except
         Raise EOutlineTreeError.Create('Error on '+ Node.Text+' property');
     end;
   end;
 end;

end;

procedure TPropertiesTree.ColWidthsChanged;
begin
    inherited;
    if HandleAllocated then
      FitToWidth;
end;

procedure TPropertiesTree.Lockstate;
begin
    inc(FStateLock);
end;
procedure TPropertiesTree.Unlockstate;
begin
    Dec(FStateLock);
end;
//-----------------------------------------------------------------------------
procedure TPropertiesTree.Deselect;
var Node : TFlyNode;
begin
  //freeing Property editors
 Lockstate;
 try
  Node :=items.GetFirstNode;
  while Node<>nil do
  begin
     if Node.data <> nil then
     try
       TRTPropertyEditor(Node.data).Free;
     except
     end;
     Node :=Node.Getnext;
  end;

  Items.Clear;
  
  if Complist<>nil then    // CEDR 04.10.02 - sometimes Complist may be nil => AV!!!
    Complist.Clear;
  Editormode:=false;
 finally
    UnlockState;
 end;
end;

procedure TPropertiesTree.InitColumns;
begin
    StructureCol :=0;
    UpdateColumns;
end;

procedure TPropertiesTree.Loaded;
begin
   FitTowidth;
   inherited Loaded;
   InitColumns;//prepare columns for operation
   //ShowLines :=false;
   Items.clear;
end;


//-----------------------------------------------------------------------------
procedure TPropertiesTree.Notification(AComponent: TComponent;
Operation: TOperation);
var i : integer;
begin
   inherited Notification(AComponent, Operation);
   if (Operation = opRemove) and (CompList<>nil) then
   begin
      for i:=0 to CompList.count-1 do
      if AComponent = CompList[i] then
        Deselect;
   end;
end;

type
  TStringListProperty = class(TClassProperty)
  public
    function GetAttributes: TRTPropertyAttributes; override;
    procedure Edit; override;
  end;

function TStringListProperty.GetAttributes: TRTPropertyAttributes;
begin
  Result := inherited GetAttributes + [rtDialog] - [rtSubProperties]-[rtReadOnly];
end;

procedure TStringListProperty.Edit;
begin
  with TStringsEditDlg.Create(Application) do
  try
    Memo.Lines := TStrings(GetOrdValue);
    UpdateStatus(nil);
    FModified := False;
    ActiveControl := Memo;
      case ShowModal of
      mrOk: SetOrdValue(Longint(Memo.Lines));
    end;
  finally
    Free;
  end;
end;

type
 TRTFlyNodesEdit = class(TRTPropertyEditor)
    function  GetAttributes : TRTPropertyAttributes; override;
    function  GetValue : string; override;
    procedure Edit; override;
 end;

function  TRtFlyNodesEdit.GetAttributes : TRTPropertyAttributes;
begin
   result := [rtDialog];
end;

function  TRTFlyNodesEdit.GetValue : string;
begin
   if GetOrdValue = 0 then
      result := '(None)'
   else
      result := '(TFlyNodes)';
end;

procedure TRTFlyNodesEdit.Edit;
begin
  EditNodes(TCustomRapidTree(GetPersistent(0)).items, Modified);
end;

//--------------------------------------------------
procedure RegisterProps;
begin
  if FPropsRegistered then exit;
  RegisterRTPropertyEditor(TypeInfo(TObject), nil, '',TClassProperty);
//  RegisterRTPropertyEditor(TypeInfo(TComponent), nil, '',TComponentProperty);
  RegisterRTPropertyEditor(TypeInfo(Longint), nil, '',TIntegerProperty);
  RegisterRTPropertyEditor(TypeInfo(Char), nil, '',TCharProperty);
  RegisterRTPropertyEditor(TypeInfo(Boolean), nil, '',TBooleanProperty);
  RegisterRTPropertyEditor(TypeInfo(TColor), nil, '',TColorProperty);
  RegisterRTPropertyEditor(TypeInfo(TFont), nil, '',TFontProperty);
  RegisterRTPropertyEditor(TypeInfo(TCursor), nil, '',TCursorProperty);
  RegisterRTPropertyEditor(TypeInfo(TCaption), nil, '',TCaptionProperty);
  RegisterRTPropertyEditor(TypeInfo(TFontName), nil, '',TFontNameProperty);
  RegisterRTPropertyEditor(TypeInfo(TModalResult), nil, '',TModalResultProperty);
  RegisterRTPropertyEditor(TypeInfo(String), nil, '',TStringProperty);

  RegisterRTPropertyEditor(TypeInfo(TGraphic), nil, '',TGraphicEdit);
  RegisterRTPropertyEditor(TypeInfo(TPicture), nil, '',TPictureEdit);
  RegisterRTPropertyEditor(TypeInfo(TBitmap), nil, '',TBitmapEdit);
  RegisterRTPropertyEditor(TypeInfo(TIcon), nil, '',TIconEdit);

  {$IFDEF DFS_COMPILER_4_UP}
  RegisterRTPropertyEditor(TypeInfo(TShortCut), nil, '',TShortCutProperty);
  {$ENDIF}
//  RegisterRTPropertyEditor(TypeInfo(TComponentName), nil, '',TComponentNameProperty);
  RegisterRTPropertyEditor(TypeInfo(TTabOrder), nil, '', TTabOrderProperty);
  RegisterRTPropertyEditor(TypeInfo(TFilename), nil, '', TMPFilenameProperty);
  RegisterRTPropertyEditor(TypeInfo(TFolderName), nil, '', TMPFolderProperty);

//IFDEF DFS_DELPHI_3_UP
  RegisterRTPropertyEditor(TypeInfo(TDateTime), nil, '',TDateTimeProperty);
  RegisterRTPropertyEditor(TypeInfo(TFontCharset), nil, '',TFontCharsetProperty);
  RegisterRTPropertyEditor(TypeInfo(TImeName), nil, '',TImeNameProperty);
  RegisterRTPropertyEditor(TypeInfo(TDate), nil, '', TDateProperty);
  RegisterRTPropertyEditor(TypeInfo(TTime), nil, '', TTimeProperty);

//ENDIF
  RegisterRTPropertyEditor(TypeInfo(TStrings), nil, '',TStringListProperty);
//IFDEF DFS_DELPHI_3_UP
  RegisterRTPropertyEditor(TypeInfo(TCollection), nil, '',TCollectionProperty);
  RegisterRTPropertyEditor(TypeInfo(TFlyNodes), nil, '',TCollectionProperty);
//endif

 //special
 RegisterRTPropertyEditor(TypeInfo(TFlyNodes), TCustomFlyTreeView, '', TRTFlyNodesEdit);

 FPropsRegistered:=true;
end;

procedure TPropertiesTree.SetEXPF(const Value: String);
begin
 if FEXPF <> Value then
 begin
   FEXPF := Value;
   AssignFilter;
 end;
end;

procedure TPropertiesTree.AssignFilter;
var aStrs : TStringList;
    aNode : TFlyNode;
begin
  if FEXPF = '' then
  begin
     Items.ShowAll;
  end
  else
  begin
    aStrs := TStringList.Create;
    try
      aStrs.CommaText := UpperCase(FEXPF);
      aNode := Items.GetFirstNode;
      While aNode <> nil do
      begin
         if aStrs.IndexOf(UpperCase(aNode.Caption)) = -1 then
         //show
          aNode.Show(True)
         else
          //hide
          aNode.Hide;

         aNode := aNode.GetNext;
      end;

    finally
     aStrs.Free;
    end;
  end;
end;

initialization
   {$IFNDEF Registered}
   {$IFNDEF ACTIVEX}
     ShowURToolTip('',false);
   {$ENDIF}
   {$ENDIF}
   FPropsRegistered:=false;
finalization

end.






