unit Treecoll;
{/*****************************************************************/
 /  Copyright (c) 1999 IMCA SYSTEMS
 /  All Rights Reserved.
 /  Creation Date: 06.01.99
 /  Last Modifications Date : 01.01.1999
 /  Version 1.52
 /  MODULE:
 /   Next generation of TTreeCollection - Polymorphic TTreeCollection with Root
 /   Polymorphic Collection with hierarchical structure
 /   TCollection and TCollectionItem united in one Class

}
interface

{$I DFS.inc}
{$ifdef DFS_COMPILER_3_UP}
{$R *.dcr}
{$endif}
 uses Windows, Messages, Classes, {$ifdef DFS_COMPILER_3_UP}IsLists,{$endif} SysUtils,TypInfo;

Const
  EStreamNA = 'Stream not assigned';
  EUnKnownProp='Unknown property type - %S';


type
  {$ifndef DFS_COMPILER_3_UP}
    TIsList = TList;
  {$endif}
  TTreeCollection = class;
  TTreeCollectionClass = class of TTreeCollection;
  TTreeCollectionSortCompare = function (Item1, Item2: TTreeCollection): Integer; stdcall;

  TTreeCollectionOnChange = procedure(item : TTreeCollection) of object;
  ETreeCollectionError = class(Exception);

  TSetObjectProc = procedure(Index: integer; aObject : TTreeCollectionClass);

  TMetaNode = class
  public
     Parent : TTreeCollectionClass;
     Index  : Integer;
     constructor Create(aParent : TTreeCollectionClass; aIndex : Integer);
  end;

  TTreeCollection = class(TPersistent)
  private
    //from item
    {$ifdef DFS_COMPILER_3_UP}
    FPropName : string;
    {$ENDIF}
    FCollection: TTreeCollection;
    Fowner : TPersistent;
    FID: Integer;
    //FType : Integer;
    FCaption : String;
    FOnChange : TTreeCollectionOnChange;
    FItems: TISList;
    FUpdateCount: Integer;
    FNextID: Integer;
    FData : Pointer;
    FClearing : Boolean;
    {$ifdef DFS_COMPILER_3_UP}
    function GetPropName: string;
    {$endif}
    function GetCount: Integer;
    procedure InsertItem(Item: TTreeCollection; AtIndex : integer);
    procedure RemoveItem(Item: TTreeCollection);
    //from item
    function GetIndex: Integer;
    procedure SetTreeCollection(Value: TTreeCollection);
    function GetCollection : TTreeCollection;
    procedure CheckListOnExist;
    //new
    function GetLevel: Integer;
    procedure SetonChange(aValue : TTreeCollectionOnChange);

  protected
    //from item
    procedure Changed(AllItems: Boolean); virtual;
    //procedure Exchanged(Index1, Index2 : integer);virtual;
    {$ifdef DFS_COMPILER_3_UP}
    function GetOwner: TPersistent; override;
    {$endif}
    function GetDisplayName: string; virtual;
    procedure SetIndex(Value: Integer); virtual;
    function GetItem(Index: Integer): TTreeCollection;
    procedure SetItem(Index: Integer; Value: TTreeCollection);
    procedure Update(Item: TTreeCollection); virtual;
    procedure SetCaption(value : string); virtual;

    function InternalGetClass(aClassname : String) : TTreeCollectionClass; virtual;
    function _Add(AtIndex : Integer): TTreeCollectionClass; virtual;//Inserts a new item to the end of a Collection.
    function ISMeta(Index : Integer) : Boolean;
    function GetPointer(Index : integer) : Pointer;
  public

    function IsList : Boolean; Virtual;//for Property editors
    procedure SetCount(Value : Integer; SetObjProc : TSetObjectProc);//callback proc for virtual add
    {$ifdef DFS_COMPILER_3_UP}
     property PropName: string read GetPropName write FPropName;// Getting property name if Coolection cantained in another object
    {$endif}

    constructor Create(aOwner: TTreeCollection); virtual;//Creates and initializes a TTreeCollection instance.
    destructor Destroy; override;//Destroys an instance of TTreeCollection .

    procedure SetOwner(aOwner: TPersistent);//Indicates which component owns the collection.

    function GetPath(Separator : String) : String;
    function GetIndexPath(Separator : String) : String;
    function GetIDPath(Separator : String) : String;

    function GetItemOnPath(aPath : string; Separator : String) : TTreeCollection;
    function GetItemOnIndexPath(aPath : string; Separator : String) : TTreeCollection;
    function GetFirstChild: TTreeCollection;//Returns the first child collection of a tree collection.
    function GetLastChild: TTreeCollection;//Returns the last child collection of a tree collection.
    function GetNext: TTreeCollection;    //Returns the next Collection after the calling Collection .
    function GetNextSibling: TTreeCollection;    //Returns the next Collection at the same level as the calling Collection.
    function GetPrev: TTreeCollection;    //Returns the previous Collection before the calling Collection.
    function getPrevSibling: TTreeCollection;    //Returns the previous Collection at the same level as the calling Collection.
    procedure Sort(Compare: TTreeCollectionSortCompare; Recursive : boolean); dynamic;    //Sorts the collections in the Collection by Compare procedure.
    procedure CustomSort(SortProc: TTreeCollectionSortCompare; Recursive : boolean);    //Sorts the collections in the Collection by SortProc procedure.
    procedure AlphaSort(Recursive : boolean);    //Sorts the collections in the Collection in alphabetical order (based on Caption) in ascending order.
    procedure Exchange(Index1, index2 : integer);    //Swaps the position of two items in the Items array.
    procedure Move(CurIndex, NewIndex: Integer);//Changes the position of an item in the Items array.

    function Add(aCaption,AClassName : string): TTreeCollectionClass; //Inserts a new item to the end of a Collection.

    procedure DeleteItem(Item : TTreeCollection);    //Deletes the first reference to the Item parameter from the Items array.
    procedure Assign(Source: TPersistent); override;    //Discards any current property information and replaces it with the information from the Source.
    procedure BeginUpdate; virtual;    //Prevents the updating of the parent Collection until the EndUpdate method is called.
    procedure BeginUpdateMinCapacity(MinCapacity : integer); virtual;    //
    procedure Clear;    //Deletes all child items.
    procedure EndUpdate; virtual;    //Re-enables Collection reindexing that was turned off with the BeginUpdate method.
    function FindItemID(ID: Integer): TTreeCollection;    //Returns the item with the specified ID.
    function Insert(Index: Integer; AclassName : String): TTreeCollection;    //Inserts a new item into the Collection before the item specified by the index parameter.
    property Clearing : Boolean read FClearing;
    property UpdateCount: Integer read FUpdateCount;
    Property OnChange : TTreeCollectiononChange read FOnChange write SetonChange;    //Occurs when the Collection changes.
    property Data : Pointer read Fdata write Fdata;    //Points to application-defined data associated with the collection.
    property TreeCollection: TTreeCollection read GetCollection write SetTreeCollection;    //Returns parent Collection
    property ID: Integer read FID;    //A unique, permanent index for the item in the Collection.
    property Index: Integer read GetIndex write SetIndex;    //Specifies the position of the item in the list of child items maintained by its parent Collection
    property Count: Integer read GetCount;    //Indicates the number of direct descendants of a item.
    property Level : integer read GetLevel;    //Indicates the level of indentation of a item within the Collection object.
    property Caption : string read Fcaption write SetCaption;    //Specifies the text that labels the item in the list.
    //replaced on direct procedures
    //property Path : string read GetPath; //determines full path to node from root collection based on captions of it's and parents
    //new 31.01.2000
    //property IndexPath : string read GetIndexPath; //determines full path to node from root collection based on indexes of it's and parents
    //new 31.01.2000
    //property IDPath : string read GetIdPath; //determines full path to node from root collection based on addresses of it's and parents
    property Items[Index: Integer]: TTreeCollection read GetItem write SetItem; default;    //Provides access to a child item by its position in the list of child items.
  end;

  //additional component for future purposes
  TRootCollection = class(TComponent)
  private
     FCollection : TTreeCollection;
     procedure WriteCollection(Stream : TStream);
     procedure ReadCollection(Stream : TStream);
  protected
     procedure DefineProperties(Filer : TFiler); override;
  public
     constructor Create(aOwner : TComponent); override;     //Creates and initializes a TRootCollection instance.
     destructor  Destroy; override;     //Destroys an instance of TRootCollection .
     Procedure SaveToStream(S : TStream);     //Writes the data in the Collection to the stream passed as the Stream parameter.
     Procedure LoadFromStream(S : TStream);     //Reads Collection data from a stream and stores the contents.
     procedure SaveToFile(const FileName: string);     //Saves a the Collection to the file specified in FileName.
     procedure LoadFromFile(const FileName: string);     //Reads the file specified in FileName and loads the data into the Collection.
   published
     property Collection : TTreeCollection read FCollection;     //Provides access to internal Collection object
  end;

  TLevels = class(TPersistent)
  private
     FoldChange : TTreeCollectionOnChange;
     FTreeCollection : TTreeCollection;
     FEndItems : TTreeCollection;
     FLevels : TTreeCollection;
     function GetDepth : integer;
     function GetLevel(index: integer) : TTreeCollection;
     procedure SetTreeCollection(ATreeCollection : TTreeCollection);
     procedure Changed(item : TTreeCollection); dynamic;
  protected
  public
      constructor Create(aOwner: TTreeCollection);     //Creates and initializes a TLevels instance.
      destructor Destroy; override;     //Destroys an instance of TLevels.
      property Collection : TTreeCollection read FTreeCollection write SetTreeCollection;     //Specifies TreeCollection instance managed by Tlevels object
      property Depth : integer read GetDepth;     //Returns depth (maximal level) of Collection
      property Levels[Index: Integer]: TTreeCollection read GetLevel; default;     //Provides access to a Collection items by its level in the Collection.
      property EndItems : TTreeCollection read FEnditems;     //Provides access to a Collection items by without childs.
  end;

  type
    TRWCallBack=procedure( Instance : TPersistent; PropInfo : PPropInfo; var OldMethod : Boolean) of object ;
    TPCallback=function( Instance : TPersistent; PropInfo : PPropInfo):string of object ;
    TSPCallBack= procedure of object;

  // filer for TreeCollection
  TTreeCollFiler=class
  private
    FStream : TStream;
    procedure SetStream(V : TStream);
  protected
    procedure CheckStream;
    procedure WriteListBegin;
    procedure WriteListEnd;
  public
     Function Seek(Offset : Longint; Origin : Word) : Longint;    //Moves the current position of the stream by Offset bytes, relative to the origin specified by Origin.
     Function  Eof : Boolean;    //Eof tests whether the file position is at the end of a file.
     Function  Bof : Boolean;    //Eof tests whether the file position is at the begin of a file.
     procedure SkipProperty;
     procedure WriteProperty(Instance : TPersistent; PropInfo : PPropInfo; var OldMethod: boolean);    //Write specified property
     procedure ReadProperty(Instance : TPersistent; PropInfo : PPropInfo; var OldMethod: boolean);    //Read specified property
     function ReadPropertyName(Instance : TPersistent; PropInfo : PPropInfo) : string;    //Read next property name
     function WritePropertyName(Instance : TPersistent; PropInfo : PPropInfo): string;    //Read next property name
     procedure WriteCollection(Value : TCollection);    //Write specified TCollection Object
     procedure ReadCollection(Value : TCollection);    //Read specified TCollection Object
     procedure ReadTreeCollection(Value : TTreeCollection);    //Write specified TTreeCollection Object
     procedure WriteTreeCollection(Value : TTreeCollection);    //Write specified TTreeCollection Object
     procedure SkipClass;    // Skips over current class
     function  NextValue : TValueType;    //Determine next value for reading
     Function  ReadChar : Char;    //Reads a character from the reader object's stream and returns that character value.
     procedure WriteChar(Value : Char);    //Writes the character passed in Value to the writer object's stream.
     Function  ReadSmallInt : SmallInt;    //Reads an Smallint-type number from the reader object's stream and returns its value.
     procedure WriteSmallInt(Value : SmallInt);    //Writes the Smallint value passed in Value to the writer object's stream.
     Function  ReadShortInt : ShortInt;    //Reads an ShortInt-type number from the reader object's stream and returns its value.
     procedure WriteShortInt(Value : ShortInt);    //Writes the Shortint value passed in Value to the writer object's stream.
     Function  ReadLongInt : LongInt;    //Reads an Longint-type number from the reader object's stream and returns its value.
     procedure WriteLongInt(Value : LongInt);    //Writes the Longint value passed in Value to the writer object's stream.
     procedure Write(const Buf; Count : Integer);    //Writes Count bytes from Buf to the writer object's associated stream.
     procedure Read(Var Buf; Count : Integer);    //Reads data from the associated buffer.
     procedure WriteString(const V : String);    //Writes tagged data to the writer object's stream.
     function  ReadString : string;    //Reads a tagged string value written by WriteString from the reader object's stream and returns its contents.
     procedure WriteStrings(Value : TStrings);    //Writes TStrings object to the writer object's stream.
     procedure ReadStrings(Value : TStrings);    //Reads a TStrings Object written by WriteStrings from the reader object's stream and returns its contents.
     Function  ReadFloat : Extended;    //Reads a floating-point number from the reader object's stream and returns its value.
     procedure WriteFloat(V : Extended);    //Writes the floating-point value passed in Value to the writer object's stream.
     procedure WriteValue(Value : TValueType);    //Writes the type of the next item on the reader object's stream.
     function  ReadValue : TValueType;    //Reads the type of the next item on the reader object's stream and returns with the stream positioned after the value-type indicator.
     procedure WriteObject(Instance : TPersistent);    //Writes the TPersistent Object passed in Instant to the writer object's stream.
     procedure ReadObject(Instance : TPersistent);    //Reads a TPersistent Object written by WriteObject from the reader object's stream and returns its contents.
     property Stream : TStream Read FStream Write SetStream;    //Specifies associated stream object
     constructor Create(AStream : TStream);    //Creates and initializes a TTreeColFiler instance.
     destructor Destroy; override;    //Destroys an instance of TTreeColFiler.
  end;



 {TTreeListPropEditor = class(TPropertyEditor)
  function  GetAttributes : TPropertyAttributes; override;
  function  GetValue : string; override;
  procedure Edit; override;
 end;}


   procedure RegisterTreeCollectionClass(AClass: TTreeCollectionClass);  //Registers a class of TTreeCollection object so that its class type can be retrieved in TTreeCollectionPropEdit.
   procedure RegisterTreeCollectionClasses(AClasses: array of TTreeCollectionClass);  //Registers a array of classes of TTreeCollection object so that its class type can be retrieved in TTreeCollectionPropEdit.
   procedure UnRegisterTreeCollectionClass(AClass: TTreeCollectionClass);  //Unregisters an object class for TTreeCollectionPropEdit.
   procedure UnRegisterTreeCollectionClasses(AClasses: array of TTreeCollectionClass);  //Unregisters an array of object classes for TTreeCollectionPropEdit.
   procedure WriteCollectionToStream(Collection : TTreeCollection; Stream : TStream);  //Writes TTreeCollection to specified stream
   procedure ReadCollectionFromStream(Collection : TTreeCollection; Stream : TStream);  //Reads TTreeCollection from specified stream
   function SavePersistentToText(Obj : TPersistent; SplitToStrings: Boolean) : string;
   procedure LoadPersistentFromText(Obj : TPersistent; Text : string);
   procedure ReadPersistentFromStream(Obj : TPersistent; Stream : TStream);
   procedure WritePersistentToStream(Obj : TPersistent; Stream : TStream);


   // new 30.01.2000
   function SaveTreeCollectionToText(Collection : TTreeCollection; SplitToStrings: Boolean) : string; //Save TTreeCollection to Text format
   // new 30.01.2000
   procedure LoadTreeCollectionFromText(Collection : TTreeCollection; Text : string); //Load TTreeCollection from Text
   // new 30.01.2000
   procedure LoadTreeCollectionFromStream(Collection : TTreeCollection; Stream : TStream);
   // new 30.01.2000
   procedure SaveTreeCollectionToStream(Collection : TTreeCollection; Stream : TStream);
   // new 30.01.2000
   procedure SaveTreeCollectionToFile(Collection : TTreeCollection; const Filename : string);
   // new 30.01.2000
   procedure LoadTreeCollectionFromFile(Collection : TTreeCollection; const Filename : string);
   // new 24.02.2000 - for reset Default properties
   procedure AssignProperties(Instance1, Instance2 : TPersistent; RecurseClasses : boolean; ExcludeFilter : string);


  function SaveComponentToText(Control : TComponent) : string;  //Converts an Component properties to String
  procedure LoadComponentFromText(Resource : String; Control : TComponent);  //Converts String for load Component's properties
  

implementation

uses IsConst, Consts,{$ifdef DFS_COMPILER_3_UP}ComStrs,{$endif} Forms, Dialogs
{$IFNDEF Registered},FTAbout {$ENDIF};

 var
  TreeCollectionClassList: {$ifdef DFS_COMPILER_3_UP}TThreadList {$else} TISList{$endif};

type
  TIntegerSet = set of 0..SizeOf(Integer) * 8 - 1;


{$ifdef DFS_COMPILER_3_UP}

  procedure TreeCollectionError(Msg: String);
  begin
    raise ETreeCollectionError.Create(Msg);
  end;

{$else}

  procedure TreeCollectionError(MsgID: Integer);

  begin
    raise ETreeCollectionError.CreateRes(MsgID);
  end;

{$endif}

procedure ReadComponent(Astream : TMemoryStream; AComponent : TComponent);
var
    Reader: TReader;
begin
    Reader := TReader.Create(Astream, 4096);
    Reader.Parent := AComponent.GetParentComponent;
    Astream.ReadResHeader;
    try
      Reader.BeginReferences;
      Reader.ReadSignature;
      Reader.ReadComponent(AComponent);
      Reader.FixupReferences;
    finally
      Reader.EndReferences;
      Reader.Free;
    end;
end;

procedure LoadComponentFromText(Resource : String; Control : TComponent);
Var
  Input, Output : TMemoryStream;
  Sts : TStrings;
begin
  Input := TMemoryStream.Create;
  Output := TMemoryStream.Create;
  sts := TStringlist.Create;
  Input.Clear;
  Output.Clear;
  try
    Sts.Text:= Resource;
    Sts.SaveToStream(Output);
    Output.Position := 0;
    ObjectTextToResource(Output, Input);
    Input.Position := 0;
    ReadComponent(Input, Control);
  finally
    Input.Free;
    Output.Free;
    Sts.Free;
  end;
end;

function SaveComponentToText(Control : TComponent) : string;
Var
  Sts : TStringList;
  Input, OutPut : TMemoryStream;
begin
  Input := TMemoryStream.Create;
  OutPut := TMemoryStream.Create;
  Sts := TStringList.Create;
  try
  Input.WriteComponentRes(Control.ClassName, Control);
  Input.Position := 0;
  ObjectResourceToText(Input, OutPut);
  OutPut.Position := 0;
  Sts.LoadFromStream(OutPut);
  Result:=Sts.Text;
  finally
    Input.Free;
    OutPut.Free;
    Sts.Free;
  end;

end;

procedure QuickSort(aCollection : TTreeCollection; SortList: TISList; L, R: Integer; SCompare: TTreeCollectionSortCompare);
var I, J: Integer;
    P, T: Pointer;
    //index1, index2 : integer;
begin

  repeat
    I:=L;
    J:=R;
    P:=SortList[(L+R) shr 1];
    repeat
      while (SCompare(TTreeCollection(SortList[I]),TTreeCollection(P)) < 0) do
       if (i < R) then Inc(I) else break;
      while (SCompare(TTreeCollection(SortList[J]),TTreeCollection(P)) > 0) do
       if  (j > L) then Dec(J) else break;
      if I <= J then
      begin

        T:=SortList[I];
        SortList[I]:=SortList[J];
        SortList[J]:=T;

        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(aCollection, SortList,L,J,SCompare);
    L := I;
  until I >= R;
end;



function DefaultSort(Item1, Item2: TTreeCollection): Integer; stdcall;
begin
   Result:=CompareStr(item1.FCaption, Item2.Caption);
end;

constructor TMetaNode.Create(aParent : TTreeCollectionClass; aIndex : Integer);
begin
    inherited Create;
    Parent := aParent;
    Index := aIndex;
end;

constructor TTreeCollection.Create(aOwner: TTreeCollection);
begin
  inherited Create;
  FCollection:=nil;
  FClearing :=false;
  if aOwner<> nil then
  try
    aOwner.InsertItem(Self, -1);
  except
   raise ETreeCollectionError.create('Unable Set Collection');
  end;
end;

destructor TTreeCollection.Destroy;
begin
  FUpdateCount := 1;
  if FItems <> nil then
  begin
   Clear;
   FItems.Free;
  end; 
  Fdata := nil;
  SetTreeCollection(nil);

  inherited Destroy;
end;

//delegating function
function TTreeCollection.InternalGetClass(aClassname : String) : TTreeCollectionClass;
begin
   Result :=TTreeCollection;
end;

function TTreeCollection.ISMeta(Index : Integer) : Boolean;
begin
   Result := TObject(FItems[Index]) is TMetaNode;
end;

function TTreeCollection.GetPointer(Index : integer) : Pointer;
begin
   Result := Fitems[Index];
end;
{
function TTreeCollection.Unpack(AtIndex : Integer): TTreeCollectionClass;
begin
  Result := TTreeCollectionClass(InternalGetClass('').Create(nil));
  TTreeCollection(Result).FCollection := Self;
  TTreeCollection(Result).FID := FNextID;
  TTreeCollection(Result).Fowner := Self;
  Inc(FNextID);
  if Fitems <> nil then
  begin
   if Fitems[AtIndex]<> nil then
    if TObject(Fitems[AtIndex]) is TMetaNode then
     TMetaNode(FItems[AtIndex]).Free;
     FItems[AtIndex] := Result;
  end;
end;
}
function TTreeCollection._Add(AtIndex : integer): TTreeCollectionClass;
begin
  Result := TTreeCollectionClass(InternalGetClass('').Create(nil));
  Inc(FUpdateCount);
  InsertItem(TTreeCollection(Result), AtIndex);
  Dec(FUpdateCount);
end;

function TTreeCollection.Add(aCaption,AClassName : string): TTreeCollectionClass;
var Aclass : TTreeCollectionClass;
begin
  //try
  try
  Aclass := InternalGetClass(aClassname);//TTreeCollectionClass(FindClass(AClassName));
  except
    {$IFDEF DFS_COMPILER_3_UP}
    raise ETreeCollectionError.CreateFmt(SClassNotFound,[aClassname]);
    {$ELSE}
    {$IFDEF DFS_CPPB_3_UP}
      raise ETreeCollectionError.CreateFmt(SClassNotFound,[aClassname]);
    {$ELSE}
      raise ETreeCollectionError.Create(FmtLoadStr(SClassNotFound, [aClassName]));
    {$ENDIF}
    {$ENDIF}
  end;

  Result := TTreeCollectionClass(Aclass.Create(Self));

  TTreeCollection(Result).Caption:=aCaption;
   if assigned(FOnchange) then TTreeCollection(Result).Onchange:=Fonchange;

end;


//new - 31.01.2000
function TTreeCollection.GetItemOnIndexPath(aPath : string; Separator : String): TTreeCollection;
var aCollection : TTreeCollection;
begin
   Result:=nil;
   aCollection := GetFirstChild;
   while aCollection<>nil do
   begin
      if ACollection.GetIndexPath(Separator) = aPath then
      begin
         Result:=aCollection;
         break;
      end;
      aCollection :=aCollection.GetNext;
   end;

end;

function TTreeCollection.GetItemOnPath(aPath : string; Separator : String) : TTreeCollection;
var aCollection : TTreeCollection;
begin
   Result:=nil;
   aCollection := GetFirstChild;
   while aCollection<>nil do
   begin
      if ACollection.GetPath(Separator) = aPath then
      begin
         Result:=aCollection;
         break;
      end;
      aCollection :=aCollection.GetNext;
   end;
end;
//new 31.01.2000
function TTreeCollection.GetIndexPath(Separator : String) : String;
var aParent : TTreeCollection;
begin
   if index<>-1 then
      Result:=inttostr(index);
   aParent:= TreeCollection;
   while (aParent<>nil) and (aParent.Index <> -1) do
   begin
      //if aParent.Index<>-1 then
      Result:=InttoStr(aParent.index) + Separator + Result;
      aParent:=aParent.TreeCollection;
   end;

end;
//new 31.01.2000
function TTreeCollection.GetIDPath(Separator : String) : String;
var aParent : TTreeCollection;
begin
   if index<>-1 then
      Result:=inttostr(LongInt(Self));
   aParent:= TreeCollection;
   while (aParent<>nil) and (aParent.Index <> -1) do
   begin
       Result:=InttoStr(LongInt(aParent))+ Separator + Result;
       aParent:=aParent.TreeCollection;
   end;

end;

function TTreeCollection.GetPath(Separator : String) : String;
var aParent : TTreeCollection;
begin
   Result:=Caption;
   aParent:= TreeCollection;
   while (aParent<>nil) and (aParent.Index <> -1) do
   begin
       Result:=aParent.Caption + Separator + Result;
       aParent:=aParent.TreeCollection;
   end;
end;

function TTreeCollection.GetDisplayName: string;
begin
  Result := ClassName;
end;

procedure TTreeCollection.DeleteItem(Item : TTreeCollection);
begin
   BeginUpdate;
   RemoveItem(Item);
   Item.Clear;
   Item.Free;
   EndUpdate;
end;

procedure TTreeCollection.Assign(Source: TPersistent);
var
  TreeColl: TTreeCollection;
  MemStream: TMemoryStream;
begin
  if Source is TTreeCollection then
  begin
    BeginUpdate;
    try
      TreeColl := TTreeCollection(Source);
      Clear;
      MemStream := TMemoryStream.Create;
     try
        WriteCollectionToStream(TreeColl, MemStream);
        MemStream.Position := 0;
        ReadCollectionFromStream(Self, MemStream);
    finally
      MemStream.Free;
    end;

    finally
      EndUpdate;
    end;
    Exit;
  end
  else
   inherited Assign(Source);
end;

procedure TTreeCollection.CheckListOnExist;
begin
   if Fitems = nil then
      Fitems := TISList.Create;
end;

procedure TTreeCollection.BeginUpdateMinCapacity(MinCapacity : integer);
begin
  CheckListOnExist;
  if MinCapacity > FItems.Capacity then
     FItems.Capacity := MinCapacity;
    BeginUpdate;
end;

procedure TTreeCollection.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

{procedure TTreeCollection.Exchanged(Index1, Index2 : integer);
begin
end;
}
procedure TTreeCollection.Changed(AllItems: Boolean);
var
  Item: TTreeCollection;
begin
  //update owner collection
  if (FCollection <> nil) and (FCollection.FUpdateCount = 0) then
  begin
    if AllItems then Item := nil else Item := Self;
    FCollection.Update(Item);
  end
  else if FUpdateCount = 0 then
   Update(Self);

end;

procedure TTreeCollection.Clear;
var i : integer;
begin

 if Fitems <> nil then
  if FItems.Count > 0 then
  begin
    BeginUpdate;
    try
      FClearing := true;
      for i:=0 to Fitems.Count-1 do
      begin
        TTreeCollection(FItems[i]).Free;
      end;
      FItems.Clear;
      FClearing := False;
    finally
      EndUpdate;
    end;
  end;

end;

procedure TTreeCollection.EndUpdate;
begin
 if FUpdateCount > 0 then
  Dec(FUpdateCount);
  Changed(false);
end;

function TTreeCollection.FindItemID(ID: Integer): TTreeCollection;
var
  I: Integer;
begin
  CheckListOnExist;
  for I := 0 to FItems.Count-1 do
  begin
    Result := TTreeCollection(FItems[I]);
    if Result.ID = ID then Exit;
  end;
  Result := nil;
end;

procedure DefaultSetObject(index : integer; aObject : TTreeCollectionClass);
begin
   if aObject<>nil then
      TTreeCollection(aObject).Caption := 'Item#'+inttostr(Index);
end;

function TTreeCollection.IsList : Boolean;
begin
   Result := False;
end;

procedure TTreeCollection.SetCount(Value : Integer; SetObjProc : TSetObjectProc);
var i : integer;
begin
   CheckListOnExist;
   BeginUpdate;
   Try
     Clear;
     for i:=0 to Value-1 do
     begin
        FItems.Add(TMetaNode.Create(nil, i));
     end;
{
     if Value > Fitems.Count then
     begin

        Fitems.Capacity := Value;
        if not Assigned(@SetObjProc) then
           SetObjProc := DefaultSetObject;
        for i:=Fitems.count to Value do
          SetObjProc(i, _Add(i));

     end
     else if Value > 0 then
     begin
        for i:=0 downto (FItems.Count - Value-1) do
           RemoveItem(TTreeCollection(FItems[0]));
     end;
}
   finally
     EndUpdate;
   end;
end;

function TTreeCollection.GetCount: Integer;
begin
 if Fitems <> nil then
  Result := FItems.Count
 else
  Result := 0;
end;

function TTreeCollection.GetItem(Index: Integer): TTreeCollection;
begin
 Result:=nil;
 if Fitems <> nil then
  if (index >= 0) and (index < FItems.Count) then
    if TObject(Fitems[Index]) is TTreeCollectionClass then
      Result := FItems[Index]
    else
      Result := TTreeCollection(_Add(Index));//AOD
end;



function TTreeCollection.Insert(Index: Integer; AclassName : String): TTreeCollection;
begin
  Result := TTreeCollection(Add('',AclassName));
  Result.Index := Index;
end;


procedure TTreeCollection.InsertItem(Item: TTreeCollection; AtIndex : integer);
begin
  CheckListOnExist;
  if AtIndex =-1 then
    FItems.Add(Item)
  else
  begin//aod
     While AtIndex > FItems.Count do
     begin
       FItems.Add(nil);
     end;
     if TObject(FItems[AtIndex]) is TMetaNode then
       TMetaNode(FItems[AtIndex]).Free;
     FItems[AtIndex] := Item;
  end;
  Item.FCollection := Self;
  Item.FID := FNextID;
  Item.Fowner := Self;
  Inc(FNextID);
  Changed(false);

end;

procedure TTreeCollection.RemoveItem(Item: TTreeCollection);
begin
  FItems.Remove(Item);
  if Item <> nil then
  begin
    Item.Changed(true);
    Item.FCollection := nil;
  end;  
end;

function TTreeCollection.GetIndex: Integer;
begin
  if FCollection <> nil then
    Result := FCollection.FItems.IndexOf(Self)
  else
    Result := -1;
end;

procedure TTreeCollection.SetItem(Index: Integer; Value: TTreeCollection);
begin
 if FItems <> nil then
  TTreeCollection(FItems[Index]).Assign(Value);
end;


procedure TTreeCollection.Update(Item: TTreeCollection);
begin
  if Assigned(FonChange) then
  begin
    if (Item <> nil)  then FOnChange(Item)
    else
    if (FCollection<>nil)  then
       FonChange(FCollection);
  end;
end;

{$ifdef DFS_COMPILER_3_UP}

function TTreeCollection.GetPropName: string;
var
  I: Integer;
  Props: PPropList;
  TypeData: PTypeData;
  aOwner: TPersistent;
begin

  Result := FPropName;
  aOwner := GetOwner;
  if (Result <> '') or (aOwner = nil) or (aOwner.ClassInfo = nil) then Exit;
  TypeData := GetTypeData(aOwner.ClassInfo);
  if (TypeData = nil) or (TypeData^.PropCount = 0) then Exit;
  GetMem(Props, TypeData^.PropCount * sizeof(Pointer));
  try
    GetPropInfos(aOwner.ClassInfo, Props);
    for I := 0 to TypeData^.PropCount-1 do
    begin
      with Props^[I]^ do
        if (PropType^^.Kind = tkClass) and
          (GetOrdProp(aOwner, Props^[I]) = Integer(Self)) then
          FPropName := Name;
    end;
  finally
    Freemem(Props);
  end;
  Result := FPropName;
end;


function TTreeCollection.GetOwner: TPersistent;
begin
   if FCollection<>nil then
     Result := FCollection
   else
   begin
     Result := FOwner;
   end;
end;

{$endif}

procedure TTreeCollection.SetOwner(aOwner: TPersistent);
begin
 if FCollection = nil then
   FOwner := AOwner
 else Fowner := FCollection;
end;

function TTreeCollection.GetCollection : TTreeCollection;
begin
   Result := FCollection;
end;

procedure TTreeCollection.SetTreeCollection(Value: TTreeCollection);
begin
  if FCollection <> Value then
  begin
    if FCollection <> nil then
     if not FCollection.FClearing then
        FCollection.RemoveItem(Self);
    if Value <> nil then Value.InsertItem(Self, -1);
  end;
end;

procedure TTreeCollection.Move(CurIndex, NewIndex: Integer);
begin
   CheckListOnexist;
   FItems.Move(CurIndex, NewIndex);
   TTreeCollection(Fitems[NewIndex]).Changed(True);
end;

procedure TTreeCollection.Exchange(Index1, index2 : integer);
begin
   if (Index1>=0) and (index2>=0) and (index1 < Count) and (index2 < Count) then
   begin
       FItems.Exchange(Index1, Index2);
       TTreeCollection(Fitems[index1]).Changed(True);
   end;
end;

procedure TTreeCollection.SetIndex(Value: Integer);
var
  CurIndex: Integer;
begin
  CurIndex := GetIndex;
  if (CurIndex >= 0) and (CurIndex <> Value) then
  begin
    FCollection.FItems.Move(CurIndex, Value);
    Changed(True);
  end;
end;

function TTreeCollection.GetLevel: Integer;
var
  Collection: TTreeCollection;
begin
  Result := 0;
  Collection := FCollection;
  while Collection <> nil do
  begin
    Inc(Result);
    Collection := Collection.FCollection;
  end;

end;

function TTreeCollection.GetFirstChild: TTreeCollection;
begin
 Result:=nil;
 if Fitems<>nil then
  if Fitems.Count>0 then
    Result := items[0];
end;

function TTreeCollection.GetLastChild: TTreeCollection;
begin
  Result:=nil;
  if Fitems <> nil then
   if Fitems.Count >0 then
     Result := items[Fitems.Count-1];
end;


function TTreeCollection.GetNext: TTreeCollection;
var
   ParentCollection, Collection : TTreeCollection;
begin
   Collection:=GetFirstChild;
   if Collection = nil then
      Collection:=GetNextSibling;
   ParentCollection:=Self;//FCollection; //parent
   while (Collection=nil) and (ParentCollection<>nil) do
   begin
       ParentCollection:=ParentCollection.Fcollection;
       if ParentCollection<> nil then
          Collection:=ParentCollection.GetNextSibling
       else Collection:=nil;
   end;
   Result:=Collection;
end;

function TTreeCollection.GetPrev: TTreeCollection;
var Collection : TTreeCollection;
begin
   Result:=GetPrevSibling;
   if Result <> nil then
   begin
    Collection := Result;
    repeat
      Result := Collection;
      Collection := Result.GetLastChild;
    until Collection = nil;
  end
  else
    Result := FCollection;

end;

function TTreeCollection.getNextSibling: TTreeCollection;
begin
  Result :=nil;
  if FCollection <> nil then
  begin
    if GetIndex < (FCollection.Count-1) then
    begin
       Result:=FCollection.items[index+1];
    end;
  end;
end;

function TTreeCollection.getPrevSibling: TTreeCollection;
var aindex : integer;
begin
   aindex:=GetIndex;
   if aindex <> 0 then
      Result:=FCollection.items[aindex-1]
   else
      Result:=nil;
end;

procedure TTreeCollection.SetCaption(value : string);
begin
   FCaption:=Value;
//   if FCollection <> nil then
//     FCollection.Update(self);
end;



procedure TTreeCollection.Sort(Compare: TTreeCollectionSortCompare; Recursive : boolean);
var i : integer;
begin
  if Fitems <> nil then
  begin
    if (FItems.Count > 0) then
      QuickSort(Self, FItems,0,Count-1,Compare);
    if Recursive then
      for i:=0 to FItems.Count-1 do
      if items[i].Count > 0 then
        items[i].Sort(Compare, Recursive);
  end;
end;

procedure TTreeCollection.CustomSort(SortProc: TTreeCollectionSortCompare; Recursive : boolean);
begin
  if not assigned(SortProc) then
      Sort(@DefaultSort,  Recursive)
  else
      Sort(@SortProc, Recursive);
end;

procedure TTreeCollection.AlphaSort(Recursive : boolean);
begin
   CustomSort(nil,Recursive);
end;

procedure TTreeCollection.SetonChange(aValue : TTreeCollectionOnChange);
var i:integer;
begin
   FonChange:=aValue;
   //recursive Change
   for i:=0 to count-1 do Items[i].onchange:=aValue;
end;

//Levels Object ----------------------------------------------------------------
constructor TLevels.Create(aOwner: TTreeCollection);
begin
   Flevels := TTreeCollection.Create(nil);
   FEnditems :=TTreeCollection.Create(nil);
   if aOwner <> nil then
   begin
        SetTreeCollection(aOwner);
        if assigned(FtreeCollection.OnChange) then
        begin//hook on Onchange
          FoldChange:=FtreeCollection.OnChange;
        end;
        FtreeCollection.OnChange:=Changed;
   end;

end;

destructor TLevels.Destroy;
begin

   if FtreeCollection <> nil then
   begin
      if Assigned(FoldChange) then
         FtreeCollection.Onchange:=FoldChange
      else FtreeCollection.Onchange:=nil;
   end;
   FLevels.Free;
   FEnditems.Free;
   SetTreeCollection(nil);
   inherited;
end;

function  TLevels.GetLevel(index: integer) : TTreeCollection;
begin
  if index > Flevels.Count-1  then
    Result:=nil
  else
    Result:=Flevels[index];
end;

procedure TLevels.SetTreeCollection(ATreeCollection : TTreeCollection);
begin
   if aTreeCollection<>nil then
   begin
      FtreeCollection:=aTreeCollection;
      Changed(FtreeCollection);
   end
   else
   begin
       Flevels.clear;
       Fenditems.Clear;
       FtreeCollection:=nil;
   end;
end;

procedure TLevels.Changed(item : TTreeCollection);
var i : integer;

procedure AddToLevels(item : TTreeCollection; Levelindex : integer);
var i: integer;
begin
   if Levelindex > Flevels.Count-1 then
        while Flevels.Count-1<Levelindex do Flevels.Add('',Classname);
    TTreeCollection((Flevels[levelindex].add('',classname))).Data:=item;
    //add To EndItems - items without Childs
    if item.Count=0 then
        TTreeCollection((FEndItems.Add('',classname))).Data:=item
    else
      for i:=0 to item.count-1 do
         AddToLevels(item[i],Levelindex+1)

end;
begin
    //scanning into
   FLevels.BeginUpdate;
   FEnditems.BeginUpdate;
   try
     FLevels.Clear;
     //clear levels
     for i:=0 to FtreeCollection.Count-1 do
     begin
        AddToLevels(FtreeCollection[i],0);
     end;
   finally
     FLevels.EndUpdate;
     FEnditems.EndUpdate;
   end;
end;

function  TLevels.GetDepth : integer;
begin
   Result:=Flevels.Count;
end;

constructor TRootCollection.Create(aOwner : TComponent);
begin
   inherited;
   FCollection := TTreeCollection.Create(Nil);
   FCollection.FOwner :=Self;
end;

destructor  TRootCollection.Destroy;
begin
   FCollection.Clear;
   FCollection.Free;
   inherited;
end;

procedure TRootCollection.DefineProperties(Filer : TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('items', ReadCollection, WriteCollection, FCollection.Count>0);
end;


procedure TRootCollection.ReadCollection(Stream : TStream);
var
  aFiler : TTreeCollFiler;
begin

  aFiler := TTreeCollFiler.Create(Stream);
  FCollection.BeginUpdate;
  try
   aFiler.ReadObject(FCollection);
  finally
    FCollection.EndUpdate;
    aFiler.Free;
  end;
end;

procedure TRootCollection.WriteCollection(Stream : TStream);
var
    aFiler : TTreeCollFiler;
begin
   aFiler := TTreeCollFiler.Create(Stream);
   try
     aFiler.WriteObject(FCollection);
   finally
     aFiler.Free;
   end;
end;

procedure TRootCollection.LoadFromStream(S: TStream);
begin
  S.ReadComponent(Self);
end;

procedure TRootCollection.SaveToStream(S: TStream);
begin
  S.WriteComponent(Self);
end;

procedure TRootCollection.LoadFromFile(const FileName: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TRootCollection.SaveToFile(const FileName: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

//-ISFiler ---------------------------------------------------------------
procedure EnumerateProps(Instance : TPersistent; CallBack : TRWCallBack; PropCallback : TPCallback; SkipCallback : TSPCallBack; OnlyPublished, Reading : Boolean; var OldMethod : Boolean);
var
  I, Count, RealCount : Integer;
  PropInfo : PPropInfo;
  PropList : PPropList;
  aname : string;
begin

  If not Assigned(CallBack) then
    exit;
  Count :=  GetTypeData(Instance.ClassInfo)^.PropCount;
  if Count > 0 then
  begin
    GetMem(PropList, Count * SizeOf(Pointer));
    try
      RealCount := GetPropList(Instance.ClassInfo,tkAny,PropList);
      i:=0;
      while i < Realcount do //      for I :=  0 to RealCount - 1 do
      begin
        PropInfo :=  PropList^[I];
        //       
        try
         if not OnlyPublished or IsStoredProp(Instance, PropInfo) or Reading then
         begin
           if not OldMethod then
             aName :=PropCallback(Instance,PropInfo);
           if aname='' then OldMethod := true;
           if (PropInfo^.Name > aName) and not OldMethod then//skip writed property
           begin
              if @SkipCallback<>nil then SkipCallback;
           end
           else if ((PropInfo^.Name < aName) and not OldMethod) or (aName ='')then//skip this property
              inc(i)
           else //compatible
           begin
             CallBack(Instance,PropInfo,OldMethod);
             inc(i)
           end;
         end
         else inc(i);
        except
         TreeCollectionError(SInvalidPropertyValue);
        end;
      end;
    finally
      FreeMem(PropList, Count * SizeOf(Pointer));
    end;
  end;
end;

Function GetOrdSize(OT : TOrdType) : Integer;
begin
  case OT of
    otSByte : Result := 1;
    otUByte : Result := 1;
    otSWord : Result := 2;
    otUWord : Result := 2;
  else Result := 4;
  end;
end;

function GetPTypeInfo(PropInfo : PPropInfo) : PTypeInfo;
begin
  {$ifdef DFS_COMPILER_3_UP}
   Result := PropInfo^.PropType^;
  {$ELSE}
   {$IFDEF DFS_CPPB_3_UP}
    Result := PropInfo^.PropType^;
   {$ELSE}
    Result := PropInfo^.PropType;
   {$ENDIF}
  {$ENDIF}
end;

Function TTreeCollFiler.Seek(Offset : Longint; Origin : Word) : Longint;
begin
  Result := Stream.Seek(OffSet,Origin);
end;

Function TTreeCollFiler.Eof : Boolean;
begin
  Result := (Stream.Position>=FStream.Size);
end;

procedure TTreeCollFiler.SetStream(V : TStream);
begin
  FStream := V;
end;

Function TTreeCollFiler.Bof : Boolean;
begin
  Result := (Stream.Position=0);
end;


Function TTreeCollFiler.ReadChar : Char;
begin
  Read(Result,sizeof(Result));
end;

procedure TTreeCollFiler.WriteChar(Value : Char);
begin
  Write(Value,SizeOF(Value));
end;

procedure TTreeCollFiler.Write(const Buf; Count : Integer);
begin
  Stream.WriteBuffer(Buf,Count);
end;

procedure TTreeCollFiler.WriteString(const V : String);
var
  L : Integer;
begin
  L :=  Length(V);
  If L=0 then
  begin
    WriteValue(vaNull);
    exit;
  end;
  if L <= 255 then
  begin
    WriteValue(vaString);
    Write(L, SizeOf(Byte));
  end else
  begin
    WriteValue(vaLString);
    Write(L, SizeOf(Integer));
  end;
  Write(Pointer(V)^, L);
end;

procedure TTreeCollFiler.WriteValue(Value : TValueType);
begin
  Write(Value, SizeOf(Value));
end;

procedure TTreeCollFiler.WriteFloat(V : Extended);
begin
  Write(V,Sizeof(V));
end;

procedure TTreeCollFiler.Read(Var Buf; Count : Integer);
begin
  Stream.ReadBuffer(Buf,Count);
end;

function TTreeCollFiler.ReadString : string;
var
  L : Integer;
  A : TValueType;
begin
  L :=  0;
  Result := '';
  A := ReadValue;
  case A of
    vaNull :
      exit;
    vaString :
      Read(L, SizeOf(Byte));
    vaLString :
      Read(L, SizeOf(Integer));
  else
      {$ifdef DFS_COMPILER_3_UP}
          raise EReadError.Create(SInvalidPropertyValue);
      {$ELSE}
        {$IFDEF DFS_CPPB_3_UP}
          raise EReadError.Create(SInvalidPropertyValue);
        {$ELSE}
          raise EReadError.CreateRes(SInvalidPropertyValue);
        {$ENDIF}

      {$ENDIF}

  end;
  SetString(Result, PChar(nil), L);
  Read(Pointer(Result)^, L);
end;

function TTreeCollFiler.ReadValue : TValueType;
begin
  Read(Result, SizeOf(Result));
end;

Function TTreeCollFiler.ReadFloat : Extended;
begin
  Read(Result,sizeof(Result));
end;

destructor TTreeCollFiler.Destroy;
begin
  inherited;
end;


constructor TTreeCollFiler.Create(AStream : TStream);
begin
  inherited Create;
  Stream := AStream;
end;

procedure TTreeCollFiler.CheckStream;
begin
  If not Assigned(Stream) then
     raise Exception.Create(EStreamNA);
end;


procedure TTreeCollFiler.SkipClass;
Var
  Size : Integer;
begin
  Size := ReadLongInt;
  Seek(Size,soFromCurrent);
end;


procedure TTreeCollFiler.WriteObject(Instance : TPersistent);
Var
  TmpStream : TMemoryStream;
  TmpFiler : TTreeCollFiler;
  aOldMethod : boolean;
begin
  CheckStream;
  TmpStream := TMemoryStream.Create;
  TmpFiler := TTreeCollFiler.Create(TmpStream);
  aOldMethod:=false;
  try
    EnumerateProps(Instance, TmpFiler.WriteProperty,TmpFiler.WritePropertyName,nil, True, False, aOldMethod);
    If Instance is TCollection then
      TmpFiler.WriteCollection(TCollection(Instance))
    else If Instance is TTreeCollection then
      TmpFiler.WriteTreeCollection(TTreeCollection(Instance))
    else if Instance is TStrings then
      TmpFiler.WriteStrings(TStrings(Instance));

    TmpStream.Position := 0;
    WriteLongInt(TmpStream.Size);
    Stream.CopyFrom(TmpStream,TmpStream.Size);
  finally
    TmpFiler.Free;
    TmpStream.Free;
  end;
end;


procedure TTreeCollFiler.ReadObject(Instance : TPersistent);
Var
  Size : Integer;
  aName : String;
  TmpStream : TMemoryStream;
  TmpFiler : TTreeCollFiler;
  aOldMethod : boolean;
begin
  CheckStream;
  TmpStream := TMemoryStream.Create;
  TmpFiler := TTreeCollFiler.Create(TmpStream);
  aOldMethod:=false;
  try
    Size := ReadLongInt;
    if Size > 0 then
    begin

       TmpStream.CopyFrom(Stream,Size);

       TmpStream.Position := 0;
       EnumerateProps(Instance,TmpFiler.ReadProperty, TmpFiler.ReadPropertyName, Tmpfiler.SkipProperty, True, True, aOldMethod);
       //skip unreaded properties
       if TmpStream.Position < Size then
       begin
        aName := TmpFiler.ReadPropertyName(Instance, nil);
        while AName <> '' do
        begin
           TmpFiler.SkipProperty;
           aName := TmpFiler.ReadPropertyName(Instance, nil);
        end;
       end;
       If Instance is TCollection then
          TmpFiler.ReadCollection(TCollection(Instance))
       else If Instance is TTreeCollection then
          TmpFiler.ReadTreeCollection(TTreeCollection(Instance))
       else if Instance is TStrings then
          TmpFiler.ReadStrings(TStrings(Instance));
    end;
  finally
    TmpFiler.Free;
    TmpStream.Free;
  end;
end;

procedure TTreeCollFiler.WriteCollection(Value : TCollection);
var
  I : Integer;
begin
  WriteLongInt(Value.Count);
  for I :=  0 to Value.Count - 1 do
    WriteObject(Value.Items[I]);
end;


procedure TTreeCollFiler.ReadCollection(Value : TCollection);
var
  Count,I : Integer;
begin
  Count :=  ReadLongInt;
  try
    Value.BeginUpdate;
    Value.Clear;
    For i :=  0 to Count-1 do
      ReadObject(Value.Add);
  finally
    Value.EndUpdate;
  end;
end;

procedure TTreeCollFiler.ReadStrings(Value : TStrings);
var
  Count,I : Integer;
begin
  Count :=  ReadLongInt;
  try
    Value.BeginUpdate;
    Value.Clear;
    For i :=  0 to Count-1 do
      Value.Add(ReadString);
  finally
    Value.EndUpdate;
  end;

end;

procedure TTreeCollFiler.WriteStrings(Value : TStrings);
var
  I : Integer;
begin
  WriteLongInt(Value.Count);
  for I :=  0 to Value.Count - 1 do
    WriteString(Value.strings[I]);
end;

procedure TTreeCollFiler.WriteTreeCollection(Value : TTreeCollection);
var
  I : Integer;
begin
  WriteLongInt(Value.Count);
  for I :=  0 to Value.Count - 1 do
  begin
    WriteString(Value[I].Classname);
    WriteObject(Value[I]);
  end;
end;

procedure TTreeCollFiler.ReadTreeCollection(Value : TTreeCollection);
var
  Count, I : Integer;
  aClassname : ShortString;
begin
  Count :=  ReadLongInt;
  try
    Value.BeginUpdate;
    //if Count>0 then
      Value.Clear;
    For i :=  0 to Count-1 do
    begin
      aClassname :=  ReadString;
      // ReadObject(TPersistent(Value.Add('', aClassName)));
        ReadObject(TPersistent(Value.Add('', Value.InternalGetClass(aClassName).Classname)));

    end;
  finally
    Value.EndUpdate;
  end;
end;

Function  TTreeCollFiler.ReadSmallInt : SmallInt;
begin
  Read(Result,Sizeof(Result));
end;

Function  TTreeCollFiler.ReadShortInt : ShortInt;
begin
  Read(Result,Sizeof(Result));
end;

Function  TTreeCollFiler.ReadLongInt : LongInt;
begin
  Read(Result,Sizeof(Result));
end;

procedure TTreeCollFiler.WriteSmallInt(Value : SmallInt);
begin
  Write(Value, SizeOf(Smallint));
end;

procedure TTreeCollFiler.WriteShortInt(Value : ShortInt);
begin
  Write(Value, SizeOf(Shortint));
end;

procedure TTreeCollFiler.WriteLongInt(Value : LongInt);
begin
  Write(Value, SizeOf(Longint));
end;

procedure TTreeCollFiler.WriteListBegin;
begin
  WriteValue(vaList);
end;

procedure TTreeCollFiler.WriteListEnd;
begin
  WriteValue(vaNull);
end;

function TTreeCollFiler.NextValue : TValueType;
begin
  Result :=  ReadValue;
  Stream.Seek(-1,soFromCurrent);
end;


procedure TTreeCollFiler.WriteProperty( Instance : TPersistent; PropInfo : PPropInfo; var OldMethod : boolean);
var
  PropType : PTypeInfo;
  procedure WriteIntProp(IntType : PTypeInfo; Value : Longint);
  Var
    P : PTypeData;
  begin
    P := GetTypeData(IntType);
    Write(Value,GetOrdSize(P^.OrdType));
  end;
  procedure WriteObjProp;
  var
    Value : TObject;
  begin
    Value :=  TObject(GetOrdProp(Instance, PropInfo));
    if (Value = nil) then
    begin
     WriteValue(vaNil);
     exit;
    end;
    if not(Value is TPersistent) then exit;
    if Value is TCollection then
    begin
      WriteValue(vaCollection);
      WriteCollection(TCollection(Value))
    end else
    begin
      WriteValue(vaTrue);
      WriteObject(TPersistent(Value));
    end;
  end;

  procedure WriteSet(Value : Longint);
  var
    I : Integer;
    BaseType : PTypeInfo;
  begin
    {$IFDEF DFS_COMPILER_3_UP}
    BaseType :=  GetTypeData(PropType)^.CompType^;
    {$ELSE}
     {$IFDEF DFS_CPPB_3_UP}
       BaseType :=  GetTypeData(PropType)^.CompType^;
     {$ELSE}
       BaseType :=  GetTypeData(PropType)^.CompType;
     {$ENDIF}
    {$ENDIF}
    WriteValue(vaSet);
    for I :=  0 to SizeOf(TIntegerSet) * 8 - 1 do
      if I in TIntegerSet(Value) then WriteString(GetEnumName(BaseType, I));
    WriteString('');
  end;
var aoldPos, anewPos : Longint;
begin
// try
  if PropInfo^.SetProc = nil then
    exit;
  PropType :=  GetPTypeInfo(PropInfo);

  //write buffer pos - for future write past position
  aoldPos := Stream.Position;
  //WritePropertyName(Instance, PropInfo);
  Write(aoldPos, SizeOf(Longint));

  case PropType^.Kind of
    tkclass :
      WriteObjProp;
    tkinteger,tkenumeration :
      WriteIntProp(PropType,GetOrdProp(Instance, PropInfo));
    tkSet :
      WriteSet(GetOrdProp(Instance, PropInfo));
    tkChar :
      WriteChar(Chr(GetOrdProp(Instance, PropInfo)));
    tkfloat :
      WriteFloat(GetFloatProp(Instance, PropInfo));
    tkString, tkLString {$IFDEF DFS_COMPILER_3_UP}, tkWString {$ENDIF} :
      WriteString(GetStrProp(Instance, PropInfo));
  //else Raise Exception.createFMT(EUnKnownProp,[PropType^.Name]);
  end;
// except
  //write next position of next property
  anewPos := Stream.Position;
  Stream.Position:= aOldPos;
  Write(aNewPos,sizeof(longint));
  Stream.Position:= anewPos;
// end;
end;

procedure ReadError(Ident : String);
begin
  raise EReadError.Create(Ident);
end;

procedure TTreeCollFiler.SkipProperty;
var apos : Longint;
begin
   ReadString;//readpropertyname
   Read(aPos,Sizeof(Longint));
   Stream.Position:=aPos;
end;

function TTreeCollFiler.ReadPropertyName(Instance : TPersistent; PropInfo : PPropInfo) : String;    //Read next property name
var aPos : Longint;
begin
   aPos := Stream.Position;
   Result:='';
   if NextValue=VaString then
   begin
     Result:=ReadString;//nextvalue
     Stream.Seek(aPos,soFromBeginning);
   end;
end;

function TTreeCollFiler.WritePropertyName(Instance : TPersistent; PropInfo : PPropInfo): String;    //Read next property name
begin
  WriteString(Propinfo^.name);
  Result := Propinfo^.name;
end;

procedure TTreeCollFiler.ReadProperty(Instance : TPersistent; PropInfo : PPropInfo; var OldMethod : Boolean);
var
  PropType : PTypeInfo;
  aPos : longint;
  aName : String;

  function ReadSet(SetType : Pointer) : Integer;
  var
    EnumType : PTypeInfo;
    EnumName : string;
    EnumValue : Integer;
  begin
    try
      if ReadValue <> vaSet then
      {$IFDEF DFS_COMPILER_3_UP}
          raise EReadError.Create(SInvalidPropertyValue);
      {$ELSE}
        {$IFDEF DFS_CPPB_3_UP}
          raise EReadError.Create(SInvalidPropertyValue);
        {$ELSE}
          raise EReadError.CreateRes(SInvalidPropertyValue);
        {$ENDIF}
      {$ENDIF}
      {$IFDEF DFS_COMPILER_3_UP}
      EnumType :=  GetTypeData(SetType)^.CompType^;
      {$ELSE}
        {$IFDEF DFS_CPPB_3_UP}
          EnumType :=  GetTypeData(SetType)^.CompType^;
        {$ELSE}
          EnumType :=  GetTypeData(SetType)^.CompType;
        {$ENDIF}

      {$ENDIF}
      Result :=  0;
      while True do
      begin
        EnumName :=  ReadString;
        if EnumName = '' then
          Break;
        EnumValue := GetEnumValue(EnumType, EnumName);
        If EnumValue<>-1 then
          Include(TIntegerSet(Result), EnumValue);
      end;
    except
      while ReadString <> '' do;
      raise;
    end;
  end;

  Procedure ReadIntProp;
  Var
    Value : Integer;
    P : PTypeData;
    IntType : PTypeInfo;
  begin
    IntType := GetPTypeInfo(PropInfo);
    P := GetTypeData(IntType);
    Read(Value,GetOrdSize(P^.OrdType));
    SetOrdProp(Instance, PropInfo, Value);
  end;

  procedure ReadObjProp;
  begin
    case NextValue of
      vaNil :
        begin
          ReadValue;
          SetOrdProp(Instance, PropInfo, 0)
        end;
      vaCollection :
        begin
          ReadValue;
          ReadCollection(TCollection(GetOrdProp(Instance, PropInfo)));
        end;
      vaTrue :
        begin
          ReadValue;
          ReadObject(TPersistent(GetOrdProp(Instance, PropInfo)));
        end;
    else
      {$IFDEF DFS_COMPILER_3_UP}
          raise EReadError.Create(SInvalidPropertyValue);
      {$ELSE}
       {$IFDEF DFS_CPPB_3_UP}
          raise EReadError.Create(SInvalidPropertyValue);
       {$ELSE}
          raise EReadError.CreateRes(SInvalidPropertyValue);
       {$ENDIF}

      {$ENDIF}

    end;
  end;

begin
  if PropInfo^.SetProc = nil then

    {$IFDEF DFS_COMPILER_3_UP}
          ReadError(SReadOnlyProperty);
    {$ELSE}
      {$IFDEF DFS_CPPB_3_UP}
         ReadError(SReadOnlyProperty);
      {$ELSE}
         EReadError.CreateRes(SInvalidPropertyValue);
      {$ENDIF}

    {$ENDIF}

  aPos:=-1;
  if not OldMethod then
  begin
     aname :=ReadString;
     Read(aPos,Sizeof(longint));
  end;
  try
   PropType :=  GetPTypeInfo(PropInfo);

   case PropType^.Kind of
      tkInteger,tkenumeration :
        ReadIntProp;
      tkSet :
        SetOrdProp(Instance, PropInfo, ReadSet(PropType));
      tkChar :
        SetOrdProp(Instance, PropInfo, Ord(ReadChar));
      tkFloat :
        SetFloatProp(Instance, PropInfo, ReadFloat);
      tkString, tkLString {$IFDEF DFS_COMPILER_3_UP}, tkWString {$ENDIF} :
        SetStrProp(Instance, PropInfo, ReadString);
      tkclass :
        ReadObjProp;
      //else Raise Exception.CreateFmt(EUnKnownProp,[PropType^.Name]);
    end;
  except
    raise;
    if aPos<>-1 then Stream.Position:=aPos;//goto next value
  end;
  //read reged properties

end;


{function  TTreeListPropEditor.GetAttributes : TPropertyAttributes;
begin
  result := [paDialog];
end;

function  TTreeListPropEditor.GetValue : string;
begin
  if GetOrdValue = 0 then
    result := '(None)'
  else
    result := '(TTreeCollection)';
end;

procedure TTreeListPropEditor.Edit;
var PropInfo : PPRopInfo;
    PropValue : TObject;
begin
   PropInfo := GetPropInfo;
   if PropInfo^.PropType^.Kind = tkClass then
   begin
      PropValue := TObject(GetOrdValue);
      EditTreeList(TTreeCollection(PropValue),TTreeCollection(PropValue).InternalGetClass,Designer);
   end;
end;}

{ENDIF}
//Class management fuctions
procedure RegisterTreeCollectionClass(AClass: TTreeCollectionClass);
begin
 if AClass <> nil then
  with TreeCollectionClassList{$ifdef DFS_COMPILER_3_UP}.LockList{$endif} do
  try
    if IndexOf(AClass) = -1 then
    begin
      //AClassName := AClass.ClassName;
      //for i:=0 to Count-1 do
      //  classreged :=TTreeCollectionClass(items[i]).Classname = aClassname;
      //if IndexOf(AClass) =-1 then
         Add(AClass);
      //if not classreged then //
      //begin
      //  Add(AClass);
      //end;
    end;
  finally
    {$ifdef DFS_COMPILER_3_UP}
    TreeCollectionClassList.UnlockList;
    {$endif}
  end;
end;


procedure RegisterTreeCollectionClasses(AClasses: array of TTreeCollectionClass);
var
  I: Integer;
begin
  for I := Low(AClasses) to High(AClasses) do
   RegisterTreeCollectionClass(AClasses[I]);
end;

procedure UnRegisterTreeCollectionClass(AClass: TTreeCollectionClass);
begin
  TreeCollectionClassList.Remove(AClass);
end;

procedure UnRegisterTreeCollectionClasses(AClasses: array of TTreeCollectionClass);
var
  I: Integer;
begin
  for I := Low(AClasses) to High(AClasses) do UnRegisterTreeCollectionClass(AClasses[I]);
end;

procedure WritePersistentToStream(Obj : TPersistent; Stream : TStream);
var
    aFiler : TTreeCollFiler;
begin
   aFiler := TTreeCollFiler.Create(Stream);
   try
     aFiler.WriteObject(Obj);
   finally
     aFiler.Free;
   end;
end;

procedure ReadPersistentFromStream(Obj : TPersistent; Stream : TStream);
var
  aFiler : TTreeCollFiler;
begin
  aFiler := TTreeCollFiler.Create(Stream);
  try
   aFiler.ReadObject(Obj);
  finally
    aFiler.Free;
  end;
end;


procedure WriteCollectionToStream(Collection : TTreeCollection; Stream : TStream);
var
    aFiler : TTreeCollFiler;
begin
   aFiler := TTreeCollFiler.Create(Stream);
   try
     aFiler.WriteObject(Collection);
   finally
     aFiler.Free;
   end;
end;

procedure ReadCollectionFromStream(Collection : TTreeCollection; Stream : TStream);
var
  aFiler : TTreeCollFiler;
begin
  aFiler := TTreeCollFiler.Create(Stream);
  try
   aFiler.ReadObject(Collection);
  finally
    aFiler.Free;
  end;
end;

procedure ConvertTextToBinary(aStream : TStream; aText : String);
var Count : integer;
    Buffer: array[0..255] of Char;
    FSourcePtr: PChar;
{$ifndef DFS_COMPILER_4_UP}
function HexToBin(Text, Buffer: PChar; BufSize: Integer): Integer; assembler;
asm
        PUSH    ESI
        PUSH    EDI
        PUSH    EBX
        MOV     ESI,EAX
        MOV     EDI,EDX
        MOV     EBX,EDX
        MOV     EDX,0
        JMP     @@1
@@0:    DB       0, 1, 2, 3, 4, 5, 6, 7, 8, 9,-1,-1,-1,-1,-1,-1
        DB      -1,10,11,12,13,14,15,-1,-1,-1,-1,-1,-1,-1,-1,-1
        DB      -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
        DB      -1,10,11,12,13,14,15
@@1:    LODSW
        CMP     AL,'0'
        JB      @@2
        CMP     AL,'f'
        JA      @@2
        MOV     DL,AL
        MOV     AL,@@0.Byte[EDX-'0']
        CMP     AL,-1
        JE      @@2
        SHL     AL,4
        CMP     AH,'0'
        JB      @@2
        CMP     AH,'f'
        JA      @@2
        MOV     DL,AH
        MOV     AH,@@0.Byte[EDX-'0']
        CMP     AH,-1
        JE      @@2
        OR      AL,AH
        STOSB
        DEC     ECX
        JNE     @@1
@@2:    MOV     EAX,EDI
        SUB     EAX,EBX
        POP     EBX
        POP     EDI
        POP     ESI
end;
{$endif}
begin
  if aText = '' then exit;
  FSourcePtr := PChar(aText);
  // for skip '{' char
  if FSourcePtr^ = '{' then
    Inc(FSourcePtr);
  while FSourcePtr^ <> '}' do
  begin
    Count := HexToBin(FSourcePtr, Buffer, SizeOf(Buffer));
    //if Count = 0 then Error(SInvalidBinary);
    aStream.Write(Buffer, Count);
    Inc(FSourcePtr, Count * 2);
  end;
end;

//remove #13#10 chars
function TrimBlanks(aString  : String) : String;
var aStrings : TStringlist;
     i : integer;
begin
   aStrings := TStringlist.Create;
   try
      aStrings.Text := aString;
      for i :=0 to aStrings.Count-1 do
        result:= Result + aStrings[i];
   finally
     aStrings.Free;
   end;

end;

function ConvertBinaryToText(aStream : TStream; SplitToStrings: Boolean) : String;
const
  BytesPerLine = 32;
var
    MultiLine: Boolean;
    I: Integer;
    Count: Longint;
    Buffer: array[0..BytesPerLine - 1] of Char;
    Text: array[0..BytesPerLine * 2 - 1] of Char;
{$ifndef DFS_COMPILER_4_UP}
procedure BinToHex(Buffer, Text: PChar; BufSize: Integer); assembler;
asm
        PUSH    ESI
        PUSH    EDI
        MOV     ESI,EAX
        MOV     EDI,EDX
        MOV     EDX,0
        JMP     @@1
@@0:    DB      '0123456789ABCDEF'
@@1:    LODSB
        MOV     DL,AL
        AND     DL,0FH
        MOV     AH,@@0.Byte[EDX]
        MOV     DL,AL
        SHR     DL,4
        MOV     AL,@@0.Byte[EDX]
        STOSW
        DEC     ECX
        JNE     @@1
        POP     EDI
        POP     ESI
end;

{$endif}

begin
    Count := aStream.Size;
    Result:='{';
    aStream.Position:=0;
    MultiLine := Count >= BytesPerLine;
    while Count > 0 do
    begin
      if MultiLine and SplitToStrings then Result:=Result + #13#10;//add new line
      if Count >= 32 then I := 32 else I := Count;
      aStream.Read(Buffer, I);
      BinToHex(Buffer, Text, I);
      Result:=Result + Text;
      Dec(Count, I);
    end;
    Result:= Result + '}';
end;

procedure LoadTreeCollectionFromText(Collection : TTreeCollection; Text : string);
Var
  a : TMemoryStream;
  TmpText : String;
begin
   a := TMemoryStream.Create;
   try
     TmpText := TrimBlanks(Text);
     ConvertTextToBinary(a, TmpText);
     a.Position:=0;
     ReadCollectionFromStream(Collection, a);
   finally
      a.Free;
   end;
end;

function SaveTreeCollectionToText(Collection : TTreeCollection; SplitToStrings: Boolean) : string;
Var
  a : TMemoryStream;
begin
   a := TMemoryStream.Create;
   try
     WriteCollectionToStream(Collection, a);
     Result:= ConvertBinaryToText(a, SplitToStrings);
   finally
      a.Free;
   end;
end;

procedure LoadPersistentFromText(Obj : TPersistent; Text : string);
Var
  a : TMemoryStream;
  TmpText : String;
begin
   a := TMemoryStream.Create;
   try
     TmpText := TrimBlanks(Text);
     ConvertTextToBinary(a, TmpText);
     a.Position:=0;
     ReadPersistentFromStream(Obj, a);
   finally
      a.Free;
   end;
end;

function SavePersistentToText(Obj : TPersistent; SplitToStrings: Boolean) : string;
Var
  a : TMemoryStream;
begin
   a := TMemoryStream.Create;
   try
     WritePersistentToStream(Obj, a);
     Result:= ConvertBinaryToText(a, SplitToStrings);
   finally
      a.Free;
   end;
end;


procedure LoadTreeCollectionFromFile(Collection : TTreeCollection; const Filename : string);
var aStrings : TStringlist;
begin
   aStrings := TStringlist.Create;
   try
     aStrings.LoadFromFile(Filename);
     LoadTreeCollectionFromText(Collection, aStrings.text);
   finally
      aStrings.Free;
   end;

end;

procedure SaveTreeCollectionToFile(Collection : TTreeCollection; const Filename : string);
var aStrings : TStringlist;
begin
   aStrings := TStringlist.Create;
   try
     aStrings.text := SaveTreeCollectionToText(Collection, True);
     aStrings.SaveToFile(Filename);
   finally
      aStrings.Free;
   end;
end;

procedure AssignProperties(Instance1, Instance2 : TPersistent; RecurseClasses : boolean; ExcludeFilter : string);
var
  I, Count : Integer;
  PropInfo1 : PPropInfo;
  PropList1 : PPropList;
//  aObject1,aObject2 : TObject;
  aTypeInfo : PTypeInfo;
  aList : TStringList;
  ToSet : boolean;
  PropName : string;
begin
  aList:=nil;
  Count :=  GetTypeData(Instance1.ClassInfo)^.PropCount;
  if Count > 0 then
  begin
    if ExcludeFilter <> '' then
    begin
      aList := TStringList.Create;
      aList.CommaText := ExcludeFilter;
    end;
    GetMem(PropList1, Count * SizeOf(Pointer));
    try
      GetPropInfos(Instance1.ClassInfo, PropList1);
      for I :=  0 to Count - 1 do
      begin
        PropInfo1 :=  PropList1^[I];
        //PropInfo2 :=  PropList2^[I];
        aTypeInfo := GetPTypeInfo(PropInfo1);//getclass
        if aList <> nil then
        begin
           PropName := PropInfo1^.name;
           ToSet := aList.IndexOf(PropName) = -1;
        end
        else ToSet := true;

        if ToSet and IsStoredProp(Instance1, PropInfo1) then
        Case aTypeInfo^.Kind of
          tkInteger, tkChar, tkEnumeration, tkSet:
          begin
             SetOrdProp(Instance2, Propinfo1, GetOrdProp(Instance1,PropInfo1));
          end;
          tkFloat:
             SetFloatProp(Instance2, Propinfo1, GetFloatProp(Instance1,PropInfo1));
          tkString{$ifdef DFS_COMPILER_3_UP}, tkLString, tkWString{$endif}:
             SetStrProp(Instance2, Propinfo1, GetStrProp(Instance1,PropInfo1));
          tkClass:
          if RecurseClasses then
          begin
             //aObject1 := TObject(GetOrdProp(Instance1, PropInfo1));
             //aObject2 := TObject(GetOrdProp(Instance2, PropInfo1));
             SetOrdProp(Instance2, Propinfo1, GetOrdProp(Instance1, PropInfo1));
             //if (aObject1 is TPersistent) and (aObject2 is TPersistent) then
             //   TPersistent(aObject2).Assign(TPersistent(aObject1));//AssignProperties(TPersistent(aObject1), TPersistent(aObject2), RecurseClasses);
          end;
        end;

      end;
    finally
      FreeMem(PropList1, Count * SizeOf(Pointer));
      if aList <> nil then
        aList.Free;
    end;
  end;

end;

procedure SaveTreeCollectionToStream(Collection : TTreeCollection; Stream : TStream);
var aStrings : TStringlist;
begin
   aStrings := TStringlist.Create;
   try
     aStrings.text := SaveTreeCollectionToText(Collection, True);
     aStrings.SaveToStream(Stream);
   finally
      aStrings.Free;
   end;
end;

procedure LoadTreeCollectionFromStream(Collection : TTreeCollection; Stream : TStream);
var aStrings : TStringlist;
begin
   aStrings := TStringlist.Create;
   try
     aStrings.LoadFromStream(Stream);
     LoadTreeCollectionFromText(Collection, aStrings.text);
   finally
      aStrings.Free;
   end;
end;

initialization
{$IFNDEF Registered}
{$IFNDEF ACTIVEX}
  ShowURToolTip('', false);
{$ENDIF}
{$ENDIF}
//  OutputDebugString(PChar('Trecol Init'));
  if GetClass('TTreeCollection') = nil then
    Classes.Registerclass(TTreeCollection);
  TreeCollectionClassList := {$ifdef DFS_COMPILER_3_UP}TThreadList.Create{$else} TISList.Create{$endif};

finalization
  TreeCollectionClassList.Free;
end.
