unit ishtmprv;
{*********************************************************}
//  Copyright (c) 2000 IMCA SYSTEMS
//  All Rights Reserved.
//  Version
//  MODULE: Preview Object with Preview abilities
{*********************************************************}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, Buttons, ExtCtrls, ishtmPrn, Printers, ToolWin,
  ComCtrls;

type
  THTMLPreviewForm = class(TForm)
    ScrollB: TScrollBox;
    PrintDialog1: TPrintDialog;
    Page: TPaintBox;
    ToolBar1: TToolBar;
    SBNextPage: TSpeedButton;
    SBPrevPage: TSpeedButton;
    SBPrint: TSpeedButton;
    SBDone: TSpeedButton;
    SaveButton: TSpeedButton;
    OpenButton: TSpeedButton;
    ZoomOut: TSpeedButton;
    ZoomIn: TSpeedButton;
    DispGrid: TSpeedButton;
    SB: TStatusBar;
    procedure SBNextPageClick(Sender: TObject);
    procedure SBDoneClick(Sender: TObject);
    procedure PagePaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure ZoomInClick(Sender: TObject);
    procedure ZoomOutClick(Sender: TObject);
    procedure SaveButtonClick(Sender: TObject);
    procedure DispGridClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure SBPrintClick(Sender: TObject);
    procedure SBPrevPageClick(Sender: TObject);

  private
    FCurPage : integer;//current page
    FScale : single;//zoom factor
    FHTMLPrinter : TISPrinter;
    procedure SetScale(aScale : single);
  public
    //ReportSystem: TReportSystem;
    FormClosed: boolean;
    procedure SeTISPrinter(aPrinter : TISPrinter);
    procedure ZoomPageHeight;
    procedure ZoomPagewidth;
    procedure ZoomPageFactor(aScale : single);
    procedure Displaypage(zoomFactor : single; aPage : integer);
    property Scale : single read Fscale write SetScale;
  end;

var
  HTMLPreviewForm: THTMLPreviewForm;

implementation

{$R *.DFM}


procedure THTMLPreviewForm.SBDoneClick(Sender: TObject);
begin
  Close;
end;

procedure THTMLPreviewForm.PagePaint(Sender: TObject);
begin
   Page.canvas.Brush.color:=Scrollb.Color;
   Page.canvas.FillRect(Rect(0,0,Page.width, Page.Height));
   DisplayPage(FScale, FCurPage);
end;

procedure THTMLPreviewForm.Displaypage(zoomFactor : single; aPage : integer);
var r : TRect;
    Step,Cur: double;
    IntCur : integer;

begin
    //
    //Draw Shadows
    //Page.Canvas.Brush.Color :=ScrollB.Color;
    //Page.Canvas.FillRect(Rect(0, 0, Page.Width, Page.Height));
    r.Left := 30;
    r.Top := 30;
    r.Right := Page.width-30;
    r.Bottom := Page.height-30;
    Page.Canvas.Brush.Color :=clWindow;
    Page.Canvas.Rectangle(r.left, r.Top,R.Right, R.Bottom);
    Page.Canvas.Brush.Color :=clBlack;
    Page.Canvas.FillRect(Rect(R.Right, R.Top+8, R.Right+8, R.Bottom+8));
    Page.Canvas.FillRect(Rect(R.Left+8, R.Bottom, R.Right, R.Bottom+8));
    //Page.Canvas.FillRect(R);
    //Escape(Printer.Handle, GETPRINTINGOFFSET, 0, nil, @p);
    Page.Canvas.StretchDraw(R, FHTMLPrinter.Metafiles[FCurPage]);
    //Draw Grid
    if DispGrid.down then
    begin
     Page.Canvas.Pen.color := clgray;
     with FHtmlPrinter do
     begin
      Step := (Page.width-60)/((Pagesize.X/XPixelPerInch)*2.542);//cm
      Cur:=30;
      While Cur < (Page.width-30) do
      begin
        IntCur := Trunc(Cur);
        Page.Canvas.MoveTo(IntCur,30);
        Page.Canvas.LineTo(IntCur,Page.Height-30);
        Cur:=Cur+Step;
      end;
      Step := (Page.Height-60)/((Pagesize.Y/YPixelPerInch)*2.542);//cm
      Cur:=30;
      While Cur < (Page.Height-30) do
      begin
        IntCur := Trunc(Cur);
        Page.Canvas.MoveTo(30,IntCur);
        Page.Canvas.LineTo(Page.Width-30,IntCur);
        Cur:=Cur+Step;
      end;
    end;
   end;
end;

procedure THTMLPreviewForm.ZoomPageFactor(aScale : single);
begin
 with FHTMLPrinter do
 begin
   Page.Height := Trunc(Pagesize.Y * FScale);
   Page.width := Trunc(Pagesize.X * FScale);
   //positioning
   Page.top := (Scrollb.Height-Page.height) div 2;
   if Page.top < 0 then Page.Top:=0;
   Page.Left:= (Scrollb.width-Page.width) div 2;
   if Page.Left < 0 then Page.Left:=0;
   //display page
   //DisplayPage(FScale, FcurPage);
   Page.Invalidate;
 end;

end;

procedure THTMLPreviewForm.ZoomPageHeight;
begin
   Scale := (ScrollB.Height) / FHTMLPrinter.PageSize.Y;
end;

procedure THTMLPreviewForm.ZoomPagewidth;
begin
   Scale := (ScrollB.Width) / FHTMLPrinter.PageSize.X;
end;

procedure THTMLPreviewForm.SetScale(aScale : single);
begin
    FScale := aScale;
    ZoomPageFactor(Fscale);
    Sb.Panels[0].Text:=' x '+CurrToStr(FScale);
end;

procedure THTMLPreviewForm.SetISPrinter(aPrinter : TISPrinter);
begin
  if aPrinter <> nil then
  begin
    // settings
    FHTMLPrinter :=aPrinter;
    FCurPage:=1;
    Sb.Panels[1].Text:=inttostr(FCurPage)+':'+inttostr(FHTMLPrinter.Pagecount);
    SBPrevPage.enabled:=false;
    SBNextPage.enabled := FHTMLPrinter.Pagecount>FcurPage;
    ZoomPagewidth;
  end;
end;





procedure THTMLPreviewForm.FormResize(Sender: TObject);
begin
   // relocate page
   Scale :=Fscale;
end;

procedure THTMLPreviewForm.ZoomInClick(Sender: TObject);
begin
   Scale:=Fscale*2;
end;

procedure THTMLPreviewForm.ZoomOutClick(Sender: TObject);
begin
  Scale:=Fscale/2;
end;

procedure THTMLPreviewForm.SaveButtonClick(Sender: TObject);
var aSd : TSavedialog;
begin
  aSd := TSavedialog.Create(Application);
  try
    with aSD do
    begin
      title:='Save To file';
      Filter:='Enhanced Windows Metafile|*.emf|All files|*.*';
      DefaultExt :='emf';
      Options :=[ofOverwritePrompt,ofPathMustExist];
      FilterIndex :=1;
      if Execute then
        FHTMLPrinter.Metafiles[FcurPage].Savetofile(Filename);
    end;
  finally
    aSD.Free;
  end;
end;

procedure THTMLPreviewForm.DispGridClick(Sender: TObject);
begin
  Page.Invalidate;
end;

procedure THTMLPreviewForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
   Case Key of
      VK_RIGHT: SendMessage(ScrollB.handle,WM_HSCROLL, SB_LINERIGHT, ScrollB.VertScrollbar.Position);
      VK_LEFT:   SendMessage(ScrollB.handle,WM_HSCROLL, SB_LINELEFT, ScrollB.VertScrollbar.Position);
      VK_UP:    SendMessage(ScrollB.handle,WM_VSCROLL, SB_LINEUP, ScrollB.HorzScrollbar.Position);
      VK_DOWN:  SendMessage(ScrollB.handle, WM_VSCROLL, SB_LINEDOWN, ScrollB.HorzScrollbar.Position);
      VK_NEXT:  SendMessage(ScrollB.handle, WM_VSCROLL, SB_PAGEDOWN, ScrollB.HorzScrollbar.Position);
      VK_PRIOR:  SendMessage(ScrollB.handle, WM_VSCROLL, SB_PAGEUP, ScrollB.HorzScrollbar.Position);
   end;
end;

procedure THTMLPreviewForm.SBPrintClick(Sender: TObject);
begin
   FHTMLPrinter.PrintIt;
end;

procedure THTMLPreviewForm.SBPrevPageClick(Sender: TObject);
begin
 if FcurPage>0 then
 begin
  Dec(FCurPage);
  If FcurPage=0 then SBPrevPage.enabled:=false;
  displayPage(Fscale, FcurPage);
  SBNextPage.Enabled:=true;
  Sb.Panels[1].Text:=inttostr(FCurPage)+':'+inttostr(FHTMLPrinter.Pagecount);
 end;
end;

procedure THTMLPreviewForm.SBNextPageClick(Sender: TObject);
begin
 if (FcurPage+1)<= FHTMLPrinter.Pagecount then
 begin
  inc(FcurPage);
  sbPrevPage.enabled:=true;
  displayPage(Fscale, FcurPage);
  if FcurPage = FHTMLPrinter.Pagecount then SBNextPage.enabled:=false;
  SBPrevPage.enabled:=True;
  Sb.Panels[1].Text:=inttostr(FCurPage)+':'+inttostr(FHTMLPrinter.Pagecount);
 end;
end;

end.
