unit ispinedit;
{/*****************************************************************/
   Copyright (c) 2001 IMCA SYSTEMS
   All Rights Reserved.
   Creation Date: 23. Apr. 2001
   Version 1.0
   MODULE: Inplace editor based on ISPlugEditor
   Description/features:
/*****************************************************************/}
interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls,
  {$ifdef DFS_COMPILER_4_UP} ImgList, {$endif} ISPlugedit;

type

  TISPlugInplaceEdit = class;
  TSelection = record
    StartPos, EndPos: Integer;
  end;

  TEditOption =(eosAlwaysShowEditor, eosGoTabs, eosReadOnly);
  TEditOptions = set Of TEditOption;

  //for future use
  TIEDataType = (iedBoolean,iedInteger,iedCurrency,iedPicklist,iedChecklist);
  TIEGetEditDataEvent = procedure(Sender: TISPlugInplaceEdit;  var AData : String) of object;
  TIEValidateDataEvent = procedure(Sender: TISPlugInplaceEdit; AData : String; var aCancel : boolean) of object;
  TIEGetPickListEvent = procedure(Sender: TISPlugInplaceEdit; var aPickList : Tstrings) of object;
  TIEGetReadOnlyEvent = procedure(Sender: TISPlugInplaceEdit; var aReadOnly : boolean) of object;

  TIEEditButtonClickEvent = procedure(Sender: TISPlugInplaceEdit) of object;
  TIEHideEditorEvent = procedure(Sender: TISPlugInplaceEdit) of object;
  TIEUpdateDataEvent= procedure(Sender: TISPlugInplaceEdit) of object;
  TIEGetClickKeyEvent=procedure(Sender: TISPlugInplaceEdit; var aKey : Word; var aShift : TShiftState) of object;
  TIEUpdateEditEvent =procedure(Sender: TISPlugInplaceEdit) of object;

  TISPlugInplaceEdit = class(TISCustomPlugEdit)
  private
    FControl: TWinControl;
    FChanged, FValidating, FShowed : boolean;
    FEditOptions : TEditOptions;
    //FPasswordChar : Char;//new
    //FAlignment: TAlignment;
    //callback from FControl events
    FOnGetEditData : TIEGetEditDataEvent;
    FOnValidateData : TIEValidateDataEvent;
    FonGetPickList : TIEGetPickListEvent;
    FonGetReadOnly : TIEGetReadOnlyEvent;

    FonEditButtonClick:TIEEditButtonClickEvent;
    FOnHideEditor : TIEHideEditorEvent;
    FOnUpdateData : TIEUpdateDataEvent;
    FonGetClickKey : TIEGetClickKeyEvent;
    FonUpdateEdit : TIEUpdateEditEvent;
    FonKeyUpParent : TKeyEvent;
    FonKeydownParent : TKeyEvent;
    FonKeyPressParent : TKeyPressEvent;
    FonDblClickParent :  TNotifyEvent;

    FClickTime: Longint;
    FCancelData:boolean;

    procedure InternalMove(const Loc: TRect; Redraw: Boolean; Scrolling : boolean);

    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMPaste(var Message); message WM_PASTE;
    procedure WMCut(var Message); message WM_CUT;
    procedure WMClear(var Message); message WM_CLEAR;
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CancelMode;
    //procedure SetPasswordChar(Value: Char);
    function GetChanged : boolean;
  protected
    //procedure CreateParams(var Params: TCreateParams); override;
    procedure SetText(Value : String); override;
    function DoTabKey : Boolean; override;
    procedure DoDropDownKeys(var Key: Word; Shift: TShiftState);
    procedure DblClick; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure BoundsChanged; virtual;
    //procedure WndProc(var Message: TMessage); override;
    //procedure CreateWnd; override;
     //virtual engine
    procedure GetEditData( var AData : String);
    procedure ValidateData(AData : String; var aCancel : boolean);
    procedure GetPickList(var aPickList : Tstrings);
    procedure GetReadOnly(var aReadOnly : boolean);

    procedure EditButtonClick(aButton : TPressedButtons);
    procedure HideEditor;
    procedure UpdateData;
    procedure UpdateEdit;//update location and data of editor
    procedure GetClickKey(var aKey : Word; var aShift : TShiftState);
    procedure KeyUpParent(var Key: Word; Shift: TShiftState);
    procedure KeydownParent(var Key: Word; Shift: TShiftState);
    procedure KeyPressParent(var Key: Char);
    procedure DblClickParent;


  public
    function OwnedHandle(CtlHandle : THandle) : Boolean;
    procedure RebuildEditor;
    procedure ButtonPressed(aButton : TPressedButtons); override;
    {$IFDEF DFS_DELPHI_4_UP}
    function UseRightToLeftAlignment: Boolean; override;
    {$ENDIF}
    procedure Hide(Scrolling : boolean);
    procedure UpdateContents; //virtual;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    //procedure Deselect;
    procedure Invalidate; reintroduce; // override;
    procedure Move(const Loc: TRect);
    function PosEqual(const Rect: TRect): Boolean;
    procedure SetFocus; override;
    procedure UpdateLoc(const Loc: TRect; Scrolling : Boolean);
    function Visible: Boolean;
    property WasChanged : boolean read GetChanged write FChanged;  // renamed Changed because of TControl name conflict
    property Validating : Boolean read FValidating write FValidating;
    //property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    property EditOptions : TEditOptions read FeditOptions write FeditOptions;
    property CancelData : boolean read FcancelData write FcancelData;//cancelling update data
    property ClickTime : Longint read FClickTime write FClickTime;
    property  Control: TWinControl read FControl write Fcontrol;
    property  OnGetEditData : TIEGetEditDataEvent read FOnGetEditData write FOnGetEditData;
    property  OnValidateData : TIEValidateDataEvent read FOnValidateData write FOnValidateData;
    property  OnGetPickList : TIEGetPickListEvent read FonGetPickList write FonGetPickList;
    property  OnGetReadOnly : TIEGetReadOnlyEvent read FonGetReadOnly write FonGetReadOnly;

    property  OnEditButtonClick : TIEEditButtonClickEvent read FonEditButtonClick write FonEditButtonClick;
    property  OnHideEditor : TIEHideEditorEvent read FonHideEditor write FonHideEditor;
    property  OnUpdateData : TIEUpdateDataEvent read FonUpdateData write FonUpdateData;
    property  OnGetClickKey : TIEGetClickKeyEvent read FonGetClickKey write FonGetClickKey;
    property  OnUpdateEdit : TIEUpdateEditEvent read FonUpdateEdit write FonUpdateEdit;
    property  OnKeyUpParent : TKeyEvent read FonKeyUpParent write FonKeyUpParent;
    property  OnKeydownParent : TKeyEvent read FonKeydownParent write FonKeydownParent;
    property  OnKeyPressParent : TKeyPressEvent read FonKeyPressParent write FonKeyPressParent;
    property  OnDblClickParent :  TNotifyEvent read FonDblClickParent write FonDblClickParent;
    //property EditMask;
    //property Font;//passwordchar
    //Property Passwordchar : Char read FPasswordChar write SetPasswordChar default #0;
    property Color;
    property Ctl3D;
  end;

implementation

constructor TISPlugInplaceEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  IsInplace := True;
  ParentCtl3D := false;
  ParentColor :=False;
  Ctl3D := False;
  TabStop := False;
  ParentFont :=true;
  FControl := TWincontrol(Owner);
  Parent:=TWincontrol(Owner);
  Color := clWindow;
  //ControlStyle := ControlStyle-[csFramed];
  {$IFDEF DFS_DELPHI_4_UP}
  DoubleBuffered := False;
  {$endif}
end;

destructor TISPlugInplaceEdit.Destroy;
begin
  // Parent:=Fcontrol;
   inherited;
end;

{procedure TISPlugInplaceEdit.SetPasswordChar(Value: Char);
begin
  if FPasswordChar <> Value then
  begin
    FPasswordChar := Value;
    RecreateWnd;
    if HandleAllocated then
    begin
      SendMessage(Handle, EM_SETPASSWORDCHAR, Ord(FPasswordChar), 0);
      SetTextBuf(PChar(Text));
    end;
  end;

end;


procedure TISPlugInplaceEdit.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  if PasswordChar <> #0 then
      Params.Style := Params.Style or ES_PASSWORD
  else Params.Style := Params.Style or ES_MULTILINE;

  Case Ord(FAlignment) of
    ord(taLeftJustify): Params.Style := Params.Style or ES_LEFT;
    ord(taCenter): Params.Style := Params.Style or ES_CENTER;
    ord(taRightJustify): Params.Style := Params.Style or ES_RIGHT;
  end;
  Params.Style := Params.Style and not WS_BORDER;
  if NewStyleControls and Ctl3D and (csFramed in ControlStyle) then
    begin
      Params.ExStyle := Params.ExStyle or WS_EX_CLIENTEDGE;
  end;
  //Params.Style := Params.Style or WS_BORDER;
end;

}

procedure TISPlugInplaceEdit.CMShowingChanged(var Message: TMessage);
begin
  { Ignore showing using the Visible property }
end;

function TISPlugInplaceEdit.GetChanged : boolean;
begin
  Result := FChanged or Modified;
end;

procedure TISPlugInplaceEdit.SetText(Value : String);
begin
  inherited;
  FChanged := True;
end;

function TISPlugInplaceEdit.DoTabKey : Boolean;
begin
  Result := eosGoTabs in FEditOptions;
end;

procedure TISPlugInplaceEdit.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  if eosGoTabs in FEditOptions then
    Message.Result := Message.Result or DLGC_WANTTAB;
end;

procedure TISPlugInplaceEdit.WMPaste(var Message);
begin
  if (eosReadOnly in FeditOptions) then Exit;
  inherited
end;

procedure TISPlugInplaceEdit.CMCancelMode(var Message: TCMCancelMode);
begin//WM_CancelMode
//new 10/3/2
{  if (Message.Sender <> Self) and (DropDownVisible)  then
  begin
    CloseUp(False);
  end;
}
end;

procedure TISPlugInplaceEdit.WMClear(var Message);
begin
  if (eosReadOnly in FeditOptions) then Exit;
  inherited;
end;

procedure TISPlugInplaceEdit.WMCut(var Message);
begin
  if (eosReadOnly in FeditOptions) then Exit;
  inherited;
end;

procedure TISPlugInplaceEdit.DblClick;
begin
  //FControl.OnDblClick(Self);
   DblClickParent;
end;
{
procedure TISPlugInplaceEdit.SetAlignment(Value: TAlignment);
begin
  if FAlignment <> Value then
  begin
    FAlignment := Value;
    RecreateWnd;
  end;
end;
}

procedure TISPlugInplaceEdit.DoDropDownKeys(var Key: Word; Shift: TShiftState);
//var ListValue : string;
begin
  case Key of
    VK_UP, VK_DOWN:
      if ssAlt in Shift then
      begin
        if DropdownVisible then
           CloseUp(True)
        else
           DropDown;
        Key := 0;
      end;

    VK_RETURN, VK_ESCAPE:
      if DropdownVisible and not (ssAlt in Shift) then
      begin
        CloseUp(Key = VK_RETURN);
        Key := 0;
      end;
  end;
end;

procedure TISPlugInplaceEdit.KeyDown(var Key: Word; Shift: TShiftState);

  procedure SendToParent;
  begin
     KeyDownParent(Key,Shift);
  end;

  procedure ParentEvent;
  begin
    KeyDownParent(Key,Shift);
  end;

  function ForwardMovement: Boolean;
  begin
    Result := eosAlwaysShowEditor in FEditOptions;
  end;

  function Ctrl: Boolean;
  begin
    Result := ssCtrl in Shift;
  end;

  {function Selection: TSelection;
  begin
    SendMessage(Handle, EM_GETSEL, Longint(@Result.StartPos), Longint(@Result.EndPos));
  end;

  function RightSide: Boolean;
  begin
    with Selection do
      Result := ((StartPos = 0) or (EndPos = StartPos)) and
        (EndPos = GetTextLen);
   end;

  function LeftSide: Boolean;
  begin
    with Selection do
      Result := (StartPos = 0) and ((EndPos = 0) or (EndPos = GetTextLen));
  end;
  }
begin
  //Parent := Fcontrol;
  case Key of
    VK_RETURN,VK_ESCAPE:
    begin
      if  (ButtonType = tbtDropDown) then
        DoDropdownKeys(Key, Shift);

        CancelData := (Key = VK_ESCAPE) and (Shift =[]);
        if (not (ButtonType in [tbtSpin,tbtScroll])) or ((Key in [VK_RETURN, VK_ESCAPE]) and (Shift =[])) then
          SendToParent;
        exit;
    end;
    VK_ADD, VK_SUBTRACT, VK_UP, VK_DOWN, VK_PRIOR, VK_NEXT:
      begin
        if  (ButtonType = tbtDropDown) then
          DoDropdownKeys(Key, Shift);
        //CancelData :=  (Shift =[]);
        if (not (ButtonType in [tbtSpin,tbtScroll])) or ((Key in [VK_RETURN, VK_ESCAPE]) and (Shift =[])) then
        begin
          SendToParent;
          exit;
        end;
      end;
    VK_INSERT:
      if Shift = [] then
      begin
       SendToParent;
      end;
      //else if (Shift = [ssShift]) and ReadOnly then Key := 0;
    VK_LEFT,
    VK_RIGHT,
    VK_HOME,
    VK_END: SendToParent;
    {VK_LEFT: if ForwardMovement and (Ctrl or LeftSide) then SendToParent;
    VK_RIGHT: if ForwardMovement and (Ctrl or RightSide) then SendToParent;
    VK_HOME: if ForwardMovement and (Ctrl or LeftSide) then SendToParent;
    VK_END: if ForwardMovement and (Ctrl or RightSide) then SendToParent;
    }
    VK_F2:
      begin
        ParentEvent;
        if Key = VK_F2 then
        begin
          //Deselect;
          Exit;
        end;
      end;
    VK_TAB: if not (ssAlt in Shift) then SendToParent;
  end;

  if (Key = VK_DELETE) and (eosReadOnly in FeditOptions) then
    Key := 0;

  if Key <> 0 then
    inherited;// KeyDown(Key, Shift);

end;

procedure TISPlugInplaceEdit.KeyPress(var Key: Char);
var
  Selection: TSelection;
begin
  //Parent := Fcontrol;
  if Key=#27 then FcancelData:=true;
  //to Parent
  KeyPressParent(Key);

  if (Key in [#32..#255]) and (eosReadOnly in FeditOptions) then
  begin
    Key := #0;
    MessageBeep(0);
  end;
  case Key of
    #9, #27: Key := #0;
    #13:
      begin
        SendMessage(Handle, EM_GETSEL, Longint(@Selection.StartPos), Longint(@Selection.EndPos));
        {if (Selection.StartPos = 0) and (Selection.EndPos = GetTextLen) then
          Deselect
        else
          SelectAll;
        }
        Key := #0;
      end;
    ^H, ^V, ^X, #32..#255:
      if (eosReadOnly in FeditOptions) then Key := #0;
  end;
  if Key <> #0 then inherited KeyPress(Key);
end;

procedure TISPlugInplaceEdit.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited;
  KeyUpParent(Key,Shift);
end;

{
procedure TISPlugInplaceEdit.Deselect;
begin
  SendMessage(Handle, EM_SETSEL, $7FFFFFFF, Longint($FFFFFFFF));
end;
}

procedure TISPlugInplaceEdit.Invalidate;
var
  Cur: TRect;
begin
  //Parent := Fcontrol;
  ValidateRect(Handle, nil);
  InvalidateRect(Handle, nil, True);
  Windows.GetClientRect(Handle, Cur);
  MapWindowPoints(Handle, FControl.Handle, Cur, 2);
  ValidateRect(FControl.Handle, @Cur);
  InvalidateRect(FControl.Handle, @Cur, False);
end;

{$IFDEF DFS_DELPHI_4_UP}
function TISPlugInplaceEdit.UseRightToLeftAlignment: Boolean;
begin
  if Parent <> nil then
   Result := Parent.UseRightToLeftAlignment
  else
     Result := inherited UseRightToLeftAlignment;
end;
{$ENDIF}

procedure TISPlugInplaceEdit.Hide(Scrolling : boolean);
begin
 if (Parent <> nil) and Parent.HandleAllocated then
 begin
  if HandleAllocated and IsWindowVisible(Handle) then
  begin
    Invalidate;
    SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_HIDEWINDOW or SWP_NOZORDER or SWP_NOREDRAW);
//    Sections.Clear;//c26.11
    ResetButtons;
  end;
  if Focused and not Scrolling then
    Windows.SetFocus(FControl.Handle);
 end;
end;

function TISPlugInplaceEdit.PosEqual(const Rect: TRect): Boolean;
var
  Cur: TRect;
begin
  GetWindowRect(Handle, Cur);
  MapWindowPoints(HWND_DESKTOP, Fcontrol.Handle, Cur, 2);
  Result := EqualRect(Rect, Cur);
end;

function TISPlugInplaceEdit.OwnedHandle(CtlHandle : THandle) : Boolean;
var i : integer;
begin
   Result := False;
   if Handle <> CtlHandle then
   begin
     for i:=0 to ControlCount-1 do
     begin
        if TWincontrol(Controls[0]).Handle = CtlHandle then
        begin
            Result := true;
            break;
        end;
     end;
   end
   else
      Result := True;
end;

procedure TISPlugInplaceEdit.InternalMove(const Loc: TRect; Redraw: Boolean; Scrolling : boolean);
begin
  if IsRectEmpty(Loc) then Hide(Scrolling)
  else
  begin
    CreateHandle;
    Redraw := Redraw or not IsWindowVisible(Handle);
    Invalidate;
    with Loc do
      SetWindowPos(Handle, HWND_TOP, Left, Top, Right - Left, Bottom - Top,
        SWP_SHOWWINDOW or SWP_NOREDRAW);
    BoundsChanged;
    if Redraw then Invalidate;
    if Fcontrol.Focused and not Scrolling then
    begin
      //Windows.SetFocus(Handle);
      if ControlCount >0 then
        Windows.SetFocus(TWincontrol(Controls[0]).Handle);
        //TWincontrol(Controls[0]).SetFocus;
    end;
  end;
end;

procedure TISPlugInplaceEdit.BoundsChanged;
//var
//  R: TRect;
begin
{  R := Rect(2, 2, Width - 2, Height);
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@R));
  SendMessage(Handle, EM_SCROLLCARET, 0, 0);
}
end;

procedure TISPlugInplaceEdit.UpdateLoc(const Loc: TRect; Scrolling : Boolean);
begin
  InternalMove(Loc, False, Scrolling);
end;

function TISPlugInplaceEdit.Visible: Boolean;
begin
  //Parent := Fcontrol;
  Result := IsWindowVisible(Handle);
end;

procedure TISPlugInplaceEdit.Move(const Loc: TRect);
begin
  InternalMove(Loc, True, false);
end;

procedure TISPlugInplaceEdit.SetFocus;
begin
  //Parent := Fcontrol;
  if IsWindowVisible(Handle) then
    Windows.SetFocus(Handle);
end;

{procedure TISPlugInplaceEdit.CreateWnd;
begin
  inherited CreateWnd;
end;
}
procedure TISPlugInplaceEdit.UpdateContents;
var aText : string;
//    apwdChar : Char;
begin
  aText := '';
  GetEditData(aText);//virtual function, Fcontrol must be connected to this events
//  if aText <> '' then
//  begin
   Text:=aText;
   Modified := False;
   if (FocusedEditor <> nil) then
    if (FocusedEditor.Handleallocated) then
      FocusedEditor.SelectAll;
//   Invalidate;
//  end;
end;

//virtual engine
procedure TISPlugInplaceEdit.GetEditData( var AData : String);
begin
   if Assigned(FonGetEditData) then FonGetEditData(Self,aData);
end;
procedure TISPlugInplaceEdit.ValidateData(AData : String; var aCancel : boolean);
begin
   if Assigned(FonValidateData) then FonValidateData(Self,AData,aCancel);
end;

{procedure TISPlugInplaceEdit.GetEditstyle(var aEditstyle : TEditStyle);
begin
   if Assigned(FonGetEditStyle) then FonGetEditStyle(Self,aEditstyle);
end;
}
procedure TISPlugInplaceEdit.GetPickList(var aPickList : Tstrings);
begin
   if Assigned(fonGetPickList) then fonGetPickList(Self,aPickList);
end;
procedure TISPlugInplaceEdit.GetReadOnly(var aReadOnly : boolean);
begin
   if Assigned(FonGetReadOnly) then FonGetReadOnly(Self,areadOnly);
end;

procedure TISPlugInplaceEdit.RebuildEditor;
begin
  //RebuildSections;
  //ReArrangeSections;
  RebuildSections(false);
end;

procedure TISPlugInplaceEdit.ButtonPressed(aButton : TPressedButtons);
begin
   if aButton = pbEllipsis then //,pbSectionButton]
     EditButtonClick(aButton)
   else
      inherited;
end;

procedure TISPlugInplaceEdit.EditButtonClick(aButton : TPressedButtons);
begin
   FShowed :=true;
   if Assigned(FonEditButtonClick) then FonEditButtonClick(Self);
   if aButton = pbEllipsis then
     UpdateData;
   FShowed :=false;


end;

procedure TISPlugInplaceEdit.HideEditor;
begin
  if Assigned(FonHideEditor) then
    FonHideEditor(Self);
end;

procedure TISPlugInplaceEdit.UpdateData;
begin
   if Assigned(FonUpdateData) then FonUpdateData(Self);
end;
procedure TISPlugInplaceEdit.GetClickKey(var aKey : Word; var aShift : TShiftState);
begin
   if Assigned(FonGetClickKey) then FonGetClickKey(Self,aKey,aShift);
end;

procedure TISPlugInplaceEdit.UpdateEdit;
begin
   if Assigned(FonUpdateEdit) then FonUpdateEdit(Self);
end;

procedure TISPlugInplaceEdit.KeyUpParent(var Key: Word; Shift: TShiftState);
begin
   if Assigned(FonKeyUpParent) then FonKeyUpParent(Self,Key,Shift);
end;
procedure TISPlugInplaceEdit.KeyDownParent(var Key: Word; Shift: TShiftState);
begin
   if Assigned(FonKeyDownParent) then FonKeyDownParent(Self,Key,Shift);
end;
procedure TISPlugInplaceEdit.KeyPressParent(var Key: Char);
begin
   if Assigned(FonKeyPressParent) then FonKeyPressParent(Self,Key);
end;
procedure TISPlugInplaceEdit.DblClickParent;
begin
   if Assigned(FonDblClickParent) then FonDblClickParent(Self);
end;


end.
