unit BatterySettings;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, 
  Buttons, ExtCtrls, exgrid, RapTree, flytree, PropTree, battery,
  PropertiesTreeCust, functions, NumInp12, ComCtrls, cconsts, scale, ctypes;

type
  TfrmBatterytype = class(TForm)
    OKBtn: TButton;
    CancelBtn: TButton;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    rGround: TNumInp;
    Label1: TLabel;
    rUp: TNumInp;
    Label2: TLabel;
    rSense: TNumInp;
    Label3: TLabel;
    rgBatteryType: TRadioGroup;
    propTree: TPropertiesTreeCust;
    IError: TNumInp;
    Label4: TLabel;
    UError: TNumInp;
    Label5: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure rgBatteryTypeClick(Sender: TObject);
    procedure OKBtnClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
  private
    { Private declarations }
    fBattery, fBatteryEdit : TBattery;
    fBatteryList : TBatteryList;
    fMessageFunct: TAddMessage;
    procedure PrintMessage(const mtype : TMessageType; const msg : String);
  public
    { Public declarations }
    property Battery : TBattery read fBattery;
    function ReadSettings(const filename : string) : boolean;
    function StoreSettings(const filename : string) : boolean;
    property MessageFunct : TAddMessage read fMessageFunct
      write fMessageFunct;      
  end;

var
  frmBatterytype: TfrmBatterytype;

implementation


{$R *.dfm}

procedure TfrmBatterytype.FormCreate(Sender: TObject);
var i : integer;
begin
  fBattery := nil;
  fBatteryList := TBatteryList.Create();

  for i := 0 to fBatteryList.Count-1 do
    rgBatteryType.Items.Add(fBatteryList.Batteries[i].BatteryType);

  rgBatteryType.ItemIndex := 0;

  MessageFunct := nil;

end;

procedure TfrmBatterytype.rgBatteryTypeClick(Sender: TObject);
var tbi : TBattery;
    tbt : TBatteryClass;
begin
  tbi := fBatteryList.Batteries[rgBatteryType.ItemIndex];
  tbt := tbi.BatteryClass;

  fBattery.Free;
  fBatteryEdit.Free;
  fBattery := tbt.Create(nil);
  fBatteryEdit := tbt.Create(nil);
  propTree.Deselect;
  propTree.SelectComponent(fBatteryEdit);
end;

procedure TfrmBatterytype.OKBtnClick(Sender: TObject);
var tbi : TBattery;
    tbt : TBatteryClass;
    rsd : double;
begin
try
  tbi := fBatteryList.Batteries[rgBatteryType.ItemIndex];
  tbt := tbi.BatteryClass;

  fBattery.Free;
  fBattery := tbt.Create(nil);

  fBattery.Assign(fBatteryEdit);

  VCHARGER_RVALS.Rground := rGround.Value;
  VCHARGER_RVALS.Rup := rUp.Value;
  VCHARGER_RVALS.Rsense := rSense.Value;

  VMEASURING_ERRORS.UError := UError.Value;
  VMEASURING_ERRORS.IError := IError.Value;

  rsd := (UREF_VALUE/CSD_ADC_CODE_MAX)*
    (1+VCHARGER_RVALS.Rup/VCHARGER_RVALS.Rground);

  Scales[TVoltage].offset := 0;
  Scales[TVoltage].mult := (rsd/GAIN_VOLTAGE)*
    (1+VMEASURING_ERRORS.UError/100);

  Scales[TCurrent].offset := 0;
  Scales[TCurrent].mult := rsd/(GAIN_CURRENT*VCHARGER_RVALS.Rsense)*
   (1+VMEASURING_ERRORS.IError/100);

  //calculate the voltage and current scales here 

  fBattery.CalcBatterySettings;
  PrintMessage(TMInfo, 'The battery settings were calculated well.');
except
  on E: Exception do PrintMessage(TMError, E.Message);
end;
end;

function TfrmBatterytype.ReadSettings(const filename: string) : boolean;
var F : TFileStream;
    Index : integer;
begin
try
  F := nil;
  result := true;
  try
    F := TFileStream.Create(FileName, fmOpenRead);
    F.Read(Index, sizeof(Index));
    rgBatteryType.ItemIndex := Range(Index, 0, rgBatteryType.Items.Count-1);
    F.ReadComponent(fBatteryEdit);
    propTree.UpdateColumns;
    F.Read(VCHARGER_RVALS, sizeof(VCHARGER_RVALS));
    F.Read(VMEASURING_ERRORS, sizeof(VMEASURING_ERRORS));
  finally
    F.Free;
  end;
except
  result := false;
end;
end;

function TfrmBatterytype.StoreSettings(const filename: string) : boolean;
var F : TFileStream;
    Index : integer;
begin
try
  result := true;
  F := nil;
  try
    F := TFileStream.Create(FileName, fmCreate);
    Index :=  rgBatteryType.ItemIndex;
    F.Write(Index, sizeof(Index));
    F.WriteComponent(fBatteryEdit);
    F.Write(VCHARGER_RVALS, sizeof(VCHARGER_RVALS));
    F.Write(VMEASURING_ERRORS, sizeof(VMEASURING_ERRORS));
  finally
    F.Free;
  end;
except
  result := false;
end

end;

procedure TfrmBatterytype.FormActivate(Sender: TObject);
begin
  rGround.Value := VCHARGER_RVALS.Rground;
  rUp.Value := VCHARGER_RVALS.Rup;
  rSense.Value := VCHARGER_RVALS.Rsense;
  IError.Value := VMEASURING_ERRORS.IError;
  UError.Value := VMEASURING_ERRORS.UError;
end;

procedure TfrmBatterytype.PrintMessage(const mtype: TMessageType;
  const msg: String);
begin
  if Assigned(MessageFunct) then
    MessageFunct(mtype, msg);
end;

end.
