unit MainForm;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, xyGraph, ExtCtrls, Menus, ToolWin, ComCtrls, MsgForm,
  ctypes, cconsts, Buttons, GraphForm, BatterySettings, CPort,  Battery,
  commun, ALed, XPMan, ImgList, poptions, Termistor, xyData, ExtDlgs,
  About, Scale;

type
  TfrmMain = class(TForm)
    StatusBar: TStatusBar;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Exit1: TMenuItem;
    Actions1: TMenuItem;
    Settings1: TMenuItem;
    Serialport1: TMenuItem;
    BatterySettings1: TMenuItem;
    ProgramOptions1: TMenuItem;
    Window1: TMenuItem;
    ile1: TMenuItem;
    Cascade1: TMenuItem;
    help1: TMenuItem;
    ToolBar: TToolBar;
    btnQuery: TToolButton;
    btnAbort: TToolButton;
    btnStart: TToolButton;
    ToolButton5: TToolButton;
    btnSaveSettings: TToolButton;
    btnOpenSettings: TToolButton;
    OpenBatteryDialog: TOpenDialog;
    SaveBatteryDialog: TSaveDialog;
    SaveBatterySettings1: TMenuItem;
    OpenBatterySettings1: TMenuItem;
    N1: TMenuItem;
    Querycharger1: TMenuItem;
    Startcharging1: TMenuItem;
    Abortcurrentoperation1: TMenuItem;
    N2: TMenuItem;
    ToolButton1: TToolButton;
    UploadTemperatureTable: TMenuItem;
    ermistorCalibration1: TMenuItem;
    N3: TMenuItem;
    UploadBatteryData: TMenuItem;
    N4: TMenuItem;
    OpenGraphData1: TMenuItem;
    SaveGraphData1: TMenuItem;
    OpenGraphDialog: TOpenDialog;
    SaveGraphDialog: TSaveDialog;
    ImageList: TImageList;
    OpenGraphsAppended1: TMenuItem;
    OpenPictureDialog1: TOpenPictureDialog;
    About1: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure Serialport1Click(Sender: TObject);
    procedure BatterySettings1Click(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure WindowTileClick(Sender: TObject);
    procedure WindowCascadeClick(Sender: TObject);
    procedure btnOpenSettingsClick(Sender: TObject);
    procedure btnSaveSettingsClick(Sender: TObject);
    procedure btnQueryClick(Sender: TObject);
    procedure btnStartClick(Sender: TObject);
    procedure btnAbortClick(Sender: TObject);
    procedure ProgramOptions1Click(Sender: TObject);
    procedure TermistorCalibrationClick(Sender: TObject);
    procedure UploadTemperatureTableClick(Sender: TObject);
    procedure OpenGraphData1Click(Sender: TObject);
    procedure SaveGraphData1Click(Sender: TObject);
    procedure SaveGraphDataAppended1Click(Sender: TObject);
    procedure OpenGraphsAppended1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure UploadBatteryDataClick(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
  private
    { Private declarations }
    communicator :  TCommunicator;
    init_switch : boolean;
    frmGraph_I, frmGraph_U, frmGraph_T : TfrmGraph;
    Battery : TBattery;
    procedure SaveGraphs(const FileName : string);
    procedure ReadGraphs(const FileName : string);
    procedure ReadGraphsAppended(const FileName : string);
    procedure ReadGraphAppended(from : TStream; appto : Txygraph);
  public
    { Public declarations }
    procedure EnableTermMenu(sender : TObject);
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.dfm}

procedure TfrmMain.FormCreate(Sender: TObject);
begin
 init_switch := true;

 OpenBatteryDialog.DefaultExt := 'bst';
 OpenBatteryDialog.Filter := 'Battery settings files (*.bst)|*.bst';
 OpenBatteryDialog.InitialDir := GetCurrentDir;

 SaveBatteryDialog.DefaultExt := 'bst';
 SaveBatteryDialog.Filter := 'Battery settings files (*.bst)|*.bst';
 SaveBatteryDialog.InitialDir := GetCurrentDir;

 OpenGraphDialog.DefaultExt := 'dat';
 OpenGraphDialog.Filter := 'Graph data (*.dat)|*.dat';
 OpenGraphDialog.InitialDir := GetCurrentDir;

 SaveGraphDialog.DefaultExt := 'dat';
 SaveGraphDialog.Filter := 'Graph data (*.dat)|*.dat';
 SaveGraphDialog.InitialDir := GetCurrentDir;

 Battery := nil;

 InitializeScales();
end;

procedure TfrmMain.Serialport1Click(Sender: TObject);
begin
 communicator.SetComport();
end;

procedure TfrmMain.BatterySettings1Click(Sender: TObject);
begin
  if (frmBatterytype.ShowModal() = mrOk) then
  begin
    Battery := frmBatterytype.Battery;
    UploadBatteryData.Enabled := true;

    if (VBATTERY_AUTO_UPLOAD) then
      UploadBatteryDataClick(self);
    
  end
end;

procedure TfrmMain.FormShow(Sender: TObject);
begin
   if (init_switch) then
   begin
     init_switch := false;

     communicator := TCommunicator.Create;
     frmMessages := TfrmMessages.Create(self);
     communicator.MessageFunct := frmMessages.AddMessage;

     frmGraph_I := TfrmGraph.Create(self);
     frmGraph_I.Caption := 'Battery current';
     frmGraph_I.graph.XAxis.Title := 't, s';
     frmGraph_I.graph.YAxis.Title := 'I, A';
     frmGraph_I.graph.Appearance.GraphTitle := '';
     frmGraph_I.GraphNames :=  TCurrent;

     frmGraph_U := TfrmGraph.Create(self);
     frmGraph_U.Caption := 'Battery voltage';
     frmGraph_U.graph.XAxis.Title := 't, s';
     frmGraph_U.graph.YAxis.Title := 'U, V';
     frmGraph_U.graph.Appearance.GraphTitle := '';
     frmGraph_I.GraphNames :=  TVoltage;

     frmGraph_T := TfrmGraph.Create(self);
     frmGraph_T.Caption := 'Battery temperature';
     frmGraph_T.graph.XAxis.Title := 't, s';
     frmGraph_T.graph.YAxis.Title := 'T, C';
     frmGraph_T.graph.Appearance.GraphTitle := '';
     frmGraph_I.GraphNames :=  TTemperature;

     communicator.AddPoint_I := frmGraph_I.AddPoint;
     communicator.AddPoint_U := frmGraph_U.AddPoint;
     communicator.AddPoint_T := frmGraph_T.AddPoint;

     communicator.OnUpdateListValue := frmMessages.SetListItem;

     communicator.ChargerWaitAnswer := VWAIT_RESPONSE_TIMEOUT;
     communicator.ChargerQueryPeriod := VWAIT_PERIOD_TIMEOUT;

     UploadTemperatureTable.Enabled := false;
     UploadBatteryData.Enabled := false;

     frmTermistor.PrintMessage := frmMessages.AddMessage;
     frmTermistor.OnMenuEnable := EnableTermMenu;

     frmProgramOptions.OnMessage := frmMessages.AddMessage;
     frmProgramOptions.FormActivate(nil);
     Tile;

     frmBatterytype.MessageFunct :=  frmMessages.AddMessage;

   end;
end;


procedure TfrmMain.WindowCascadeClick(Sender: TObject);
begin
  Cascade;
end;

procedure TfrmMain.btnOpenSettingsClick(Sender: TObject);
begin
  if OpenBatteryDialog.Execute then
    if frmBatterytype.ReadSettings(OpenBatteryDialog.FileName)
    then frmMessages.AddMessage(TMInfo, Format('Battery settings were restored from %s well',
                                [ExtractFileName(OpenBatteryDialog.FileName)]))
    else frmMessages.AddMessage(TMError, Format('Error reading file %s',
                                [ExtractFileName(OpenBatteryDialog.FileName)]));
end;

procedure TfrmMain.btnSaveSettingsClick(Sender: TObject);
begin
  if SaveBatteryDialog.Execute then
    if frmBatterytype.StoreSettings(SaveBatteryDialog.FileName)
    then frmMessages.AddMessage(TMInfo, Format('Battery settings were stored into %s well',
                                [ExtractFileName(SaveBatteryDialog.FileName)]))
    else frmMessages.AddMessage(TMError, Format('Error saving data file %s',
                                [ExtractFileName(SaveBatteryDialog.FileName)]));
end;

procedure TfrmMain.btnQueryClick(Sender: TObject);
begin
  if not communicator.SetState(SRtoDiag) then
  frmMessages.AddMessage(TMError,'Cannot query charger');
end;

procedure TfrmMain.btnStartClick(Sender: TObject);
begin
  if communicator.SetState(SRtoStart) then
  begin
    frmGraph_I.InitializeGraph;
    frmGraph_U.InitializeGraph;
    frmGraph_T.InitializeGraph;
  end
  else frmMessages.AddMessage(TMError,'Cannot start changing');
end;

procedure TfrmMain.btnAbortClick(Sender: TObject);
begin
  if not communicator.SetState(SRtoAbort) then
  frmMessages.AddMessage(TMError,'Cannot abort operation');
end;

procedure TfrmMain.ProgramOptions1Click(Sender: TObject);
begin
  if (frmProgramOptions.ShowModal = mrOk) then
  begin
    communicator.ChargerWaitAnswer := VWAIT_RESPONSE_TIMEOUT;
    communicator.ChargerQueryPeriod := VWAIT_PERIOD_TIMEOUT;
  end;
end;

procedure TfrmMain.TermistorCalibrationClick(Sender: TObject);
begin
  frmTermistor.Show;
  if (frmTermistor.CalibrationPresent and VBATTERY_AUTO_UPLOAD) then
    UploadTemperatureTableClick(self);
end;

procedure TfrmMain.UploadTemperatureTableClick(Sender: TObject);
var i, t : integer;
begin
  communicator.FlashStartBlock :=  CTERMS_START_BLOCK;
  communicator.FlashLenght := CTERM_BUFFER_LENGHT;

  for i := 0 to frmTermistor.CalibrationDataCount-1 do
  begin
    t := frmTermistor.CalibrationData[i];
    communicator.FlashBuffer[2*i] := (t shr 8) and $FF;
    communicator.FlashBuffer[2*i+1] := t and $FF;
  end;

  //calculate the Termistor table here

  if not communicator.SetState(SRtoFlashStart) then
    frmMessages.AddMessage(TMError,'Cannot start programming flash memory');
end;

procedure TfrmMain.EnableTermMenu(sender: TObject);
begin
  UploadTemperatureTable.Enabled := true;
end;

procedure TfrmMain.ReadGraphs(const FileName: string);
var f : TFileStream;
    g : TxyGraph;
begin
    f := Tfilestream.create(FileName,fmOpenRead);
  try
    g := frmGraph_I.graph;
    g.readfromStream(f);
    g := frmGraph_U.graph;
    g.readfromStream(f);
    g := frmGraph_T.graph;
    g.readfromStream(f);
  finally
    f.Free;
  end;

end;

procedure TfrmMain.SaveGraphs(const FileName: string);
var f : TFileStream;
    g : TxyGraph;
begin
    f := Tfilestream.create(FileName,fmCreate);
  try
    g := frmGraph_I.graph;
    g.savetostream(f, true);
    g := frmGraph_U.graph;
    g.savetostream(f, true);
    g := frmGraph_T.graph;
    g.savetostream(f, true);
  finally
    f.Free;
  end;

end;

procedure TfrmMain.OpenGraphData1Click(Sender: TObject);
begin
   if OpenGraphDialog.Execute then
     ReadGraphs(OpenGraphDialog.FileName);
end;

procedure TfrmMain.SaveGraphData1Click(Sender: TObject);
begin
  if SaveGraphDialog.Execute then
     SaveGraphs(SaveGraphDialog.FileName);
end;

procedure TfrmMain.SaveGraphDataAppended1Click(Sender: TObject);
begin
  if OpenGraphDialog.Execute then
     ReadGraphsAppended(OpenGraphDialog.FileName);
end;

procedure TfrmMain.ReadGraphsAppended(const FileName: string);
var f : TFileStream;
begin
    f := TFileStream.Create(FileName, fmOpenRead);
  try
    ReadGraphAppended(f, frmGraph_I.graph);
    ReadGraphAppended(f, frmGraph_U.graph);
    ReadGraphAppended(f, frmGraph_T.graph);
  finally
    f.Free;
  end;
end;

procedure TfrmMain.ReadGraphAppended(from: TStream; appto: Txygraph);
var g : TxyGraph;
    s_orig, s_app : TSeries;
    scount_orig, scount_app, i : integer;
    x, y, r : double;
    index : integer;
    b : boolean;
begin
  //append the graph data to the existing graphs
  g := TxyGraph.CreateParented(handle);

  try
    g.Plotting := false;
    appto.Plotting := false;

    g.readfromStream(from);

    scount_orig := 0;
    s_orig := g.serieshandle;
    while (s_orig <> nil) do
    begin
      inc(scount_orig);
      s_orig := s_orig.NextSeries;
    end;

    scount_app := 0;
    s_app := appto.serieshandle;
    while (s_app <> nil) do
    begin
      inc(scount_app);
      s_app := s_app.NextSeries;
    end;

    for i := 1 to scount_orig do
    begin
      s_app := appto[scount_app+i];
      s_orig := g[i];

      s_app.DrawPoints := s_orig.DrawPoints;
      s_app.DrawLine := s_orig.DrawLine;
      s_app.LineStyle := s_orig.LineStyle;

      b := s_orig.GetFirstPoint(x, y, r, index);
      while (b) do
      begin
        s_app.Add(x, y);
        b := s_orig.GetNextPoint(x, y, r, Index);
      end;

    end;

    appto.Plotting := true;
  finally
    g.Free
  end;

end;


procedure TfrmMain.OpenGraphsAppended1Click(Sender: TObject);
begin
  if OpenGraphDialog.Execute then
    ReadGraphsAppended(OpenGraphDialog.FileName);
end;


procedure TfrmMain.About1Click(Sender: TObject);
begin
  frmAbout.ShowModal;
end;

procedure TfrmMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  btnAbortClick(self);
end;

procedure TfrmMain.UploadBatteryDataClick(Sender: TObject);
var i : integer;
begin
  if Assigned(Battery) then
  begin
    communicator.FlashLenght := Battery.BatteryFlashLenght;
    communicator.FlashStartBlock := CBATTERY_FLASH_START_BLOCK;

    for i := 0 to  Battery.BatteryFlashLenght - 1 do
      communicator.FlashBuffer[i] := Battery.BatterySettingsData[i];
  end;

  if not communicator.SetState(SRtoFlashStart) then
    frmMessages.AddMessage(TMError,'Cannot start programming flash memory');
end;

procedure TfrmMain.WindowTileClick(Sender: TObject);
begin
  Tile;
end;

procedure TfrmMain.Exit1Click(Sender: TObject);
begin
  Close;
end;

end.
