unit Marquandt;

interface
uses Types, Constants, Models, Math, GlobVar, Linear, Opmetods, Dialogs, SysUtils;
const MarquardtMessage : array[0..5] of String = (
 '',
 '³   ',
 '    ',
 '   ',
 '     ',
 '     '
                                                  );
type
 TMarquardt = class(TOptObject)
private
 Alamda : Double;
 Np : Integer;
 Bt, Kt : double;
 InternalIteration : Integer;
 function OneIteration : Integer;
 function S(const V : double) : double;
protected
 function CalcFunct : Double; override;
 function CalcElementTerm (const Index : TPIndex) : double; override;
 function CalcElementHesse (const IndexX, IndexY : TPIndex) : double; override;
public
 StopCondition : Double;
 MaxInternalIteration : Integer;
 procedure Optimize; override;
 procedure SetOptimMethodParams; override;
                  end;

implementation

function TMarquardt.CalcElementHesse(const IndexX,
  IndexY: TPIndex): double;
var i,j : integer;
    ds, dx : double;
begin
 dx := 0;
 if LimitsCheck then
 for I := 1 to OptimDim do
 begin
  j := FixToNormal[i];
  dx := dx + S(Params[j]-MaxParams[j])*(Params[j]-MaxParams[j]) +
             S(MinParams[j]-Params[j])*(MinParams[j]-Params[j]);
 end;

{ if dx = 0 then ds := 0
 else
 ds := ChiOld*Kt*IntPower(dx, np-2)*exp(bt*dx)*
 ((np+bt*dx)*(np-1+bt*dx)+bt*dx*dx)*
 (S(Params[IndexX]-MaxParams[IndexX])+S(MinParams[IndexX]-Params[IndexX]))*
 (S(Params[IndexY]-MaxParams[IndexY])+S(MinParams[IndexY]-Params[IndexY]));
 }ds := 0;
 result := inherited CalcElementHesse(IndexX, IndexY) + ds;

end;

function TMarquardt.CalcElementTerm(const Index: TPIndex): double;
var i,j : integer;
    ds, dx : double;
begin
 dx := 0;
 if LimitsCheck then
 for I := 1 to OptimDim do
 begin
  j := FixToNormal[i];
  dx := dx + S(Params[j]-MaxParams[j])*(Params[j]-MaxParams[j]) +
             S(MinParams[j]-Params[j])*(MinParams[j]-Params[j]);
 end;

{ if dx = 0 then ds := 0
 else
  ds := ChiOld*Kt*IntPower(dx, np-1)*exp(bt*dx)*(np+bt*dx)*
  (S(Params[Index]-MaxParams[Index])+S(MinParams[Index]-Params[Index]));
} ds := 0;
 result := inherited CalcElementTerm(Index) - ds;
end;

function TMarquardt.CalcFunct: Double;
var i,j : integer;
    ds, dx : double;
begin
 dx := 0;
 if LimitsCheck then
 for I := 1 to OptimDim do
 begin
  j := FixToNormal[i];
  dx := dx + S(Params[j]-MaxParams[j])*(Params[j]-MaxParams[j]) +
             S(MinParams[j]-Params[j])*(MinParams[j]-Params[j]);
 end;

{ if dx = 0 then ds := 0
           else ds := ChiOld*Kt*IntPower(dx,np)*exp(bt*dx);
}ds := 0;
 result := inherited CalcFunct + ds;
end;

function TMarquardt.OneIteration : Integer;
const Delta : Double = 10.0;
var i, j : Integer;
begin
 OldParams := Params;
 ChiOld := Chi;
 InternalIteration := 0;
repeat
 if InternalIteration > MaxInternalIteration then
  raise EMethodError.Create(MarquardtMessage[4]);

 for i := 1 to OptimDim do
  FreeTerm[i] := CalcElementTerm(FixToNormal[i]);
 for i := 1 to OptimDim do
   for J := i to OptimDim do
    begin
     Hesse[i,j] := CalcElementHesse (FixToNormal[i], FixToNormal[j]);
     Hesse[j,i] := Hesse[i,j];
    end;
 for I := 1 to OptimDim do
   Hesse[i,i] := Hesse[i,i]*(1 + Alamda);

 symmatrix (OptimDim, Hesse, FreeTerm, FixDelta, result);

 if Result <> 0 then raise EMethodError.Create(MarquardtMessage[result]);

 for I := 1 to OptimDim do
  Params[FixToNormal[i]] := Params[FixToNormal[i]] + FixDelta[i];
 Chi := CalcFunct;
 if Chi < ChiOld then
  begin
   Inc (IterationCount);
   ALamda := ALamda/Delta;
   break;
  end
                  else
  begin
   Params := OldParams;
   ALamda :=  ALamda*Delta;
   Inc (InternalIteration);
   continue;
  end;
until false;
end;//OneIteration

procedure TMarquardt.Optimize;
var ResultOneIteration : Integer;
begin
 ChiOld := 0;
 Chi := CalcFunct;
 IterationCount := 0;
 Alamda := 0.001;
 MethodMessage := '';

 repeat
  ResultOneIteration := OneIteration;
  if ResultOneIteration <> 0 then break;
  if IterationCount > MaxIteration then
  raise EMethodError.Create(MarquardtMessage[5]);

  if Abs(ChiOld - Chi)/Chi < (StopCondition/100) then
    break;
 until Terminated;
end;

function TMarquardt.S(const V: double): double;
begin
 if V > 0 then result := 1
          else result := 0;
end;

procedure TMarquardt.SetOptimMethodParams;
begin
 inherited SetOptimMethodParams;
 with GenParams, MarqParams do
 begin
  MaxIterationsCount := oc_Marq_MaxIterationsCount;
  MaxInternalIteration := oc_Marq_MaxInternalIteration;
  StopCondition := oc_Marq_StopCondition;
  Kt := oc_Marq_K;
  Bt := oc_Marq_BETA;
  Np := oc_Marq_POWER;
  LimitsCheck := oc_LimitsCheck;
 end;
end;

initialization


end.
