unit MsgForm;

interface

uses
  Windows,Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ctypes, cconsts, commun, Menus;

type
  TfrmMessages = class(TForm)
    PageControl: TPageControl;
    TabMessages: TTabSheet;
    TabValues: TTabSheet;
    messages: TRichEdit;
    ListView: TListView;
    PopupMenu: TPopupMenu;
    ClearMessages1: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure ClearMessages1Click(Sender: TObject);
  private

    { Private declarations }

  public
    { Public declarations }
    procedure ClearMessages();
    procedure AddMessage(const mtype : TMessageType = TMError; const msg : String = '');
    procedure SetListItem(const Item: TListNames; const Value : double);
  end;

var
  frmMessages: TfrmMessages;

implementation

{$R *.dfm}

procedure TfrmMessages.AddMessage(const mtype: TMessageType = TMError;
  const msg: String = '');
begin
 messages.SelAttributes.Color := CMsgColor[mtype];
 messages.Lines.Add(CPREFICS[mtype] + msg); 
 Update;
end;

procedure TfrmMessages.ClearMessages;
var i : TListnames;
begin
  messages.Lines.Clear;

  for I := Low(LIST_ITEMS_NAMES) to High(LIST_ITEMS_NAMES) do
    ListView.Items[integer(i)].SubItems.Insert(0, CINIT_SYMBOL);

end;

procedure TfrmMessages.FormCreate(Sender: TObject);
var NewColumn : TListColumn;
    ListItem : TListItem;
    I : TListnames;
begin
  messages.ReadOnly := true;
  messages.Lines.Clear;

  with ListView do
  begin

    ViewStyle := vsReport;

    NewColumn := Columns.Add;
    NewColumn.Caption := 'Variable';
    NewColumn.Width := 100;
    NewColumn := Columns.Add;
    NewColumn.Caption := 'Value';

    for I := Low(LIST_ITEMS_NAMES) to High(LIST_ITEMS_NAMES) do
    begin
      ListItem := Items.Add;
      ListItem.Caption := LIST_ITEMS_NAMES[i];
      ListItem.SubItems.Add(CINIT_SYMBOL);
    end;
  end;

  PageControl.Pages[0].Show;
end;

procedure TfrmMessages.SetListItem(const Item: TListNames; const Value: double);
var s : string;
    vi, secs, mins, hours : integer;
    st : TChargerStateName;
    i : TErrorCode;
begin
  case Item of

    TTime:
    begin
      vi := round(value);
      hours := vi div 3600;
      vi := vi - hours * 3600;
      mins := vi div 60;
      secs := vi - mins*60;

      s := Format('%d:%d.%d',[hours, mins, secs]);
    end;

    TChState:
    begin
      st := TChargerStateName(round(value));
      if ( (st >= Low(TChargerStateName)) and (st <= High(TChargerStateName)))
      then s := CChargerStateName[st]
      else s := 'Unknown charger state!!!'
    end;

    TError:
    begin
      s := '';
      vi := round(value);
      for i := Low(TErrorCode) to High(TErrorCode) do
        if (vi = CERROR_CODES[i]) then s := CERROR_MESSAGES[i];

      if (s = '') then s := Format('Unknown error code, %xh', [vi]);   
    end;
    else
      s := Format('%3.4g', [Value]);
  end;

  ListView.Items[integer(Item)].SubItems.Insert(0, s);
end;

procedure TfrmMessages.ClearMessages1Click(Sender: TObject);
begin
  messages.Lines.Clear;
end;

end.
