unit Opmetods;

interface
uses Types, Constants, Models, GlobVar, Dialogs, Linear,
     SpshFunct, SysUtils, Math, Classes, Windows;

type TOptObject = class (TThread)
constructor Create(TerminateProcedure : TNotifyEvent;
                   UpdateProcedure : TNotifyString);
private
 fLimitsCheck : Boolean;
 fStartTime, fStopTime :  TTimeStamp;
 fMethodMessage : string;
 fProbability : Double;
 fChi2Method : boolean;
 fOnUpdateProcedure : TNotifyString;
 fCalcParamsUncertaines : boolean;
 function GetChi2 : double;
 function GetResultOptimization : boolean;
 function GetParameterUncertain (Index : TPIndex) : Double;
 procedure CalculateCovarMartix;
 procedure CalculateParamsUncert;
 procedure SetPropability (P : Double);
 procedure SetModel (const AModel : TCSubModel);
 function GetAverageError : Double;
 function GetProcessTime : Double;
 procedure ClearModelUncertaines;
 procedure UpdateMessage;
 procedure Optimization;
protected
 FNumberFunctCalc : Integer;
 FixToNormal : TConvertArray;
 Model : TCSubModel;
 Params, OldParams, MinParams, MaxParams, Delta, FreeTerm,
 TempParams, FixParams, FixDelta : TParamsArray;
 ParamsUncerains : TParamsArray;
 Hesse, Covar  : TMatrixArray;
 Chi, ChiOld : Double;
 OptimDim : TPIndex;
 TotalDim : TPIndex;
 IterationCount, MaxIteration : Integer;
 function GetError (const Index : Integer) : double;
 procedure Optimize; virtual; abstract;
 function ToFix (const AInput : TParamsArray) : TParamsArray;
 function UnFix (const AInput : TParamsArray) : TParamsArray;
 function CalcFunct : Double; virtual;
 function CalcElementTerm (const Index : TPIndex) : double; virtual;
 function CalcElementHesse (const IndexX, IndexY : TPIndex) : double; virtual;
 function CalcFunctFromVector (const Vector : TParamsArray) : double; virtual;

 procedure SetOptimParams; virtual;
 procedure GetOptimParams; virtual;

 procedure CheckAllValuesLimits;
 procedure CheckValuesLimitsInVector (var V : TParamsArray);

 procedure UpdateUnFixValues;

 procedure PreOptimization; virtual;
 procedure PostOptimization; virtual;

 procedure FixStartTime;
 procedure FixStopTime;
public
 property LimitsCheck : boolean read fLimitsCheck write fLimitsCheck;
 property MethodMessage : string read fMethodMessage write fMethodMessage;
 property Chi2 : Double read GetChi2;
 property OptimSucsess : boolean read GetResultOptimization;
 property NumberIterations : Integer read IterationCount;
 property MaxIterationsCount : Integer read MaxIteration write MaxIteration ;
 property RelParameteterUncertain [Index : TPindex] : Double read
          GetParameterUncertain;
 property Probability : Double read fProbability write SetPropability;
 property SubModel : TCSubModel read Model write SetModel;
 property AverageError : Double read GetAverageError;
 property OptimTime : Double read GetProcessTime;
 property Chi2Method : boolean read fChi2Method write fChi2Method;
 property OnUpdateProcedure : TNotifyString read fOnUpdateProcedure
                                            write fOnUpdateProcedure;
 property NumberFuncEvaluations : Integer read fNumberFunctCalc;
 procedure SetOptimMethodParams; virtual;
 procedure Execute; override;
 function CalcResult  : double;

                  end;
implementation

constructor TOptObject.Create (TerminateProcedure : TNotifyEvent;
                               UpdateProcedure : TNotifyString);
begin
 inherited Create(True);
 OnTerminate := TerminateProcedure;
 OnUpdateProcedure := UpdateProcedure;
 Priority := tpHigher;
 FreeOnTerminate := true;

 IterationCount := 0;
 fLimitsCheck := false;
 fMethodMessage := '';
 fProbability := 0.95;
 Model := nil;
 fChi2Method := false;
 fCalcParamsUncertaines := false;
 fNumberFunctCalc := 0;
end;

function TOptObject.GetResultOptimization : boolean;
begin
 result := Model.OptimSucsess;
end;

function TOptObject.CalcFunct : Double;
var sum : Extended;
    i, j : Integer;
begin
 inc(fNumberFunctCalc);
 sum := 0;
 with Model do
  for i := 0 to Lines - 1 do
   for j := DataLowIndex[i] to DataHighIndex [i] do
    sum := sum + W[i]*W[i]*Sqr((Y_C[j] - ftFunction(Params, X_C[j], Z_C[i])))/GetError(j);
 result := sum;
end;

function TOptObject.CalcElementTerm(const Index : TPIndex) : double;
var sum : Extended;
    i, j : Integer;
begin
inc(FNumberFunctCalc, 2);
 sum := 0;
 with Model do
  for i := 0 to Lines - 1 do
   for j := DataLowIndex[i] to DataHighIndex [i] do
   sum := sum + ftVectorFunction(Index, Params, X_C[j], Z_C[i])*W[i]*
       (Y_C[j] - ftFunction (Params, X_C[j], Z_C[i]))/(GetError(j));
 result := sum;
end;

function TOptObject.CalcElementHesse (const IndexX, IndexY : TPIndex) : double;
var sum : Extended;
    i, j : Integer;
begin
 inc(FNumberFunctCalc, 2);
 sum := 0;
 with Model do
  for i := 0 to Lines - 1 do
   for j := DataLowIndex[i] to DataHighIndex [i] do
   sum := sum + (ftVectorFunction (IndexX, Params, X_C[j], Z_C[i])
               * ftVectorFunction(IndexY, Params, X_C[j], Z_C[i])
          )/(GetError(j));
 result := sum;
end;

function TOptObject.CalcFunctFromVector (const Vector : TParamsArray) : double;
var sum : Extended;
    i, j : Integer;
begin
//Attention!!! Vector - Fix Array;
 inc(fNumberFunctCalc);
 sum := 0;

for I := 1 to OptimDim do
  Params[FixToNormal[i]] := Vector[i];

 with Model do
  for i := 0 to Lines - 1 do
   for j := DataLowIndex[i] to DataHighIndex [i] do
    sum := sum + W[i]*W[i]*Sqr((Y_C[j] - ftFunction(Params, X_C[j], Z_C[i])))/GetError(j);
 result := sum;
end;

procedure TOptObject.SetOptimParams;
var I : TPIndex;
begin
 fNumberFunctCalc := 0;

 OptimDim := 0;
 for I := 1 to TotalDim do
  begin
   if MP[i-1+Model.Offset].State = stOptimize then
    begin
     Inc (OptimDim);
     FixToNormal[OptimDim] := I;
    end;
   Params[i] := MP[i-1+Model.Offset].StartValue;
   MinParams[i] := MP[i-1+Model.Offset].MinStartValue;
   MaxParams[i] := MP[i-1+Model.Offset].MaxStartValue;
  end;
end;

procedure TOptObject.GetOptimParams;
var I : TPIndex;
begin
 for I := 1 to TotalDim do
  MP[i-1+Model.OffSet].OV := Params[i];
end;

procedure TOptObject.CheckAllValuesLimits;
var i : TPIndex;
begin
if fLimitsCheck then
 for I := 1 to TotalDim do
  begin
   if Params[i] > MaxParams[i] then Params[i] := MaxParams[i];
   if Params[i] < MinParams[i] then Params[i] := MinParams[i];
  end;
end;

procedure TOptObject.CheckValuesLimitsInVector (var V : TParamsArray);
var i,J : TPIndex;
begin
//Attention!!! V - Fix Array;
if fLimitsCheck then
 for I := 1 to OptimDim do
  begin
   J := FixToNormal[i];
   if V[i] > MaxParams[j] then V[i] := MaxParams[j];
   if V[i] < MinParams[j] then V[i] := MinParams[j];
  end;
end;

procedure TOptObject.UpdateUnFixValues;
var i : TPIndex;
begin
 for I := 1 to OptimDim do FixParams[i] := FixParams[i] + FixDelta[i];
end;

function TOptObject.ToFix (const AInput : TParamsArray) : TParamsArray;
var I : TPIndex;
begin
 for I := 1 to OptimDim do result[i] := AInput[FixToNormal[i]];
end;

function TOptObject.UnFix (const AInput : TParamsArray) : TParamsArray;
var I : TPindex;
begin
 for I := 1 to OptimDim do result[FixToNormal[i]] := AInput[i];
end;

procedure TOptObject.PreOptimization;
begin
 with Model do
  begin
  Model.OptimSucsess := false;

   if not DataSet then
    raise EProcessError.Create(
     '³  ,  !');

   if not CheckDataSucsess then
    raise EProcessError.Create(
     '       !');

   if not PreprocessSucsess then
    raise EProcessError.Create(
     '    ,  !');

  SetOptimParams;
  end;//with
end;

procedure TOptObject.PostOptimization;
begin
with Model do
begin
 try
   GetOptimParams;
   if not PostprocessSucsess then raise EProcessError.Create(
     '     !');

  ClearModelUncertaines;
  if fCalcParamsUncertaines then
  begin
   CalculateCovarMartix;
   CalculateParamsUncert;
  end;
 except
  on E:EMathError do
  begin
   raise EProcessError.Create(
    '       !');
     Model.OptimSucsess := false;
  end
   else raise;
 end;//except;
 OptimSucsess := true;
end;//with
end;

procedure TOptObject.CalculateCovarMartix;
var I, J : Integer;
    InverseResult : Integer;
begin

if OptimDim = 1 then
 begin
  Covar[1,1] := 1/CalcElementHesse(FixToNormal[1], FixToNormal[1]);
  Exit;
 end;

  for i := 1 to OptimDim do
   for J := i to OptimDim do
    begin
     Hesse[i,j] := CalcElementHesse (FixToNormal[i], FixToNormal[j]);
     Hesse[j,i] := Hesse[i,j];
    end;
invrealmatrix (OptimDim, Hesse, Covar, InverseResult);
if InverseResult <> 0 then
 begin
  if GenParams.CheckCovarMatrix then
  raise EFunctionError.Create('   !');
  for i := 1 to OptimDim do
   for J := i to OptimDim do
    begin
     Covar[i,j] := 0;
     Covar[j,i] := 0;
    end;
 end

end;

procedure TOptObject.CalculateParamsUncert;
var I : Integer;
    M : Double;
    T : ItemParam;
    L : Double;
begin
if fChi2Method then L := 1
               else L := CalcFunct/(Model.TotalPoints-OptimDim);

 for I := 1 to TotalDim do ParamsUncerains[i] := 0;

 M := 100*CalcStudentDistrib(Model.TotalPoints, fProbability);

 for I := 1 to OptimDim do
 ParamsUncerains[i] := sqrt(L*Covar[i,i])*M/Max(1,abs(Params[i]));


 for I := 1 to OptimDim do
 begin
  T := Model.ParamItem[FixToNormal[i]-1];
  T.ErrorVal := ParamsUncerains[i];
  Model.ParamItem[FixToNormal[i]-1] := T;
 end;
end;

function TOptObject.GetParameterUncertain (Index : TPIndex) : Double;
begin
 result := ParamsUncerains[FixToNormal[Index]];
end;

procedure TOptObject.SetPropability (P : Double) ;
begin
 fProbability := P;
 if (Model <> nil) and  (Model.InstanceSize <> 0) then
 CalculateParamsUncert;
end;

procedure TOptObject.SetModel (const AModel : TCSubModel);
begin
 Model := AModel;
 TotalDim := Model.NumberParams;
 OptimDim := TotalDim;
 Model.OptimSucsess := false;
end;

function TOptObject.GetAverageError : Double;
var sum, sum1 : Double;
    i, j : Integer;
begin
 sum := 0;
 sum1 := 0;
 with Model do
  for i := 0 to Lines - 1 do
   for j := DataLowIndex[i] to DataHighIndex [i] do
    begin
     sum := sum + abs(Y_C[j]);
     sum1 := sum1 + Sqr(Y_C[j] - ftGrafic(X_C[j], Z_C[i]));
    end;

 result := 100*sqrt(sum1){*(Model.TotalPoints-1)}/sum;
end;

procedure TOptObject.FixStartTime;
begin
 fStartTime := DateTimeToTimeStamp(Time);
end;

procedure TOptObject.FixStopTime;
begin
 fStopTime := DateTimeToTimeStamp(Time);
end;

function TOptObject.GetProcessTime : Double;
begin
 result := (fStopTime.Time - fStartTime.Time)/1000;
end;

procedure TOptObject.ClearModelUncertaines;
var I : Integer;
    T : ItemParam;
begin
 for I := 0 to Model.NumberParams - 1 do
 begin
  T := Model.ParamItem[i];
  T.ErrorVal := 0;
  Model.ParamItem[i] := T;
 end;
end;

function TOptObject.GetError (const Index : Integer) : double;
begin
 if fChi2Method then result := Model.Y_E2[Index]
                else result := 1;
end;

function TOptObject.GetChi2 : double;
begin
 if fChi2Method then result := CalcFunct
                else result := CalcFunct/(Model.TotalPoints - OptimDim);
end;

procedure TOptObject.Optimization;
begin
 try
  FixStartTime;
  Optimize;
 finally
  FixStopTime;
 end; 
end;

procedure TOptObject.SetOptimMethodParams;
begin
with GenParams do
 begin
  Probability := oc_Probability;
  Chi2Method := oc_Chi2Method;
  fCalcParamsUncertaines := CalcParamsUncertaines;
 end;
end;

procedure TOptObject.Execute;
begin
try
 try
  PreOptimization;
  MethodMessage := '    ...';
  Synchronize(UpdateMessage);
  MethodMessage := '';
  Optimization;

  PostOptimization;

 if MethodMessage = '' then
  MethodMessage := '    !';

 except
 on EZeroDivide do
   MethodMessage := 'ĳ  0   !';
 on EOverflow do
   MethodMessage := '   !';
 on EUnderflow do
   MethodMessage := ' !';
 on EInvalidOp do
   MethodMessage :='     !';
 on EInvalidArgument do
   MethodMessage := '  !';
 on E:EMethodError do
   MethodMessage := E.Message;
 on E:EFunctionError do
   MethodMessage := E.Message;
 else
   MethodMessage := '     ';
 
 end;
finally Synchronize(UpdateMessage);
end;
end;

procedure TOptObject.UpdateMessage;
begin
 if Assigned(OnUpdateProcedure) then OnUpdateProcedure(self, MethodMessage);
end;

function TOptObject.CalcResult : double;
var i, j : integer;
    sum : double;
begin
 sum := 0;
 with Model do
  for i := 0 to Lines - 1 do
   for j := DataLowIndex[i] to DataHighIndex [i] do
     sum := sum + W[i]*W[i]*Sqr(Y_C[j] - ftGrafic(X_C[j], Z_C[i]));

 result := sum;
end;

end.


