// classes to convert the charger values in the graph and vica verse

unit Scale;

interface

uses math, classes, ctypes, cconsts;

type TScale = class
  private
    fforward_mult, fforward_offset, fback_mult, fback_offset : double;
    procedure SetMult(const Value: double);
    procedure SetOffset(const Value: double);
  public
    constructor Create(const mult, offset : double); overload;
    constructor Create(); overload;
    function scaled(const value : double): double;
    function scalei(const value : double): integer;
    function reversescaled(const value : double): double;
    function reversescalei(const value : double): integer;
    property Mult : double read fforward_mult write SetMult;
    property Offset : double read fforward_offset write SetOffset;
  end;

var Scales : array [TListNames] of TScale;

procedure InitializeScales();
implementation

{ TScale }

constructor TScale.Create;
begin
  fforward_mult := 1;
  fback_mult := 1;
  fforward_offset := 0;
  fback_offset := 0;
end;

constructor TScale.Create(const mult, offset: double);
begin
  fforward_mult := mult;
  fforward_offset := offset;

  fback_mult := 1/fforward_mult;
  fback_offset := -fforward_offset/fforward_mult;
end;

function TScale.reversescalei(const value: double): integer;
begin
  result := round (fback_mult*value + fback_offset);
end;

function TScale.reversescaled(const value: double): double;
begin
  result := fback_mult*value + fback_offset;
end;

function TScale.scalei(const value: double): integer;
begin
  result := round(fforward_mult*value + fforward_offset);
end;

function TScale.scaled(const value: double): double;
begin
  result := fforward_mult*value + fforward_offset;
end;

procedure TScale.SetMult(const Value: double);
begin
  fforward_mult := Value;

  fback_mult := 1/fforward_mult;
  fback_offset := -fforward_offset/fforward_mult;
end;

procedure TScale.SetOffset(const Value: double);
begin
  fforward_offset := value;

  fback_mult := 1/fforward_mult;
  fback_offset := -fforward_offset/fforward_mult;
end;

procedure InitializeScales();
var rsd : double;
begin

  Scales[TVoltage] := TScale.Create();
  Scales[TCurrent] := TScale.Create();

  rsd := (UREF_VALUE/CSD_ADC_CODE_MAX)*
    (1+VCHARGER_RVALS.Rup/VCHARGER_RVALS.Rground);

  Scales[TVoltage].offset := 0;
  Scales[TVoltage].mult := (rsd/GAIN_VOLTAGE)*
    (1+VMEASURING_ERRORS.UError/100);

  Scales[TCurrent].offset := 0;
  Scales[TCurrent].mult := rsd/(GAIN_CURRENT*VCHARGER_RVALS.Rsense)*
   (1+VMEASURING_ERRORS.IError/100);

  Scales[TTime] := TScale.Create(CHARGER_TIMER_PERIOD, 0);

  Scales[TTemperature] :=
    TScale.Create((CMAX_TEMPERATURE_CHARGER-CMIN_TEMPERATURE_CHARGER)/(CN_TABCHARGER_POINTS-1), CMIN_TEMPERATURE_CHARGER);

  Scales[TChState] := nil;
  Scales[TError] := nil;
end;


end.
