unit cconsts;

interface

uses Graphics, ctypes, windows;

const CMSGCOLOR : array[TMessageType] of TColor = (clBlack, clBlue, clRed);
const CPREFICS : array[TMessageType] of string = ('[Info] ', '[Warning] ', '[Error] ');

const CBATCAPMIN : double = 0.001; //battery capacitance
const CBATCAPMAX : double = 10;

const TMAXMIN : double = 10;//battery maximum temperature
const TMAXMAX : double = 100;

const TMINMIN : double = 0; //battery start-up temperature
const TMINMAX : double = 60;

const TIMEOUTMIN : integer = 1; //timeout
const TIMEOUTMAX : integer = 6000;

const CCRMIN = 0.005;
const CCRMAX = 5; //charge ratio

const NCELLSMIN = 1;
const NCELLSMAX = 10;

const REG_COM_PORT : string =
  'HKEY_LOCAL_MACHINE\Software\Kremin\BatteryCharger\ComPort';

const REG_PRG_SETTINGS : string =
  'Software\Kremin\BatteryCharger\ProgramOptions';

const ROOT_KEY : HKEY  = HKEY_LOCAL_MACHINE;

const CCOMSETTINGS : string = 'Please set serial port settings';

const CCOM_START : string = 'S';
const CCOM_STOP  : string = 'P';

const CWAIT_FOR_SUSPEND : integer = 500;
const TIMEOUT_QUANTUM : integer = 100;

const TYPE_BYTE_START = 0;
const TYPE_BYTE_STOP  = 0;

const CODE_BYTE_START = 1;
const CODE_BYTE_STOP  = 1;

const TIME_START_BYTE = 2;
const TIME_STOP_BYTE = 3;

const U_START_BYTE = 4;
const U_STOP_BYTE = 5;

const I_START_BYTE = 6;
const I_STOP_BYTE = 7;

const T_START_BYTE = 8;
const T_STOP_BYTE = 8;

const GDATA_BYTE_START = 2;
const GDATA_BYTE_STOP = 2;

const ERROR_BYTE_START = 9;
const ERROR_BYTE_STOP = 9;

//NOERRORS, INVALID_VOLTAGE, TIMEOUT_EXPIRED, FLASH_WRITE);
const CERROR_CODES : array[TErrorCode] of byte = ($00,$A0,$50,$FF);
const CERROR_MESSAGES : array[TErrorCode] of string =
('No Errors',
 'Invalid battery voltage',
 'Operation Timeout Expired',
 'Flash Write Error');

const CHARGER_INVALID_COMMAND = 0;

const CHARGER_ERROR_CONTROL = 99;
const CHARGER_ERROR_BATTERY = 100;
const CHARGER_CONFDIAG = 101;
const CHARGER_CONFSTOP = 102;
const CHARGER_CONFSTART = 103;
const CHARGER_ADD_CURRDATA = 104;
const CHARGER_CHARGING_COMPLETE = 105;
const CHARGER_DISCHARGING_COMPLETE = 106;
const CHARGER_DISCHARGING_STOP = 107;

const CHARGER_FLASH_START = 110;
const CHARGER_FLASH_WRITING = 111;

const CMAX_PACKET_SIZE : integer = 100;

const CWAIT_WAKEUP : integer = 1000;

const CFLASH_SIZE_BLOCK : integer = 64;

const SetStateMsg : array [TComState] of string = (
 '',
 'AA',
 'CC',
 'BB',
 'DD',
 'EE',
 'FF'
);


{TComState = (
SReady,
SRtoDiag,
SRtoStart,
SRtoAbort,
SCharging,
SRtoStartCal,
SRToWriteBlock);
}

const CNEXT_STATE : array [TComState] of TComState = (
 SReady,
 SReady,
 SCharging,
 SReady,
 SCharging,
 SRToWriteBlock,
 SRToWriteBlock
 );

const CWANT_TIMEOUT : array [TComState] of boolean = (
 false,
 false,
 false,
 false,
 true,
 false,
 false
);

//TListNames = (TTime, TVoltage, TCurrent, TTemperature, TReserved1, TReserved2);
const LIST_ITEMS_NAMES : array[TListNames] of string =
('Time, h:m.s', 'Voltage, V', 'Current, A', 'Temperature, C', 'State', 'Error');

const CINIT_SYMBOL : string = '-';


const CN_APROX_PARAMS : integer = 3;

const CDELTA_TEMP : double = 273.15;

const CN_TABCHARGER_POINTS = 256;
const CMIN_TEMPERATURE_CHARGER : double = 0;
const CMAX_TEMPERATURE_CHARGER : double = 63.75;
const CMAX_CHARGER_OPERATION_TEMPERATURE : integer = 80;

const CTERMS_START_BLOCK = 194;
const CTERM_BUFFER_LENGHT : integer = 2*CN_TABCHARGER_POINTS;

const CN_TAB_GRAPH_POINTS : integer = 256;
const CMIN_TEMPERATURE_GRAPH : integer = -40;
const CMAX_TEMPERATURE_GRAPH : integer = 125;

const UREF_VALUE : double = 1.307;
const CSD_ADC_CODE_MIN : double = 0;
const CSD_ADC_CODE_MAX : double = 1023;
const CSD_SCALE_VALUE : double = 8192;

const GAIN_CURRENT : double = 16.0;
const GAIN_VOLTAGE : double = 1;

const CTIMER_BYTE_MIN : integer = 0;
const CTIMER_BYTE_MAX : integer = 255;

const CTIMER_WORD_MIN : integer = 0;
const CTIMER_WORD_MAX : integer = 32767;

const CWRITE_FLASH_OK    = $00;
const CWRITE_FLASH_ERROR = $FF;
const CWRITE_FLASH_END   = $55;

const CBATTERY_FLASH_LEN : integer = 64;
const CBATTERY_FLASH_START_BLOCK : integer = 192;

const  CPRODUCT_NAME : string =
         'NiCd, NiMH, Li-ion, Li-Polymer and SLA'#10#13'universal charger';
       CCHARGER_VERSION : string = 'Version: 0.7 beta';
       CCOPYRIGHT : string = 'Author: Victor Kremin';
       CRELEASE_DATE : string = 'Release date: Jan, 05, 2002';
       CCONTACT_INFO : string = 'For suggestions and bugs report please contact '#10#13'vkremin@polynet.lviv.ua';


const CChargerStateName : array [TChargerStateName] of string =
( 'STARTING',
  'STOPPING',
  'STOPPED',
  'WAITING',
  'TRICLE',
  'RAPID',
  'ACTIVATE',
  'DISCHARGE'
 );

const CSTOPPED_VALUE : integer = integer(STOPPED);

var VWAIT_RESPONSE_TIMEOUT : integer = 1000;
var VWAIT_PERIOD_TIMEOUT : integer = 500;
var VBATTERY_AUTO_UPLOAD : boolean = false;

var VCHARGER_RVALS : TRStruct = (
  Rterms : 1;
  Rground : 10;
  Rup : 30;
  Rsense : 0.25;
);

var VMEASURING_ERRORS : TErrorStruct =(
  UError : 0;
  IError : 0;
);


const CHARGER_TIMER_PERIOD : double = 0.25;

const TIMER1_INTERVAL : double = $0004;
const TIMER2_INTERVAL : double = $0100;
const TIMER3_INTERVAL : double = $0400;

const SECS_PER_MINUTE : double = 60.0;
const TEMP_MEASURING_POINTS_PER_MINUTE = 16.0;

const LIION_CHARGE_TERM_VOLTAGE : double = 4.3;
const LIION_CKECKINGS_DEACTIVE_TIME : byte = 2;

const NICD_TRICLE_CHARGE_VOLTAGE : double = 2.0;
const NICD_BATTERY_MAX_VOLTAGE : double = 1.9;
const NICD_RAPID_STOP_VOLTAGE : double = 1.8;

const SLA_CHARGE_TERM_VOLTAGE : double = 3.0;
const SLA_CKECKINGS_DEACTIVE_TIME : byte = 2;


implementation
end.

