
//non-linear approximation by Levenberg-Marguandtr method

unit lm;

interface

uses math, classes, sysutils;

type VectorDouble = array of double;
     VectorInt    = array of integer;
     MatrixDouble = array of VectorDouble;


type TEqFunct = procedure (const x: double; const a: VectorDouble; var y: double;
       var dyda: VectorDouble; const na: integer);

procedure funcs(const x: double; const a: VectorDouble; var y: double;
       var dyda: VectorDouble; const na: integer);

function MinimizeFunct(const x, y, sig : VectorDouble;
                       const DataPoints : integer;
                       const Eqfunct : TEqFunct;
                       const ma : integer;
                       const mfit : integer;
                       const Fix_vals : VectorInt;
                       const InitPoint : VectorDouble;
                       var   Solution : VectorDouble;
                       var   sqerror  : double
                       ) : boolean;

implementation

procedure funcs(const x: double; const a: VectorDouble; var y: double;
       var dyda: VectorDouble; const na: integer);
var t : double;
    yt : double;
    dy : double;
begin
  t := ln(x);
  yt := a[1] + a[2]*t + a[3]*t*sqr(t);
  y := 1/yt - 273.15;

  dy := -1/(yt*yt);
  dyda[1] := dy;
  dyda[2] := t*dy;
  dyda[3] := t*sqr(t)*dy;
end;


procedure covsrt(var covar: MatrixDouble; ma: integer;
       lista: VectorInt; mfit: integer);

var
   j,i: integer;
   swap: double;
begin
   for j := 1 to ma-1 do begin
      for i := j+1 to ma do begin
         covar[i,j] := 0.0
      end
   end;
   for i := 1 to mfit-1 do begin
      for j := i+1 to mfit do begin
         IF (lista[j] > lista[i])  then begin
            covar[lista[j],lista[i]] := covar[i,j]
         end else begin
            covar[lista[i],lista[j]] := covar[i,j]
         end
      end
   end;
   swap := covar[1,1];
   for j := 1 to ma do begin
      covar[1,j] := covar[j,j];
      covar[j,j] := 0.0
   end;
   covar[lista[1],lista[1]] := swap;
   for j := 2 to mfit do begin
      covar[lista[j],lista[j]] := covar[1,j]
   end;
   for j := 2 to ma do begin
      for i := 1 to j-1 do begin
         covar[i,j] := covar[j,i]
      end
   end
end;


procedure mrqcof(x,y,sig: VectorDouble; ndata: integer;
       var a: VectorDouble; mma: integer; lista: VectorInt;  const funct : TEqFunct;
       mfit: integer; var alpha: MatrixDouble;
       var beta: VectorDouble;  var chisq: double);

var
   k,j,i: integer;
   ymod,wt,sig2i,dy: double;
   dyda: VectorDouble;
begin
try
  SetLength(dyda, mma+1);

   for j := 1 to mfit do begin
      for k := 1 to j do begin
         alpha[j,k] := 0.0
      end;
      beta[j] := 0.0
   end;
   chisq := 0.0;
   for i := 1 to ndata do begin
      funct(x[i],a,ymod,dyda,mma);
      sig2i := 1.0/(sig[i]*sig[i]);
      dy := y[i]-ymod;
      for j := 1 to mfit do begin
         wt := dyda[lista[j]]*sig2i;
         for k := 1 to j do begin
            alpha[j,k] := alpha[j,k]+wt*dyda[lista[k]]
         end;
         beta[j] := beta[j]+dy*wt
      end;
      chisq := chisq+dy*dy*sig2i
   end;
   for j := 2 to mfit do begin
      for k := 1 to j-1 do begin
         alpha[k,j] := alpha[j,k]
      end
   end
finally
  SetLength(dyda, mfit+1);
end;
end;

procedure symmatrix (n      : Integer;
                     var a  : MatrixDouble;
                     const b: VectorDouble;
                     var x  : VectorDouble;
                     var st : Integer);
{---------------------------------------------------------------------------}
{                                                                           }
{  The procedure symmatrix solves a system of linear equations with a       }
{  symmetric matrix.                                                        }
{  Data:                                                                    }
{    n - number of equations = number of unknowns,                          }
{    a - two-dimensional array containing elements of the matrix of the     }
{        system (the elements a[i,j] for i>=j are changed on exit),         }
{    b - one-dimensional array containing free terms of the system.         }
{  Result:                                                                  }
{    x - an array containing the solution.                                  }
{  Other parameters:                                                        }
{    st - a variable which within the procedure symmatrix is assigned the   }
{         value of:                                                         }
{           1, if n<1,                                                      }
{           2, if the matrix of the system is not symmetric,                }
{           3, if at least one determinant of the form                      }
{                  | a[1,1] a[1,2] ... a[1,i] |                             }
{                  | a[2,1] a[2,2] ... a[2,i] |                             }
{                  | ........................ |                             }
{                  | a[i,1] a[i,2] ... a[i,i] |                             }
{              (i=1,2,...,n) is equal to zero,                              }
{           0, otherwise.                                                   }
{         Note: If st<>0, then the elements of array x are not calculated.  }
{  Unlocal identifiers:                                                     }
{    vector - a type identifier of extended array [q1..qn], where q1<=1 and }
{             qn>=n,                                                        }
{    matrix - a type identifier of extended array [q1..qn,q1..qn], where    }
{             q1<=1 and qn>=n.                                              }
{                                                                           }
{---------------------------------------------------------------------------}
const CheckMin = 1e-5;
var i,j,k : Integer;
    u,v,z : Extended;
begin
  st:=0;
  if n<1
    then st:=1;
  {for i:=2 to n do
    for j:=1 to i-1 do
      if abs(a[i,j]-a[j,i]) > CheckMin
        then st:=2;}
  if st=0
    then begin
           i:=0;
           repeat
             i:=i+1;
             for j:=1 to i do
               begin
                 z:=a[j,i];
                 if i=j
                   then begin
                          for k:=j-1 downto 1 do
                            begin
                              u:=a[i,k];
                              v:=u*a[k,k];
                              a[i,k]:=v;
                              z:=z-u*v
                            end;
                          if z=0
                            then st:=3
                            else a[i,i]:=1/z
                        end
                   else begin
                          for k:=j-1 downto 1 do
                            z:=z-a[i,k]*a[j,k];
                          a[i,j]:=z
                        end
               end
           until (i=n) or (st=3);
           if st=0
             then begin
                    for i:=1 to n do
                      begin
                        z:=b[i];
                        for j:=i-1 downto 1 do
                          z:=z-a[i,j]*x[j];
                        x[i]:=z
                      end;
                    for i:=n downto 1 do
                      begin
                        z:=x[i]*a[i,i];
                        for j:=i+1 to n do
                          z:=z-a[j,i]*x[j];
                        x[i]:=z
                      end
                  end
         end
end;

procedure gaussj(var a: MatrixDouble; n,np: integer;
       var b: MatrixDouble; m,mp: integer);
(* Programs using GAUSSJ must define the types
TYPE
   VectorDouble = ARRAY [1..np,1..np] OF double;
   MatrixDouble = ARRAY [1..np,1..mp] OF double;
   glnp = ARRAY [1..np] OF integer;
in the main routine. *)
var
   big,dum,pivinv: double;
   i,icol,irow,j,k,l,ll: integer;
   indxc,indxr,ipiv: VectorInt;

begin
try
   SetLength(indxc, np+1);
   SetLength(indxr, np+1);
   SetLength(ipiv, np+1);

   icol := 0;
   irow := 0;

   for j := 1 to n do begin
      ipiv[j] := 0
   end;
   for i := 1 to n do begin
      big := 0.0;
      for j := 1 to n do begin
         IF (ipiv[j] <> 1) then begin
            for k := 1 to n do begin
               IF (ipiv[k] = 0) then begin
                  IF (abs(a[j,k]) >= big) then begin
                     big := abs(a[j,k]);
                     irow := j;
                     icol := k
                  end
               end else IF (ipiv[k] > 1) then begin
                  raise EMathError.Create('pause 1 in GAUSSJ - singular matrix');
               end
            end
         end
      end;
      ipiv[icol] := ipiv[icol]+1;
      IF (irow <> icol) then begin
         for l := 1 to n do begin
            dum := a[irow,l];
            a[irow,l] := a[icol,l];
            a[icol,l] := dum
         end;
         for l := 1 to m do begin
            dum := b[irow,l];
            b[irow,l] := b[icol,l];
            b[icol,l] := dum
         end
      end;
      indxr[i] := irow;
      indxc[i] := icol;
      IF (a[icol,icol] = 0.0) then begin
         raise EMathError.Create('pause 2 in GAUSSJ - singular matrix');
      end;
      pivinv := 1.0/a[icol,icol];
      a[icol,icol] := 1.0;
      for l := 1 to n do begin
         a[icol,l] := a[icol,l]*pivinv
      end;
      for l := 1 to m do begin
         b[icol,l] := b[icol,l]*pivinv
      end;
      for ll := 1 to n do begin
         IF (ll <> icol) then begin
            dum := a[ll,icol];
            a[ll,icol] := 0.0;
            for l := 1 to n do begin
               a[ll,l] := a[ll,l]-a[icol,l]*dum
            end;
            for l := 1 to m do begin
               b[ll,l] := b[ll,l]-b[icol,l]*dum
            end
         end
      end
   end;
   for l := n doWNTO 1 do begin
      IF (indxr[l] <> indxc[l]) then begin
         for k := 1 to n do begin
            dum := a[k,indxr[l]];
            a[k,indxr[l]] := a[k,indxc[l]];
            a[k,indxc[l]] := dum
         end
      end
   end;
finally
   indxc := nil;
   indxr := nil;
   ipiv := nil;
end;
end;


procedure mrqmin(x,y,sig: VectorDouble; ndata: integer;
       var a: VectorDouble; mma: integer; lista: VectorInt; const funct : TEqFunct;
       mfit: integer; var covar,alpha: MatrixDouble;
       var chisq,alamda: double;
       var glbeta: VectorDouble; var glochisq : double);

var
   k,kk,j,ihit, st: integer;
   atry,da: VectorDouble;
begin
try
  SetLength(atry, mma+1);
  SetLength(da, mma+1);

   IF (alamda < 0.0) then begin
      kk := mfit+1;
      for j := 1 to mma do begin
         ihit := 0;
         for k := 1 to mfit do begin
            IF (lista[k] = j) then ihit := ihit+1
         end;
         IF (ihit = 0) then begin
            lista[kk] := j;
            kk := kk+1
         end else IF (ihit > 1) then begin
            raise EMathError.Create('pause 1 in routine MRQMIN');
         end
      end;
      IF (kk <> (mma+1)) then begin
         raise EMathError.Create('pause 2 in routine MRQMIN');
      end;
      alamda := 0.001;
      mrqcof(x,y,sig,ndata,a,mma,lista, funct, mfit,alpha,glbeta,chisq);
      glochisq := chisq;
      for j := 1 to mma do begin
         atry[j] := a[j]
      end
   end;
   for j := 1 to mfit do begin
      for k := 1 to mfit do begin
         covar[j,k] := alpha[j,k]
      end;
      covar[j,j] := alpha[j,j]*(1.0+alamda);
   end;

     symmatrix(mfit, covar, glbeta, da, st);
     if (st <> 0) then
       raise EMathError.Create('cannot solve linear equestion set');

   IF (alamda = 0.0) then begin
      covsrt(covar,mma,lista,mfit);
      exit;
   end;
   for j := 1 to mfit do begin
      atry[lista[j]] := a[lista[j]]+da[j]
   end;
   mrqcof(x,y,sig,ndata,atry,mma,lista,funct, mfit,covar,da,chisq);
   IF (chisq < glochisq) then begin
      alamda := 0.1*alamda;
      glochisq := chisq;
      for j := 1 to mfit do begin
         for k := 1 to mfit do begin
            alpha[j,k] := covar[j,k]
         end;
         glbeta[j] := da[j];
         a[lista[j]] := atry[lista[j]]
      end
   end else begin
      alamda := 10.0*alamda;
      chisq := glochisq
   end;

 finally
   atry := nil;
   da := nil;
 end
end;

function MinimizeFunct(const x, y, sig : VectorDouble;
                       const DataPoints : integer;
                       const Eqfunct : TEqFunct;
                       const ma : integer;
                       const mfit : integer;
                       const Fix_vals : VectorInt;
                       const InitPoint : VectorDouble;
                       var   Solution : VectorDouble;
                       var   sqerror  : double
                       ) : boolean;
var lista : VectorInt;
    alamda, chisq, chiold, glochisq : double;
    a, glbeta : VectorDouble;
    covar, alpha : MatrixDouble;
    iter_count, i : integer;

const CMAX_ITERS : integer = 1000;
      CMIN_CHI : double = 1e-10;
      CPROC_TERM : double = 1e-8;
begin
 try
   result := false;

   SetLength(lista, ma+1);
   SetLength(covar, ma+1, ma+1);
   SetLength(alpha, ma+1, ma+1);
   SetLength(glbeta, ma+1);
   SetLength(a, ma+1);

   for i := 1 to ma do a[i] := InitPoint[i];

   for i := 1 to ma do
     if (fix_vals[i] <> 0) then lista[i] := i
                           else lista[i] := 0;

   alamda := -1;
   mrqmin(x,y,sig,datapoints,
         a, ma,lista, EqFunct, mfit, covar,alpha,
         chisq, alamda,
         glbeta, glochisq);

   chiold := chisq;
   iter_count := 0;

   while (true) do
   begin
       mrqmin(x,y,sig,datapoints,
         a, ma,lista, EqFunct, mfit,covar,alpha,
         chisq, alamda,
         glbeta, glochisq);

       if (abs(chiold - chisq)/max(chisq, CMIN_CHI) <  CPROC_TERM)
       then begin
         result := true;

         alamda := 0;
         mrqmin(x,y,sig,datapoints,
                a, ma, lista, EqFunct, mfit,covar,alpha,
                chisq, alamda, glbeta, glochisq);

         for i := 1 to ma do Solution[i] := a[i];
         break;
       end;

       if (iter_count > CMAX_ITERS) then
         raise EMathError.Create('Not convergence into LM method');
       inc(iter_count);
       chiold := chisq;
   end;

  sqerror :=  chisq;
  finally
    lista := nil;
    covar := nil;
    alpha := nil;
    glbeta := nil;
    a := nil;
  end;
end;

end.
