unit NonLinear;
interface
uses Classes, SysUtils, Math;

type ANFunction = function (const X : double; const P : array of double) : double;

function SecantSolve       (const Min, Max : double;
                            const Eps : double;
                            F : ANFunction;
                            const P : array of double) : double;

function HalfDividingSolve (const Min, Max : double;
                            const Eps : double;
                            F : ANFunction;
                            const P : array of double) : double;

function ZBrentSolve       (const Min, Max : double;
                            const Eps : double;
                            F : ANFunction;
                            const P : array of double) : double;


implementation

function HalfDividingSolve (const Min, Max : double;
                            const Eps : double;
                            F : ANFunction;
                            const P : array of double) : double;
const itmax = 100;
var iter : Integer;
    A, B, C : Double;
begin
 if F(Min, P) * F(Max, P) > 0 then
 raise EMathError.Create(' "  ,       !');
 iter := 0;
 A := Min;
 B := Max;
 repeat
  C := (A + B)/2;
  if F(C,P)*F(A,P) > 0 then A := C
                       else B := C;
  inc (iter);
  if iter > itmax then raise
 EMathError.Create(' "  ,      !');
 until Abs(F(C,P)) < Eps;
result := C;
end;

function    SecantSolve    (const Min, Max : double;
                            const Eps : double;
                            F : ANFunction;
                            const P : array of double) : double;
const itmax = 200;
var   A, B, C : double;
      iter : Integer;
      fn, fnp, fx : double;
begin
A := Min;
B := Max;
fn := F(A,P);
fnp := F(B,P);
if fn*fnp >= 0 then
 raise EMathError.Create('The non-linear equestion cannot be solved');
iter := 0;

repeat
C := A - fn*(B-A)/(fnp-fn);
fx := F(C,P);
if fx*fn > 0 then
 begin
  A := C;
  fn := fx;
 end
              else
 begin
  B := C;
  fnp := fx;
 end;
inc (iter);
  if iter > itmax then raise
 EMathError.Create('The non-linear equestion cannot be solved');
until Abs(fx) < eps;

result := C;
end;


function ZBrentSolve       (const Min, Max : double;
                            const Eps : double;
                            F : ANFunction;
                            const P : array of double) : double;
const itmax=100;
      tol=3.0e-8;
VAR
   a,b,c,d,e: double;
   min1,min2,mine: double;
   fa,fb,fc,ps,q,r: double;
   s,tol1,xm: double;
   iter: integer;
BEGIN
   a := Min;
   b := Max;
   fa := f(a,P);
   fb := f(b,P);
   d := 0;
   e := 0;
   c := 0;

   IF (fb*fa > 0.0) THEN
    raise EMathError.Create('The non-linear equestion cannot be solved!');
   fc := fb;
   FOR iter := 1 TO itmax DO BEGIN
      IF (fb*fc > 0.0) THEN BEGIN
         c := a;
         fc := fa;
         d := b-a;
         e := d
      END;
      IF (abs(fc) < abs(fb)) THEN BEGIN
         a := b;
         b := c;
         c := a;
         fa := fb;
         fb := fc;
         fc := fa
      END;
      tol1 := 2.0*tol*abs(b)+0.5*eps;
      xm := 0.5*(c-b);
      IF ((abs(xm) <= tol1) OR (fb = 0.0)) THEN
      BEGIN
       result := b;
       exit;
      END;
      IF ((abs(e) >= tol1) AND (abs(fa) > abs(fb))) THEN BEGIN
         s := fb/fa;
         IF (a = c)  THEN BEGIN
            ps := 2.0*xm*s;
            q := 1.0-s
         END ELSE BEGIN
            q := fa/fc;
            r := fb/fc;
            ps := s*(2.0*xm*q*(q-r)-(b-a)*(r-1.0));
            q := (q-1.0)*(r-1.0)*(s-1.0)
         END;
         IF (ps > 0.0) THEN  q := -q;
         ps := abs(ps);
         min1 := 3.0*xm*q-abs(tol1*q);
         min2 := abs(e*q);
         IF (min1 < min2) THEN mine := min1 ELSE mine := min2;
         IF (2.0*ps < mine) THEN BEGIN
            e := d;
            d := ps/q;
         END ELSE BEGIN
            d := xm;
            e := d
         END
      END ELSE BEGIN
         d := xm;
         e := d
      END;
      a := b;
      fa := fb;
      IF (abs(d) > tol1) THEN BEGIN
         b := b+d
      END ELSE BEGIN
         IF (xm > 0) THEN BEGIN
            b := b+abs(tol1)
         END ELSE BEGIN
            b := b-abs(tol1)
         END
      END;
      fb := f(b,P)
   END;
 result := b;
end;


end.

