unit poptions;

interface

uses Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, 
  Buttons, ExtCtrls, NumInp12, ctypes, cconsts, registry;

type
  TfrmProgramOptions = class(TForm)
    OKBtn: TButton;
    CancelBtn: TButton;
    GroupBox1: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    nimWaitForResponse: TNumInp;
    niDataperiod: TNumInp;
    cbAutoUpload: TCheckBox;
    procedure FormActivate(Sender: TObject);
    procedure OKBtnClick(Sender: TObject);
  private
    { Private declarations }
    fAddMessage: TAddMessage;
  procedure PrintMessage(const mestype : TMessageType; const str : string);
  public
    { Public declarations }
    property OnMessage : TAddMessage read fAddMessage write fAddMessage;
  end;

var
  frmProgramOptions: TfrmProgramOptions;

implementation

const SWAIT_RESPONSE_TIMEOUT : string = 'WaitResponse';
const SWAIT_QUERY_PERIOD : string = 'WaitQueryPeriod';
const SAUTO_BATTERY_UPLOAD : string = 'BatteryAutoUpload';

{$R *.dfm}

procedure TfrmProgramOptions.FormActivate(Sender: TObject);
var res_timeout_tmp,  period_timeout_tmp : integer;
    auto_upload_tmp : boolean;
    reg_data : TRegistry;
begin
  reg_data := TRegistry.Create();
  try
    try
      reg_data.RootKey := ROOT_KEY;
      if (reg_data.OpenKey(REG_PRG_SETTINGS, false)) then
      begin
        res_timeout_tmp := reg_data.ReadInteger(SWAIT_RESPONSE_TIMEOUT);
        period_timeout_tmp := reg_data.ReadInteger(SWAIT_QUERY_PERIOD);
        auto_upload_tmp := reg_data.ReadBool(SAUTO_BATTERY_UPLOAD);
      end
        else raise ERegistryException.Create('');

      VWAIT_RESPONSE_TIMEOUT := res_timeout_tmp;
      VWAIT_PERIOD_TIMEOUT := period_timeout_tmp;
      VBATTERY_AUTO_UPLOAD := auto_upload_tmp;
    finally
      reg_data.Free;
    end;
  except
    PrintMessage(TMWarning,
      'Cannot restore program options from registry, use defauls.');
  end;

  nimWaitForResponse.Value := VWAIT_RESPONSE_TIMEOUT;
  niDataperiod.Value := VWAIT_PERIOD_TIMEOUT;
  cbAutoUpload.Checked := VBATTERY_AUTO_UPLOAD;
end;

procedure TfrmProgramOptions.PrintMessage(const mestype: TMessageType;
  const str: string);
begin
    if Assigned(OnMessage) then OnMessage(mestype, str);
end;

procedure TfrmProgramOptions.OKBtnClick(Sender: TObject);
var reg_data : TRegistry;
begin
  VWAIT_RESPONSE_TIMEOUT := Round(nimWaitForResponse.Value);
  VWAIT_PERIOD_TIMEOUT := Round(niDataperiod.Value);
  VBATTERY_AUTO_UPLOAD := cbAutoUpload.Checked;

  reg_data := TRegistry.Create();
  try
    try
      reg_data.RootKey := ROOT_KEY;
      if (reg_data.OpenKey(REG_PRG_SETTINGS, true)) then
      begin
        reg_data.WriteInteger(SWAIT_RESPONSE_TIMEOUT, VWAIT_RESPONSE_TIMEOUT);
        reg_data.WriteInteger(SWAIT_QUERY_PERIOD, VWAIT_PERIOD_TIMEOUT);
        reg_data.WriteBool(SAUTO_BATTERY_UPLOAD, VBATTERY_AUTO_UPLOAD);
        reg_data.CloseKey();
      end
        else
          raise ERegistryException.Create('');

    finally
      reg_data.Free;
    end;

  except
    PrintMessage(TMError, 'The program options cannot be stored in the registry');
  end;
end;

end.
