unit IsPlugEdit;
{/*****************************************************************/
   Copyright (c) 2001 IMCA SYSTEMS
   All Rights Reserved.
   Creation Date: 2. 02. 2001
   Version 2.0
   Get more 45 editors in ONE!
   ISPlugEditor - editor with editable sections of different types
   Description/features:
     1. 45+ types of predefened types of editors - such as unique
        DropDownTree, IPAddress, Color
     2. Four types of input - manual, spinbutton, dropdown, ellipsis, scroll
     3  Like DropdownList style in ComboBox types
     4. Autocomplete and AutoDropdown features
     5. Custom organize sections of editor
     6. Multicolumnar dropdowns!!!
     7  Alignment and custom widths in dropdowns
     8. Dropdowns can be sized- DropDownsized feature
     9  Full RightToLeft support
     10. Shift editors on get Focus (allow little size of editor)
     11 Unlimited number of data sections
     12. Custom colors, borders, separators for each section of data
     13. Easy work - simply set needed type of Editor in EditorType
     14. More professional - work with Sections of Editor
     15. Custom dropdowns - any Delphi control as Dropdown!

   History:
   23 Mar 2001 - First Release 1.0
   2. Apr 2001 - 1.05
     Some corrections:
         1. Addition of new section at design/Run time -> change EditorType on tetCustom
         2. Improved/Corrected UpdateSelNode
   28.Apr 2001 - 1.3
      Some improvements
      Added Text property R/W ability
    14.11.01 - 2.01
      Corrected work of dropdown window
      Corrected work of spin, scroll buttons
      Added more complete Currency support - added currency symbol and it position
   ----------------------------------------------------------------
    Format:
    EditorType
    ButtonType
    Items
    Sections
    DropDownRows
    Images
    Colcount
    ValueCol
    ItemSeparator -  dropdowntree
}

//     -
//  -0.99 -     -  -0
//       ,  ..

{$I DFS.inc}
{$R *.res}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls,
  {$ifdef DFS_COMPILER_4_UP} ImgList, {$endif} ExGrid, RapTree, TreeColl;

const
  InitRepeat = 400;
  RepeatPause     = 100;  { pause (ms)}
  CM_CLOSEUP = WM_USER+1;//CN_BASE - 1;
  EmptySeparator = '|';
type
  TEditorType = (tetCustom, tetSimple, tetMemo, tetYesNo, tetTrueFalse, tetCheckBox,
  tetInteger, tetCurrency, tetPercent,
  tetImage, tetImageCaption, tetImageMemo,
  tetColor,
  tetDropDownCheckList, tetDropDownList, tetDropDown, tetDropdownTree,
  tetIPAddress,
  tetGeneralDate, tetLongDate, tetMediumDate, tetShortDate, tetLongTime, tetMediumTime, tetShortTime,
  tetD, tetDD, tetDDD, tetDDDD,
  tetM, tetMM, tetMMM, tetMMMM,
  tetYY, tetYYYY, tetH, tetHH, tetN, tetNN, tetS, tetSS, tetPassword,
  tetProgress, tetValueProgress, tetPercentProgress, tetHyperLink, tetButton);

  //Format
  TIsDataType = (tdtInteger, tdtString, tdtMultiline, tdtD, tdtDD, tdtDDD, tdtDDDD,
    tdtM, tdtMM, tdtMMM, tdtMMMM,  tdtYY, tdtYYYY, tdt24H, tdt24HH, tdt12H, tdt12HH, tdtAP, tdtAMPM, tdtN, tdtNN, tdtS, tdtSS,
    tdtROList{read Only List - Dropdown}, tdtColor, tdtCheck, tdtImage, tdtImageCaption, tdtImageMemo,
    tdtProgress, tdtValueProgress, tdtPercentProgress, tdtHyperLink, tdtButton,tdtLeadingZeroesInteger);

  TISDataTypes = set of TIsDataType;
  TISPlugSections = class;
  TISCustomPlugEdit = class;

  PColInfo = ^TColInfo;
  TColInfo = Record
     Width : Integer;
     Alignment : TAlignment;
  end;

  TISPlugSection = class;
  {$ifdef DFS_COMPILER_3_UP}
   ISectionAdapter = interface
    ['{01779B33-E3D1-11D3-86A2-00E04CAAC9BF}']
    procedure ReferenceSection(Section: TISPlugSection);
    procedure ReleaseSection;
   end;
   {$ENDIF}

  TISPlugSection = class(TCollectionItem)
   private
      {$ifdef DFS_COMPILER_3_UP}
      FSectionAdapter          : ISectionAdapter;
      {$ENDIF}
      //FNotChangeMinMax : Boolean;
      FMaxLength, FWidth : Integer;
      FItemSeparator : String;
      FPasswordChar : Char;
      FColInfos : TList;
      FNodes : TFlyNodes;
      FSelected : TFlyNode;
      FLSeparator, FRSeparator : String;
      FType : TIsDataType;
      FAlignment : TAlignment;
      FundoInternalIndex, FinternalIndex, FDropDownColCount, FDropDownStructureCol : integer;
      FundoInternalValue, FInternalValue : String;
      FMinValue, FMaxValue : integer;
      FPressed, FReadOnly, FShowBorder : Boolean;
      FProgressColor, FBorderColor, FColor, FFontcolor : TColor;
      {$ifdef DFS_COMPILER_3_UP}
      procedure SetSectionAdapter(const aValue: ISectionAdapter);
      {$ENDIF}

      procedure ItemsChanged(Item : TTreeCollection);
      procedure UpdateSelNode;
      function FullNodeText(aNode : TFlyNode) : String;
      function GetRealMaxValue : integer;

      procedure SetFontcolor(Value : TColor);
      procedure SetColor(Value : TColor);
      procedure SetReadOnly(Value : Boolean);
      procedure SetMinValue(Value : integer);
      procedure SetMaxValue(Value : integer);
      procedure SetMaxLength(const Value : integer);
      procedure SetType(const Value : TIsDataType);
      procedure SetAlignment(const Value : TAlignment);
      procedure SetLSeparator(const Value : String);
      procedure SetRSeparator(const Value : String);
      procedure SetShowBorder(Value : Boolean);
      function GetItems : TFlyNodes;
      procedure SetItems(Value : TFlyNodes);
      function IsRSeparatorStored : Boolean;
      function IsLSeparatorStored : Boolean;
      function ISValueStored : Boolean;
      function ISMaxLengthStored  : Boolean;
      procedure SetInternalValue(aValue : String);
      procedure SetInternalIndex(const aValue : Integer);
      procedure SetBorderColor(Value : TColor);
      procedure SetItemSeparator(const Value : String);

      procedure SetDropDownStructureCol(const Value : integer);
      procedure SetDropDownColCount(const Value : integer);
      //procedure SetColAlignments(Alignments : TStringlist);
      //procedure SetColWidths(Widths : TStringlist);
      function GetColWidth(Index : integer) : integer;
      procedure SetColWidth(Index, Value : integer);
      function GetColAlignment(Index : integer) : TAlignment;
      procedure SetColAlignment(Index : integer; Value : TAlignment);

      procedure UndoValue;
      procedure SetUndoValue;
      procedure SetWidth(Value : integer);
      procedure SetPasswordChar(Value : Char);
      procedure SetProgressColor(Value : TColor);
   protected
      procedure Validate;
      procedure WriteColAligns(Writer: TWriter);
      procedure WriteColWidths(Writer: TWriter);
      procedure ReadColWidths(Reader: TReader);
      procedure ReadColAligns(Reader: TReader);
      procedure ReadData(Stream: TStream);
      procedure WriteData(Stream: TStream);
      procedure DefineProperties(Filer: TFiler); override;
   public
      constructor Create(Collection: TCollection); override;
      destructor Destroy; override;
      procedure Assign(Source: TPersistent); override;
      {$ifdef DFS_COMPILER_3_UP}
      property SectionAdapter : ISectionAdapter read FSectionAdapter write SetSectionAdapter;
      {$ENDIF}
      property DropDownColwidth[index : integer] : integer read GetColWidth write SetColWidth;
      property DropDownColAlignment[index : integer] : TAlignment read GetColAlignment write SetColAlignment;
   published
      property DataType : TIsDataType read FType write SetType default tdtInteger;
      property DropDownColCount : integer read FDropDownColCount write SetDropDownColCount default 1;
      property DropDownStructureCol : integer read FDropDownStructureCol write SetDropDownStructureCol default 0;
      property ItemSeparator : String read FItemSeparator write SetItemSeparator;
      property BorderColor : TColor read FBorderColor write SetBorderColor default clBlack;
      property ProgressColor : TColor read FProgressColor write SetProgressColor default clHighlight;
      property Color : TColor read FColor Write SetColor default clWindow;
      property FontColor : TColor read FFontcolor write SetFontColor default clBlack;
      //new
      property PasswordChar : Char read FPasswordChar write SetPasswordChar default #0;
      property ReadOnly : Boolean read FreadOnly write SetReadOnly default False;
      property LeftSeparator : String read FLSeparator write SetLSeparator stored IsLSeparatorStored;
      property RightSeparator : String read FRSeparator write SetRSeparator stored IsRSeparatorStored;
      property ShowBorder : Boolean  read FShowborder write SetShowBorder default false;
      property MinValue : integer read FMinValue write SetMinValue stored ISValueStored;
      property MaxValue : integer read FMaxValue write SetMaxValue stored ISValueStored;
      property MaxLength : integer read FMaxLength write SetMaxLength stored ISMaxLengthStored;
      property Alignment : TAlignment read FAlignment write SetAlignment default taCenter;
      property Items : TFlyNodes read GetItems write SetItems;
      property Value : String read FInternalValue write SetInternalValue; //it property should be last
      property Width : integer read FWidth write SetWidth default 0;
      property ItemIndex : integer read FinternalIndex write SetInternalIndex default -1;
  end;



  TSizedPopupTree = class(TCustomRapidTree)
  private
    //sizegrip
    FSizeGrip : Boolean;
    FSizing : Boolean;
    Offset, Sizingstart : TPoint;
    procedure SetSizeGrip(const Value : Boolean);
    function  IsSizeGrip(Pt : TSmallPoint) : Boolean;
    procedure WMNCCalcSize(var Message: TWMNCCalcSize); message WM_NCCALCSIZE;
    procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
    procedure WMNCLMouseDown(var Message: TMessage); message WM_NCLBUTTONDOWN;
    procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
   protected
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    {$ifdef DFS_COMPILER_4_UP}
    procedure ValidateState;
    {$endif}
    function GetAlignment(Col : integer) : TAlignment; override;
   public
    property SizeGrip : Boolean read FSizeGrip write SetSizeGrip default False;
  end;


  TDropDownStyle = (ddsSized, ddsOK, ddsOKCancel, dds3d, ddsSizeToControlSize);
  TDropdownStyles = set of TDropDownStyle;
  TDropDownType = (ddtStandart, ddtCustom);
   //Memo, Calendar, Picture(?)
  // 
  //1.   dropdown
  //2. , 
  //3.      dropdown
  TISDropDown = class(TCustomPanel) //tcustomPanel
  private
    FSizing, CloseUpStarted : Boolean;
    FNonClientHeight : integer;
    FLeftTop, Sizingstart : TPoint;
    FPlugEdit : TISCustomPlugEdit;
    FContainedControl : TWinControl;
    OKR, CancelR : TRect;
    FOnTop, OKUC, CancelUC, OKDown, CancelDown : boolean;
    FStyles : TDropDownStyles;
    procedure ClearControls;
    function IsSizeGrip(Pt : TSmallPoint) : Boolean;
    procedure WMNCCalcSize(var Message: TWMNCCalcSize); message WM_NCCALCSIZE;
    procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
    procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
    procedure WMNCLMouseDown(var Message: TMessage); message WM_NCLBUTTONDOWN;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure CMCloseUp(var Message: TMessage); message CM_CLOSEUP;
    procedure WMActivate(var Message: TWMActivate); message WM_ACTIVATE;
    function Contain(aHandle : HWnd): Boolean;
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetStyles(Value : TDropDownStyles);
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;

    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure CreateParams(var Params: TCreateParams); override;
    //procedure CreateWnd; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Styles : TDropDownStyles read FStyles write SetStyles;
    property ContainedControl : TWinControl read FContainedControl;
  end;


  TPopupTree = Class(TSizedPopupTree)
  private
    FDonotChangeSection : Boolean; //   contents  
    FPlugEdit : TISCustomPlugEdit;
    function GetStateIndex(const Node: TFlyNode): Integer;
    procedure NodeStateRefreshChildren(Node: TFlyNode);
    procedure NodeStateRefreshParent(const Node: TFlyNode);
    procedure NodeStateRefreshSiblings(const Node: TFlyNode);
    procedure Toggle(Node : TFlyNode);
    function GetCheckText(aItemSeparator : String) : String;
    procedure SetCheckText(Value, aItemSeparator : String);
    procedure WMKillFocus(var Msg: TWMKillFocus); message WM_KILLFOCUS;
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CancelMode;
  protected
    function SelectCell(var ACol, ARow: Longint; MoveAnchor : Boolean;  Shift: TShiftState; SelWay : TSelectionWay): Boolean; override;
    procedure GetBitmap(aNode : TFlyNode; index :  integer; Image : TBitmap); override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure KeyPress(var Key: Char); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure LocateItem(Text : String);
    function GetAlignment(Col : integer) : TAlignment; override;
  public
    constructor Create(AOwner: TComponent); override;
  end;

  //TISMemo = Class(TCustomMemo)
  //protected
  //  procedure CreateParams(var Params: TCreateParams); override;
  //public
  //end;
  // With Alignment support
  //
  TISAlignedEdit = Class(TCustomEdit)
  private
    //FCanvas : TCanvas;
    FDeletePressed, FCustomDraw{, FLastPos, FFirstPos} : Boolean;
    FAlignment: TAlignment;
    FMultiline : Boolean;
    FIndex : integer;
    procedure SetAlignment(Value : TAlignment);
    procedure SetMultiline(Value : Boolean);

    {$ifdef DFS_COMPILER_4_UP}
    procedure WMTimer(var Msg: TWMTimer); message WM_TIMER;
    {$ENDIF}
    //procedure WMSetText(var Message: TMessage); message WM_SETTEXT;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Msg: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMPaint(var Msg : TWMPaint); message WM_PAINT;
    procedure WMNCCalcSize(var Message: TWMNCCalcSize); message WM_NCCALCSIZE;
    procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
    procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
    procedure WMPaste(var Message: TMessage); message WM_PASTE;
    procedure DrawEditor;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure WndProc(var Message: TMessage); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure Click; override;
    procedure DblClick; override;
    procedure KeyPress(var Key: Char); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    {$ifdef DFS_COMPILER_4_UP}
    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean; override;
    {$endif}
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure EmulateMouseDown;
    procedure EmulateMouseUp;
    property Index : integer read FIndex;
    property Alignment: TAlignment read FAlignment write SetAlignment;
    property MultiLine : Boolean read FMultiline write SetMultiLine;
  end;

  TButtonType = (tbtNone, tbtEllipsis, tbtSpin, tbtDropDown, tbtScroll);
  TPressedButtons = (pbEllipsis, pbDownButton, pbUpButton, pbDropDown, pbSectionButton, pbProgress);
  TMoveFocus = (tmfFirst, tmfLast, tmfNext, tmfPrev);
  TButtonPressEvent = procedure(Sender: TObject; Button : TPressedButtons) of object;
  // event handlers
  TPrepareDropDownEvent = procedure(Sender: TObject; Section : TISPlugSection; DropDown : TISDropDown) of object;
  TGetDropDownControlEvent = procedure(Sender: TObject; Section : TISPlugSection; DropDown : TISDropDown; var DropDownCtl : TWinControl) of object;
  TCloseUpEvent = procedure(Sender: TObject; Section : TISPlugSection; DropDown : TISDropDown; var Accept : Boolean) of object;
  TAcceptCloseUpEvent = procedure(Sender: TObject; var Accept : Boolean) of object;
  //TValueChangingEvent = procedure(Sender: TObject; Section : TISPlugSection; Var Allow : Boolean) of object;
  TValueChangedEvent = procedure(Sender: TObject; Section : TISPlugSection) of object;


//  TDropDownType = (ddtStandard, ddtCustom);//new 26/10/2001
  TCrackWinCtl = class(TWinControl);
//  TPlugStoreState = (pssText, pssEditorType, pssSections);
//  TPlugStoreStates = set of TPlugStoreState;


  TISCustomPlugEdit = class(TCustomPanel)
  private
    { Private declarations }
    FStyles : TDropdownStyles;
    FDropDown : TISDropdown;//TPopupTree;
    FDrawOffset, FDropDownRows : integer;
    FInit, FAutoComplete, FAutoDropDown, {FSizeGrip,} FDropDownVisible : Boolean;
    FReadOnly, {FRightAligned,} FClearing, FRearrange, FUp, FDown : Boolean;
    {$IFDEF DFS_COMPILER_4_UP}
       Fimages : TCustomImagelist;
    {$ELSE}
       Fimages                                            : TImagelist;
    {$ENDIF}
    FListtoAdd, FSpacing : integer;
    FLockChangeEditorType, FAutoSize : Boolean;
    //FShowSpin : Boolean;
    FEditorsList : TList;
    FFocusedEditor : TIsAlignedEdit;
    FButtonWidth : Integer;

    FOnButtonPress : TButtonPressEvent;
    FOnPrepareDropDown : TPrepareDropDownEvent;
    FOnGetDropDownControl : TGetDropDownControlEvent;
    FOnCloseUp: TCloseUpEvent;
    FOnAcceptCloseUp : TAcceptCloseUpEvent;
    //FOnValueChanging : TValueChangingEvent;
    FOnValueChanged : TValueChangedEvent;

    FButtonType : TButtonType;
    FEditorType : TEditorType;

    FSections : TISPlugSections;
    procedure AddColors(const S : String);
    procedure ShiftControls(aEdit : TISAlignedEdit);
    function DaysInMonth (const Month, Year : integer): Byte;
    procedure ValidateAndSynchronize(EditorIndex : integer);
    procedure WMSettingChange(var Message: TMessage); message WM_SETTINGCHANGE;
    procedure WMNCCalcSize (var Message: TWMNCCalcSize); message WM_NCCALCSIZE;
    procedure WMNCHittest (var Message: TMessage); message WM_NCHITTEST;
    procedure WMNCLButtonUp(var Message: TMessage); message WM_NCLBUTTONUP;//WM_NCLBUTTONUP;
    procedure WMNCLButtonDown(var Message: TMessage); message WM_NCLBUTTONDOWN;
    procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
    procedure WMTimer(var Msg: TWMTimer); message WM_TIMER;
    procedure CMFontChanged(var Message); message CM_FONTCHANGED;
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CancelMode;
    procedure CMCloseUp(var Message: TMessage); message CM_CLOSEUP;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMSize(var Message: TMessage); message WM_SIZE;
    //procedure WMActivate(var Message: TWMActivate); message WM_ACTIVATE;
    procedure SetValue(Index :integer; Value : String);
    function GetValue(Index : Integer) : String;
    //procedure SetShowSpin(Value : Boolean);
    function GetSectionCount : integer;
    function GetText : String;

    procedure SetButtonWidth(const Value : integer);
    procedure SetButtonType(Value: TButtonType);
    procedure SetSpacing(const Value : integer);
    procedure SetEditorType(Value : TEditorType);
    procedure SetSections(Value : TISPlugSections);

    // btn management functions
    function UpBtnRect : TRect;
    function DownBtnRect : TRect;
    function BtnRect : TRect;
    function MinusBtnRect : TRect;
    function PlusBtnRect : TRect;

    procedure DrawButton;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure SetAutoSize(Value : Boolean); reintroduce;
    function ValidateInput(Sender : TISAlignedEdit;  var Key : Char) : Boolean;
    procedure ProceedButton(aButton : TPressedButtons);
    procedure SetEditorText(Index : integer);

    procedure SetImages(Value: {$IFDEF DFS_COMPILER_4_UP} TCustomImageList {$ELSE} TImagelist {$ENDIF});
//    procedure SetRightAligned(const Value : Boolean);
    //dropdown management
    procedure PrepareDropDown;
    procedure CreateDropdown;
    //function GetPopupTree : TPopupTree;
    function GetCheckText(Index : Integer) : String;


{    function GetAsDateTime : TDateTime;
    Function GetAsInteger : Integer;
    function GetAsCurrency : Currency;
    function GetAsFloatValue : Double;
}
    function GetModified : Boolean;
    procedure SetModified(Value : Boolean);
    procedure SetReadOnly(Value : Boolean);
    function GetAlignment : TAlignment;
    procedure SetAlignment(Value : TAlignment);

  protected
    { Protected declarations }
    IsInplace: Boolean;
    procedure SetText(Value : String); virtual;
    procedure RebuildSections(ClearSections : Boolean);
    function DoTabKey : Boolean; virtual; //abstract;
    procedure ResetButtons;
    procedure ArrangeSection(index : integer);
    procedure Loaded; override;
    procedure CreateWnd; override;
    procedure Paint; override;
    procedure ReArrangeSections;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;

    procedure ClearEditors;
    procedure ButtonPressed(aButton : TPressedButtons); dynamic;
    procedure PrepareDropDownEvent(Section : TISPlugSection; DropDown : TISDropDown); dynamic;
    procedure CloseUpEvent(Section : TISPlugSection; DropDown : TISDropDown; var Accept : Boolean); dynamic;
    procedure AcceptCloseUpEvent(var Accept : Boolean);
    procedure GetDropDownControl(Section : TISPlugSection; DropDown : TISDropDown; var DropDownCtl : TWinControl); dynamic;
    //procedure ValueChanging(Sender: TObject; Section : TISPlugSection; Var Allow : Boolean); dynamic;
    procedure SetStyles(Value : TDropdownStyles);
    procedure ValueChanged(Section : TISPlugSection); dynamic;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure MakeCloseUp(Accept : Boolean);
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure SetUpdateState(State : Boolean);//FLockChangeEditorType
    procedure FocusEditor(Position : TMoveFocus);
    procedure SetActiveEditor(Value : Integer);
    function Focused: Boolean; {$IFDEF DFS_COMPILER_4_UP}override;{$endif} //new 29.05.2001
    procedure CloseUp(Accept : Boolean);
    procedure DropDown;
    function DropdownVisible : Boolean;
    function DropdownCtl : TISDropDown;//TPopupTree;
    property DropdownStyles : TDropdownStyles read FStyles write SetStyles default [dds3d];
    property Alignment : TAlignment read GetAlignment write SetAlignment default taLeftJustify;
    property Modified : Boolean Read GetModified write SetModified;
    property AutoComplete : Boolean read FAutoComplete write FAutoComplete default True;
    property AutoDropDown : Boolean read FAutoDropDown write FAutoDropdown default True;
    property Images : {$IFDEF DFS_COMPILER_4_UP} TCustomImageList {$ELSE} TImagelist {$ENDIF} read FImages write SetImages;
    property AutoSize : boolean read FAutoSize write SetAutoSize default False;
    property Values[index : integer] : String read GetValue write SetValue;
    property FocusedEditor : TISAlignedEdit read FFocusedEditor;
    property ButtonWidth : integer read FButtonWidth write SetButtonWidth default 15;
    property SectionCount : integer read GetSectionCount;
    //property ShowSpin : Boolean read FShowSpin write SetShowSpin default True;

    property Text : String read GetText write SetText;
    property Font;
    property BorderStyle;
    property Spacing : integer read FSpacing write SetSpacing default 1;
    property ButtonType : TButtonType read FButtonType write SetButtonType default tbtNone;
    //!!!!! editor type should be defined before Sections
    property EditorType : TEditorType read FEditorType write SetEditorType default tetSimple;
    property Sections : TISPlugSections read FSections write SetSections;
    //    property RightAligned : Boolean read FRightAligned write SetRightAligned;
    //property DropDownTree : TPopupTree read GetPopupTree;
    //property DropDownSized : Boolean read FsizeGrip write FSizeGrip default True;
    property DropDownRows : integer read FDropDownRows write FDropDownRows default 8;
    property ReadOnly : Boolean read FreadOnly write SetReadOnly default False;

    property OnButtonPress : TButtonPressEvent read FOnButtonPress write FOnButtonPress;
    property OnPrepareDropDown : TPrepareDropDownEvent read FOnPrepareDropDown Write FOnPrepareDropDown;
    property OnCloseUp: TCloseUpEvent read FOnCloseUp write FOnCloseUp;
    property OnAcceptCloseUp: TAcceptCloseUpEvent read FOnAcceptCloseUp write FOnAcceptCloseUp;
    property OnGetDropDownControl : TGetDropDownControlEvent read FOnGetDropDownControl write FOnGetDropDownControl;//new
    //property OnValueChanging : TValueChangingEvent read FOnValueChanging write FOnValueChanging;
    property OnValueChanged : TValueChangedEvent read FOnValueChanged write FOnValueChanged;
  end;

  TISPlugSections = class(TCollection)
   private
      FEdit : TISCustomPlugEdit;
      Fowner : TPersistent;
      FEmptying, FEmpty, FUpdating : Boolean;
      FEditorType : TEditorType;
      procedure ScanDateTimeString(const Format: string; MonthMedium, WithoutAMPM , ClearSections : Boolean);
      procedure SetEditorType(Value : TEDitorType);
      function  DrawColor(aCanvas : TCanvas; aColor : Longint; R : TRect) : TRect;
      procedure Drawimage(aCanvas : TCanvas; aImages : {$IFDEF DFS_COMPILER_4_UP}TCustomImagelist {$ELSE} TImagelist {$ENDIF};
                Index : integer; Var ImageRect: TRect; RTL : Boolean);
      procedure ScanText(Value : String);
   protected
      FRebuilding : Boolean;
      procedure SetSection(Index : integer; Value : TISPlugSection);
      function Getsection(Index : integer) : TISPlugSection;
      function GetOwner: TPersistent; override;
      procedure Update(Item: TCollectionItem); override;
      procedure SetText(Value : String);
      procedure DrawEditorImage(aCanvas : TCanvas; aValue : String; R : TRect; aIndex : Integer; Focused, aCtl3D, RTL, UseValue, AnotherCanvas : Boolean; aImages : {$IFDEF DFS_COMPILER_4_UP}TCustomImagelist {$ELSE} TImagelist {$ENDIF});
   public
      procedure SetUpdateState(Value : Boolean);
      constructor Create;//(aOwner : TISCustomPlugEdit);
      destructor Destroy; override;
      procedure Empty;
      function ISEmpty : boolean;
      procedure Assign(Source: TPersistent); override;
      function GetText: String;//get text for any editor type exclude tetCustom (for it type - special format)
      function GetTranslatedText : String; //use for get masked text for any editorType - do not save
      procedure SetOwner(aOwner : TPersistent);
      procedure RebuildEditors(aEditorType : TEditorType; ClearSections : Boolean);//very urgent
      function AddSection : TISPlugSection;
      procedure RemoveSection(Value : TISPlugSection);
      property EditorType : TEditorType read FeditorType write SetEditorType;
      property Item[index : integer] : TISPlugSection read GetSection Write Setsection; default;
  end;


  TISPlugEdit = class(TISCustomPlugEdit)
  public
    {$ifdef DFS_COMPILER_4_UP}
    property DockManager;
    {$endif}
  published
    property DropdownStyles;
    property AutoComplete;
    property AutoDropDown;
    //property DropDownSized;
    property Images;
    property ButtonType;
    property Spacing;

    property ButtonWidth;
    property Align;
    property Alignment;
    property AutoSize;
    //events
    property OnPrepareDropDown;
    property OnGetDropDownControl;
    property OnCloseUp;
    property OnAcceptCloseUp;
    //property OnValueChanging;
    property OnValueChanged;
    property OnButtonPress;
    {$ifdef DFS_COMPILER_4_UP}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property UseDockManager default True;
    property DockSite;
    property DragKind;
    property ParentBiDiMode;
    property OnConstrainedResize;
    property OnDockDrop;
    property OnDockOver;
    property OnEndDock;
    property OnGetSiteInfo;
    property OnStartDock;
    property OnUnDock;
    property OnCanResize;
    {$endif}
    property BorderStyle default bsSingle;
    property Color;
    property Ctl3D;
    property Visible;
    property DragCursor;
    property DragMode;
    property Enabled;
    property FullRepaint;
    property Font;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnKeyDown;
    property OnKeyUp;
    property OnKeyPress;
    property OnResize;
    property OnStartDrag;
    property EditorType;
    property Sections;
    property Text;//new
  end;


implementation

uses Math,{$IFNDEF Registered}FTAbout, {$ENDIF} comctrls, ISConst, ShellApi;

var
   HookHandle : HHOOK = 0 ;
   DropdownHandle : TISDropdown = nil;



function CallWndProcHook (Code: Integer; WParam: WPARAM; LParam: LPARAM): LRESULT;
stdcall;
begin
  if Code = HC_ACTION then
  begin
      Case  WParam of
{     case PCWPSTRUCT(LParam)^.Message of//Case  WParam of
        WM_NCACTIVATE:
          if (DropDownHandle <> nil) then
          begin
            if PCWPSTRUCT(LParam)^.Hwnd <> DropDownHandle.Handle then
            begin
              PCWPSTRUCT(LParam)^.WParam := Longint(True);
              //SendMessage(PCWPSTRUCT(LParam)^.Hwnd, WM_NCACTIVATE, Longint(True), 0);
//              exit;//Result := Ord(False);
            end;
          end;
}

        WM_NCRBUTTONDOWN, WM_NCLBUTTONDOWN ,WM_LBUTTONDOWN,WM_RBUTTONDOWN:
          if (DropDownHandle <> nil) and (PMouseHookStruct(LParam)^.Hwnd <> DropDownHandle.Handle) then
          begin //ParentWindow                            MouseHookStruct
           if (not DropDownHandle.Contain(PMouseHookStruct(LParam)^.Hwnd)) then
           begin
            //first
            {PostMessage(DropDownHandle.Handle, CM_CANCELMODE, 0, 0);
            SetFocus(PMouseHookStruct(LParam)^.Hwnd);
            }
            DropDownHandle.FPlugedit.CloseUp(false);
            Result :=1;
            exit;
           end;
           {else
           begin
             SetFocus(PMouseHookStruct(LParam)^.Hwnd);
             //PostMessage(PMouseHookStruct(LParam)^.Hwnd, WParam, PCWPStruct(LParam)^.WParam, PCWPStruct(LParam)^.LParam);
             Result :=1;
             //exit;
           end;}
           
         end;
     end;
  end;
  Result := CallNextHookEx(HookHandle, Code, WParam, LParam);
end;

procedure InstallHook;
begin
  HookHandle := SetWindowsHookEx({WH_CALLWNDPROC}WH_MOUSE, CallWndProcHook, 0, GetCurrentThreadId);
end;

procedure UnInstallHook;
begin
 UnhookWindowsHookEx (HookHandle);
end;

{ TISPlugSections }

constructor TISPlugSections.Create;//(aOwner : TISCustomPlugEdit);
begin
   inherited Create(TISPlugSection);
end;

destructor TISPlugSections.Destroy;
begin
   FEdit := nil;
   Clear;
   inherited Destroy;
end;

function TISPlugSections.GetTranslatedText : String;
var i: integer;
begin
 Result := '';
 if not FEmpty then
  for i:=0 to Count-1 do
     with Item[i] do
      Result := Result + FLSeparator + FinternalValue+FRSeparator;
end;

procedure TISPlugSections.Assign(Source: TPersistent);
var i : integer;
begin
 if Source Is TISPlugSections then
 begin
   BeginUpdate;
   try
    Clear;
    for i:=0 to TISPlugSections(Source).Count-1 do
      with AddSection do
         Assign(TISPlugSections(Source)[i]);
   finally
     EndUpdate;
   end;
 end
 else
   inherited

end;

function TISPlugSections.GetText: String;
var i : integer;
    ls, rs : string;

    function NextSepExists : boolean;
    begin
       result := false;
       if (i < Count-1) then
         Result := Item[i+1].FLSeparator <> '';
    end;

    function PrevSepExists : boolean;
    begin
       result := false;
       if (i > 0) then
         Result := Item[i-1].FRSeparator <> '';
    end;
begin
   Result := '';
   rs:='';
   ls :='';
   if not FEmpty then
     for i:=0 to Count-1 do
      with Item[i] do
      begin

        if FLSeparator <> '' then
        begin
          ls := FLSeparator;
        end
        else
        begin
         if (Count=1) or (i=0) or PrevSepExists or (rs <> '') then
          ls:=''
         else
          ls := EmptySeparator; //special separate symbol
        end;

        if FRSeparator <> '' then
          rs := FRSeparator
        else
        begin
         if (Count =1) or (i=Count-1) or NextSepExists then
          rs := ''
         else
          rs := EmptySeparator;
        end;

        Result := Result + ls + FinternalValue + rs;
      end;
end;

procedure TISPlugSections.ScanText(Value : String);
var P : PChar;
    i : integer;
    ls, rs : string;

  procedure ScanSeparator(Sep : String);
  var Tmp : String;
      found : boolean;
      OldP : PChar;
  begin
     tmp :='';
     found := false;
     OldP := P;
     while P^ <> #0 do
     //while (Tmp <> Sep) do
     begin
      if (Tmp <> Sep) then
      begin
       if Length(tmp) < Length(Sep) then
         tmp:=tmp+P^
       else
          break;  
      end
      else
      begin
        found := true;
        break;
      end;
      inc(P);
     end;
     if not found then
       P := OldP;
  end;

  procedure ScanValue(Section : TISPlugSection);
  var atmp : String;
      Numeric : Boolean;
      NextSep : String;
      NsTmp : PChar;
  begin
     if Section.FType in [tdtcheck, tdtInteger, tdtD, tdtDD, tdtM, tdtMM, tdtYY, tdtYYYY, tdt24H, tdt24HH, tdt12H, tdt12HH, tdtN, tdtNN, tdtS, tdtSS, tdtProgress, tdtValueProgress, tdtPercentProgress,tdtLeadingZeroesInteger] then
        Numeric :=True
     else Numeric :=False;
     if Numeric then
     begin
        aTmp := '';
        //include sign char before scan numeric->
        if P^ = '-' then
        begin
           aTmp := aTmp + P^;
           inc(P);
        end;
        While P^ in ['0','1','2','3','4','5','6','7','8','9'] do
        begin
          aTmp := aTmp + P^;
          inc(P);
        end;
        Section.Value := aTmp;
     end
     else
     begin
        if Section.RightSeparator <>'' then
          NextSep := Section.RightSeparator
        else if Section.Index < Section.Collection.Count-1 then //not last
        begin
          if TISPlugSection(Section.Collection.Items[Section.Index+1]).LeftSeparator <>'' then
            NextSep := TISPlugSection(Section.Collection.Items[Section.Index+1]).LeftSeparator;
          //else if <<-          - 
        end
        else
           NextSep := #0;//EmptySeparator;

        aTmp := '';
        nsTmp := P;
        While P^ <> #0 do
        begin
          if StrLComp(nsTmp, Pchar(NextSep), Length(NextSep))<>0 then
          begin
            aTmp := aTmp + P^;
            inc(P);
            nsTmp := P;
          end
          else Break;
        end;
        Section.Value := aTmp;
     end;
  end;

  function NextSepExists : boolean;
  begin
    result := false;
    if (i < Count-1) then
      Result := Item[i+1].FLSeparator <> '';
  end;

  function PrevSepExists : boolean;
  begin
    result := false;
    if (i > 0) then
     Result := Item[i-1].FRSeparator <> '';
  end;
  
begin
   P := PChar(Value);
   for i:=0 to Count-1 do
   begin
     with Item[i] do
     begin
       if LeftSeparator <> '' then
          ls := LeftSeparator
       else
       begin
        if (i=0) or (Count=1) or PrevSepExists or (rs <> '')then
         ls := ''
        else
         ls := EmptySeparator;
       end;

       if RightSeparator <> '' then
          rs := RightSeparator
       else
       begin
         if (i=Count-1) or (Count=1) or NextSepExists then
           rs := ''
         else
           rs := EmptySeparator;
       end;
     end;

     //if item[i].LeftSeparator <> '' then
     // ScanSeparator(item[i].LeftSeparator);
     if ls <>'' then
      ScanSeparator(ls);
      ScanValue(item[i]);
     //if item[i].RightSeparator <> '' then
     // ScanSeparator(item[i].RightSeparator);
     if rs <> '' then
      ScanSeparator(rs);
   end;
end;

procedure TISPlugSections.SetText(Value : String);
var tmpValue : DWord;
    i, Param : Integer;
    aDate : TDateTime;
    Day, Mo, Year, h,m,s,ms : Word;
    scanned, Pm : Boolean;
    P : Pchar;
    aColor : TColor;
    TmpStr : String;
    sList : TStringList;
begin
  if Value = '' then
    Empty
  else
   Case FEditorType of
       tetButton, tetHyperLink,tetPassword, tetMemo, tetSimple, tetCurrency, {tetMediumDate, tetLongDate,} tetCustom:
         ScanText(Value);
       tetImage, tetImageMemo, tetImageCaption:
       begin
         sList := TStringList.Create;
         try
           sList.CommaText := Value;
           if SList.Count >= Count then
           for i:=0 to Count-1 do
             Item[i].Value := sList[i];
         finally
           slist.Free;
         end;
       end;
       //Texts
       //  TIsPlugSection(Items[0]).Value := Value;
        tetYesNo, tetTrueFalse, tetCheckBox:
        begin
           if (Value = '1') or (Value='Yes') or (Value='True')  then
              TIsPlugSection(Items[0]).ItemIndex := 1
           else
              TIsPlugSection(Items[0]).ItemIndex := 0;
        end;
        //, , , , tetButton
        tetPercentProgress, tetProgress, tetInteger,tetValueProgress:
           TIsPlugSection(Items[0]).Value := InttoStr(strToInt(Value));
         tetPercent: //strtocurr
           if Pos('%', Value) <> 0 then
            TIsPlugSection(Items[0]).Value := IntToStr(StrToInt(Copy(Value,1,length(Value)-1)))
           else
            TIsPlugSection(Items[0]).Value := IntToStr(StrToInt(Value));
        //tetCurrency:
        //  begin
        //   TIsPlugSection(Items[0]).Value := CurrToStr(Int(StrToCurr(Value)));
        //   TIsPlugSection(Items[1).Value := CurrToStr(Frac(StrToCurr(Value)));
        //  end;
        tetColor:
           begin
            aColor := StringToColor(Value);
            TIsPlugSection(Items[0]).Value := ColorToString(aColor);
           end;
        //tetImage, tetImageCaption, tetImageMemo:
        //   TIsPlugSection(Items[0]).ItemIndex := strToInt(Value);
        tetDropDownCheckList,
        tetDropDownList, tetDropDown, tetDropdownTree:
           TIsPlugSection(Items[0]).Value := Value;
        tetIPAddress:
              if Strscan(PChar(Value), '.') = nil then
              begin
               TmpValue := strToint(Value);// if it integer value
               Param := (TmpValue shl 16) shr 16;
               TIsPlugSection(Items[0]).Value := inttostr(Hi(Param));
               TIsPlugSection(Items[1]).Value := inttostr(Lo(Param));
               Param := (TmpValue shr 16);
               TIsPlugSection(Items[2]).Value := inttostr(Hi(Param));
               TIsPlugSection(Items[3]).Value := inttostr(Lo(Param));
              end
              else
              begin
                TIsPlugSection(Items[0]).Value := '';
                TIsPlugSection(Items[1]).Value := '';
                TIsPlugSection(Items[2]).Value := '';
                TIsPlugSection(Items[3]).Value := '';
                tmpStr:='';
                P := Pchar(Value);
                i:=0;
                Scanned := False;
                while not Scanned do
                begin
                 if not (P^ in ['.', #0]) then
                   tmpStr := tmpStr+P^
                 else
                 begin
                  TIsPlugSection(Items[i]).Value := tmpStr;
                  tmpStr :='';
                  inc(i);
                  Scanned := (P^ = #0);
                 end;
                 inc(P);
                end;
              end;
        tetGeneralDate, tetMediumDate, tetLongDate, tetShortDate,
        tetLongTime, tetMediumTime, tetShortTime:
        begin
        aDate := strToDateTime(Value);
        if (FEditorType in [tetGeneralDate, tetLongDate, tetMediumDate, tetShortDate]) then
        begin
           DecodeDate(aDate, Year, Mo, Day);
           for i:=0 to Count -1 do
           begin
               Case TIsPlugSection(Items[i]).Ftype of
                  tdtD, tdtDD:
                     TIsPlugSection(Items[i]).Value := InttoStr(Day);
                  tdtDDD:
                     TIsPlugSection(Items[i]).Value := ShortDayNames[DayOfWeek(aDate)];
                  tdtDDDD:
                     TIsPlugSection(Items[i]).Value := LongDayNames[DayOfWeek(aDate)];
                  tdtM, tdtMM:
                     TIsPlugSection(Items[i]).Value := InttoStr(Mo);
                  tdtMMM:
                     TIsPlugSection(Items[i]).Value := ShortMonthNames[Mo];
                  tdtMMMM:
                     TIsPlugSection(Items[i]).Value := LongMonthNames[Mo];
                  tdtYY, tdtYYYY:
                     TIsPlugSection(Items[i]).Value := InttoStr(Year);
               end;
           end;
        end;
        if (FEditorType in [tetGeneralDate, tetLongTime, tetMediumTime, tetShortTime]) then
        begin
          //aDate := StrToTime(Value);
          DecodeTime(aDate, H,m,s,ms);
          Pm := False;
          for i:=0 to Count -1 do
           begin
               Case TIsPlugSection(Items[i]).Ftype of
                  tdt24H,tdt24HH:
                     TIsPlugSection(Items[i]).Value := InttoStr(h);
                  tdt12H, tdt12HH:
                    if h >= 12 then
                    begin
                     TIsPlugSection(Items[i]).Value := Inttostr(h-12);
                     Pm :=true;
                    end;
                  tdtAP, tdtAMPM:
                    if Pm = True then
                       TIsPlugSection(Items[i]).ItemIndex := 1
                    else
                       TIsPlugSection(Items[i]).ItemIndex := 0;
                  tdtN, tdtNN:
                       TIsPlugSection(Items[i]).Value := Inttostr(m);
                  tdtS, tdtSS:
                       TIsPlugSection(Items[i]).Value := Inttostr(s);
               end;
           end;
        end;
        end;
        tetYY, tetYYYY, tetH, tetHH, tetN, tetNN, tetS, tetSS, tetD, tetDD, tetM, tetMM:
          TIsPlugSection(Items[0]).Value := inttostr(strToint(value));
        tetMMM, tetMMMM, tetDDD, tetDDDD:
          TIsPlugSection(Items[0]).Value := Value;
   end;
end;

function TISPlugSections.ISEmpty : boolean;
begin
  Result := FEmpty;
end;

procedure TISPlugSections.Empty;
var i : integer;
begin
   FEmpty := True;
   FEmptying := true;
   try
     for i:=0 to Count -1 do
      TISPlugSection(items[i]).Value := '';
   finally
     FEmptying := false;
   end;
end;

procedure TISPlugSections.SetEditorType(Value : TEDitorType);
begin
   FEditorType := Value;
end;

procedure TISPlugSections.SetSection(Index : integer; Value : TISPlugSection);
begin
   if (Index < Count)  then
      TISPlugSection(Items[Index]).Assign(Value);
end;

procedure TISPlugSections.SetOwner(aOwner : TPersistent);
begin
   Fowner := aOwner;
end;

function TISPlugSections.GetOwner: TPersistent;
begin
     Result := FOwner;//FEdit;
end;

function TISPlugSections.Getsection(Index : integer) : TISPlugSection;
begin
  if (Index < Count)  then
     Result:= TISPlugSection(inherited Items[Index])
  else
     Result := nil;
end;


procedure TISPlugSections.SetUpdateState(Value : Boolean);
begin
  FUpdating := Value;
  if (Fupdating = false) and (Fedit <> nil) then
    Fedit.RebuildSections(false);

end;

procedure TISPlugSections.Update(Item: TCollectionItem);
begin
 if (not FUpdating) and (FEdit<>nil) then
 begin
//  if not (csLoading in FEdit.ComponentState) then
   if Item <> nil then
      FEdit.ArrangeSection(Item.Index)
   else
      FEdit.RebuildSections((FEdit.EditorType <> tetCustom ) {and (not FEdit.FLockChangeEditorType)});//(false);//new 22.03.2
 end;
end;


function TISPlugSections.AddSection : TISPlugSection;
begin
  Result := TISPlugSection(Add);
end;

procedure TISPlugSections.RemoveSection(Value : TISPlugSection);
begin
 if Value <> nil then
 begin
     Value.Collection := nil;//remove
     Value.Free;
     Update(nil);
 end;

end;

procedure TISPlugSections.ScanDateTimeString(const Format: string; MonthMedium, WithoutAMPM , ClearSections : Boolean);
var
  AppendLevel: Integer;
  AddedSection : TISPlugsection;
  procedure AppendFormat(Format: PChar);
  var
    Starter, Token, LastToken: Char;
    Use12HourClock,
    BetweenQuotes: Boolean;
    P: PChar;
    sCount: Integer;

    function FindSection(aTypes : TISDataTypes) : TISPlugsection;
    var i : integer;
    begin
      Result := nil;
       for i:=0 to self.Count-1 do
         if Item[i].FType in aTypes then
         begin
            Result := Item[i];
            break;
         end;
       if Result = nil then
         Result := AddSection;
    end;
    
    procedure GetCount;
    var
      P: PChar;
    begin
      P := Format;
      while Format^ = Starter do Inc(Format);
      sCount := Format - P + 1;
    end;

  begin
    if (Format <> nil) and (AppendLevel < 2) then
    begin
      Inc(AppendLevel);
      LastToken := ' ';
      Use12HourClock := False;

      while Format^ <> #0 do
      begin
        Starter := Format^;
        Inc(Format);
        if Starter in LeadBytes then
        begin
          if Format^ = #0 then Break;
          Inc(Format);
          LastToken := ' ';
          Continue;
        end;
        Token := Starter;
        if Token in ['a'..'z'] then Dec(Token, 32);
        if Token in ['A'..'Z'] then
        begin
          if (Token = 'M') and (LastToken = 'H') then Token := 'N';
          LastToken := Token;
        end;

        case Token of
          'Y':
            begin
              GetCount;
              if ClearSections then
                 AddedSection := AddSection
              else
                AddedSection := FindSection([tdtYY, tdtYYYY]);
              with AddedSection do
              begin
               Case sCount of
                 2: FType := tdtYY;
                 4: FType := tdtYYYY;
               end;//case
               FAlignment := taLeftJustify;
               FLSeparator:='';
               //if Fedit <> nil then
               //  Fedit.FSpacing :=1;
               Validate;
              end;
            end;
          'M':
            begin
              GetCount;
              if MonthMedium then sCount := 3;
              if ClearSections then
                 AddedSection := AddSection
              else
                AddedSection := FindSection([tdtM, tdtMM, tdtMMM, tdtMMMM]);
              with AddedSection do
              begin
               Case sCount of
                 1: FType := tdtM;
                 2: FType := tdtMM;
                 3: FType := tdtMMM;
               else
                FType := tdtMMMM;

               end;//case
               FLSeparator := '';
               //if Fedit <> nil then
               //  Fedit.FSpacing :=1;
               FAlignment := taLeftJustify;
               Validate;
              end;
           end;
          'D':
            begin
              GetCount;
              if ClearSections then
                 AddedSection := AddSection
              else
                AddedSection := FindSection([tdtD, tdtDD, tdtDDD, tdtDDDD]);
              with AddedSection do
              begin
               Case sCount of
                 1: FType := tdtD;
                 2: FType := tdtDD;
                 3: FType := tdtDDD;
                 4: FType := tdtDDDD;
               end;//case
               FLSeparator := '';
               //if Fedit <> nil then
               //  Fedit.FSpacing :=1;
               FAlignment := taLeftJustify;
               Validate;
              end;
            end;
          'H':
            begin
              GetCount;
              BetweenQuotes := False;
              P := Format;
              while P^ <> #0 do
              begin
                if P^ in LeadBytes then
                begin
                  Inc(P);
                  if P^ = #0 then Break;
                  Inc(P);
                  Continue;
                end;
                case P^ of
                  'A', 'a':
                    if not BetweenQuotes then
                    begin
                      if ((StrLIComp(P, 'AM/PM', 5) = 0)or (StrLIComp(P, 'A/P',   3) = 0) or (StrLIComp(P, 'AMPM',  4) = 0))  then
                        Use12HourClock := True;
                      Break;
                    end;
                  'H', 'h':
                    Break;
                  '''', '"': BetweenQuotes := not BetweenQuotes;
                end;
                Inc(P);
              end;

              if sCount > 2 then sCount := 2;
              if WithoutAMPM then
                 Use12HourClock := False;
              if ClearSections then
                 AddedSection := AddSection
              else
                AddedSection := FindSection([tdt12H, tdt12HH, tdt24H, tdt24HH]);

              with AddedSection do
              begin
               Case sCount of
                 1: if Use12HourClock then FType := tdt12H else FType := tdt24H;
                 2: if Use12HourClock then FType := tdt12HH else FType := tdt24HH;
               end;//case
               FLSeparator := '';
               FAlignment := taLeftJustify;
               //if Fedit <> nil then
               //  Fedit.FSpacing :=1;
               Validate;
              end;
            end;
          'N':
            begin
              GetCount;
              if sCount > 2 then sCount := 2;
              if ClearSections then
                 AddedSection := AddSection
              else
                AddedSection := FindSection([tdtN, tdtNN]);

              with AddedSection do
              begin
               Case sCount of
                 1: FType := tdtN;
                 2: FType := tdtNN;
               end;//case
               FAlignment := taLeftJustify;
               FLSeparator := '';
               //if Fedit <> nil then
               //  Fedit.FSpacing :=1;
               //FMinvalue := 0; FMaxValue := 59;
               Validate;
              end;
            end;
          'S':
            begin
              GetCount;
              if ClearSections then
                 AddedSection := AddSection
              else
                AddedSection := FindSection([tdtS, tdtSS]);

              with AddedSection do
              begin
               Case sCount of
                 1: FType := tdtS;
                 2: FType := tdtSS;
               end;//case
               FAlignment := taLeftJustify;
               FLSeparator := '';
               //if Fedit <> nil then
               //  Fedit.FSpacing :=1;
               //FMinvalue := 0; FMaxValue := 59;
               Validate;
              end;
            end;
          'A':
            begin
              P := Format - 1;
              if (StrLIComp(P, 'AM/PM', 5) = 0) then
              begin
                 if not WithoutAMPM then
                 if ClearSections then
                   AddedSection := AddSection
                 else
                   AddedSection := FindSection([tdtAMPM]);

                 with AddedSection do
                 begin
                  FType := tdtAMPM;
                  FreadOnly := True;
                  FAlignment := taLeftJustify;
                  FLseparator := '';
                  FMinvalue := 0; FMaxValue := 1;
                  //if Fedit <> nil then
                  //   Fedit.FSpacing :=1;
                  Items.Add(nil, 'AM');
                  Items.Add(nil, 'PM');
                  Validate;
                 end;
                Inc(Format, 4);
              end
              else if (StrLIComp(P, 'AMPM', 4) = 0) then
              begin
                 if not WithoutAMPM then
                  if ClearSections then
                   AddedSection := AddSection
                  else
                   AddedSection := FindSection([tdtAMPM]);

                 with AddedSection do
                 begin
                  FType := tdtAMPM;
                  FAlignment := taLeftJustify;
                  FreadOnly := True;
                  FLSeparator := '';
                  FMinvalue := 0; FMaxValue := 1;
                  Items.Add(nil, TimeAMString);
                  Items.Add(nil, TimePMString);
                  //if Fedit <> nil then
                  //  Fedit.FSpacing :=1;
                  Validate;
                 end;
                Inc(Format, 4);
              end
              else
              if not WithoutAMPM then
              if StrLIComp(P, 'A/P', 3) = 0 then
              begin
                 if ClearSections then
                   AddedSection := AddSection
                 else
                   AddedSection := FindSection([tdtAP]);

                with AddedSection do
                 begin
                  FType := tdtAP;
                  FReadOnly := True;
                  FAlignment := taLeftJustify;
                  FLSeparator := '';
                  FMinvalue := 0; FMaxValue := 1;
                  Items.Add(nil, 'A');
                  Items.Add(nil, 'P');
                  //if Fedit <> nil then
                  //  Fedit.FSpacing :=1;
                  Validate;
                end;
                Inc(Format, 2);
              end;
              if not WithoutAMPM then
               Use12HourClock := True;
              //else
              //AppendChars(@Starter, 1);

            end;

          '/':      //treecollection
            //TISPlugSection(Items[Count-1])
            AddedSection.RightSeparator:=DateSeparator;//AppendChars(@DateSeparator, 1);
          ':':
            //TISPlugSection(Items[Count-1])
            AddedSection.RightSeparator:=TimeSeparator;//AppendChars(@TimeSeparator, 1);
          '''', '"':
            begin
              //P := Format;
              while (Format^ <> #0) and (Format^ <> Starter) do
              begin
                if Format^ in LeadBytes then
                begin
                  Inc(Format);
                  if Format^ = #0 then Break;
                end;
                Inc(Format);
              end;
              //AppendChars(P, Format - P);
              if Format^ <> #0 then Inc(Format);
            end;
        else
          //TISPlugSection(Items[Count-1])
          AddedSection.FRSeparator:=AddedSection.FRSeparator+Starter//AppendChars(@Starter, 1);
        end;
      end;
      Dec(AppendLevel);
    end;
  end;

begin
//  BufPos := 0;
  AppendLevel := 0;
  if Format <> '' then
     AppendFormat(Pointer(Format))
end;

procedure TISPlugSections.RebuildEditors(aEditorType : TEditorType; ClearSections : Boolean);//big-big function
var i : integer;
//    aNode : TFlyNode;
 function DecR(Base : integer; Mul : integer) : integer;
 var i : integer;
 begin
    Result := Base;
    for i:=1 to Mul-1 do
      Result := Result * Base;
 end;

begin
 FUpdating := True;

  if (aEditorType <> tetCustom) and ClearSections then
  begin
   FRebuilding := True;
   Clear; //31/3/2
  end;
 FEditorType := aEditorType;
 if Fedit <> nil then
   FEdit.FeditorType := aEditorType;

 try

// if FEdit <> nil then
//   FEdit.FRightAligned := False;

  case aEditorType of
   tetImage, tetImageMemo,
   tetImageCaption:
   if ClearSections then
   begin
    with AddSection do
    begin
       Case aEditorType of
        tetImage:
          FType := tdtImage;
        tetImageCaption:
          FType := tdtImageCaption;
        tetImageMemo:
          FType := tdtImageMemo;
       end;
       FAlignment := taLeftJustify;
       FLSeparator:='';
       //if Fedit <> nil then
       //  Fedit.FSpacing :=0;
       Validate;
    end;
  end;

  tetButton, tetDropDownCheckList,
  tetDropDownList, tetDropdownTree,
  tetDropDown, tetMemo, tetHyperLink, tetSimple,tetPassword:
  if ClearSections then
  begin
    with AddSection do
    begin
       Case aEditorType of
       tetMemo:
         FType := tdtMultiline;

       tetDropDownCheckList, tetDropDown,tetDropdownTree:
       begin
         FType := tdtRoList;
       end;
       tetHyperLink:
         FType := tdtHyperLink;
       tetButton:
         FType := tdtButton;
       else
         FType := tdtString;
         if aEditorType = tetPassword then
            FPasswordChar := '*';
       end;
       FAlignment := taLeftJustify;
       FLSeparator:='';
       //if Fedit <> nil then
       //  Fedit.FSpacing :=0;
       Validate;
    end;
  end;
  tetPercent, tetInteger:
   //how to setup min/max Value
   if ClearSections then
   with AddSection do
   begin
       FType := tdtInteger;
       FAlignment := taRightJustify;
       FLSeparator:='';
       //if Fedit <> nil then
       //  FEdit.FSpacing :=1;
       if aEditorType = tetPercent then
         FRSeparator := '%'
       else
         FRSeparator:='';
       FMaxValue := MaxInt;
       FMinvalue := -MaxInt;
       Validate;
   end;

   tetProgress, tetValueProgress, tetPercentProgress:
   if ClearSections then
   with AddSection do
   begin
      case aEditorType of
       tetProgress: FType := tdtProgress;
       tetValueProgress: FType := tdtValueProgress;
       tetPercentProgress: FType := tdtPercentProgress;
      end;
      FMaxValue := MaxInt;
      FMinvalue := 0;
      Validate;
   end;

   tetCheckBox, tetYesNo, tetTrueFalse: //
   if ClearSections then
   with AddSection do
   begin
       Case aEditorType of
         tetCheckBox: FType := tdtCheck;
         tetYesNo:
         begin
            FType := tdtROList;
            Items.Add(nil, 'No');
            Items.Add(nil, 'Yes');
         end;
         tetTrueFalse:
         begin
            FType := tdtROList;
            Items.Add(nil, 'False');
            Items.Add(nil, 'True');
         end;
       end;
       FAlignment := taCenter;
       FLSeparator:='';
       //if FEdit <> nil then
       //  Fedit.FSpacing :=1;
       FRSeparator:='';

       Validate;
   end;
  tetIPAddress:
  begin
   if ClearSections then
   for i:=0 to 3 do
    with AddSection do
     begin
       FType := tdtInteger;
       FAlignment := taCenter;
       if i = 0 then
         FLSeparator:=''
       else FLSeparator:='.';
       FRSeparator :='';
       //if FEdit<> nil then
       // FEdit.FSpacing := 8;
       FMinvalue := 1;
       FMaxValue := 255;
       Validate;
     end;
  end;
  tetCurrency:
   for i:=0 to 1 do
   begin
       if ClearSections then AddSection;
       Case i of
        0://integer part
        with item[i] do
        begin
          FType := tdtInteger;
          FAlignment := taRightJustify;
          FLSeparator:='';
          FMaxValue := MaxInt;
          FMinvalue := -MaxInt;
          if CurrencyFormat = 0 then
           FLSeparator := CurrencyString
          else if CurrencyFormat = 2 then
           FLSeparator := CurrencyString+' '
          else
           FLSeparator := '';
          FRSeparator:=DecimalSeparator;
          Validate;
        end;
        1:
        with item[i] do
        begin
          FType := tdtLeadingZeroesInteger;
          FAlignment := taLeftJustify;
          FMinvalue := 0;
          FMaxvalue := DecR(10, CurrencyDecimals)-1;
          FLSeparator:='';
          if CurrencyFormat = 1 then
            FRSeparator := CurrencyString
          else if CurrencyFormat = 3 then
            FRSeparator := ' '+CurrencyString
          else
            FRSeparator:='';
          Validate;
        end;
       end;//case
   end;
  tetGeneralDate: ScanDateTimeString(ShortDateFormat+' '+LongTimeFormat, False, False, ClearSections);
  tetLongDate:    ScanDateTimeString(LongDateFormat, False, False, ClearSections);
  tetLongTime:    ScanDateTimeString(LongTimeFormat, False, False, ClearSections);
  tetShortDate:   ScanDateTimeString(ShortDateFormat, False, False, ClearSections);
  tetShortTime:   ScanDateTimeString(ShortTimeFormat, False, False, ClearSections);
  tetMediumDate:  ScanDateTimeString(ShortDateFormat, True, False, ClearSections);
  tetMediumTime:  ScanDateTimeString(LongTimeFormat, False, True, ClearSections);
  tetD:           ScanDateTimeString('D', False, True, ClearSections);
  tetDD:          ScanDateTimeString('DD', False, True, ClearSections);
  tetDDD:         ScanDateTimeString('DDD', False, True, ClearSections);
  tetDDDD:        ScanDateTimeString('DDDD', False, True, ClearSections);
  tetM:           ScanDateTimeString('M', False, True, ClearSections);
  tetMM:          ScanDateTimeString('MM', False, True, ClearSections);
  tetMMM:         ScanDateTimeString('MMM', False, True, ClearSections);
  tetMMMM:        ScanDateTimeString('MMMM', False, True, ClearSections);
  tetYY:          ScanDateTimeString('YY', False, True, ClearSections);
  tetYYYY:        ScanDateTimeString('YYYY', False, True, ClearSections);
  tetH:           ScanDateTimeString('H', False, True, ClearSections);
  tetHH:          ScanDateTimeString('HH', False, True, ClearSections);
  tetN:           ScanDateTimeString('N', False, True, ClearSections);
  tetNN:          ScanDateTimeString('N', False, True, ClearSections);
  tetS:           ScanDateTimeString('S', False, True, ClearSections);
  tetSS:          ScanDateTimeString('SS', False, True, ClearSections);
  tetColor:
  begin
     if ClearSections then
     with AddSection do
     begin
       FType := tdtColor;
       FAlignment := taLeftJustify;
       FLSeparator:='';
       FRSeparator :='';
       //if FEdit<> nil then
       //  Fedit.FSpacing := 0;
       Validate;
     end;
  end;

 end;//case
finally
  FUpdating :=False;
  FRebuilding := False;
end;
end;

{ TISPlugSection }
constructor TISPlugSection.Create(Collection: TCollection);
begin
   inherited Create(nil);
   {$ifdef DFS_COMPILER_3_UP}
   FSectionAdapter := nil;
   {$ENDIF}
   FitemSeparator :='.';
   FInternalValue := '';
   FPasswordChar := #0;
   FinternalIndex := -1;
   FMaxLength := 0;
   Fwidth :=0;
   FAlignment := taCenter;
   FType := tdtInteger;
   FLSeparator := '';
   FRSeparator := '';
   FDropDownColCount := 1;
   FDropDownStructureCol := 0;

   FColor := clWindow;
   FFontcolor := clBlack;
   FProgressColor := clHighlight;
   FBorderColor := clBlack;

   FShowBorder := False;
   FColInfos := nil;

   if Collection <> nil then
   with TISPlugSections(Collection) do
   begin
    if FEdit <> nil then
    begin
     FColor := Fedit.Color;
     FFontColor := Fedit.Font.Color;
     //prevent of EditorType from
     if (not FRebuilding) and (not FEdit.FLockChangeEditorType) then
      with FEdit do
      begin
       if not (csLoading in ComponentState) then
          EditorType := tetCustom;
      end;
    end;
   end;
   Self.Collection := Collection;
end;

destructor TISPlugSection.Destroy;
begin
   {$ifdef DFS_COMPILER_3_UP}
   SectionAdapter := nil;
   {$ENDIF}

  if FNodes<>nil then FNodes.Free;
  if FColInfos <> nil then
  begin
    while FColInfos.Count> 0 do
    begin
     Dispose(FColInfos[0]);
     FColInfos.Delete(0);
    end;
    FColInfos.Clear;
    FColInfos.Free;
  end;
  if Collection <> nil then
  with TISPlugSections(Collection) do
  begin
   if FEdit <> nil then    //prevent change of EditorType
    if (not FRebuilding) and (not FEdit.FLockChangeEditorType) and
      not (csLoading in FEdit.ComponentState) then
        FEdit.EditorType := tetCustom;
  end;

  inherited Destroy;
end;


function TISPlugSection.GetColWidth(Index : integer) : integer;
begin
  if FColInfos = nil then
     Result := ExGrid.DefColWidth
  else
  begin
    if (Index>=0 )and (index < FColInfos.Count) then
     Result := PColInfo(FColInfos[Index])^.Width
    else
     Result := ExGrid.DefColWidth;
  end;
end;

procedure TISPlugSection.SetColWidth(Index, Value : integer);
var aPColInfo : PColInfo;
begin
   if FColInfos = nil then
      FColInfos := TList.Create;

   while index > FColInfos.Count-1 do
   begin
      aPColInfo := new(PColInfo);
      aPColInfo.Width := ExGrid.DefColWidth;
      aPColInfo.Alignment := taLeftJustify;
      FColInfos.Add(aPColInfo);
   end;
   if (Index>=0) {and (index < FColInfos.Count)} then
      PColInfo(FColInfos[index])^.Width := Value;
end;

function TISPlugSection.GetColAlignment(Index : integer) : TAlignment;
begin
  if FColInfos = nil then
     Result := taLeftJustify
  else
  begin
    if (Index>=0)and (index < FColInfos.Count) then
     Result := PColInfo(FColInfos[Index])^.Alignment
    else
     Result := taLeftJustify;
  end;

end;

procedure TISPlugSection.SetColAlignment(Index : integer; Value : TAlignment);
var aPColInfo : PColInfo;
begin
   if FColInfos = nil then
      FColInfos := TList.Create;

   while index > FColInfos.Count-1 do
   begin
      aPColInfo := new(PColInfo);
      aPColInfo.Width := ExGrid.DefColWidth;
      aPColInfo.Alignment := taLeftJustify;
      FColInfos.Add(aPColInfo);
   end;

   if (Index>=0) {and (index < FColInfos.Count)} then
      PColInfo(FColInfos[index])^.Alignment := Value;
end;

function TISPlugSection.IsLSeparatorStored : Boolean;
begin
   Result := not (FLSeparator = '');
end;

function TISPlugSection.IsRSeparatorStored : Boolean;
begin
   Result := not (FRSeparator = '');
end;

function TISPlugSection.ISValueStored : Boolean;
begin
   Result := (FType in [tdtInteger, tdtProgress, tdtValueProgress, tdtPercentProgress, tdtLeadingZeroesInteger]);
end;

function TISPlugSection.ISMaxLengthStored : Boolean;
begin
   Result := FType = tdtString;
end;

procedure TISPlugSection.SetMinValue(Value : integer);
begin
   if (FType in [tdtInteger,tdtProgress, tdtValueProgress, tdtPercentProgress,tdtLeadingZeroesInteger]) and (Value <> FMinValue) then
   begin
      FMinvalue := Value;
      Changed(False);
   end;
   Validate;
end;

procedure TISPlugSection.SetMaxValue(Value : integer);
begin
   if (FType in [tdtInteger,tdtProgress, tdtValueProgress, tdtPercentProgress, tdtLeadingZeroesInteger]) and (Value <> FMaxValue) {and (Value > FMinValue)} then
   begin
      FMaxvalue := Value;
      //if FinternalValue <> '' then
      //   SetInternalIndex(StrToInt(FinternalValue));
      Changed(False);
   end;
   Validate;
end;

procedure TISPlugSection.SetDropDownStructureCol(const Value : integer);
begin
   if FDropDownStructureCol <> Value then
   begin
      FDropDownStructureCol := Value;
   end;
end;

procedure TISPlugSection.SetDropDownColCount(const Value : integer);
begin
   if FDropDownColCount <> Value then
   begin
      FDropDownColCount := Value;
   end;
end;

procedure TISPlugSection.SetInternalIndex(const aValue : Integer);
var i : integer;
    aNode : TFlyNode;
    aFounded : Boolean;
begin
    aFounded := False;
    if TISPlugSections(Collection).FEditorType <> tetDropdowncheckList then
    begin
       i:=FMinValue;
       aNode := Items.GetFirstNode;
       while aNode <> nil do
       begin
          if i = aValue then
          begin
            FSelected := aNode;
            TISPlugSections(Collection).FEmpty := false;
            if TISPlugSections(Collection).FEditorType <> tetDropDownTree then
              FInternalValue := aNode.Text
            else
              FInternalValue := ANode.GetPath(ItemSeparator);

            if DataType = tdtYYYY then
              FInternalIndex := strToInt(aNode.Text)
            else
              FInternalIndex := i;
            Changed(False);
            aFounded :=True;
            Break;
          end;
          inc(i);
          aNode := aNode.GetNext;
       end;
       if (not aFounded) and (FinternalIndex <> aValue) then
       begin
          FinternalIndex := aValue;
          if not ((FInternalIndex >= FMinvalue) and (FInternalIndex <= FMaxValue)) then
            FInternalIndex := FMinValue;
       end;
    end;
end;

procedure TISPlugSection.SetInternalValue(aValue : String);
 function AddZeroes(aVal : string) : string;
   var i : integer;
 begin
   Result := aVal;
   for i:=Length(aVal) to FMaxLength-1 do
     Result := Result + '0';
 end;

  var ToChange : Boolean;
begin
  if not TISPlugSections(Collection).FEmptying then
     TISPlugSections(Collection).FEmpty := False;

  if FType in [tdtInteger, tdtD, tdtDD, tdtM, tdtMM, tdtYY, tdtYYYY, tdt24H, tdt24HH, tdt12H, tdt12HH, tdtN, tdtNN, tdtS, tdtSS, tdtCheck,
               tdtProgress, tdtValueProgress, tdtPercentProgress, tdtLeadingZeroesInteger] then
  begin
   if aValue <>'' then
   begin
    if (FType = tdtLeadingZeroesInteger)  then
       aValue := AddZeroes(aValue);
    SetInternalIndex(StrToInt(aValue));
   end
   else
      SetInternalIndex(-1);
   end;
   ToChange := aValue <> FinternalValue;
   FInternalValue := aValue;
   UpdateSelNode;
   if ToChange then
   begin
     Changed(False);
   end;
end;

{$ifdef DFS_COMPILER_3_UP}
procedure TISPlugSection.SetSectionAdapter(const aValue: ISectionAdapter);
begin
  if FSectionAdapter <> nil then FSectionAdapter.ReleaseSection;
  FSectionAdapter := aValue;
  if FSectionAdapter <> nil then FSectionAdapter.ReferenceSection(Self);

end;
{$ENDIF}

procedure TISPlugSection.ItemsChanged(Item : TTreeCollection);
var aNode : TFlyNode;
begin
 if (Items.Count > 0) then
 begin
  if not (FType in [tdtD, tdtDD, tdtDDD, tdtDDDD, tdtM, tdtMM, tdtMMM, tdtMMMM,  tdtYY]) then
  begin
    aNode := Items.GetFirstNode;
    FMinValue := 0;
    FMaxValue := -1;//FMinValue;
    while aNode <> nil do
    begin
     inc(FMaxValue);
     aNode := aNode.GetNext;
    end;
    if FMaxValue =-1 then FMaxValue :=0;;
  end;
  if TIsPlugSections(Collection).FEdit <> nil then
  if  (TIsPlugSections(Collection).FEdit.EditorType = tetDropDowncheckList) then
  with TIsPlugSections(Collection) do
  begin
     SetInternalValue(Fedit.GetCheckText(Index));
  end;
   SetInternalIndex(FinternalIndex);
 end;
end;

procedure TISPlugSection.SetMaxLength(const Value : integer);
begin
   if (FType = tdtString) and (Value <> FMaxLength) then
   begin
     FMaxLength := Value;
     Changed(False);
     Validate;
   end;
end;

function TISPlugSection.FullNodeText(aNode : TFlyNode) : String;
 var aP : TFlyNode;
begin
 Result := '';
 if aNode <> nil then
 begin
     aP := aNode.Parent;
     Result := aNode.Text;
     while aP <> nil do
     begin
       Result := aP.Text+FItemSeparator+Result;
       aP := aP.Parent;
     end;
 end;
end;

procedure TISPlugSection.UpdateSelNode;
var
    aNode : TFlyNode;
    i : integer;
    Found : Boolean;
begin
 if Items.Count > 0 then
 begin
   aNode := Items.GetfirstNode;
   i :=0;
   Found := False;
   While aNode <> nil do //I <> FinternalValue do //strcomp
   begin
     if StrComp(Pchar(aNode.GetPath(Itemseparator)), PChar(FInternalValue)) =0 then
     begin
       Found:=True;
       Break;
     end;
     inc(i);
     aNode := aNode.GetNext;
   end;
   FSelected := aNode;
   if Found then
     FInternalIndex := FminValue+i
   else FInternalIndex := -1;
 end;
end;

function TISPlugSection.GetRealMaxValue : integer;
var aNode : TFlyNode;
begin
      Result := 0;
      if Items.Count > 0 then
      begin
         aNode := Items[0];
         while aNode <> nil do
         begin
           aNode := aNode.GetNext;
           Inc(Result);
         end;
      end;
end;

procedure TISPlugSection.Assign(Source: TPersistent);
var i : integer;
begin
 if Source Is TISPlugSection then
 begin
  //if Assigned(Collection) then Collection.BeginUpdate;
  //try
   FDropDownColCount := TISPlugSection(Source).FDropDownColCount;
   FDropDownStructureCol := TISPlugSection(Source).FDropDownStructureCol;
   for i:=0 to FDropDownColCount-1 do
   begin
     DropDownColwidth[i] := TISPlugSection(Source).DropDownColwidth[i];
     DropDownColAlignment[i] := TISPlugSection(Source).DropDownColAlignment[i];
   end;
   
   FMaxLength := TISPlugSection(Source).FMaxlength;
   Ftype := TISPlugSection(Source).FType;
   FWidth := TISPlugSection(Source).Width;
   FAlignment := TISPlugSection(Source).FAlignment;
//   PickList.Assign(TISPlugSection(Source).PickList);
   if FminValue > FMaxValue then
   begin //some incorrectly input values >!!!
     FMinvalue := TISPlugSection(Source).Maxvalue;
     FMaxvalue := TISPlugSection(Source).Minvalue;
   end
   else
   begin
     FMinvalue := TISPlugSection(Source).Minvalue;
     FMaxvalue := TISPlugSection(Source).Maxvalue;
   end;
   FShowBorder := TISPlugSection(Source).ShowBorder;
   FLSeparator := TISPlugSection(Source).LeftSeparator;
   FRSeparator := TISPlugSection(Source).RightSeparator;
   FReadOnly := TISPlugSection(Source).ReadOnly;
   FFontcolor := TISPlugSection(Source).Fontcolor;
   Fcolor := TISPlugSection(Source).Color;
   FInternalValue := TISPlugSection(Source).Value;
   FinternalIndex := TISPlugSection(Source).FinternalIndex;
   FProgressColor := TISPlugSection(Source).FProgressColor;
   FBorderColor := TISPlugSection(Source).FBorderColor;
   Items.Assign(TISPlugSection(Source).Items);
   Fselected :=nil;
   Validate;
   Changed(False);
  //finally
  //  if Assigned(Collection) then Collection.EndUpdate;
  //end;
 end;
end;

procedure TISPlugSection.SetPasswordChar(Value : Char);
begin
   if Value <> FPasswordChar then
   begin
     FPasswordChar := Value;
     Changed(True);
   end;

end;

procedure TISPlugSection.SetWidth(Value : Integer);
begin
   if Value <> FWidth then
   begin
     FWidth := Value;
     Changed(True);
   end;
end;

procedure TISPlugSection.SetType(const Value : TIsDataType);
begin
   if Value <> FType then
   begin
     FType := Value;
     ItemsChanged(nil);
     Validate;
     Changed(True);
   end;
end;

procedure TISPlugSection.SetItems(Value: TFlyNodes);
begin
  if Value = nil then
  begin
    FNodes.Free;
    FNodes := nil;
    Exit;
  end
  else
  begin
    FNodes.Assign(Value);
    FMinvalue :=0;
    FMaxValue:= Value.Count-1;
  end;  
  Changed(false);
end;

function TISPlugSection.GetItems: TFlyNodes;
begin
  if FNodes = nil then
  begin
    FNodes :=  TFlyNodes.Create(nil);
    FNodes.OnChange := ItemsChanged;
  end;
  Result := FNodes;
end;

procedure TISPlugSection.SetReadOnly(Value : Boolean);
begin
 if Value <> FReadOnly then
 begin
   FReadOnly := Value;
   Changed(False);
 end;
end;

procedure TISPlugSection.SetItemSeparator(const Value : String);
begin
   if Value <> FItemSeparator then
   begin
      FItemSeparator := Value;
      if FType = tdtRoList then
      begin
         if TIsPlugSections(Collection).FEdit <> nil then
         with TIsPlugSections(Collection).FEdit do
         begin
           if EditorType = tetDropDownCheckList then
              FinternalValue := GetCheckText(Index)
           else if (EditorType = tetDropdownTree) and (FSelected <> nil) then
             FinternalValue := Fselected.GetPath(ItemSeparator);
         end;
      end;
      Changed(False);
   end;
end;

procedure TISPlugSection.SetProgressColor(Value : TColor);
begin
   if Value <> FProgressColor then
   begin
     FProgressColor := Value;
     Changed(False);
   end;
end;

procedure TISPlugSection.SetBorderColor(Value : TColor);
begin
   if Value <> FBorderColor then
   begin
     FBorderColor := Value;
     Changed(False);
   end;

end;

procedure TISPlugSection.SetColor(Value : TColor);
begin
   if Value <> FColor then
   begin
     FColor := Value;
     Changed(False);
   end;
end;

procedure TISPlugSection.SetFontcolor(Value : TColor);
begin
   if Value <> FFontColor then
   begin
     FFontColor := Value;
     Changed(False);
   end;
end;

procedure TISPlugSection.SetShowBorder(Value : Boolean);
begin
   if Value <> FShowBorder then
   begin
     FShowborder := Value;
     Changed(False);
   end;
end;

procedure TISPlugSection.SetLSeparator(const Value : String);
begin
 if Value <> FLSeparator then
   begin
     FLSeparator := Value;
     Changed(True);
     
   end;
end;

procedure TISPlugSection.SetRSeparator(const Value : String);
begin
 if Value <> FRSeparator then
   begin
     FRSeparator := Value;
     Changed(True);
   end;
end;

procedure TISPlugSection.SetAlignment(const Value : TAlignment);
begin
   if Value <> FAlignment then
   begin
     FAlignment := Value;
     Changed(False);
   end;
end;

procedure TISPlugSection.SetUndoValue;
begin
   FUndoInternalValue := FInternalValue;
   FUndoInternalIndex := FInternalIndex;
end;

procedure TISPlugSection.UndoValue;
begin
   FInternalValue := FundoInternalValue;
   FInternalIndex := FUndoInternalIndex;
   //UpdateSelNode;
   TISPlugSections(Collection).Fedit.SetEditorText(index);
end;

procedure TISPlugSection.ReadColAligns(Reader: TReader);
var
  I: Integer;
begin
  with Reader do
  begin
    ReadListBegin;
    i :=0;
    while not Reader.EndOfList do
    begin
      DropDownColAlignment[I] := TAlignment(ReadInteger);
      inc(i);
    end;
    Reader.ReadListEnd;
  end;
end;

procedure TISPlugSection.ReadColWidths(Reader: TReader);
var
  I: Integer;
begin
  with Reader do
  begin
    ReadListBegin;
    i :=0;
    while not Reader.EndOfList do
    begin
      DropDownColWidth[I] := ReadInteger;
      inc(i);
    end;
    Reader.ReadListEnd;
  end;
end;

procedure TISPlugSection.WriteColWidths(Writer: TWriter);
var
  I: Integer;
begin
  with Writer do
  begin
    WriteListBegin;
    for I := 0 to DropDownColCount - 1 do
    begin
      WriteInteger(Ord(DropDownColWidth[i]));
//      Showmessage('Width'+ IntToStr(Ord(DropDownColWidth[i])));
    end;  
    WriteListEnd;
  end;
end;

procedure TISPlugSection.WriteColAligns(Writer: TWriter);
var
  I: Integer;
begin
  with Writer do
  begin
    WriteListBegin;
    for I := 0 to DropDownColCount - 1 do
    begin
      WriteInteger(Ord(DropDownColAlignment[i]));
    end;
    WriteListEnd;
  end;
end;

procedure TISPlugSection.DefineProperties(Filer: TFiler);
begin
   inherited DefineProperties(Filer);
   with Filer do
   begin
     DefineBinaryProperty('Data', ReadData, WriteData, Ftype <> tdtColor);
     DefineProperty('ColWidths', ReadColWidths, WriteColWidths, FColInfos <> nil);
     DefineProperty('ColAlignments', ReadColAligns, WriteColAligns, FColInfos <> nil);
   end;
end;

procedure TISPlugSection.ReadData(Stream: TStream);
begin
   try
     TreeColl.ReadCollectionFromStream(Items, Stream);
     Validate;
     UpdateSelNode;
   finally
     Itemschanged(nil);
   end;
end;

procedure TISPlugSection.WriteData(Stream: TStream);
begin
   TreeColl.WriteCollectionToStream(Items, Stream);
end; //

procedure TISPlugSection.Validate;
begin
   //Validate MaxLength
   case FType Of
      tdtD, tdtDD, tdtMM, tdtMMM, tdt12H, tdt12HH, tdt24H, tdt24HH, tdtN, tdtNN, tdtS, tdtSS, tdtYY:
         FMaxLength := 2;
      tdtYYYY : FMaxLength := 4;
      tdtProgress, tdtValueProgress, tdtPercentProgress, tdtInteger, {tdtString,} tdtHyperLink, tdtDDD, tdtDDDD, tdtMMMM : FMaxLength := 0;
      tdtLeadingZeroesInteger:
      begin
         FMaxLength := Length(inttoStr(FMaxValue));//new 26/04/2
      end;
   end;
   //Validate MinValue, MaxValue
   case FType Of
      tdtColor:
      if TisPlugSections(Collection).Fedit <> nil then
      begin
       TisPlugSections(Collection).Fedit.FListToAdd := Index;
       try
        Items.Clear;
        GetColorValues(TisPlugSections(Collection).FEdit.AddColors);
        FMinValue := 0; FMaxValue := Items.Count;
       finally
        TisPlugSections(Collection).Fedit.FListToAdd := -1;
       end;

      end;
      tdtAP, tdtAMPM, tdtCheck:
      begin
         FMinValue := 0; FMaxValue := 1;
      end;
      tdtButton, tdtImage, tdtImageCaption, tdtImageMemo, tdtRoList, tdtHyperLink, tdtString:
      if Items.Count > 0 then
      begin
         FMinValue := 0; FMaxValue := GetRealMaxValue;
         UpdateSelNode;
      end;
      tdtDDD, tdtDDDD:
      begin
        FMinValue := 1; FMaxValue := 7;
        FAlignment := taCenter;

      end;
      tdtD, tdtDD:
      begin
        FMinValue := 1; FMaxValue := 31; //getsystemmetrics
        FAlignment := taRightJustify;
      end;
      tdtM, tdtMM:
      begin
        FMinValue := 1; FMaxValue := 12;
        FAlignment := taRightJustify;
      end;
      tdtMMM, tdtMMMM :
      begin
        FMinValue := 1; FMaxValue := 12;
        FAlignment := taCenter;
      end;
      tdt24H, tdt24HH:
      begin
        FMinValue := 0; FMaxValue := 23;
        FAlignment := taRightJustify;
      end;
      tdt12H, tdt12HH:
      begin
        FMinValue := 1; FMaxValue := 12;
        FAlignment := taRightJustify;
      end;
      tdtN, tdtNN, tdtS, tdtSS:
      begin
        FMinValue := 0; FMaxValue := 59;
        FAlignment := taRightJustify;
      end;
      tdtYY:
      begin
        FMinValue := 2000; FMaxValue := 2099;
        FAlignment := taRightJustify;
      end;
      tdtYYYY :
      begin
        FMinValue := 1; FMaxValue := 9999;
        FAlignment := taRightJustify;
      end;
   end;
  //if not ((FInternalIndex >= FMinvalue) and (FInternalIndex <= FMaxValue)) then
  //begin
  if (FType in [tdtString, tdtMultiline,tdtROList, tdtColor, tdtImage, tdtImageCaption, tdtImageMemo, tdtHyperLink, tdtButton]) then
    FInternalIndex := -1//FMinValue;
  else // numerical values ->>
    FInternalIndex := 0;
  //end;
end;

{ TISAlignedEdit }
constructor TISAlignedEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
//  FCanvas := TControlCanvas.Create;
//  TControlCanvas(FCanvas).Control := Self;
end;

destructor TISAlignedEdit.Destroy;
begin
//  FCanvas.Free;
  inherited Destroy;
end;

procedure TISAlignedEdit.CreateParams(var Params: TCreateParams);
const
  Alignments: array[Boolean, TAlignment] of DWORD =
    ((ES_LEFT, ES_RIGHT, ES_CENTER),(ES_RIGHT, ES_LEFT, ES_CENTER));
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or  Alignments[{$ifdef DFS_COMPILER_4_UP}UseRightToLeftAlignment,{$ELSE}False,{$ENDIF} FAlignment];
    if FMultiline then
       Style := Style or ES_MULTILINE;// or WS_VSCROLL;
  end;
end;

procedure TISAlignedEdit.WMNCCalcSize (var Message: TWMNCCalcSize);
begin
  inherited;
  if FCustomDraw then
  with Message do
  begin
    if (BorderStyle<>BSNone) then
      inc(CalcSize_Params^.rgrc[0].right,2);
    CalcSize_Params^.rgrc[0].Right := CalcSize_Params^.rgrc[0].Left;
    CalcValidRects:=true;
 end;
end;

procedure TISAlignedEdit.WMNCPaint(var Message: TMessage);
begin
  if FCustomDraw then
     DrawEditor;
  inherited;
end;

procedure TISPlugSections.Drawimage(aCanvas : TCanvas; aImages : {$IFDEF DFS_COMPILER_4_UP}TCustomImagelist {$ELSE} TImagelist {$ENDIF};
          Index : integer; Var ImageRect: TRect; RTL : Boolean);
   var Pt      : TPoint;
   aBitmap : TBitmap;
   XOffset : integer;
   MaskDC :  HDC;
   Save: THandle;
begin
  if Index < 0 then exit;
  aBitmap := TBitmap.Create;
  try
   //with TISCustomPlugedit(Parent) do
     aImages.Getbitmap(Index, aBitmap);

   with ImageRect do
   begin
      if RTL then
        Pt.X:= Right-aBitmap.Width-3
      else
      begin
        Pt.X:=left+3;
      end;

     XOffset:= aBitmap.Width+6;
     Pt.Y:=Top + (((Bottom - Top) - aBitmap.Height) div 2);
   end;

   //with TISCustomPlugedit(Parent) do
   if (Index<>-1) or (index < aImages.Count) then
   begin
      aBitmap.Transparent:=true;
      aBitmap.TransparentColor:= aBitmap.Canvas.Pixels[aBitmap.Width-1,aBitmap.height-1];//Canvas.Pixels[Pt.X,Pt.y];//
      aBitmap.TransparentMode := tmAuto;
      Save := 0;
      MaskDC := 0;
      try
        MaskDC := CreateCompatibleDC(0);
        Save := SelectObject(MaskDC, aBitmap.MaskHandle);
        TransparentStretchBlt(aCanvas.Handle, Pt.X, Pt.Y, aBitmap.Height,
          aBitmap.Height, aBitmap.Canvas.Handle, 0, 0, aBitmap.Height,
          aBitmap.Height, MaskDC, 0, 0);

        //aCanvas.StretchDraw(ImageRect, aBitmap);
      finally
        if Save <> 0 then SelectObject(MaskDC, Save);
        if MaskDC <> 0 then DeleteDC(MaskDC);
      end;
      if RTL then
        Dec(ImageRect.right,Xoffset)
      else
        Inc(ImageRect.Left,Xoffset);
    end;
 finally
    aBitmap.Free
 end;
end;

function TISPlugSections.DrawColor(aCanvas : TCanvas; aColor : Longint; R : TRect) : TRect;
var aOldColor : TColor;
begin
 with aCanvas do
 begin
   FillRect(R);
   Result := Rect(R.Left, R.Top, R.Left+(R.Bottom-R.Top), R.Bottom);
   InflateRect(Result,-1,-1);
   aOldColor := Brush.Color;
   Brush.Color := clBlack;
   FrameRect(Result);

   Brush.Color := clWindow;
   inflateRect(Result,-1,-1);
   FrameRect(Result);
   inflateRect(Result,-1,-1);

   Brush.Color := TColor(aColor);
   FillRect(Result);
   Brush.Color := aOldColor;
 end;
end;


procedure TISPlugSections.DrawEditorImage(aCanvas : TCanvas; aValue : String;
  R : TRect; aIndex : Integer; Focused, aCtl3D, RTL, UseValue, AnotherCanvas : Boolean;
  aImages : {$IFDEF DFS_COMPILER_4_UP}TCustomImagelist {$ELSE} TImagelist {$ENDIF});

  function isInteger(aValue : string) : boolean;
  begin
      try
       strtoInt(aValue);
       Result:=true;
      except
        Result:=false;
      end;
  end;
var
  X, aFlags : integer;
  BoxRect : TRect;
  aNode : TFlyNode;
  CurVal : Double;
  StrVal : string;
  aBmp : TBitmap;
  Oldcolor : TColorRef;
  OldBkMode : integer;
const
   RTLFlags: array [Boolean] of Integer = (DT_LEFT, DT_RIGHT);//(0, DT_RTLREADING);
begin
with aCanvas do
begin
    aNode := nil;
    if not AnotherCanvas then
    begin
      if Item[aIndex].FShowborder then //
      begin
       Brush.Color := Item[aIndex].FBorderColor;
       FrameRect(R);
       InflateRect(R,-1,-1);
      end;
      Brush.Color := Item[aIndex].Color;
      FillRect(R);  //
    end;

    case Item[aIndex].Ftype of
      tdtColor:
      begin
        with Item[aIndex] do
        begin
         if UseValue then
         begin
           if aValue <>'' then
           begin
            BoxRect := DrawColor(aCanvas, Longint(stringToColor(aValue)), R);
            inc(R.Left,(R.Bottom-R.Top)+4);
           end;
         end
         else
         begin
          if (FSelected <> nil) and ((FSelected.Data<>nil) or (FSelected.ImageIndex >= 0) )then //25/05/2001
          begin
            if Fselected.Data <> nil then
              BoxRect := DrawColor(aCanvas, Longint(Fselected.Data), R)
            else BoxRect := DrawColor(aCanvas, Longint(Fselected.ImageIndex), R);
          end
          else
           if Value <>'' then
            BoxRect := DrawColor(aCanvas, Longint(stringToColor(Value)), R)
           else
            BoxRect := DrawColor(aCanvas, Longint(clNone), R);
          inc(R.Left,(R.Bottom-R.Top)+4);
         end;

         //aCanvas.Font := Self.Font;

         //if Focused and (not TISCustomPlugedit(Parent).FDropDownvisible) then
         if Focused  then
         begin
           Brush.Color := clHighLight;
           Font.Color := clHighLightText;
         end
         else if not AnotherCanvas then
         begin
           Brush.Color := Color;
           Font.Color := FontColor;
         end;
         InflateRect(R,-1,-1);
         FillRect(R);
         InflateRect(R,-1,-1);
         if UseValue then
           DrawText(aCanvas.Handle, PChar(aValue), -1, R, DT_END_ELLIPSIS+DT_SINGLELINE+DT_VCENTER or RTLFlags[RTL])
         else
         begin
          if FSelected <> nil then
           DrawText(aCanvas.Handle, PChar(FullnodeText(Fselected)), -1, R, DT_END_ELLIPSIS+DT_SINGLELINE+DT_VCENTER or RTLFlags[RTL])
          else
           DrawText(aCanvas.Handle, PChar(Value), -1, R, DT_END_ELLIPSIS+DT_SINGLELINE+DT_VCENTER or RTLFlags[RTL]);
         end;
         //BoxRect:= R;
        end;
      end;
      tdtROList:
      begin
        with Item[aIndex] do
        begin
         //aCanvas.Font := Self.Font;
         if Focused then
         begin
           Brush.Color := clHighLight;
           Font.Color := clHighLightText;
         end
         else
         begin
           Brush.Color := Color;
           Font.Color := FontColor;
         end;

         FillRect(R);
         InflateRect(R,-1,-1);

         //if FEdit.EditorType = tetDropdownCheckList then
            DrawText(aCanvas.Handle, PChar(FInternalValue), -1, R, DT_END_ELLIPSIS+DT_SINGLELINE+DT_VCENTER or RTLFlags[RTL])
         //else
         // if FSelected <> nil then
         //   DrawText(aCanvas.Handle, PChar(FInternalValue), -1, R, DT_END_ELLIPSIS+DT_SINGLELINE+DT_VCENTER or RTLFlags[RTL]);

         //BoxRect:= R;
        end;
      end;
      tdtCheck:
      with Item[aIndex] do
      begin
        if not aCtl3d then
           aFlags := DFCS_FLAT
        else aFlags := 0;

        FillRect(R);
        InflateRect(R,-1,-1);
        if UseValue then
        begin
           if (aValue ='') or (aValue ='0') or (aValue ='False') then
             DrawFrameControl(aCanvas.Handle, R, DFC_BUTTON, DFCS_BUTTONCHECK+aFlags)
           else if (aValue='-1') or (aValue ='1') or (aValue ='True') then
             DrawFrameControl(aCanvas.Handle, R, DFC_BUTTON, DFCS_BUTTONCHECK+ (DFCS_CHECKED)+aFlags)
           else
             DrawFrameControl(aCanvas.Handle, R, DFC_BUTTON, DFCS_BUTTONCHECK+ (DFCS_INACTIVE)+aFlags)
        end
        else
          DrawFrameControl(aCanvas.Handle, R, DFC_BUTTON, DFCS_BUTTONCHECK+ (DFCS_CHECKED*FinternalIndex)+aFlags);
      end;
      tdtProgress, tdtValueProgress, tdtPercentProgress:
      with Item[aIndex],aCanvas do
      begin
        CurVal := 0;
        if UseValue then
        begin
           if (aValue = '') or (not isInteger(aValue)) then aValue := '0';
           if (FMaxValue - FMinValue) > 0 then
             CurVal := ((StrToInt(aValue) - FMinValue)/(FMaxValue - FMinValue));
        end
        else
          if (FInternalIndex - FMinValue) > 0 then
           CurVal := ((StrToInt(Value) - FMinValue)/(FMaxValue - FMinValue));

        BoxRect := R;
        if Ftype = tdtPercentProgress then
          StrVal := InttoStr(Trunc(CurVal*100))+'%'
        else
         if UseValue then
           StrVal := aValue
         else
           StrVal := Value;

        if CurVal >0 then
        begin
         aBmp := TBitmap.Create;
         aBmp.Width := R.Right-R.Left;
         aBmp.Height := R.Bottom-r.Top;
         try
           BoxRect := Rect(0,0,aBmp.Width,aBmp.Height);
           aBmp.Canvas.Brush := aCanvas.Brush;
           aBmp.Canvas.FillRect(BoxRect);
           aBmp.Canvas.Pen.Style := psSolid;  //createpen
           //SetTextColor(aBmp.Canvas.Handle, ColortoRGB(FontColor));
           if Ftype <> tdtProgress then
           begin
             aBmp.Canvas.Font := aCanvas.Font;
             aBmp.Canvas.Font.Color := FontColor;
             SetBkMode(aBmp.Canvas.Handle, TRANSPARENT);//getnearestcolor
             DrawText(aBmp.Canvas.Handle, PChar(StrVal), -1, BoxRect, DT_SINGLELINE or DT_CENTER or DT_VCENTER);
           end;
           aBmp.Canvas.Pen.Mode := pmNotXor;
           aBmp.Canvas.Pen.Color := FProgressColor;
           aBmp.Canvas.Pen.Width := Trunc(aBmp.Width*CurVal);
           X := {R.Left+}((aBmp.Canvas.Pen.Width div 2)-1);
           if Odd(aBmp.Canvas.Pen.Width) then
             Dec(X);
           aBmp.Canvas.MoveTo(X, BoxRect.Top);
           aBmp.Canvas.LineTo(X, BoxRect.Bottom);
           aCanvas.CopyRect(R, aBmp.Canvas, BoxRect);
         finally
           aBmp.Free;
         end;
        end;

        inflateRect(R,-1,-1);

      end;
      tdtButton:
      with Item[aIndex],aCanvas do
      begin
        //if not aCtl3d then
        //  aFlags := DFCS_FLAT
        //else
        aFlags := 0;
        if FPressed then
          aFlags := aFlags + DFCS_PUSHED;
        DrawFrameControl(Handle, R, DFC_BUTTON, DFCS_BUTTONPUSH+DFCS_ADJUSTRECT	+aFlags);

        Oldcolor := SetTextColor(aCanvas.Handle, ColortoRGB(FontColor));
        OldBkMode := SetBkMode(aCanvas.Handle, TRANSPARENT);//getnearestcolor
        try
         StrVal := Value;
         if (aImages <> nil) and (aNode <> nil) then
         begin
           BoxRect := R;
           InflateRect(boxRect,-1,-1);
           if aImages <> nil then
           begin
            if (aNode.ImageIndex>=0) and (aNode.ImageIndex < aImages.Count) then
             Drawimage(aCanvas, aImages, aNode.ImageIndex, BoxRect, RTL);
             if RTL then
               R.Right:= BoxRect.Right
             else
              R.Left:= BoxRect.Left;
           end;
           DrawText(aCanvas.Handle, PChar(FullnodeText(aNode)), -1, R, DT_END_ELLIPSIS+DT_SINGLELINE+DT_VCENTER or RTLFlags[RTL]);
         end
         else
           DrawText(aCanvas.Handle, PChar(StrVal), -1, R, DT_SINGLELINE or DT_CENTER or DT_VCENTER);
        finally
          //restore
          SetTextColor(aCanvas.Handle, Oldcolor);
          SetBkMode(aCanvas.Handle, OldBkMode);
        end;
       end;

      tdtImage,
      tdtImageCaption,tdtImageMemo:
      with Item[aIndex] do
      begin
       if UseValue then
       begin
         try
         if aValue <> '' then
            aNode := TFLyNode(Items.GetItemOnPath(aValue, ItemSeparator));
         except
         end
       end
       else if FSelected = nil then
         aNode := TFLyNode(Items.GetItemOnPath(Value, ItemSeparator))
       else
         aNode := FSelected;

      if aNode <> nil then
      begin
       BoxRect := R;
       InflateRect(boxRect,-1,-1);
       if aImages <> nil then
       begin
        if (aNode.ImageIndex>=0) and (aNode.ImageIndex < aImages.Count) then
         Drawimage(aCanvas, aImages, aNode.ImageIndex, BoxRect, RTL);

         if RTL then
           R.Right:= BoxRect.Right
         else
           R.Left:= BoxRect.Left;
       end;

       //aCanvas.Font := Self.Font;
       if Ftype in [tdtImageCaption,tdtImageMemo] then
       begin
         if Focused  then
         begin
           Brush.Color := clHighLight;
           Font.Color := clHighLightText;
         end
         else if not AnotherCanvas then
         begin
           Brush.Color := Item[aIndex].Color;
           Font.Color := FontColor;
         end;

         //R.Left := BoxRect.Left;
         //if (Items.Count > 0) and (FinternalValue < Items.Count) then
          FillRect(R);
          InflateRect(R,-1,-1);
          //InflateRect(boxRect,-1,-1);
          if Ftype  = tdtImageMemo then
            DrawText(aCanvas.Handle, PChar(FullnodeText(aNode)), -1, R, DT_END_ELLIPSIS+DT_LEFT+DT_WORDBREAK or RTLFlags[RTL])
          else
            DrawText(aCanvas.Handle, PChar(FullnodeText(aNode)), -1, R, DT_END_ELLIPSIS+DT_SINGLELINE+DT_VCENTER or RTLFlags[RTL]);
         end;
        end;
      end;
    end;
    if Focused then
    begin
      inflateRect(R,1,1);
      DrawFocusRect(R); //DrawFocusRect(FR);
    end;
 end;
end;

procedure TISAlignedEdit.DrawEditor;
var
  DC: HDC;
  r : TRect;
  Bmp : TBitmap;
begin
  inherited;
  if not (HandleAllocated and IsWindowVisible(Handle)) then exit;
  //double buffered to avoid flickering 
  R := Rect(0,0, Width, Height);
  Bmp := TBitmap.Create;
  Bmp.Width := Width;
  Bmp.Height := Height;
  DC := GetWindowDC(Handle);
  //FCanvas.Handle := DC;
  with TISCustomPlugedit(Parent) do
  try
     FSections.DrawEditorImage(Bmp.Canvas,  '', R, FIndex,  Self.Focused and (not FDropDownvisible), Ctl3D, False, False,
       {$ifdef DFS_DELPHI_4_UP}UseRightToLeftAlignment{$else}False{$endif}, FImages);
     BitBlt(DC, 0, 0, Width, Height, Bmp.Canvas.handle, 0, 0, SRCCOPY);
  finally
    //FCanvas.Handle := 0;
    ReleaseDC(Handle, DC);
    Bmp.Free;
  end;

end;

procedure TISAlignedEdit.WMPaint(var Msg : TWMPaint);
var
  DC : HDC;
  NewPen,OldPen : HPEN;
begin
  inherited;
  if (Parent <> nil) then
  with TISCustomPlugEdit(Parent).FSections[FIndex] do
   if FShowborder then //and (FBordercolor <> clBlack) then
   begin
      DC := GetWindowDC(Handle);
      NewPen := CreatePen(PS_SOLID, 1, ColorToRGB(FBorderColor));
      OldPen := SelectObject(DC, NewPen);
      try
        MoveToEx(DC, 0, 0, nil);
        LineTo(DC, ClientWidth-1, 0);
        MoveToEx(DC, ClientWidth-1, 0, nil);
        LineTo(DC, ClientWidth-1, ClientHeight-1);
        MoveToEx(DC, 0, 0, nil);
        LineTo(DC, 0, ClientHeight-1);
        MoveToEx(DC, 0, clientHeight-1, nil);
        LineTo(DC, clientWidth, clientHeight-1);
      finally
        SelectObject(DC, OldPen);
        DeleteObject(NewPen);
        ReleaseDC(Handle, DC);
      end;
   end;

end;


procedure TISAlignedEdit.WMKillFocus(var Msg: TWMKillFocus);
begin
 if not TISCustomPlugEdit(Parent).IsInplace then
 begin
  if (Parent <> nil) and HandleAllocated then
   if not TISCustomPlugEdit(Parent).FClearing then
   with TISCustomPlugEdit(Parent) do
   begin
    if FButtonType <> tbtNone then
      DrawButton;
    if FCustomdraw then
      DrawEditor;

{    if FDropDownVisible and (Msg.FocusedWnd<>0) then
      if not FDropDown.Contain(Msg.FocusedWnd) then //<> FDropDown.Handle then
         CloseUp(False);

      else
        Windows.SetFocus(Msg.FocusedWnd);
       Msg.Result:=0;
}       
   end;
end;
   inherited;
end;


procedure TISAlignedEdit.WMSetFocus(var Message: TWMSetFocus);
begin
  if (Parent <> nil) and HandleAllocated then
  begin
    TISCustomPlugEdit(Parent).FFocusedEditor := Self;
    if (Left < 0 ) or ((Left+Width) > TISCustomPlugEdit(Parent).ClientWidth) then
    begin
      TISCustomPlugEdit(Parent).ShiftControls(Self);
    end;
    if FCustomdraw then DrawEditor;
  end;
  Inherited;
end;

procedure TISAlignedEdit.WMSetCursor(var Message: TWMSetCursor);
begin
 with TISCustomPlugEdit(Parent), Message do
 if FSections[Findex].FType in [tdtROList, tdtColor, tdtButton, tdtCheck, tdtImage, tdtImageCaption, tdtImageMemo, tdtProgress, tdtValueProgress, tdtPercentProgress] then
   Windows.SetCursor(LoadCursor(0, IDC_ARROW))
 else if FSections[Findex].FType = tdtHyperLink then
   Windows.SetCursor(LoadCursor(0, IDC_HAND))
 else inherited;
end;

procedure TISAlignedEdit.WMNCHittest(var Message: TWMNCHitTest);
begin
 with TISCustomPlugEdit(Parent), Message do
 if FSections[Findex].FType in [tdtROList, tdtColor, tdtButton, tdtCheck, tdtImage, tdtImageCaption, tdtImageMemo, tdtProgress, tdtValueProgress, tdtPercentProgress] then
  Result := HTCLIENT
 else inherited;
end;

procedure TISAlignedEdit.WndProc(var Message: TMessage);
const
  AltMask = $20000000;
begin
  case Message.Msg of
    WM_GETDLGCODE:
      with TMessage(Message) do
      if (TISCustomPlugEdit(Parent).DoTabKey) and (not TISCustomPlugEdit(Parent).FDropDownvisible) then
      begin
         Message.Result := Message.Result or DLGC_WANTTAB or DLGC_WANTARROWS;
         Exit;
      end;
    wm_KeyDown, wm_SysKeyDown://,wm_Char:
     with TWMKey(Message) do //WM_GETDLGCODE
      if (CharCode <> 0) and TISCustomPlugEdit(Parent).FDropdownvisible then
      begin
       with TMessage(Message) do
        if (CharCode = VK_UP) and ((KeyData and AltMask) <> 0) then
          PostMessage(TISCustomPlugEdit(Parent).FDropdown.FContainedControl.Handle, CM_CANCELMODE, 0, 0)
        else if (CharCode in [VK_TAB, VK_RETURN, VK_ESCAPE]) then
          SendMessage(TISCustomPlugEdit(Parent).Handle, Msg, WParam, LParam)
        else
          SendMessage(TISCustomPlugEdit(Parent).FDropdown.FContainedControl.Handle, Msg, WParam, LParam);
        Exit;
       end
       else if (not (CharCode in [VK_LEFT, VK_RIGHT, VK_DELETE,  VK_HOME, VK_END])) and
        not ((GetKeyState(VK_CONTROL) < 0) or (GetKeyState(VK_SHIFT) < 0) )
        and not ((CharCode = VK_RETURN) and (TISCustomPlugEdit(Parent).FSections[FIndex].FType = tdtButton)) then
       begin
        try
        with TMessage(Message) do
          SendMessage(TISCustomPlugEdit(Parent).Handle, Msg, WParam, LParam);//first
         Exit;
        except
          inherited WndProc(Message);
        end;
      end;
  end;
  inherited WndProc(Message);
end;

procedure TISAlignedEdit.DblClick;
begin
  inherited;
  with TISCustomPlugEdit(Parent).FSections[FIndex] do
  if {(Items.Count > 1) and} (FType <> tdtCheck) then
  begin
    TISCustomPlugEdit(Parent).ProceedButton(pbUpButton); 
  end;
end;

procedure TISAlignedEdit.Click;
begin
  inherited;
  if TISCustomPlugEdit(Parent).FSections[FIndex].Ftype = tdtHyperLink then
  begin
      if Trim(TISCustomPlugEdit(Parent).FSections[FIndex].Value) <> '' then
       ShellExecute(0, 'OPEN', PChar(TISCustomPlugEdit(Parent).FSections[FIndex].Value), nil, nil, SW_SHOWMAXIMIZED);
  end;
end;

procedure TISAlignedEdit.MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited;
  with TISCustomPlugEdit(Parent).FSections[Findex] do
  if (FType = tdtButton) and FPressed then
  begin
     if FPressed then
       TISCustomPlugEdit(Parent).ButtonPressed(pbSectionButton);
     FPressed := False;
     MouseCapture := False;
     DrawEditor;
  end;
end;

procedure TISAlignedEdit.EmulateMouseUp;
begin
  MouseUp(mbLeft, [], 1,1);
end;

procedure TISAlignedEdit.EmulateMouseDown;
begin
  MouseDown(mbLeft, [], 1,1);
end;

procedure TISAlignedEdit.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
   inherited;
   with TISCustomPlugEdit(Parent).FSections[Findex] do
   begin
    case fType of
      tdtCheck:
       TISCustomPlugEdit(Parent).ProceedButton(pbDownButton);
      tdtButton:
      begin
       FPressed := True;
       MouseCapture := true;
       DrawEditor;
      end;
      tdtProgress, tdtValueProgress, tdtPercentProgress:
      begin
        FinternalIndex := Trunc( ((FMaxValue - FMinValue) /Self.Width) * X);
        TISCustomPlugEdit(Parent).ProceedButton(pbProgress);
      end;
    end;
  end;
end;

procedure TISAlignedEdit.SetMultiline(Value : Boolean);
begin
   if Value <> FMultiline then
   begin
     FMultiline := Value;
     RecreateWnd;
   end;
end;

procedure TISAlignedEdit.SetAlignment(Value: TAlignment);
begin
  if FAlignment <> Value then
  begin
    FAlignment := Value;
    RecreateWnd;
  end;
end;

{procedure TISAlignedEdit.WMSetText(var Message: TMessage);
var Key : Char;
begin
 Key :=#0;
 Message.Result := Ord(TISCustomPlugEdit(Parent).ValidateInput(Self, Key));
end;}

procedure TISAlignedEdit.WMPaste(var Message: TMessage);
begin
  inherited;
  if Parent <> nil then
   with TISCustomPlugEdit(Parent) do
   begin
      FSections[FIndex].Value := Self.Text;
      Modified:=true;
   end;
end;

procedure TISAlignedEdit.KeyPress(var Key: Char);
begin
  if Parent <> nil then
  with TISCustomPlugEdit(Parent) do
  begin
     FSections[FIndex].SetUndoValue;
     if not ValidateInput(Self, Key) then
     begin
      Fsections[FIndex].UndoValue;
      Key := #0;
     end
     else
       inherited;
  end;
end;

type
  TSelection = record
    StartPos, EndPos: Integer;
  end;


procedure TISAlignedEdit.KeyDown(var Key: Word; Shift: TShiftState);
var //aPos : integer;
    aKey : Char;

  function Selection: TSelection;
  begin
    SendMessage(Handle, EM_GETSEL, Longint(@Result.StartPos), Longint(@Result.EndPos));
  end;

  function RightSide: Boolean;
  begin
    if FCustomDraw then
      result := true
    else
     with Selection do
      Result := ((StartPos = GetTextLen) and (StartPos = EndPos));
      //Result := (not FCustomDraw) and (((StartPos = 0) or (EndPos = StartPos)) and
      //  (EndPos = GetTextLen));
   end;

  function LeftSide: Boolean;
  begin
    if FCustomDraw then
     Result := true
    else
    with Selection do
      Result := ((StartPos = 0) and (StartPos = EndPos));//and (EndPos = 0));// or (EndPos = GetTextLen)));
  end;
//  TMoveFocus = (tmfFirst, tmfLast, tmfNext, tmfPrev);
begin
 if Parent<>nil then
 begin
   if (not TISCustomPlugEdit(Parent).FDropDownVisible) then
   begin
{     aPos := LongRec(SendMessage(Handle, EM_GETSEL, 0, 0)).Hi;
     if aPos = 0 then
       FFirstPos := True
     else FFirstPos := False;
     if aPos = Length(Text) then
       FLastPos := True
     else FLastPos := False;
}
     Case Key Of
       VK_RETURN:
       begin
        if TISCustomPlugEdit(Parent).FSections[FIndex].Ftype = tdtButton then
            EmulateMouseDown
        else if (TISCustomPlugEdit(Parent).ButtonType = tbtEllipsis) and (ssCtrl in Shift) then
           TISCustomPlugEdit(Parent).ButtonPressed(pbEllipsis);
       end;
       VK_LEFT:
       if LeftSide then
       begin
         if (Findex > 0) then
         //       -   
         //    -     
           TISCustomPlugEdit(Parent).FocusEditor(tmfPrev)
         else //if LeftSide then
         begin //       
           TISCustomPlugEdit(Parent).KeyDown(Key,Shift);
           exit;
         end;
       end;
       VK_RIGHT:
       if RightSide then
       begin
         if (Findex < TISCustomPlugEdit(Parent).FSections.Count-1) then
           TISCustomPlugEdit(Parent).FocusEditor(tmfNext)
         else //if RightSide then
         begin
           TISCustomPlugEdit(Parent).KeyDown(Key,Shift);
           exit;
         end;
       end;
     end;
     inherited;
   end;
   if Key = VK_DELETE then
   begin
      aKey := #0;
      FdeletePressed := True;
      KeyPress(aKey);
      TISCustomPlugEdit(Parent).Modified := true;
      FdeletePressed := False;
   end;

//   if not ((Key in [VK_LEFT, VK_RIGHT]) and (not ReadOnly)) then
//      TISCustomPlugEdit(Parent).Keydown(Key, Shift);
  end;
end;


procedure TISAlignedEdit.KeyUp(var Key: Word; Shift: TShiftState);
begin
   if Parent<>nil then
   begin
      TISCustomPlugEdit(Parent).KeyUp(Key, Shift);
      if not TISCustomPlugEdit(Parent).FDropDownVisible then
        inherited;
      if TISCustomPlugEdit(Parent).FSections[Findex].FPressed then
        EmulateMouseUp;
   end
   else
     inherited;
end;

{$ifdef DFS_COMPILER_4_UP}
procedure TISAlignedEdit.WMTimer(var Msg: TWMTimer);
begin
     Killtimer(Handle, 1);
     TISCustomPlugEdit(Parent).DrawButton;
end;


function TISAlignedEdit.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean;
var Key : Word;
    Plug : TISCustomPlugEdit;
begin
  Result := inherited DoMouseWheelDown(Shift, MousePos);
  Plug := TISCustomPlugEdit(Parent);
  if not Result then
  with Plug do
  if ButtonType in [tbtSpin, tbtScroll] then
  begin
    Key := VK_DOWN;
    Plug.Keydown(Key, Shift);
    Plug.FDown:=false;
    Killtimer(Plug.Handle, 1);
    SetTimer(Plug.Handle, 1, RepeatPause, nil);
    Result:=True;
  end
  else if (ButtonType = tbtDropDown) and FDropDownVisible then
  begin
     Result := TCrackWinCtl(FDropDown.FContainedControl).DoMouseWheelDown(Shift, MousePos);
  end;
end;

function TISAlignedEdit.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean;
var Key : Word;
    Plug : TISCustomPlugEdit;
begin
  Result := inherited DoMouseWheelUp(Shift, MousePos);
  Plug := TISCustomPlugEdit(Parent);
  if not Result then
  with Plug do
  begin
    if ButtonType in [tbtSpin, tbtScroll] then
    begin
      Key := VK_UP;
      Plug.Keydown(Key, Shift);
      FUp:=false;
      Killtimer(Plug.Handle, 1);
      SetTimer(Plug.Handle, 1, RepeatPause, nil);
      Result:=True;
    end
    else if (ButtonType = tbtDropDown) and FDropDownVisible then
    begin
       Result := TCrackWinCtl(FDropDown.FContainedControl).DoMouseWheelUp(Shift, MousePos);
    end;
  end;
end;
{$endif}

{ TISCustomPlugEdit }

constructor TISCustomPlugEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csAcceptsControls, csSetCaption]+[csReflector, csReplicatable];
  inherited Alignment := taLeftJustify;
  IsInplace := false;
  FDropDownRows := 8;
  //FSizeGrip := True;
  FReadOnly := False;
  FAutoComplete :=True;
  FAutoDropdown := True;
  BevelInner := bvNone;
  BevelOuter := bvNone;
  FEditorsList := TList.Create;
  BorderStyle := bsSingle;
  FSpacing := 1;
  Feditortype := tetSimple;
  FButtonType := tbtNone;
  FButtonWidth := 15;
  Color := clWindow;
  FautoSize := False;
  FImages := nil;
  FDropDown := nil;
  FSections := TISPlugSections.Create;//(Self);
  FSections.Fedit := Self;
  FSections.Fowner :=Self;
end;

destructor TISCustomPlugEdit.Destroy;
begin
  FSections.Free;//it must be before -> EditorsList.Free
  if FDropDown <> nil then
     FDropDown.Free;
  FEditorsList.Free;

  inherited Destroy;
end;

procedure TISCustomPlugEdit.SetSections(Value : TISPlugSections);
begin
 if Value <> nil then
  FSections.Assign(Value)
 else
  FSections.Clear;
end;

procedure TISCustomPlugEdit.Loaded;
begin
  inherited Loaded;
  //RebuildSections(false);
  RearrangeSections;

end;


procedure TISCustomPlugEdit.CreateWnd;
begin
   inherited CreateWnd;
end;

function TISCustomPlugEdit.DropdownCtl : TISDropDown;
begin
   if FDropdownVisible then
      Result := FDropdown
   else Result := nil;
end;

function TISCustomPlugEdit.DropdownVisible : Boolean;
begin
   Result := FDropdownVisible;
end;

procedure TISCustomPlugEdit.Notification(AComponent: TComponent;
Operation: TOperation);
begin
   inherited Notification(AComponent, Operation);
   if Operation = opRemove then
   begin
      if AComponent = FImages then
        FImages := nil;
   end;
end;

procedure TISCustomPlugEdit.SetImages(Value: {$IFDEF DFS_COMPILER_4_UP} TCustomImageList {$ELSE} TImagelist {$ENDIF});
begin
   FImages := Value;
{   if FImages <> nil then
   begin
      if (FDropDown<>nil) then
           FDropDown.Images:=Fimages;
   end;
}   
end;

procedure TISCustomPlugEdit.SetAutoSize(Value : Boolean);
begin
   if FAutoSize <> Value then
   begin
      FAutoSize := Value;
      if Value then
         RearrangeSections;
   end;
end;


procedure TISCustomPlugEdit.WMNCPaint(var Message: TMessage);
begin
   if (FButtonType <> tbtNone) and (FButtonWidth > 0) then
   begin
      DrawButton;
   end;
   inherited;
end;

procedure TISCustomPlugEdit.WMNCHittest (var Message: TMessage);
var R,Br : TRect;
begin
  inherited;
  if (FButtonType <> tbtNone) and (FButtonWidth > 0) then
  begin
   GetwindowRect(Handle,R);
   Br:=BtnRect;
   offsetRect(Br,r.left,r.top);
   With Message do
   begin
    if Ptinrect(Br, Point(loword(lparam),hiword(lparam))) then
     result:=HTBORDER;
    if FButtonType = tbtScroll then
    begin
     Br:=MinusBtnRect;
     offsetRect(Br,r.left,r.top);
     if Ptinrect(Br, Point(loword(lparam),hiword(lparam))) then
      result:=HTBORDER;
    end;
  end;
 end;

end;

procedure TISCustomPlugEdit.WMNCCalcSize (var Message: TWMNCCalcSize);
begin
   inherited;
   if (FButtonType <> tbtNone) and (FButtonWidth > 0) then
   begin
     {$ifdef DFS_COMPILER_4_UP}
     if UseRightToLeftAlignment then
     begin
      if FButtonType = tbtScroll then
       Dec(Message.CalcSize_Params^.rgrc[0].Right, FButtonWidth);
       inc(Message.CalcSize_Params^.rgrc[0].left, FButtonWidth);
     end
     else
     {$endif}
       if FButtonType = tbtScroll then
         Inc(Message.CalcSize_Params^.rgrc[0].Left, FButtonWidth);
       Dec(Message.CalcSize_Params^.rgrc[0].Right, FButtonWidth);

     Message.CalcValidRects:=true;
   end;
end;

procedure TISCustomPlugEdit.CMFontChanged(var Message);
begin
  inherited;
  if HandleAllocated then Perform(WM_SIZE, 0, 0);
end;


procedure TISCustomPlugEdit.CMCloseUp(var Message: TMessage);
begin
  MakeCloseUp(Boolean(Message.WParam));
end;

procedure TISCustomPlugEdit.CMCancelMode(var Message: TCMCancelMode);
begin//WM_CancelMode
  inherited;
  if (Message.Sender <> Self) and (FDropDownVisible) and (Message.Sender <> FDropDown) then
    CloseUp(False);
end;

procedure TISCustomPlugEdit.CMCtl3DChanged(var Message: TMessage);
begin
  inherited;
  Invalidate;
  ReArrangeSections;
end;

procedure TISCustomPlugEdit.WMTimer(var Msg: TWMTimer);
begin
   if FInit then
   begin
      FInit :=False;
      Killtimer(Handle, 1);
      if FUp or FDown then
        SetTimer(Handle, 1, RepeatPause, nil);
   end;

   //check global mouse
{   if not FMousedown then
   begin
     FUp := false;
     FDown := false;
   end;
}
   if Not FUp and not FDown or (FButtonType in [tbtDropdown, tbtEllipsis]) then
   begin
     Killtimer(Handle, 1);
     DrawButton;//Exit;
   end
   else if FUp then //activecontrol
     ButtonPressed(pbUpButton)
   else if FDown then
     ButtonPressed(pbDownButton);
end;


procedure TISCustomPlugEdit.ProceedButton(aButton : TPressedButtons);
begin
 if FFocusedEditor<>nil then
 begin
  //if FFocusedEditor.Text <> '' then
  with FSections[FFocusedEditor.Findex] do
  begin
   //if FinternalIndex = -1 then
   //  FinternalIndex:=FminValue
   //else
   Case aButton of
    pbUpButton:
      if FInternalIndex >= FMaxValue then
        FInternalIndex:= FMinValue
      else
        Inc(FInternalIndex);
    pbDownButton:
     if FInternalIndex <= FMinValue then
        FInternalIndex:= FMaxValue
     else
        Dec(FInternalIndex);
   end;
   //UpdateSelNode;
  end;
 end;
 Sections.FEmpty := False;
 with FSections[FFocusedEditor.Findex] do
  SetInternalIndex(FinternalIndex);
 ValidateAndSynchronize(FFocusedEditor.Findex);
 SetEditorText(FFocusedEditor.FIndex);
 FFocusedEditor.Modified:=true;
 if not (Fsections[FFocusedEditor.Findex].FType = tdtHyperLink) then
   FFocusedEditor.SelectAll;

end;

procedure TISCustomPlugEdit.ValidateAndSynchronize(EditorIndex : integer);
var i, aDSection, DIM : integer;
    aDay, aMonth, aYear : Word;
    aAMPM : integer;
    aDate : TDatetime;
begin
   // sunchronyze same sections and validate Min/Max Values
   //if not FSynchronize then
   //begin
    //FSynchronize := True;
    Try
      aYear := 1;
      aMonth := 1;
      aDay := 1;
      aAMPM := -1;
      aDSection := -1;
      for i := 0 to Fsections.Count -1 do
       Case FSections[i].Ftype of
         tdtAP, tdtAMPM:
         begin
           aAMPM := i;
         end;
         tdtD, tdtDD:
         begin
           aDay := Fsections[i].FInternalIndex;
           aDSection := i;
         end;
         tdtM, tdtMM,tdtMMM, tdtMMMM:
           aMonth := Fsections[i].FInternalIndex;
         tdtYY, tdtYYYY:
           aYear := Fsections[i].FInternalIndex;
       end;

       //synchronize hour /ampm
       if (FSections[EditorIndex].Ftype in [tdt24H, tdt24HH, tdt12H, tdt12HH]) and (aAMPM <>-1) then
       with FSections[EditorIndex] do
       begin
         if Ftype in [tdt24H, tdt24HH] then
         begin
            if (FInternalIndex >= 0) and (FInternalIndex<12) then
               FSections[aAmpm].ItemIndex := 0
            else FSections[aAmpm].ItemIndex := 1;
         end
         else
         begin
            if ((FInternalIndex = 12) and FUp) or ((FInternalIndex = 11) and FDown) then
            begin
             if (FSections[aAmpm].ItemIndex=0) then
                FSections[aAmpm].ItemIndex := 1
             else FSections[aAmpm].ItemIndex := 0;
            end
         end;
       end;

       //synchronize day
       DIM := DaysInMonth(aMonth,aYear);
       if aDsection <> -1 then
       if FSections[aDSection].FMaxValue <> DIM then
       begin
          if FSections[aDSection].FinternalIndex > DIM then
          begin
            FSections[aDSection].FinternalIndex := DIM;
            aDay := DIM;
            FSections[aDSection].Changed(False);
          end;
          FSections[aDSection].FMaxValue := DIM;
          FSections[aDSection].FMinValue := 1;
       end;

       aDate := Encodedate(aYear, aMonth, aDay);
       for i := 0 To FSections.count-1 do
         if (i<> EditorIndex) and (FSections[i].Ftype in [tdtDDD, tdtDDDD]) then
         begin
             Fsections[i].FInternalIndex := DayOfWeek(aDate);
             Fsections[i].Changed(False);
         end;


      for i := 0 To FSections.count-1 do
        if i <> EditorIndex then
        begin
           Case FSections[EditorIndex].Ftype of
           tdtD, tdtDD:
              if FSections[i].Ftype in [tdtDDD, tdtDDDD] then
                 Fsections[i].SetInternalIndex(DayOfWeek(aDate));
           //tdtDDD, tdtDDDD,
           tdtM, tdtMM:
           begin
            case FSections[i].Ftype of
              tdtMMM, tdtMMMM :
                 FSections[i].SetInternalValue(FSections[EditorIndex].FinternalValue);
              tdtD, tdtDD:
                 begin
                   FSections[i].FMaxValue := DIM;
                   if FSections[i].FInternalIndex > DIM then
                   begin
                      FSections[i].FinternalIndex := DIM;
                      ValidateAndSynchronize(i);
                   end;
                 end;
            end;
           end;
           tdtMMM, tdtMMMM:
              if FSections[i].Ftype in [tdtM, tdtMM] then
                 Fsections[i].SetInternalValue(FSections[EditorIndex].FinternalValue);
           //tdtYY, tdtYYYY
           end;
        end;

        ValueChanged(Fsections[EditorIndex]);
    finally
      //FSynchronize := False;
    end;
   //end;
end;

procedure TISCustomPlugEdit.SetEditorText(Index : integer);
begin
with Fsections[Index] do
begin
    if FSections.FEmpty then
    begin
     FInternalValue := ''
    end
    else
     Case Ftype of
       //tdtString:
       // TISAlignedEdit(Controls[Index]).Text := FInternalValue;
       tdtLeadingZeroesInteger:
         FinternalValue := Format('%.'+inttostr(FMaxLength)+'d',[FInternalIndex]);
       tdtYY:
         if FInternalIndex <> 0 then
          FinternalValue := FormatDateTime('yy',EncodeDate(FInternalIndex, 1, 1));
       tdtYYYY:
         if FInternalIndex <> 0 then
          FinternalValue := FormatDateTime('yyyy',EncodeDate(FInternalIndex, 1, 1));
       tdtProgress, tdtValueProgress, tdtPercentProgress, tdtInteger, tdtM, tdtD, tdtN, tdtS, tdt12H, tdt24H:
         FinternalValue := inttostr(FInternalIndex);
       tdtMM, tdtDD, tdtSS, tdtNN, tdt12HH, tdt24HH:
         FinternalValue := Format('%.2d',[FInternalIndex]);
       tdtDDD:
         if FInternalIndex <> 0 then
           FinternalValue := ShortDayNames[FInternalIndex];
       tdtDDDD:
         if FInternalIndex <> 0 then
          FinternalValue := LongDayNames[FInternalIndex];
       tdtMMM:
         if FInternalIndex <> 0 then
           FinternalValue := ShortMonthNames[FInternalIndex];
       tdtMMMM:
        if FInternalIndex <> 0 then
         FinternalValue := LongMonthNames[FInternalIndex];
       tdtAMPM, tdtAP:
         if FinternalIndex<>-1 then
           FinternalValue := Items[FInternalIndex].Text;
      end;
      //finalyze it
      TISAlignedEdit(Controls[Index]).Text := FinternalValue;
      TISAlignedEdit(Controls[Index]).Modified := False;
      
      if TISAlignedEdit(Controls[Index]).FCustomDraw and not FSections.FRebuilding then
         TISAlignedEdit(Controls[Index]).DrawEditor;
      if not (csLoading in Componentstate) then
        ValueChanged(Fsections[Index]);
    end;
end;

procedure TISCustomPlugEdit.PrepareDropDownEvent(Section : TISPlugSection; DropDown : TISDropDown);
begin
  if Assigned(FonPrepareDropdown) then
     FonPrepareDropDown(Self, Section, DropDown);
end;


procedure TISCustomPlugEdit.AcceptCloseUpEvent(var Accept : Boolean);
begin
  if Assigned(FOnAcceptCloseUp) then
    FOnAcceptCloseUp(Self, Accept);
end;

procedure TISCustomPlugEdit.CloseUpEvent(Section : TISPlugSection; DropDown : TISDropDown; var Accept : Boolean);
begin
  if Assigned(FonCloseUp) then
     FonCloseUp(Self, Section, DropDown, Accept);
end;

procedure TISCustomPlugEdit.ValueChanged(Section : TISPlugSection);
begin
  if Assigned(FonValueChanged) then
     FonValueChanged(Self, Section);
end;

procedure TISCustomPlugEdit.ButtonPressed(aButton : TPressedButtons);
begin
  if FFocusedEditor<>nil then
  begin
      ProceedButton(aButton);
  end;
  if Assigned(FonButtonPress) then
     FonButtonPress(Self, aButton);
end;

procedure TISCustomPlugEdit.WMSetFocus(var Message: TWMSetFocus);
begin
   if FFocusedEditor <> nil then
      windows.SetFocus(FFocusededitor.Handle)
   else FocusEditor(tmfFirst);  
end;

procedure TISCustomPlugEdit.WMSize(var Message: TMessage);
begin
  inherited;
  if not FRearrange then
    ReArrangeSections;
  DrawButton;  
end;


function TISCustomPlugEdit.DownBtnRect : TRect;
var Xoffset, Tp : integer;
begin
  XOffset := (Height-ClientHeight) div 2;
   if Odd(Height) then
    Tp := (Height div 2)+1
   else Tp := (Height div 2);

 {$IFDEF DFS_COMPILER_4_UP}
 if UseRightToLeftAlignment then
   Result :=Rect(XOffset, Tp, XOffset+FButtonWidth, Height-Xoffset)
 else
   Result :=Rect(Width - XOffset-FButtonWidth, Tp, Width - XOffset, Height-XOffset);
 {$ELSE}
  Result :=Rect(Width - XOffset-FButtonWidth, Tp, Width - XOffset, Height-Xoffset);
 {$ENDIF}
end;


function TISCustomPlugEdit.UpBtnRect : TRect;
var Xoffset, Btm : integer;
begin
 XOffset := (Height-ClientHeight) div 2;
// if Odd(ClientHeight) then
   Btm := (Height div 2);
// else
// Btm := (ClientHeight div 2)+1;
 {$IFDEF DFS_COMPILER_4_UP}
 if UseRightToLeftAlignment then
   Result :=Rect(XOffset, XOffset, XOffset+FButtonWidth, Btm)
 else
   Result :=Rect(Width - XOffset-FButtonWidth, XOffset, Width - XOffset, Btm);
 {$ELSE}
  Result :=Rect(Width - XOffset-FButtonWidth, XOffset, Width - XOffset, Btm);
 {$ENDIF}
end;

function TISCustomPlugEdit.PlusBtnRect : TRect;
begin
  Result := BtnRect;
end;

function TISCustomPlugEdit.MinusBtnRect : TRect;
var Xoffset : integer;
begin
 XOffset := (Height-ClientHeight) div 2;
 {$IFDEF DFS_COMPILER_4_UP}
 if UseRightToLeftAlignment then
   Result :=Rect(Width-FButtonWidth-XOffset, XOffset, Width - Xoffset, Height-XOffset)
 else
   Result :=Rect(XOffset, XOffset, FButtonWidth+XOffset, Height-XOffset)
 {$ELSE}
  Result :=Rect(XOffset, XOffset, FButtonWidth+XOffset, Height-XOffset)
 {$ENDIF}
end;

function TISCustomPlugEdit.BtnRect : TRect;
var Xoffset : integer;
begin
 XOffset := (Height-ClientHeight) div 2;
 {$IFDEF DFS_COMPILER_4_UP}
 if UseRightToLeftAlignment then
   Result :=Rect(XOffset, XOffset, FButtonWidth+XOffset, Height-XOffset)
 else
   Result :=Rect(Width-FButtonWidth-XOffset, XOffset, Width - Xoffset, Height-XOffset);
 {$ELSE}
  Result :=Rect(Width-FButtonWidth-XOffset, XOffset, Width - Xoffset, Height-XOffset);
 {$ENDIF}

end;

procedure TISCustomPlugEdit.DrawButton;
var aRect : TRect;
    W, X, Y, Flags : integer;
    DC: HDC;
begin
 DC := GetWindowDC(Handle);
 try
   Flags :=0;
   if not Enabled then
      Flags := Flags + DFCS_INACTIVE;
   if not (Ctl3D or (dds3d in FStyles)) then
      Flags := Flags + DFCS_FLAT;

  case FButtonType of
  tbtDropdown:
  begin
    aRect := BtnRect;
    if FDown then Flags := Flags+ DFCS_PUSHED;
    DrawFrameControl(DC, aRect, DFC_SCROLL,DFCS_SCROLLDOWN+Flags);
  end;
  tbtEllipsis:
  begin
    aRect := BtnRect;
    if FDown then Flags := DFCS_PUSHED;
    DrawFrameControl(DC, aRect, DFC_BUTTON, DFCS_BUTTONPUSH+Flags);

    with aRect do
    begin
      X := Left + ((Right - Left) shr 1) - 1 + Ord(FDown);
      Y := Top + ((Bottom - Top) shr 1) - 1 + Ord(FDown);
    end;
    W := FButtonWidth shr 3;
    if W = 0 then W := 1;
    PatBlt(DC, X, Y, W, W, BLACKNESS);
    PatBlt(DC, X - (W * 2), Y, W, W, BLACKNESS);
    PatBlt(DC, X + (W * 2), Y, W, W, BLACKNESS);
  end;
  tbtSpin:
  begin
    FillRect(DC,BtnRect, GetSysColorBrush(COLOR_BTNFACE));
    aRect := UpBtnRect;
    if Fup then Flags := Flags+ DFCS_PUSHED;
    DrawFrameControl(DC, aRect, DFC_SCROLL,DFCS_SCROLLUP+Flags);
    if Fup then Flags := Flags- DFCS_PUSHED;//clear flag
    aRect := DownBtnRect;
    if FDown then
       Flags := Flags+ DFCS_PUSHED;
    DrawFrameControl(DC, aRect, DFC_SCROLL,DFCS_SCROLLDOWN+Flags);
  end;
  tbtScroll:
  begin
    aRect := PlusBtnRect;
    if Fup then Flags := Flags+ DFCS_PUSHED;
    DrawFrameControl(DC, aRect, DFC_SCROLL,DFCS_SCROLLRIGHT+Flags);
    if Fup then Flags := Flags- DFCS_PUSHED;//clear flag
    aRect := MinusBtnRect;
    if FDown then
       Flags := Flags+ DFCS_PUSHED;
    DrawFrameControl(DC, aRect, DFC_SCROLL,DFCS_SCROLLLEFT+Flags);
  end;

 end;//case

 finally
    ReleaseDC(Handle, DC);
 end;
end;

procedure TISCustomPlugEdit.Paint;
var i : integer;
    aRect : TRect;
begin
    inherited;
    if not Enabled then
    with Canvas do
    begin
      Brush.Color := cl3DLight;
      Pen.color := ClGray;
    end;
    Canvas.Fillrect(ClientRect);
    aRect := clientRect;
    //X:=FDrawOffset;
    //aRect.Left := FDrawOffset;
    if FeditorsList.Count > 0 then
      aRect.Left := TIsAlignedEdit(FeditorsList[0]).Left - Canvas.TextWidth(Fsections[0].LeftSeparator)-Fspacing
    else
      aRect.Left := FDrawOffset;

    for i:=0 to FEditorsList.Count-1 do
    begin
      inc(aRect.Left, Fspacing);
      if Fsections[i].FLSeparator<>'' then
        DrawText(Canvas.Handle, Pchar(Fsections[i].FLSeparator),Length(Fsections[i].FLSeparator), aRect, DT_EDITCONTROL);

      with TISAlignedEdit(FEditorsList[i]) do
      //  aRect := Rect(X, Top, Left, Top+Height);
        aRect.Left := Left+Width;

      if FSections[i].Showborder then
         OffsetRect(aRect, 0, 2);

      //X := TISAlignedEdit(FEditorsList[i]).left + TISAlignedEdit(FEditorsList[i]).Width;
      //aRect.Left := X;
      //aRect.Right := ClientWidth;
      if Fsections[i].FRSeparator<>'' then
        DrawText(Canvas.Handle, Pchar(Fsections[i].FRSeparator),Length(Fsections[i].FRSeparator), aRect, DT_EDITCONTROL);
      inc(aRect.Left, Canvas.TextWidth(Fsections[i].FRSeparator));
    end;

 //   if FButtonType<>tbtNone then
 //     DrawButton;
end;

{procedure TISCustomPlugEdit.SetRightAligned(const Value : Boolean);
begin
   if Value <> FRightAligned then
   begin
      FRightAligned := Value;
      ReArrangeSections;
   end;

end;
}

procedure TISCustomPlugEdit.SetActiveEditor(Value : Integer);
begin
   if (Value <= FSections.Count-1) and (FeditorsList.Count = FSections.Count)then
   begin
      //FocusedEditor := TIsAlignededit(FeditorsList[]);
      //FFocusedEditor.Setfocus;
    if not (csDesigning in ComponentState) then
    begin
      Windows.SetFocus(TISAlignedEdit(FEditorsList[Value]).Handle);
      FFocusedEditor := TISAlignedEdit(FEditorsList[Value]);
    end;  
   end;
end;

procedure TISCustomPlugEdit.ClearEditors;
var i : integer;
    aEditor : TISAlignedEdit;
begin
 FClearing :=true;
 try
  for i:=0 to ControlCount-1 do
   begin
     aEditor := TISAlignedEdit(Controls[0]);
     aEditor.Parent := nil;
     aEditor.Free;
   end;
   FEditorsList.Clear;
 finally
   Fclearing := False;
 end;
end;

function TISCustomPlugEdit.Focused: Boolean;
var i : integer;
begin
  Result :=False;
  for i:=0 to ControlCount-1 do
    if TWinControl(Controls[i]).Focused then
    begin
       Result :=True;
       break;
    end;
end;

procedure TISCustomPlugEdit.RebuildSections(ClearSections : Boolean);
var aEditor : TISAlignedEdit;
    i : integer;
begin
 FSections.FRebuilding := True;
 try
   //clear FeditorsList
//   Application.ProcessMessages;
   ClearEditors;
   //build new list
   //  
   if (not (csLoading in Componentstate)) and (EditorType <> tetCustom ) then
    FSections.RebuildEditors(EditorType, ClearSections);
   FSections.FRebuilding := True;
   for i:=0 to FSections.Count-1 do
   begin
      aEditor := TISAlignedEdit.Create(Self);
      if FSections[i].FType in [tdtROList, tdtColor, tdtButton, tdtCheck, tdtImage, tdtImageCaption, tdtImageMemo, tdtProgress, tdtValueProgress, tdtPercentProgress] then
      begin
        aEditor.FCustomDraw := True;
      end;
      aEditor.Parent := Self;
      aEditor.Color := Color;
      aEditor.PasswordChar := FSections[i].PasswordChar;
      aEditor.FIndex := i;
      FEditorsList.Add(aEditor);
      ArrangeSection(I);
      //aEditor.BorderStyle := bssingle;
   end;
   RearrangeSections;
   {for i:=0 to FEditorsList.Count-1 do
   if TISAlignedEdit(FeditorsList[i]).FCustomDraw = True then
   begin
      TISAlignedEdit(FeditorsList[i]).DrawEditor;
   end;
   }
//   if FeditorsList.Count >0 then
//      FocusEditor(tmfFirst);//FFocusedEditor := TISAlignedEdit(FEditorsList[0]);
 finally
   FSections.FRebuilding := False;
 end;
end;



function TISCustomPlugEdit.GetValue(Index: Integer): String;
begin
   Result := TISAlignedEdit(FEditorsList[index]).Text;
end;


procedure TISCustomPlugEdit.FocusEditor(Position : TMoveFocus);
var aCtl : integer;
begin
 aCtl := -1;
 if (ControlCount > 0) then
 begin
  case Position of
    tmfFirst:
       aCtl := 0;
    tmfLast:
      aCtl := ControlCount-1;
    tmfNext:
    begin
      if (FFocusedEditor <> nil) then
      begin
         if (FFocusedEditor.FIndex+1) < ControlCount then
            aCtl := FFocusedEditor.FIndex+1
          else aCtl := 0;
      end
      else aCtl := 0;
    end;
    tmfPrev:
    begin
      if (FFocusedEditor <> nil) then
      begin
         if FFocusedEditor.FIndex > 0 then
            aCtl := FFocusedEditor.FIndex-1
          else aCtl := ControlCount-1;
      end
      else aCtl := 0;
    end;
  end;
 end;

 if (aCtl <> -1) and (ControlCount >0) then
 begin
    SetActiveEditor(aCtl);
    //Windows.SetFocus(TISAlignedEdit(FEditorsList[ACtl]).Handle);
    //FFocusedEditor := TISAlignedEdit(FEditorsList[ACtl]);
 end;

end;

function TISCustomPlugEdit.GetSectionCount : integer;
begin
   Result := FSections.Count;
end;

procedure TISCustomPlugEdit.WMSettingChange(var Message: TMessage);
begin
  inherited;
  if FEditorType <> tetCustom then
  begin
     Rebuildsections(False);
  end;
end;

procedure TISCustomPlugEdit.SetUpdateState(State : Boolean);//
begin
  FLockChangeEditorType := State;
end;

procedure TISCustomPlugEdit.SetEditorType(Value : TEditorType);
begin
   if (Value <> FEditortype) or (FSections.Count = 0) then
   begin
       FEditorType := Value;
       if FEditorType <> tetCustom then
       begin
        Fsections.FEditorType := Value;
        FLockChangeEditorType:= True;
        if not (csLoading in ComponentState) then
          RebuildSections(True);
         //ReArrangeSections;
         Fsections.Empty;
        FLockChangeEditorType:= False; 
       end;
   end;
end;

function TISCustomPlugEdit.DoTabKey : Boolean;
begin
   Result := False;
end;

procedure TISCustomPlugEdit.ResetButtons;
begin
   FUp :=False;
   FDown := False;
end;

function TISCustomPlugEdit.GetAlignment : TAlignment;
begin
   Result := inherited Alignment;
end;

procedure TISCustomPlugEdit.SetAlignment(Value : TAlignment);
begin
   inherited Alignment := Value;
   if not FRearrange then
     ReArrangeSections;
end;

procedure TISCustomPlugEdit.ArrangeSection(index : integer);
begin
 if index < FeditorsList.Count then
  with TISAlignedEdit(FeditorsList[Index]) do
  begin
       if FSections[Index].ShowBorder then
         BorderStyle := bsSingle
       else
         BorderStyle := bsNone;

       ReadOnly := Sections[Index].ReadOnly;
       Alignment := Sections[Index].Alignment;//taCenter;
       MaxLength := Sections[Index].MaxLength;
       if not (FSections[index].FType = tdtHyperLink) then
          AutoSelect := True;
       AutoSize := False;
       Ctl3d := False;
       Font.Assign(Self.Font);
       if FSections[index].FType = tdtHyperLink then
         Font.Style := Font.Style+ [fsUnderLine];

       Font.Color := Sections[Index].FontColor;
       Color := Sections[Index].Color;
       {if Sections[Index].Items.count >0 then
       begin
         Sections[Index].FMinValue := 0;
         Sections[Index].FMaxValue := Sections[Index].Items.Count-1;
       end;
       }
       SetEditorText(Index);
  end;
end;



procedure TISCustomPlugEdit.ReArrangeSections;
var
    aNumCharWidth, i, aWidth, w1, w2, Offset : integer;

    function GetMaxString(i : integer) : integer;
    var W : integer;
        aNode : TFlyNode;
    begin
       Result :=0;
       aNode := Fsections[i].Items.getFirstnode;
       while aNode <> nil do
       begin
          W := Canvas.TextWidth(aNode.Text)+2;
          if Result < W then Result := W;
          aNode := aNode.GetNext;
       end;
    end;

    function GetMaxName(Names : Array of String) : integer;
    var j,w : integer;
    begin
       Result :=0;
       for j:= Low(Names) to High(Names) do
       begin
         W := Canvas.TextWidth(Names[j])+2;
         if Result < W then Result := W;
       end;
    end;

    function GetMaxList(Names : TFlyNodes) : integer;
    var j,w : integer;
    begin
       Result :=0;
       for j:= 0 to Names.Count-1 do
       begin
         W := Canvas.TextWidth(Names[j].Text)+2;
         if Result < W then Result := W;
       end;
    end;

begin
 if not ((not HandleAllocated) or (Sectioncount=0)) then
 begin
   //calculate widths of editors
   Offset := 0;
   //FDrawOffset := 0;
   Canvas.Font.Assign(Font);
   aWidth :=0;
   aNumCharWidth := Canvas.TextWidth('9');
   for i:=0 to FEditorsList.count-1 do
   begin
     if Fsections[i].FWidth > 0 then
        aWidth := Fsections[i].FWidth
     else
     with TISAlignedEdit(FEditorsList[i]) do
     Case Fsections[i].Ftype of
       tdtImage, tdtImageCaption, tdtImageMemo:
       begin
          if Images <> nil then
             aWidth := Images.Width;
          if Fsections[i].Ftype in [tdtImageCaption, tdtImageMemo] then
            inc(aWidth, 4+GetMaxList(FSections[i].Items));
          if Fsections[i].Ftype = tdtImageMemo then
             TISAlignedEdit(FEditorsList[i]).Multiline :=True;
       end;
       tdtColor: aWidth := ClientHeight+4+GetMaxList(FSections[i].Items);
       tdtRoList,tdtAP,tdtAMPM: aWidth := GetMaxList(FSections[i].Items);
       tdtProgress, tdtValueProgress, tdtPercentProgress, tdtInteger, tdtLeadingZeroesInteger:
       begin
        W1 := Length(IntTostr(Fsections[i].FMaxValue));
        W2 := Length(IntTostr(Fsections[i].FMinValue));
        aWidth := (Max(W1, w2)* aNumCharWidth) +2;
        {if then//Fsections[i].FMaxValue > Fsections[i].FMinValue then
         aWidth := Length(IntTostr(Fsections[i].FMaxValue)) * aNumCharWidth+2
        else
         aWidth := Length(IntTostr(Fsections[i].FMinValue)) * aNumCharWidth+2;
        }
       end;

       tdtHyperLink, tdtString: aWidth := GetMaxString(i);// On maxlength or PickList
       tdtMultiline:
       begin
         aWidth := GetMaxString(i);// On maxlength or PickList
         TISAlignedEdit(FEditorsList[i]).Multiline :=True;
       end;
       tdtYY, tdtM,
       tdtMM, tdtD,
       tdtDD, tdt12H, tdt12HH, tdt24H, tdt24HH,
       tdtN,  tdtNN, tdtS,
       tdtSS:       begin aWidth := 2 * aNumCharWidth;MaxLength:=2;end;//2 digit
       tdtYYYY:     begin aWidth := 4 * aNumCharWidth;MaxLength:=4;end;//4 digit
       //on SisUtils base
       tdtDDD:      aWidth := GetMaxName(ShortDayNames);
       tdtDDDD:     aWidth := GetMaxName(LongDayNames);
       tdtMMM:      aWidth := GetMaxName(ShortMonthNames);
       tdtMMMM:     aWidth := GetMaxName(LongMonthNames);
       tdtCheck:
           if FSections.Count > 1 then
           begin
              aWidth := TISAlignedEdit(FEditorsList[i]).Height;
              if Odd(aWidth) then
                Dec(aWidth);
           end;

       else aWidth :=0;
     end;

     ArrangeSection(i);

     Inc(Offset, Canvas.TextWidth(FSections[i].LeftSeparator) + FSpacing);

     TISAlignedEdit(FEditorsList[i]).Left := Offset;
     //if TISAlignedEdit(FEditorsList[i]).Multiline then
       TISAlignedEdit(FEditorsList[i]).Height := ClientHeight;
     //else if FSections[i].FType = tdtCheck then
     //  TISAlignedEdit(FEditorsList[i]).Height := aWidth;

     TISAlignedEdit(FEditorsList[i]).Top := (ClientHeight-TISAlignedEdit(FEditorsList[i]).Height) div 2;

     if SectionCount > 1 then
     begin
      if Fsections[i].ShowBorder then Inc(awidth,2);
       TISAlignedEdit(FEditorsList[i]).Width := aWidth;
     end
     else
     begin
         if FButtonType <> tbtNone then
           TISAlignedEdit(FEditorsList[i]).Width:=ClientWidth{-FButtonwidth}-TISAlignedEdit(FEditorsList[i]).Left
         else
           TISAlignedEdit(FEditorsList[i]).Width:= ClientWidth - TISAlignedEdit(FEditorsList[i]).Left;
         with TISAlignedEdit(FEditorsList[i]) do
          Width:= Width - (Canvas.TextWidth(FSections[i].RightSeparator)+FSpacing);

          //TISAlignedEdit(FEditorsList[i]).Height:=ClientHeight;
     end;

     inc(Offset, TISAlignedEdit(FEditorsList[i]).Width+Canvas.TextWidth(FSections[i].RightSeparator)+FSpacing);
   end;

   if FAutoSize and (FEditorsList.Count>1)then
   begin
      FRearrange := True;
      try
      with TISAlignedEdit(FEditorsList[FEditorsList.Count-1]) do
        aWidth := Left+Width;
      //Height := TISAlignedEdit(FEditorsList[FEditorsList.Count-1]).Height;
      if FButtonType<>tbtNone then
         inc(aWidth, FButtonWidth);
      ClientWidth := aWidth + FSpacing;
      finally
        FRearrange := False;
      end;
   end;

   {if UseRightToLeftAlignment and (FButtonType <> tbtNone) then
   begin
     for i:=0 to FEditorsList.Count-1 do
        with TISAlignedEdit(FEditorsList[i]) do
          Left:=Left+ FButtonWidth;
   end
   else}
   Case Alignment of
     taCenter, taRightJustify:
     begin
     with TISAlignedEdit(FEditorsList[FEditorsList.Count-1]) do
      aWidth := Left+Width+(Canvas.TextWidth(FSections[FEditorsList.Count-1].RightSeparator)+FSpacing);//right margin
      OffSet := ClientWidth - aWidth;
      if Alignment = taCenter then
        Offset := Offset div 2;
        
      if Offset > 0 then
        for i:=0 to FEditorsList.Count-1 do
        with TISAlignedEdit(FEditorsList[i]) do
          Left:=Left+ Offset;
      end;
   end;

 end;
 Invalidate;
end;



procedure TISCustomPlugEdit.ShiftControls(aEdit : TISAlignedEdit);
var i, aOffset : Integer;
begin
   aOffset :=0;
   if aEdit.Left < 0 then
   begin
      aOffset :=  - aEdit.Left;
   end
   else if (aEdit.Left+aEdit.Width) > ClientWidth then
   begin
      aOffset := ClientWidth - (aEdit.Left+aEdit.Width);
   end;

   for i:= 0 to FEditorsList.count-1 do
     TIsAlignedEdit(FeditorsList[i]).Left:= TIsAlignedEdit(FeditorsList[i]).Left+aOffset;

   if FeditorsList.Count > 1 then
     FDrawOffset := TIsAlignedEdit(FeditorsList[0]).Left - Canvas.TextWidth(Fsections[0].LeftSeparator)
   else
     FDrawOffset :=0;

   Invalidate;
end;

procedure TISCustomPlugEdit.AddColors(const S : String);
begin
   if FListToAdd <> -1 then
     with FSections[FListToAdd].Items.Add(nil, S) do
     begin
        ImageIndex := ColorToRGB(StringToColor(S) );
        SelectedIndex := ImageIndex;
     end;
end;

procedure TISCustomPlugEdit.SetStyles(Value : TDropdownStyles);
begin
  if (Value <> FStyles) then
  begin
    FStyles := Value;
    if HandleAllocated and (not IsInplace) then
      invalidate;//DrawButton;
  end;
end;

procedure TISCustomPlugEdit.GetDropDownControl(Section : TISPlugSection; DropDown : TISDropDown; var DropDownCtl : TWinControl);
var  aTree : TPopupTree;
begin
  if Assigned(FOnGetDropDownControl) then
     FOnGetDropDownControl(Self, Section, DropDown, DropDownCtl);
  if not assigned(DropDownCtl) then
  begin
     //Standard PopupTree
    //DropDown.Styles := [];//[ddsSized, ddsOK, ddsOKCancel];
    aTree := TPopupTree.Create(DropDown);
    aTree.FPlugEdit := DropDown.FPlugEdit;
    DropDownCtl := aTree;//register
  end;
 if Assigned(DropDownCtl) then
 begin
  //if not DropDown.Handleallocated then
  //   DropDown.handleNeeded;
  //DropDownCtl.Parent := nil;
  //DropDownCtl.ParentWindow := FDropDown.Handle;
  DropDown.FContainedControl := DropDownCtl;
  DropDown.FContainedControl.Parent := FDropDown;//18/03/2
//  if not DropDownCtl.Handleallocated then
//     DropDownCtl.handleNeeded;

 end;
end;

function GetParentFormHandle(aWinHandle : THandle) : THandle;
begin
  Result:=aWinHandle;
  While GetParent(Result) > 0 do
    Result:= GetParent(Result);
end;

//new - 2.2.2
procedure TISCustomPlugedit.CreateDropdown;
begin
  //first
  {$IFDEF ACTIVEX}
    FDropDown := TISDropDown.CreateParented(GetTopParentHandle);//GetDesktopWindow);//
  {$else}
    FDropDown := TISDropDown.CreateParented(Application.Handle);//GetDesktopWindow);
  {$endif}
  FDropDown.FPlugEdit := Self;
  FDropDown.Visible := false;
  {$ifdef DFS_COMPILER_4_UP}
  FDropDown.Bidimode := BidiMode;
  {$endif}
end;

procedure TISCustomPlugEdit.DropDown;
var
  P: TPoint;
  DDBounds : TRect;
  Y: Integer;
  Node : TFlyNode;
  DropDownCtl : TWinControl;
//  aForm : TCustomForm;
  ahwnd : THandle;
begin
  if FFocusedEditor = nil then exit;
  if FDropDown = nil then
   begin
      CreateDropdown;
   end
   else
   begin
     FDropDown.ClearControls;
   end;

  FDropDown.Fstyles := FStyles;
  FDropDown.Font := Font;
  FDropDown.Color := clWindow;//Color;

  FSections[FFocusedEditor.FIndex].SetUndoValue;
  //Here - >
  //if FSizeGrip then FDropDown.Styles := FDropDown.Styles + [ddsSized];
  //if Ctl3D then FDropDown.Styles := FDropDown.Styles + [dds3D];
  DropDownCtl := nil;
  GetDropDownControl(Fsections[FFocusedEditor.FIndex], FDropDown, DropDownCtl);
  if DropDownCtl is TPopupTree then
  begin
    FDropDown.Styles := FDropDown.Styles + [ddsSizeToControlSize];
    DropDownCtl.Width := Self.Width;
    PrepareDropDown;
  end
  else
    PrepareDropDownEvent(Fsections[FFocusedEditor.FIndex], FDropDown);

  //Here ->
  with FDropDown do
  begin
   //Styles := [ddsSized, ddsOK, ddsOKCancel];
   BorderStyle :=bsSingle;
   Ctl3d := True;
   BevelInner := bvNone;
   BevelOuter := bvNone;

   if not FContainedControl.Handleallocated then
     FContainedControl.handleNeeded;
   if ddsSizeToControlSize in Styles then
      DDBounds := Rect(0, 0, FContainedControl.Width, FContainedControl.Height+FNonClientHeight)
   else
      DDBounds := FDropDown.BoundsRect;
      //FDropDown.SetBounds(0, 0, FContainedControl.Width, FContainedControl.Height+FNonClientHeight);

   P := Point(0, 0);
   P := Self.ClientToScreen(P);
   if Self.BorderStyle = bsSingle then
   begin
      if Self.Ctl3D then
      begin
        Dec(P.Y,2);
        Dec(P.X, 2);
      end
      else
        Dec(P.Y,1);
        Dec(P.X,1);
   end;
   Y := P.Y + Self.Height;

   if Y + (DDBounds.Bottom -DDBounds.Top) > Screen.Height then
   begin
    Y := P.Y - (DDBounds.Bottom -DDBounds.Top);//FDropDown.Height;
    FonTop := True;
    FDropDown.RecreateWnd;
   end
   else
     FOnTop := False;
     //FContainedControl.Align := alclient;
  end;

  if FDropDown.FContainedControl is TPopupTree then
  with TPopupTree(FDropDown.FContainedControl) do
  begin
   if ColCount = 1 then
     FitColumnToClientWidth:=true;
    UpdateScrollRange;
   Node := GetNodeAtRow(Row);
   if Node <> nil then
      Selected := Node;
  end;

  //FDropDown.BoundsRect := Rect(P.X, Y, P.X+(DDBounds.Right-DDBounds.Left), Y+(DDBounds.Bottom -DDBounds.Top));

  if GetParentForm(Parent) <> nil then
   ahwnd := GetParentForm(Parent).Handle
  else
   ahwnd := GetParentFormHandle(Parent.Handle);
  FDropDown.FLeftTop := Point(P.X, Y);
  FDropDown.SetBounds(P.X, Y, (DDBounds.Right-DDBounds.Left), (DDBounds.Bottom -DDBounds.Top));
  FDropDown.FContainedControl.SetBounds(0,0, FDropDown.ClientWidth, FDropDown.ClientHeight);
  SetWindowPos(FDropDown.Handle, HWND_TOP, P.X, Y, DDBounds.Right-DDBounds.Left{FDropDown.Width}, (DDBounds.Bottom -DDBounds.Top){FDropDown.Height},
      SWP_NOSIZE or SWP_NOACTIVATE  or SWP_SHOWWINDOW);
  //init
  DropDownHandle := FDropDown;

  FDropDownVisible := True;
  FDropDown.Visible := True;
//  Windows.SetFocus(Handle);


  //SendMessage(ahwnd, WM_SETREDRAW, Ord(false), 0);
  Windows.SetFocus(FDropDown.FContainedControl.Handle); //static
  PostMessage(ahwnd,WM_NCACTIVATE, Longint(True), 0);
  //SendMessage(ahwnd, WM_SETREDRAW, Ord(True), 0);
  Application.ProcessMessages;
end;

function IsLeapYear (Year: Word): Boolean;
begin
   Result := ((Year and 3) = 0) and ((Year mod 100 > 0) or (Year mod 400 = 0))
end;

function TISCustomPlugEdit.DaysInMonth (const Month, Year : integer): Byte;
begin
case Month of //month
  2: if IsLeapYear (Year) then
       Result := 29
     else
       Result := 28;
  4, 6, 9, 11: Result := 30;
  else
   Result := 31;
 end;
end;

procedure TISCustomPlugEdit.PrepareDropDown;
  var I,W : integer;
  procedure AddDays(Nodes : TFlyNodes);
  var i, aMonth, aYear : Integer; //decodedate
      M,D,Y : Word;
  begin
     aMonth := -1;
     aYear  := -1;
     for i:=0 to FSections.Count-1 do
     with Fsections[i] do
     begin
        if Ftype in [tdtM, tdtMM, tdtMMM, tdtMMMM] then
        begin
          aMonth := FInternalIndex;
          Break;
        end;
     end;

     if aMonth = -1 then
     begin
       DecodeDate(Now, Y,M,D);
       aMonth := M;
     end;

     for i:=0 to FSections.Count-1 do
     with Fsections[i] do
     begin
        if Ftype in [tdtYY, tdtYYYY] then
        begin
          aYear := FInternalIndex;
          Break;
        end;
     end;

     if aYear = -1 then
     begin
       DecodeDate(Now, Y,M,D);
       aYear := Y;
     end;

     D:=DaysInMonth(aMonth, aYear);
     Nodes.Clear;
     for i:=1 to D do
       with Nodes.Add(nil, inttostr(i)) do
         ImageIndex := i;

  end;

  procedure AddArrayName(Nodes : TFlyNodes; Names : Array of String);
  var j : integer;
  begin
     Nodes.Clear;
     for j := Low(Names) to High(Names) do
        Nodes.Add(nil, Names[j]);
  end;

  procedure AddCounter(Nodes : TFlyNodes; Min, Max : Integer);
  var i : integer;
  begin
     Nodes.Clear;
     for i := Min to Max do
     begin
        with Nodes.Add(nil, InttoStr(i)) do
          ImageIndex := i;
     end;
  end;

  function FindNode(Items : TFlyNodes; MinVal, Value : Integer) : integer;
  var aNode : TFlyNode;
      i : integer;
  begin
    Result := 0;
    if Items.Count > 0 then
    begin
     aNode := Items.GetFirstNode;
     I := MinVal;
     While  aNode <> nil do //I <> Value do
     begin
        if i = Value then
          break;
        aNode := anode.GetNext;
        inc(i);
     end;
     if (I = Value) and (aNode <> nil) then
     begin
        Result := aNode.GetRow;
     end;
    end;
  end;
begin
   if FFocusedEditor <> nil then
   with TPopupTree(FDropDown.FContainedControl) do
   begin
     //HandleNeeded;
     FitToHeight := FFocusedEditor.Multiline;//(Fsections[FFocusedEditor.FIndex].FType = tdtImageMemo);
     //SizeGrip := DropdownSized;
     WordWrap := FitToHeight;
     Showimages :=  FSections[FFocusedEditor.FIndex].FType in [tdtColor, tdtImageMemo, tdtImage, tdtImageCaption];
     if ShowImages then
       Options := Options-[goRangeSelect]
     else
       Options := Options-[goRangeSelect]+[goRowSelect];
     ColCount := FSections[FFocusedEditor.FIndex].FDropDownColCount;
     StructureCol := FSections[FFocusedEditor.FIndex].FDropDownStructureCol;

     Ctl3d := False;
     BorderStyle := bsNone;
     // these settings necessarily for determine internal value on selecting string in TPopupTree.Selectcell
     Items.DefaultImageIndex :=-1;
     Items.DefaultSelectedIndex := -1;

     if ColCount =1 then
       DefaultColwidth := Width
     else
     begin
       Options := Options+[goVertLine];
       W:=0;
       for i:=0 to ColCount-1 do
       begin
          ColWidths[i] := Fsections[FFocusedEditor.FIndex].DropDownColWidth[i];
          inc(W,ColWidths[i] );
       end;
       if Width < W then
         Width := W + GetsystemMetrics(SM_CXHTHUMB);
     end;
     if (Self.Images <> nil) and ShowImages then
        Images := Self.Images
     else Images := nil;

     ShowLines := False;
     ShowLogic := EditorType = tetDropdownCheckList; //boundchanged
     //ShowButtons := EditorType = tetDropDownTree;
     ShowRoot := not (EditorType in [tetDropDownList, tetDropDown, tetColor, tetImage, tetImageCaption, tetImageMemo]);

     //Ctl3d := Self.Ctl3d;

     if not (ShowLogic or ShowImages or ShowButtons) then
        Indent := 3

     else if ShowImages and (Self.Images <> nil) then
     begin
        TransParentMode := True;
        Indent := Self.Images.Width;
        if not Wordwrap then
           DefaultRowheight := Self.Images.Height;
     end
     else
     begin
        Indent := GetCheckSize + 4;
        DefaultRowheight := Canvas.TextHeight('Wg')+4;
     end;

   end;
  // prepare dropdownlist:

  if FFocusedEditor <> nil then
  with Fsections[FFocusedEditor.FIndex] do
  begin
    Items.OnChange := nil;
    Case Ftype of
      tdtYY: AddCounter(Items, 0,99);
      tdtYYYY: AddCounter(Items, 1930, 2029);
      tdtD,tdtDD: AddDays(Items);
      tdtDDD: AddArrayName(Items, ShortDayNames);
      tdtDDDD: AddArrayName(Items, LongDayNames);
      tdtM, tdtMM: AddCounter(Items, 1,12);
      tdtMMM: AddArrayName(Items, ShortMonthNames);
      tdtMMMM: AddArrayName(Items, LongMonthNames);
      tdt24H, tdt24HH : AddCounter(Items, 0,23);
      tdt12H, tdt12HH:  AddCounter(Items, 1,12);
      tdtN, tdtNN,
      tdtS, tdtSS: AddCounter(Items, 0, 59);
    end; //case
    ItemsChanged(nil);
    Items.OnChange := ItemsChanged;
  end;
  PrepareDropdownEvent(Fsections[FFocusedEditor.FIndex], FDropDown);

  with TPopupTree(FDropDown.FContainedControl) do
  begin
     //Items := Fsections[FFocusedEditor.FIndex].Items;
     Items.BeginUpdate;
     try
       Items.Clear;
       for i:=0 to Fsections[FFocusedEditor.FIndex].Items.Count-1 do
          with Items.Add(nil,'') do
            Assign(Fsections[FFocusedEditor.FIndex].Items[i]);
     finally
       Items.EndUpdate;
     end;

   
//     if FitToHeight then
//     begin
//        for i:=0 to RowCount-1 do
//        with TSizedPopupTree(FDropdown) do
//          RecalcRowHeight(i);//, ColWidths[StructureCol]);
//     end;

      if Items.Count > FDropDownrows then
        ClientHeight := FDropDownRows * DefaultRowHeight
      else ClientHeight := Items.Count * DefaultRowHeight+4;

     if SizeGrip then
       Height := Height+ GetSystemMetrics(SM_CYVSCROLL)+2;

      if EditorType = tetDropDownCheckList then
      begin
      with Fsections[FFocusedEditor.FIndex] do
       SetCheckText(FinternalValue, FItemseparator);
      end
      else
       with Fsections[FFocusedEditor.FIndex] do
       begin
         UpdateSelNode;
         TPopupTree(FDropdown.FContainedControl).FDoNotChangeSection := False;
         Row := FindNode(TPopupTree(FDropdown.FContainedControl).Items, FMinValue, FInternalIndex);
       end;
     Col := 0;

  end;

end;

procedure TISCustomPlugEdit.CloseUp(Accept : Boolean);
begin
  PostMessage(Handle, CM_CLOSEUP, ord(Accept),0);
end;

procedure TISCustomPlugEdit.MakeCloseUp(Accept : Boolean);
var DoCloseUp, Accepted : Boolean;
begin
//----------- close up control for aggregated dropdowns, such as ActiveX controls
 DoCloseUp := True;
 AcceptCloseUpEvent(DoCloseUp);
 if not DoCloseUp then
 begin
  //Windows.SetFocus(FDropDown.FContainedControl.Handle); //static
  //SetWindowPos(FDropDown.Handle, HWND_TOP, FDropDown.Left, FDropDown.Top, FDropDown.Width, FDropDown.Height,
  //    SWP_NOSIZE or SWP_NOACTIVATE  or SWP_SHOWWINDOW);
  exit;
 end;
//-----------
 if (FDropdown <> nil) then //latch
   if FDropdown.CloseUpStarted then
     Exit
   else
     FDropdown.CloseUpStarted := True;

    Accepted := Accept;
    if {Accepted and} (FFocusedEditor <> nil) then
      CloseUpEvent(Fsections[FFocusedEditor.FIndex], FDropdown, Accepted);

   if GetCapture <> 0 then
     SendMessage(GetCapture, WM_CANCELMODE, 0, 0);

    if (FFocusedEditor <> nil) then
    begin
     if (not Accepted) then
       Fsections[FFocusedEditor.FIndex].UndoValue
     else
     begin
       Fsections[FFocusedEditor.FIndex].Value := FFocusedEditor.Text;
       Modified := true;
     end;
    end;

    if FDropDownVisible then
    begin
      //FDropdown.ClearControls;
      DropDownHandle := nil;
      FDropdown.Visible := False;
      SetWindowPos(FDropdown.Handle, 0, 0, 0, 0, 0, SWP_NOZORDER or SWP_NOMOVE  or SWP_HIDEWINDOW);
      //SWP_HIDEWINDOW or SWP_NOZORDER or SWP_NOREDRAW
      {or SWP_NOSIZE or SWP_NOACTIVATE}
      FDropDownVisible := False;
      FDropdown.Free;
      FDropDown := nil;

    end;
    if FFocusedEditor <> nil then
       Windows.SetFocus(FFocusedEditor.Handle);
    Invalidate;
end;



procedure TISCustomPlugEdit.SetSpacing(const Value : integer);
begin
 if (Value <> FSpacing) and (Value>0) then
 begin
    FSpacing := Value;
    RearrangeSections;
 end;
end;

procedure TISCustomPlugEdit.SetButtonWidth(const Value : integer);
begin
 if (Value <> FButtonWidth) and (Value>0) then
 begin
    FButtonWidth := Value;
    //RecreateWnd;//sizechanged
    Perform(CM_BORDERCHANGED, 0, 0);
    //if FButtonType<>tbtNone then
    //  RearrangeSections;
 end;
end;

procedure TISCustomPlugEdit.SetButtonType(Value: TButtonType);
begin
    if Value <> FButtonType then
    begin
        //new 2.2.2
        //if (Value <> tbtDropDown) and (FDropDown <> nil) then
        //begin
        //   FDropDown.Free;
        //   FDropdown :=nil;
        //end;

        FButtonType := Value;
        Perform(CM_BORDERCHANGED, 0, 0);//
        //RecreateWnd;

        //if not (csLoading in ComponentState) then
        // Rebuildsections;
        ReArrangeSections;

        DrawButton;
    end;
end;

procedure TISCustomPlugEdit.SetValue(Index: integer; Value: String);
begin
   TISAlignedEdit(FEditorsList[index]).Text := Value;
end;

function TISCustomPlugEdit.ValidateInput(Sender : TISAlignedEdit; Var Key : Char) : Boolean;
var
  i, InsPos,SelEnd : Integer;
  aKeyText, aFullText : String;
  aNode : TFlyNode;

  function AutoCompleteArray(Names : Array of String) : Boolean;
  var j : integer;
  begin
     Result := false;
     for j := Low(Names) to High(Names) do
     begin
       if Pos(UpperCase(aFullText), UpperCase(Names[j])) = 1 then //string founded
       begin
         Sender.Text := Names[j];
         Sender.SelStart := InsPos;
         Sender.SelLength := Length(Sender.Text) - InsPos;
         FSections[i].FInternalIndex := j;
         FSections[i].FInternalValue := Names[j];
         Result := True;
         Key:=#0;
         Break;
       end;
       Key :=#0;
     end;
  end;

begin
  i := Sender.FIndex;
  aKeyText := '';
  // get full text
  Result := true;

  if Word(Key) in [VK_RETURN, VK_ESCAPE] then
  begin
    Key :=#0;
    exit;
  end;

  InsPos := LongRec(SendMessage(Sender.Handle, EM_GETSEL, 0, 0)).Hi;
  SelEnd := LongRec(SendMessage(Sender.Handle, EM_GETSEL, 0, 0)).Lo;
  if (Sender.SelLength=0) and (SelEnd = Length(Sender.Text)) then
  begin //     -     
    if ((Key = FSections[i].RightSeparator)) then
    begin
       FocusEditor(tmfNext);
       Result := false;
    end
    else if (i < Fsections.Count-1) then
     if ((Key = FSections[i+1].LeftSeparator)) then
     begin
        FocusEditor(tmfNext);
        Result := false;
     end;
    if not Result then exit; 
  end;

  if Sender.SelLength = Length(Sender.Text) then
  begin
     if (Word(Key) = VK_BACK) or (Key = #0)then
      aFullText:=''
     else
       aFullText:= Key;
     InsPos :=1;
  end
  else if Sender.SelLength > 0 then
  begin
    aFullText := Sender.Text;
    aKeyText := aKeyText + Key;

    if not ((Word(Key) = VK_BACK) or (Key=#0))then
    begin
      if SelEnd > InsPos then
       insert(aKeyText, aFullText, SelEnd+1)
      else
       insert(aKeyText, aFullText, InsPos+1);
    end;

    if SelEnd > InsPos then
      Delete(aFullText, InsPos+1, Sender.SelLength)
    else
      Delete(aFullText, SelEnd+1, Sender.SelLength);

    //if not ((Word(Key) = VK_BACK) or (Key=#0))then
    //  aFullText := aFullText + Key;
    InsPos :=InsPos-Sender.SelLength+1;
  end
  else if (Word(Key) <> VK_BACK) and (Key <> #0)then
    aFullText := Copy(Sender.Text, 0, InsPos) + Key+Copy(Sender.Text, InsPos+1, MAXINT )
  else if (Word(Key) = VK_BACK) then
  begin
    aFullText := Sender.Text;
    Delete(aFullText, InsPos, 1);
  end
  else if Sender.FDeletePressed then
  begin
    aFullText := Sender.Text;
    Delete(aFullText, InsPos+1, 1);
  end;
  if (Sender.MaxLength > 0) and (Length(aFullText) > Sender.MaxLength) then
    aFullText := Sender.Text;
  if aFullText<>'' then
  begin
  try
   with Fsections[i] do
   Case Ftype of
       tdtLeadingZeroesInteger, tdtInteger:
       begin
         Result := (aFullText = '-') or (strToInt(aFullText) >= FMinValue) and (strToInt(aFullText) <= FMaxValue);
         if Result then
          begin
            FInternalValue := aFullText;
            if aFullText <> '-' then
              FinternalIndex := StrToInt(aFullText);
          end;
       end;
       tdtcolor, tdtHyperLink, tdtString, tdtMultiline:
       begin
         if FAutoComplete then
          if FSections[i].Items.Count > 0 then//tstringlist
          begin
            // AutoComplete support
            Result := True;
            if FAutoDropDown and (not FDropdownvisible) then
            begin
              FDown := True;//prevent selection
              try
               DropDown;
              finally
                FDown :=False;
              end;
            end;
            if FDropDownVisible  then
              aNode := TPopupTree(FDropDown.FContainedControl).Items.GetFirstNode
            else
              aNode := FSections[i].Items.GetFirstNode;
            if Key <>#8 then
            while aNode <> nil do
            begin
               if (Pos(UpperCase(aFullText), UpperCase(aNode.Text)) = 1) then //and (UpperCase(aFullText)<> UpperCase(aNode.Text)) then //string founded
               begin
                Sender.Text := aNode.Text;
                aFullText := Sender.Text;
                Sender.SelStart := InsPos;
                Sender.SelLength := Length(Sender.Text) - InsPos;
                if FDropDownvisible then
                   TPopupTree(FDropDown.FContainedControl).Selected := aNode;
                Key :=#0;
                Break;
               end
               else
                 aNode := aNode.GetNext;//Visible;
            end;
          end
          else
            Result := true;
          if Result then
             FInternalValue := aFullText;
       end;
       tdtM, tdtMM :
       begin
         Result := (StrToInt(aFullText)>=1) and (StrToInt(aFullText) <= 12);
         if Result then
          begin
            FInternalValue := aFullText;
            FinternalIndex := StrToInt(aFullText);
          end;
       end;
       tdtD, tdtDD :
       begin
         Result := (StrToInt(aFullText)>=1) and (StrToInt(aFullText) <= 31);
         if Result then
          begin
            FInternalValue := aFullText;
            FinternalIndex := StrToInt(aFullText);
          end;
       end;
       tdt24H, tdt24HH :
       begin
         Result := (StrToInt(aFullText)>=0) and (StrToInt(aFullText) <= 23);
         if Result then
          begin
            FInternalValue := aFullText;
            FinternalIndex := StrToInt(aFullText);
          end;
       end;
       tdt12H, tdt12HH :
       begin
         Result := (StrToInt(aFullText)>=1) and (StrToInt(aFullText) <= 12);
          if Result then
          begin
            FInternalValue := aFullText;
            FinternalIndex := StrToInt(aFullText);
          end;
       end;
       tdtYY       :
       begin
          Result := (StrToInt(aFullText)>=0) and (StrToInt(aFullText) <= 99);
          if Result then
          begin
            FInternalValue := aFullText;
            FinternalIndex := StrToInt(aFullText);
          end;
       end;
       tdtN, tdtNN,
       tdtS, tdtSS:
       begin
          Result := (StrToInt(aFullText)>=0) and (StrToInt(aFullText) <= 59);
          if Result then
          begin
            FInternalValue := aFullText;
            FinternalIndex := StrToInt(aFullText);
          end;
       end;
       tdtYYYY:
       begin
          Result := (StrToInt(aFullText)>=0);
          if Result then
          begin
            FInternalValue := aFullText;
            FinternalIndex := StrToInt(aFullText);
          end;
       end;
       tdtDDD:
        if Key in ['1'..'7'] then // numeric input
        begin
          Sender.Text := ShortDayNames[strtoint(Key)];
          Result := True;
          FinternalValue := Key;
          if Result then
            FinternalIndex := StrToInt(Key);
          Key :=#0;
        end
        else
        begin
         if FAutoComplete then
          Result := AutoCompleteArray(ShortDayNames)
         else Result := True;
        end;
       tdtDDDD:
        if Key in ['1'..'7'] then // numeric input
        begin
          Sender.Text := ShortDayNames[strtoint(Key)];
          Result := True;
          FinternalValue := Key;
          if Result then
            FinternalIndex := StrToInt(Key);
          Key :=#0;
        end
        else
        begin
         if FAutoComplete then
          Result := AutoCompleteArray(LongDayNames)
         else
           Result := True;
        end;
       tdtMMM:
       if Key in ['1'..'9'] then // numeric input
        begin
          Sender.Text := ShortMonthNames[strtoint(Key)];
          Result := True;
          FinternalValue := Key;
          if Result then
            FinternalIndex := StrToInt(Key);
          Key :=#0;
        end
        else
        begin
         if FAutoComplete then
          Result := AutoCompleteArray(ShortMonthNames)
         else Result := True;
        end;
       tdtMMMM:
       if Key in ['1'..'9'] then // numeric input
        begin
          Sender.Text := LongMonthNames[strtoint(Key)];
          Result := True;
          FinternalValue := Key;
          if Result then
            FinternalIndex := StrToInt(Key);
          Key :=#0;
        end
        else
        begin
         if FAutoComplete then
          Result := AutoCompleteArray(LongMonthNames)
         else Result := True;
        end;
     end;
     if Result then
     begin
       Sections.FEmpty := False;
       ValidateAndSynchronize(Sender.Findex);
     end;
  except
     Result := False;
  end;
 end
 else
   Sections.Empty;
end;

procedure TISCustomPlugEdit.WMNCLButtonUp(var Message: TMessage);
var R,Cr : TRect;
    X, Y : integer;
begin
  With Message do
  if (WParam<>HTCLIENT) and (FButtonType <> tbtNone) then
   begin
    GetwindowRect(Handle,R);
    Cr := BtnRect;
    offsetRect(Cr,r.left,r.top);
    if Ptinrect(cr, Point(loword(lparam), hiword(lparam))) then
    begin
      X := Loword(lparam) - R.Left;
      Y := Hiword(lparam) - R.Top;
      MouseUp(mbLeft, [], X, Y);
    end;
    if FButtonType = tbtScroll then
    begin
      CR := MinusBtnRect;
      offsetRect(Cr,r.left,r.top);
      if Ptinrect(cr, Point(loword(lparam), hiword(lparam))) then
      begin
        X := Loword(lparam) - R.Left;
        Y := Hiword(lparam) - R.Top;
        MouseUp(mbLeft, [], X, Y);
      end;
    end;
   end;
   inherited;
end;

procedure TISCustomPlugEdit.WMNCLButtonDown(var Message: TMessage);
var R,Cr : TRect;
    X, Y : integer;
begin
 if not (csDesigning in ComponentState) then
 begin
  With Message do
  if (WParam<>HTCLIENT) and (FButtonType <> tbtNone) then
   begin
    GetwindowRect(Handle,R);
    Cr:= BtnRect;
    offsetRect(Cr,r.left,r.top);
    if Ptinrect(cr, Point(loword(lparam), hiword(lparam))) then
    begin
      X := Loword(lparam) - R.Left;
      Y := Hiword(lparam) - R.Top;
      MouseDown(mbLeft, [], X, Y);
    end;
    if FButtonType = tbtScroll then
    begin
      CR := MinusBtnRect;
      offsetRect(Cr,r.left,r.top);
      if Ptinrect(cr, Point(loword(lparam), hiword(lparam))) then
      begin
        X := Loword(lparam) - R.Left;
        Y := Hiword(lparam) - R.Top;
        MouseDown(mbLeft, [], X, Y);
      end;
    end;
   end;
 end;
 inherited;
end;

procedure TISCustomPlugEdit.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var i : integer;
    CFocused : boolean;
begin
 if (Button = mbLeft) then
 begin
  CFocused := false;
  for i:=0 to ControlCount-1 do
   if Controls[i] is TISAlignedEdit then
   begin
    if TISAlignedEdit(Controls[i]).Focused then
    begin
      CFocused := true;
      Break;
    end;
   end;
   if (not CFocused) then //and (ControlCount>0) then
     FocusEditor(tmfFirst);
     //TISAlignedEdit(Controls[0]).SetFocus;

  case FButtonType of
  tbtDropdown, tbtEllipsis:
  if (PtInRect(BtnRect,Point(X,Y))) then
  begin
     FDown :=true;
     DrawButton;
     if FButtonType = tbtDropdown then
     begin
        if FDropDownVisible = true then
         CloseUp(True)
        else
         DropDown;
     end

//    MouseUp  -    MouseUp  
{     else
     begin
       ButtonPressed(pbEllipsis);
     end;
}
  end;
  tbtScroll:
  if (PtInRect(PlusBtnRect,Point(X,Y))) then
  begin
     FUp :=True;
     DrawButton;
     ButtonPressed(pbUpButton);
     FInit := True;
     SetTimer(Handle, 1, InitRepeat, nil);

  end
  else if (PtInRect(MinusBtnRect, Point(X,Y))) then
  begin
     FDown :=true;
     DrawButton;
     ButtonPressed(pbDownButton);
     FInit := True;
     SetTimer(Handle, 1, InitRepeat, nil);
  end;
  tbtSpin:
  if (PtInRect(UpBtnRect,Point(X,Y))) then
  begin
     FUp :=True;
     DrawButton;
     ButtonPressed(pbUpButton);
     FInit := true;
     SetTimer(Handle, 1, InitRepeat, nil);
  end
  else if (PtInRect(DownBtnRect, Point(X,Y))) then
  begin
     FDown :=true;
     DrawButton;
     ButtonPressed(pbDownButton);
     SetTimer(Handle, 1, InitRepeat, nil);
     FInit := true;
  end;
  end;//case
 end;
 inherited;
end;

procedure TISCustomPlugEdit.GetChildren(Proc: TGetChildProc; Root: TComponent);
begin
end;

procedure TISCustomPlugEdit.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if FUp or FDown then
  begin
     DrawButton;
     if FButtonType = tbtEllipsis then
     begin
       ButtonPressed(pbEllipsis);
     end;
     KillTimer(Handle, 1);
     Fup :=   False;
     FDown := False;
     DrawButton;
  end;
  inherited;
end;

procedure TISCustomPlugEdit.KeyDown(var Key: Word; Shift: TShiftState);
var
//ControlIndex : integer;
    FDropdowned : Boolean;
begin

  FDropdowned := False;

  Case Key of
     VK_INSERT, VK_SPACE:
       if FSections[FFocusedEditor.FIndex].FType = tdtCheck then
          ProceedButton(pbDownButton)
       else if (Key = VK_SPACE) and (FSections[FFocusedEditor.FIndex].FType = tdtButton) then
          FFocusedEditor.EmulateMousedown;
     VK_RETURN, VK_ESCAPE:
       if FDropdownVisible then
       begin
          CloseUp(Key = VK_RETURN);
          Key :=0; Shift :=[];
          //exit;
       end;
     VK_DOWN:
     if ((ssAlt in Shift) and (FButtonType = tbtDropDown)) or (FButtonType in [tbtSpin, tbtScroll]) then
     begin
        FDown := True;
        DrawButton;
        if FButtonType = tbtDropDown then
        begin
         if FDropDownVisible  then
            CloseUp(False);
           DropDown;
           FDropdowned :=True;
           FDown := False;
           DrawButton;
        end
        else
        begin
         ButtonPressed(pbDownButton);
        end;
        Key :=0; Shift :=[];
     end;
     VK_UP:
     if FButtonType in [tbtSpin, tbtScroll] then
     begin
        FUp := True;
        DrawButton;
        ButtonPressed(pbUpButton);
        Key :=0;
     end
     else if (FButtonType = tbtDropdown) and FDropdownVisible and (ssAlt in Shift) then
     begin
        CloseUp(True);
        Key :=0; Shift :=[];
     end;
  end;
  if FDropDownVisible and (not FDropdowned) then
     FDropDown.KeyDown(Key, Shift);
 inherited;

end;

procedure TISCustomPlugEdit.CMEnabledChanged(var Message: TMessage);
var
  I: Integer;
begin
  inherited;
  for I := 0 to FEditorsList.Count - 1 do
  begin
    TISAlignedEdit(FEditorsList[I]).Enabled := Enabled;
    if Enabled then
    begin
      TISAlignedEdit(FEditorsList[I]).Color := FSections[i].Color;
      TISAlignedEdit(FEditorsList[I]).Font.color := FSections[i].FontColor;
    end
    else
    begin
      TISAlignedEdit(FEditorsList[I]).Color := cl3DLight;
      TISAlignedEdit(FEditorsList[I]).Font.color := ClGray;
    end;
  end;
  Invalidate;
end;

procedure TISCustomPlugEdit.KeyUp(var Key: Word; Shift: TShiftState);
begin
  Case Key of
     VK_DOWN, VK_UP:
     begin
        FDown := False;
        FUp := False;
        DrawButton;
     end;
  end;
  inherited;
end;

procedure TISCustomPlugEdit.SetText(Value : String);
begin
 try
   FSections.SetText(Value);
 except
   Fsections.SetText('');
 end;
 Invalidate;
end;

function TISCustomPlugEdit.GetText : String;
begin
   Result := FSections.GetText;
end;

function TISCustomPlugEdit.GetModified : Boolean;
var i : integer;
begin
   Result := False;
   for i:=0 to  FEditorsList.Count-1 do
   begin
     Result := TCustomEdit(FeditorsList[i]).Modified;
     if Result then Break;
   end;
end;

procedure TISCustomPlugEdit.SetReadOnly(Value : Boolean);
var i : integer;
begin
   for i:=0 to  FSections.Count-1 do
     FSections[i].ReadOnly := Value;
end;

procedure TISCustomPlugEdit.SetModified(Value : Boolean);
var i : integer;
begin
   for i:=0 to  FEditorsList.Count-1 do
     TCustomEdit(FeditorsList[i]).Modified := Value;
end;

function TISCustomPlugEdit.GetCheckText(Index : Integer) : String;
var aNode : TFlyNode;
    aList : TStringList;
    aPath : String;
begin
  aList := TStringList.Create;
  with FSections[Index] do
  try
  aNode := Items.GetfirstNode;
  while aNode <> nil do
  begin
     if aNode.StateIndex = STATE_CHECKED then
     begin
        aPath := aNode.GetPath(ItemSeparator);
        //aPath := Copy(aPath, 2, Length(aPath)-1);
        aList.Add(aPath);
     end;
     aNode := aNode.GetNext;
  end;
  Result  := aList.CommaText;
  finally
    aList.Free;
  end;
end;


{ TSizedPopupTree }
procedure TSizedPopupTree.SetSizeGrip(const Value: Boolean);
begin
  if FSizeGrip <> Value then
  begin
     FSizeGrip := Value;
     {$ifdef DFS_COMPILER_4_UP}
     ValidateState;
     {$endif}
  end;
end;

procedure TSizedPopupTree.CMColorChanged(var Message: TMessage);
begin
   inherited;
   {$ifdef DFS_COMPILER_4_UP}
   ValidateState;
   {$endif}
end;

{$ifdef DFS_COMPILER_4_UP}
procedure TSizedPopupTree.ValidateState;
begin
   Perform(CM_BORDERCHANGED, 0, 0);
end;
{$endif}

procedure TSizedPopupTree.WMNCCalcSize(var Message: TWMNCCalcSize);
begin
  inherited;
  if FSizeGrip then
  begin
    Dec(Message.CalcSize_Params^.rgrc[0].Bottom, GetSystemMetrics(SM_CYVSCROLL));
//    Message.CalcValidRects:=true;
//    Message.Result :=0;
  end;
end;



procedure TSizedPopupTree.WMNCPaint(var Message: TMessage);
var
  MaskDC, DC: HDC;
  sR, RC, RW, SaveRW: TRect;
  EdgeSize: Integer;
//  WinStyle: Longint;
  Save: THandle;
  aBmp : TBitmap;
  Flags : integer;

 function ScrollBarVisible(Code: Word): Boolean;
  var
    Min, Max: Integer;
  begin
    Result := False;
    if (ScrollBars = ssBoth) or
      ((Code = SB_HORZ) and (ScrollBars = ssHorizontal)) or
      ((Code = SB_VERT) and (ScrollBars = ssVertical)) then
    begin
      GetScrollRange(Handle, Code, Min, Max);
      Result := Min <> Max;
    end;
  end;
begin
  { Get window DC that is clipped to the non-client area }
  inherited;
  if FSizeGrip then
  begin
    DC := GetWindowDC(Handle);
    try
      Windows.GetClientRect(Handle, RC);
      GetWindowRect(Handle, RW);
      MapWindowPoints(0, Handle, RW, 2);
      OffsetRect(RC, -RW.Left, -RW.Top);
      ExcludeClipRect(DC, RC.Left, RC.Top, RC.Right, RC.Bottom);

      SaveRW := RW;
      EdgeSize := GetSystemMetrics(SM_CYVSCROLL);
      Inc(RC.Bottom, EdgeSize);
      RW := RC;
      with RW do
        begin
          {WinStyle := GetWindowLong(Handle, GWL_STYLE);
          if (WinStyle and WS_VSCROLL) <> 0 then
            Inc(Right, GetSystemMetrics(SM_CYVSCROLL));
          if (WinStyle and WS_HSCROLL) <> 0 then
            Inc(Bottom, GetSystemMetrics(SM_CXHSCROLL));
          }
          if ScrollBarVisible(SB_HORZ) then
            Inc(Bottom, GetSystemMetrics(SM_CXHSCROLL));
          if ScrollBarVisible(SB_VERT) then
            Inc(Right, GetSystemMetrics(SM_CYVSCROLL));
        end;

      with RW do
        sr := Rect(Right-EdgeSize, Bottom-EdgeSize, Right, Bottom);
      IntersectClipRect(DC, RW.Left, RW.Top, RW.Right, RW.Bottom);
      RW := SaveRW;
      { Erase parts not drawn }
      OffsetRect(RW, -RW.Left, -RW.Top);
      Windows.FillRect(DC, RW, Brush.Handle);
      aBmp := TBitmap.Create;
      abmp.Width := EdgeSize;
      abmp.Height := EdgeSize;
      try
        Flags := DFCS_SCROLLSIZEGRIP{$ifdef DFS_COMPILER_4_UP}+DFCS_TRANSPARENT{$endif};
        if (not Ctl3d) and (BordersTyle= bsNone) then
           Flags := Flags + DFCS_FLAT;

        DrawFrameControl(aBmp.Canvas.Handle, Rect(0,0, Edgesize, EdgeSize),DFC_SCROLL,Flags);
        aBmp.Transparent:=true;
        aBmp.TransparentColor:=aBmp.Canvas.Pixels[0,0];//
        aBmp.TransparentMode := tmAuto;
        Save := 0;
        MaskDC := 0;
        try
          MaskDC := CreateCompatibleDC(0);
          Save := SelectObject(MaskDC, aBmp.MaskHandle);
          //FillRect(DC, RW, Color);
          TransparentStretchBlt(DC, SR.Left, Sr.Top, EdgeSize,
          EdgeSize, aBmp.Canvas.Handle, 0, 0, EdgeSize,
          EdgeSize, MaskDC, 0, 0);
        finally
          if Save <> 0 then SelectObject(MaskDC, Save);
          if MaskDC <> 0 then DeleteDC(MaskDC);
        end;
      finally
        aBmp.Free;
      end;
    finally
      ReleaseDC(Handle, DC);
    end;
  end;

end;

function TSizedPopupTree.IsSizeGrip(Pt : TSmallPoint) : Boolean;
var RW : TRect;
    EdgeSize : integer;
begin
  GetWindowRect(Handle, RW);
  Dec(Pt.X, RW.Left);
  Dec(Pt.Y, RW.Top);
  EdgeSize := GetSystemMetrics(SM_CYVSCROLL);
  Result := ((Pt.X)>(RW.Right-RW.Left)-EdgeSize) and ((Pt.X)<(RW.Right-RW.Left)) and
     ((Pt.Y)>(RW.Bottom-RW.Top)-EdgeSize) and ((Pt.Y)<(RW.Bottom-RW.Top));
end;

procedure TSizedPopupTree.WMNCHitTest(var Message: TWMNCHitTest); //message WM_NCHITTEST;
begin
  inherited;
  if FSizeGrip then
  with Message do
  begin
     if IsSizeGrip(Pos) then
        Result := HTGROWBOX;
  end;
end;

procedure TSizedPopupTree.WMSetCursor(var Message: TWMSetCursor);
begin
  if (not (csDesigning in ComponentState)) and FSizeGrip then
  begin
  with Message do
   case Smallint(HitTest) of
     HTGROWBOX:
        Windows.SetCursor(Screen.Cursors[crSizeNWSE]);
     else inherited;
   end;
   end
   else inherited;
end;

procedure TSizedPopupTree.WMNCLMouseDown(var Message: TMessage);
var aPt : TSmallPoint;
begin
  inherited;
   With Message do
   if FSizeGrip then
   begin
      aPt := TSmallPoint(Lparam);
      if IsSizeGrip(aPt) then
      begin
        FSizing :=True;
        SizingStart.X := aPt.X;
        SizingStart.Y := aPt.Y;
        SizingStart := ScreenToClient(SizingStart);
        Offset.X := Width -SizingStart.X;
        Offset.Y := Height -SizingStart.Y;
        MouseCapture := True;
      end;
   end;
end;


procedure TSizedPopupTree.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
   if FSizing then
   begin
     if (Abs(Width - (X+Offset.X))>2) or (Abs(Height - (Y+Offset.Y))>2) then
     begin
      //SetWindowPos(FDropDown.Handle, HWND_TOP, P.X, Y, FDropDown.Width, FDropDown.Height,
      //SWP_NOSIZE or SWP_NOACTIVATE or SWP_SHOWWINDOW);
      //Width := X+Offset.X;
      //Height :=Y+Offset.Y;
      SetBounds(Left, Top, X+Offset.X, Y+Offset.Y);
      InvalidateGrid;
     end;
   end
   else
     Inherited;
end;

function TSizedPopupTree.GetAlignment(Col : integer) : TAlignment;
begin
   Result := inherited GetAlignment(Col);
end;

{
procedure TSizedPopupTree.WMLButtonDown(var Message: TWMLButtonDown);
begin
  inherited;
end;
}
procedure TSizedPopupTree.WMLButtonUp(var Message: TWMLButtonUp);
begin
   if FSizing then
   begin
      Fsizing := False;
      MouseCapture := False;
   end
   else
     inherited;
end;

//-------TPopupTree-------------------------------------------
constructor TPopupTree.Create(AOwner: TComponent);
begin
   inherited;
   FDonotChangeSection := true;
   Parent := TISCustomPlugEdit(aOwner);
   FAsFocused :=True;
   FixedCols:=0;
   Fixedrows:=0;
   ColCount:=1;
   Structurecol:=0;
   ShowImages  := true;
   ShowRoot    := true;
   ShowButtons := true;
   ScrollBars  := ssVertical;
   ShowLogic   := false;
   Tooltips:=true;
   Options:=[goDrawFocusSelected,goThumbTracking];
end;


procedure TPopupTree.GetBitmap(aNode : TFlyNode; index :  integer; Image : TBitmap);
var  R : TRect;
     aIntNode : TFlyNode;
begin
  if (FPlugEdit <> nil) then
  with FPlugEdit do
  begin
    if FSections[FFocusedEditor.Findex].FType = tdtColor then //FPlugEdit.EditorType = tetColor then
    begin
      R := Rect(0,0, Indent, RowHeights[aNode.Getrow]);
      Image.Height := R.Bottom;
      Image.Width := R.Right;
      with FFocusedEditor do
      begin
        aIntNode := TFlyNode(Fsections[Findex].Items.GetItemOnIndexPath(aNode.GetIndexPath('.'),'.'));
      if (aIntNode <> nil) and (aIntNode.ImageIndex >= 0) then
        FSections.DrawColor(Image.Canvas, Longint(aIntNode.ImageIndex), R)
      else if (aIntNode <> nil) and (aIntNode.Data<>nil) then
        FSections.DrawColor(Image.Canvas, Longint(aIntNode.Data), R)
      else if aIntNode<>nil then
        FSections.DrawColor(Image.Canvas, ColorToRGB(StringToColor(aIntNode.Caption)), R)
      end;
    end
    else if FPlugEdit.Images <> nil then
    begin
     with FPlugEdit.Images do
       GetBitmap(index, Image);
    end;
  end
  else
    inherited;
end;

procedure TPopupTree.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or WS_BORDER;
    ExStyle := WS_EX_TOOLWINDOW;
    WindowClass.Style := CS_SAVEBITS;
  end;
end;


function TPopupTree.GetAlignment(Col : integer) : TAlignment;
begin
  with FPlugEdit.FSections[FPlugEdit.FFocusedEditor.FIndex] do
     Result := DropDownColAlignment[Col];
end;


procedure TPopupTree.KeyPress(var Key: Char);
begin
  inherited;
  if Key in ['+','-'] then Key :=#0;
end;

procedure TPopupTree.CMCancelMode(var Message: TCMCancelMode);
begin
   inherited;
   if FPlugEdit <> nil then
      FPlugEdit.CloseUp(False);
end;

procedure TPopupTree.WMKillFocus(var Msg: TWMKillFocus);
begin
   inherited;
//   if FPlugEdit <> nil then
//      FPlugEdit.CloseUp(False);
end;


procedure TPopupTree.KeyDown(var Key: Word; Shift: TShiftState);
var aKey : Word;
    aShift : TShiftState;
begin
  aKey := Key;
  aShift := Shift;

  if (Shift = []) then
  begin
    if (aKey = VK_INSERT) then
     Toggle(GetnodeAtrow(Row))
    else if aKey in [VK_RIGHT, VK_LEFT, VK_ADD, VK_SUBTRACT] then
       Key :=0
    else if (aKey in [VK_RETURN, VK_ESCAPE]) and (FPlugEdit <> nil) then
    begin
       FPlugEdit.Keydown(aKey, aShift);
       Key :=0;//c26.11
    end;

  end
  else if (aKey = VK_UP) and (ssAlt in aShift) then
  begin
     //FPlugEdit.Closeup(True);
     FPlugEdit.Keydown(aKey, aShift);
     Key := 0;
 end;
 if (Key <> 0) then
   inherited;
 //TISDropDown(Parent).Keydown(Key, Shift);//send to container for filtering
end;

procedure TPopupTree.Toggle(Node : TFlyNode);
begin
      begin
         if (nil = Node) then exit;
         //if Assigned(FOnStateChanging) then FOnStateChanging(Node, AllowChange);
         if {(AllowChange) and} (Node.StateIndex<>STATE_EMPTY) then
         begin
            case Node.StateIndex of
               STATE_OPTIONED : ChangeState(Node, STATE_UNOPTIONED);
               STATE_UNOPTIONED: ChangeState(Node, STATE_OPTIONED);
            else
              ChangeState(Node, 1+((Node.StateIndex) mod 2));
            end;
            if HandleAllocated then
                InvalidateCell(StructureCol, Node.Getrow);

            if Node.StateIndex in [STATE_OPTIONED,STATE_UNOPTIONED] then
              NodeStateRefreshSiblings(Node);
            //if totRefreshChilds in FtreeOptions then
            NodeStateRefreshChildren(Node);
            //if totrefreshParents in FtreeOptions then
            NodeStateRefreshParent(Node);
         end;
      end;

   if FPlugEdit.FFocusedEditor <> nil then
   with FPlugEdit.FSections[FPlugEdit.FFocusedEditor.FIndex] do
   begin
     FPlugEdit.FFocusedEditor.Text := GetCheckText(Itemseparator);
     FInternalValue := FPlugEdit.FFocusedEditor.Text;
     FPlugEdit.SetEditorText(FPlugEdit.FFocusedEditor.FIndex);
     if FPlugEdit.FFocusedEditor.FCustomdraw then
        FPlugEdit.FFocusedEditor.DrawEditor;

   end;
end;

function TPopupTree.GetCheckText(aItemSeparator : String) : String;
var aNode : TFlyNode;
    aList : TStringList;
    aPath : String;
begin
  aList := TStringList.Create;
  aNode := Items.GetfirstNode;
  try
  while aNode <> nil do
  begin
     if aNode.StateIndex = STATE_CHECKED then
     begin
        aPath := aNode.GetPath(aItemSeparator);
        //aPath := Copy(aPath, 2, Length(aPath)-1);
        aList.Add(aPath);
     end;
     aNode := aNode.GetNext;
  end;
  Result  := aList.CommaText;
  finally
    aList.Free;
  end;
end;

procedure TPopupTree.SetCheckText(Value, AItemSeparator : String);
var aNode : TFlyNode;
    aList : TStringList;
    i : integer;
begin
  aList := TStringList.Create;
  try
     aList.CommaText := Value;
     aNode := Items.GetFirstNode;
     while aNode <> nil do
     begin
       aNode.StateIndex := STATE_UNCHECKED;
       aNode := aNode.GetNext;
     end;
     for i:=0 to aList.Count-1 do
     begin
       aNode := TFlyNode(Items.GetItemOnPath(aList[i],AItemSeparator));
       if aNode <> nil then
       begin
          aNode.StateIndex := STATE_CHECKED;
          NodeStateRefreshChildren(aNode);
          NodeStateRefreshParent(aNode);
          aNode.MakeVisible;
       end;
     end;
  finally
    aList.Free;
  end;
end;

procedure TPopupTree.NodeStateRefreshSiblings(const Node: TFlyNode);
var aNode : TFlyNode;
begin
   if Node.StateIndex = STATE_OPTIONED then
   begin
      // 
      aNode := Node;
      while Anode<>nil do
      begin
         if (Anode <> Node) and (Anode.StateIndex = STATE_OPTIONED) then
         begin
            //Anode.StateIndex := STATE_UNOPTIONED;
            if ChangeState(ANode, STATE_UNOPTIONED) then
              NodeStateRefreshChildren(Anode);
         end;
         Anode:=Anode.GetNextSibling;
      end;
      // 
      aNode := Node;
      while Anode<>nil do
      begin
         if (Anode <> Node) and (Anode.StateIndex = STATE_OPTIONED) then
         begin
            //Anode.StateIndex := STATE_UNOPTIONED;
            if ChangeState(ANode, STATE_UNOPTIONED) then
              NodeStateRefreshChildren(Anode);
         end;
         Anode:=Anode.GetPrevSibling;
      end;
   end;

end;

function TPopupTree.GetStateIndex(const Node: TFlyNode): Integer;
var
   i                                   : Integer;
   nodeChild                           : TFlyNode;
   intStateIndex                       : Integer;
   bolNone, bolAll, bolSome, BolOption : Boolean;
begin
   Result := STATE_UNCHECKED;
   bolNone := FALSE;
   bolAll := FALSE;
   bolSome := FALSE;
   BolOption := false;
   intStateIndex := STATE_UNCHECKED;
   for i := 0 to Node.Count-1 do begin
      nodeChild := Node.Item[i];
      if NOT(bolNone) then bolNone :=(nodeChild.StateIndex=STATE_UNCHECKED);
      if NOT(bolAll) then bolAll :=(nodeChild.StateIndex=STATE_CHECKED);
      if NOT(bolSome) then bolSome :=(nodeChild.StateIndex=STATE_GRAYEDCHECKED);
      if not(BolOption) then BolOption:=(nodeChild.StateIndex=STATE_OPTIONED);
      if (bolSome) OR (bolNone AND bolAll) then break;

   end;
   Case Node.StateIndex of
      STATE_UNCHECKED,STATE_CHECKED,STATE_GRAYEDCHECKED:
      begin
         if (bolSome) OR (bolNone AND bolAll) then intStateIndex := STATE_GRAYEDCHECKED
         else if (bolNone) then intStateIndex := STATE_UNCHECKED
            else if (bolAll) then intStateIndex := STATE_CHECKED;
         //if BolOption then intStateIndex := STATE_GRAYED;
         //else intStateIndex := 0;
         Result := intStateIndex;
      end;
      STATE_OPTIONED,STATE_UNOPTIONED:
      begin
         if BolOption then Result := STATE_OPTIONED else Result := STATE_UNOPTIONED;
      end;
   end;
end;

procedure TPopupTree.NodeStateRefreshParent(const Node: TFlyNode);
var
   nodeParent    : TFlyNode;
   intStateIndex : Integer;
   //allow         : Boolean;
begin
   nodeParent := Node.Parent;
   if (nil = nodeParent) then exit;
   //if CheckBox
   //allow:=totRefreshParents in FTreeOptions;   //totRefreshChilds,totRefreshParents
   //fired Event
   //RefreshState(NodeParent,trmRefreshByChild,allow);
   //if Allow then
   //begin
      if (Node.StateIndex <> nodeParent.StateIndex) and
      ((nodeParent.Stateindex >= STATE_UNCHECKED) and (nodeParent.Stateindex <= STATE_GRAYEDCHECKED)) then
      begin
         intStateIndex := GetStateIndex(nodeParent);
         if (intStateIndex <> nodeParent.StateIndex) then begin
            //nodeParent.StateIndex := intStateIndex;
            ChangeState(nodeParent, intStateIndex);
         end;
      end
      else if (nodeParent.Stateindex = STATE_UNOPTIONED) and (Node.Stateindex = STATE_OPTIONED) then
         begin
            //nodeParent.Stateindex :=   STATE_OPTIONED;
            if ChangeState(nodeParent, STATE_OPTIONED) then
              NodeStateRefreshSiblings(nodeParent);
         end;
   //end;
   NodeStateRefreshParent(nodeParent);
end;

procedure TPopupTree.NodeStateRefreshChildren(Node: TFlyNode);
var
   nodeChild : TFlyNode;
   i         : Integer;
//   allow     : Boolean;
begin
   //if CheckBox
   //fired Event
   with Node do
   begin
      case StateIndex of
         STATE_UNCHECKED, STATE_CHECKED:
            for i := 0 to Count-1 do
            begin
               nodeChild:= Item[i];
               //allow    :=totRefreshChilds in FTreeOptions;//init before
               //RefreshState(nodeChild,trmRefreshByParent,allow);
               //if Allow then
               //begin
                  if nodeChild.StateIndex < STATE_OPTIONED then
                     ChangeState(nodeChild, StateIndex);
                  //nodeChild.StateIndex := StateIndex;
               //end;
               NodeStateRefreshChildren(nodeChild);
            end;
         STATE_UNOPTIONED :
            for i := 0 to Count-1 do begin
               nodeChild := Item[i];//   
               if nodeChild.StateIndex = STATE_OPTIONED then
                  if ChangeState(nodeChild, StateIndex) then
                     NodeStateRefreshChildren(nodeChild);
            end;
      end; //case
   end; //with
end;


procedure TPopupTree.LocateItem(Text : String);
var aNode : TFlyNode;
begin
   aNode:=items.GetFirstNode;
   while aNode<>nil do
     if aNode.Text=Text then
       begin
          aNode.Selected:=true;
          break;
       end
       else
        aNode:=aNode.GetNext;
end;

function TPopupTree.SelectCell(var ACol, ARow: Longint; MoveAnchor : Boolean;  Shift: TShiftState; SelWay : TSelectionWay): Boolean;
var aNode : TFlyNode;
   function AbsoluteCoord(aNode : TFlyNode) : integer;
   begin
      Result :=0;
      aNode := aNode.GetPrev;
      while aNode <> nil do
      begin
          inc(Result);
          aNode := aNode.GetPrev;
      end;
   end;
begin
 Result := inherited SelectCell(aCol, aRow, Moveanchor, Shift, SelWay);
 if not FDoNotChangeSection then
 if Result and (FPlugEdit <> nil) then
   if (FPlugEdit.FFocusedEditor <> nil) and (FPlugEdit.Fdropdownvisible){ and (not FPlugEdit.Fdown)} then
   with FPlugEdit.FSections[FPlugEdit.FFocusedEditor.FIndex] do
   begin
      aNode := GetNodeatRow(aRow);

      if aNode <> nil then
      begin
         if (aNode.ImageIndex <> -1) and (aNode.SelectedIndex=-1) then//is system specified indexes
           //FPlugEdit.FSections[FPlugEdit.FFocusedEditor.FIndex].
           ItemIndex := aNode.ImageIndex
         else
           //FPlugEdit.FSections[FPlugEdit.FFocusedEditor.FIndex].
           ItemIndex := FMinValue + AbsoluteCoord(aNode);
         if FPlugEdit.EditorType <> tetDropDowncheckList then
         begin
          //FPlugEdit.SetEditorText(FPlugEdit.FFocusedEditor.FIndex);
          FPlugEdit.ValidateAndSynchronize(FPlugEdit.FFocusedEditor.FIndex);
          if FPlugEdit.FFocusedEditor.FCustomdraw then
            FPlugEdit.FFocusedEditor.DrawEditor;
         end;
      end;
   end;
end;

procedure TPopupTree.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
var aNode : TFlyNode;
    aHittests : THitTests;
begin
   inherited;
   if Button = mbLeft then
   begin
   aNode := GetNodeAt(X,Y);
   aHittests:= GetHitTestInfoAt(X,Y);
   if aNode <> nil then
   begin
    if htOnStateIcon in aHitTests then
      Toggle(ANode)
    else
      Selected := aNode;
   end;   
  end;
end;

procedure TPopupTree.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
//var FOnMouseUp_f : TMouseEvent;
var  aHittests : THitTests;
begin
  inherited;// MouseUp(Button, Shift, X, Y);
  aHittests := GetHitTestInfoAt(X,Y);
  if (Button = mbLeft) and (htOnlabel in aHitTests) then
    FPlugEdit.CloseUp(True);
end;

constructor TISDropDown.Create(AOwner: TComponent);
begin
  Inherited Create(AOwner); //controlstyle
//  ControlStyle := ControlStyle + [csReflector];
  //[{csAcceptsControls, csCaptureMouse, csClickEvents,}
  //  csSetCaption, csOpaque, csDoubleClicks, csReplicatable];
  OKR := Rect(0,0,0,0);
  CancelR := OKR;
end;

{procedure TISDropDown.CreateWnd;
begin
  inherited CreateWnd;
  Windows.SetParent(Handle, 0);
//  CallWindowProc(DefWndProc, Handle, wm_SetFocus, 0, 0);
end;
}
procedure TISDropDown.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or WS_BORDER +WS_POPUP; //or WS_CHild;// OR WS_GROUP +WS_POPUP;// ;//
    WindowClass.style := CS_SAVEBITS + CS_NOCLOSE;
    ExStyle := {ExStyle or} WS_EX_TOOLWINDOW;
  end;
end;

destructor TISDropDown.Destroy;
begin
  inherited Destroy;
end;

procedure TISDropDown.ClearControls;
begin
 if FContainedControl <> nil then
 try
  FContainedControl.Free;
  FContainedControl := nil;
 except
  FContainedControl := nil;
 end;
end;

function TISDropDown.Contain(aHandle : HWnd) : boolean;
 function FindHandle (aHandle : HWnd) : Boolean;
 var awnd : HWnd;
 begin
    Result := False;
    awnd := Windows.GetParent(aHandle);
    while aWnd <> 0 do
    begin
      if aWnd = Self.Handle then
      begin
        Result := True;
        Break;
      end;
      awnd := Windows.GetParent(aWnd);
    end;

 end;
 {procedure FindHandle(aCtl : TWinControl; var Founded : Boolean);
 var i: integer;
 begin
  for i:=0 to aCtl.ControlCount-1 do
  begin
     if aCtl.Controls[i] is TWinControl then
     begin
        if TWincontrol(aCtl.Controls[i]).Handle = aHandle then
        begin
           //atCtl := TWincontrol(aCtl.Controls[i]);
           Founded := True
        end
        else
           FindHandle(TWinControl(aCtl.Controls[i]), Founded);
        if Founded then break;
     end;
  end;
 end;}
begin
  Result := (aHandle = Self.Handle);
  if not Result then
     Result := FindHandle(aHandle);//FindHandle(Self, Result);
end;

procedure TISDropDown.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
    if FContainedControl = AComponent then
      FContainedControl := nil;
end;

procedure TISDropDown.SetStyles(Value : TDropDownStyles);
begin
  if Value <> FStyles then
  begin
     FStyles := Value;
  end;
end;

procedure TISDropDown.WMNCCalcSize(var Message: TWMNCCalcSize);
var FHeight, GripHeight : integer;
begin
  inherited;
  if (FStyles * [ddsSized, ddsOK, ddsOKCancel]) <> [] then
  begin
   if ddsSized in FStyles then
     GripHeight := GetSystemMetrics(SM_CYVSCROLL)
   else
     GripHeight :=0;

   if (FStyles * [ddsOK, ddsOKCancel]) <> [] then
     FHeight :=  Canvas.TextHeight('OK')+14
   else
     FHeight := 0;

   FNonClientHeight := Max(FHeight, GripHeight);
   if FOnTop then
     Inc(Message.CalcSize_Params^.rgrc[0].Top, FNonClientHeight)
   else
     Dec(Message.CalcSize_Params^.rgrc[0].Bottom, Max(FHeight, FNonClientHeight));
  end;
end;

procedure TISDropDown.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
  if ((Shift =[]) and (Key in [VK_ESCAPE, VK_RETURN])) or ((Key= VK_UP) and (Shift =[ssAlt])) then
  begin
    if FPlugedit.FFocusedEditor <> nil then
     Windows.SetFocus(FPlugedit.FFocusedEditor.Handle)
    else
     Windows.SetFocus(FPlugedit.Handle);
  end;
//  else if FContainedControl <> nil then
//  begin
//    TCrackWinCtl(FContainedControl).KeyDown(Key, Shift);
//  end;
end;

procedure TISDropDown.WMLButtonDown(var Message: TWMLButtonDown);
begin
   inherited;
end;

procedure TISDropDown.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited;
  if FContainedControl <> nil then
  begin
   TCrackWinCtl(FContainedControl).MouseDown(Button, Shift, X, Y);
  end;
end;


procedure TISDropDown.WMNCPaint(var Message: TMessage);
var
  MaskDC, DC: HDC;
  RC, BR, RW, SaveRW: TRect;
  Save: THandle;
  aBmp : TBitmap;

  EdgeSize: Integer;
  Flags : integer;
begin
  { Get window DC that is clipped to the non-client area }
  inherited;
  if (FStyles * [ddsSized, ddsOK, ddsOKCancel]) <> [] then
  begin
    DC := GetWindowDC(Handle);
    try
      Windows.GetClientRect(Handle, RC);
      GetWindowRect(Handle, RW);
      MapWindowPoints(0, Handle, RW, 2);
      OffsetRect(RC, -RW.Left, -RW.Top);
      OffsetRect(RW, -RW.Left, -RW.Top);

      if FOnTop then
        RW.Bottom := RW.Top + ((RW.Bottom- RW.Top) - (RC.Bottom - RC.Top))
      else
        RW.Top := RW.Bottom - ((RW.Bottom- RW.Top) - (RC.Bottom - RC.Top));
      ExcludeClipRect(DC, RC.Left, RC.Top, RC.Right, RC.Bottom);

      inflateRect(RW, -1,-1);
      //SaveRW := RW;
      EdgeSize := GetSystemMetrics(SM_CYVSCROLL);
      { Erase parts not drawn }

      FillRect(DC, RW, GetSysColorBrush(COLOR_BTNFACE));

      if BorderStyle = bsSingle then
      begin
       //  if Ctl3d then
       //     InflateRect(RW,-2,-2)
       //  else
       InflateRect(RW,-1,-1);
      end;

      if Ctl3d or (dds3d in FStyles) then
        DrawEdge(DC, RW, EDGE_SUNKEN, BF_RECT or  BF_ADJUST);

      if (FStyles * [ddsOK, ddsOKCancel]) <> [] then
      begin
       SelectObject(DC, FPlugEdit.Font.Handle);
       SetBkColor(DC,ColortoRGB(clBtnFace));
       SaveRW := RW;
       OffsetRect(RW, 4, 2);
       InflateRect(RW,-1, -1);
       BR := Rect(RW.Left, RW.Top, Rw.Left, RW.Bottom);
       DrawText(DC, PChar(teOK), Length(teOK), BR,DT_CENTER+DT_SINGLELINE+DT_VCENTER+DT_CALCRECT);
       inc(BR.Right, 14);
       if BR.Right > SaveRW.Right then
          BR.Right := RW.Right;

       if OKDown and OKUC then
         DrawEdge(DC, BR, BDR_SUNKENINNER, BF_RECT or BF_ADJUST)
       else
         DrawEdge(DC, BR, BDR_RAISEDINNER, BF_RECT or BF_ADJUST);
        DrawText(DC, PChar(teOK), Length(teOK), BR,DT_CENTER+DT_SINGLELINE+DT_VCENTER);
        OKR := BR;

       if ddsOKCancel in FStyles then
       begin
         OffsetRect(RW, (BR.Right - BR.Left)+4, 0);
         BR := Rect(RW.Left, RW.Top, Rw.Left+Canvas.TextWidth(teCancel)+4, RW.Bottom);
         DrawText(DC, PChar(teCancel), Length(teCancel), BR,DT_CENTER+DT_SINGLELINE+DT_VCENTER+DT_CALCRECT);
         inc(BR.Right, 14);

         if BR.Right > SaveRW.Right then
            BR.Right := SaveRW.Right;

         if CancelDown and CancelUC then
           DrawEdge(DC, BR, BDR_SUNKENINNER, BF_RECT or BF_ADJUST)
         else
           DrawEdge(DC, BR, BDR_RAISEDINNER, BF_RECT or BF_ADJUST);

         DrawText(DC, PChar(teCancel), Length(teCancel), BR,DT_CENTER+DT_SINGLELINE+DT_VCENTER);
         CancelR := BR;
       end;
       RW := SaveRW;
      end;

     if ddsSized in FStyles then
     begin
      Flags := DFCS_SCROLLSIZEGRIP{$ifdef DFS_COMPILER_4_UP}+DFCS_TRANSPARENT{$endif};
      if (not Ctl3d) and (BorderStyle = bsNone) then
         Flags := Flags + DFCS_FLAT;
      if FonTop then
      begin
        aBmp := TBitmap.Create;
        abmp.Width := EdgeSize;
        abmp.Height := EdgeSize;
        DrawFrameControl(aBmp.Canvas.Handle, Rect(0,0, Edgesize, EdgeSize),DFC_SCROLL, Flags);
        aBmp.Transparent:=true;
        aBmp.TransparentColor:=aBmp.Canvas.Pixels[0,0];//
        aBmp.TransparentMode := tmAuto;
        Save := 0;
        MaskDC := 0;
        try
          MaskDC := CreateCompatibleDC(0);
          Save := SelectObject(MaskDC, aBmp.MaskHandle);
          TransparentStretchBlt(DC, RW.Right-EdgeSize, RW.Top+EdgeSize, EdgeSize,
          -EdgeSize, aBmp.Canvas.Handle, 0, 0, EdgeSize,
          EdgeSize, MaskDC, 0, 0);
        finally
          if Save <> 0 then SelectObject(MaskDC, Save);
          if MaskDC <> 0 then DeleteDC(MaskDC);
        end;
      end
      else
        DrawFrameControl(DC, Rect(RW.Right-EdgeSize,RW.Top, RW.Right, RW.Bottom),DFC_SCROLL,Flags);
     end;

     finally
      ReleaseDC(Handle, DC);
    end;
  end;

end;

function TISDropDown.IsSizeGrip(Pt : TSmallPoint) : Boolean;
var RW : TRect;
    //EdgeSize : integer;
begin
  GetWindowRect(Handle, RW);
  Dec(Pt.X, RW.Left);
  Dec(Pt.Y, RW.Top);
  //EdgeSize := GetSystemMetrics(SM_CYVSCROLL);
  if FOnTop then
   Result := ((Pt.X)>(RW.Right-RW.Left)-FNonClientHeight) and ((Pt.X)<(RW.Right-RW.Left)) and
     ((Pt.Y)>0) and ((Pt.Y)<FNonClientHeight)
  else
   Result := ((Pt.X)>(RW.Right-RW.Left)-FNonClientHeight) and ((Pt.X)<(RW.Right-RW.Left)) and
     ((Pt.Y)>(RW.Bottom-RW.Top)-FNonClientHeight) and ((Pt.Y)<(RW.Bottom-RW.Top));
end;

procedure TISDropDown.WMNCHitTest(var Message: TWMNCHitTest); //message WM_NCHITTEST;
var PTN : TPoint;
begin
  inherited;
  if (FStyles * [ddsSized, ddsOK, ddsOKCancel]) <> [] then
  with Message do
  begin
     PTN := ScreenToClient(Point(Pos.X, Pos.Y));
     if (ddsSized in FStyles) and IsSizeGrip(Pos) then
        Result := HTGROWBOX
     else
     begin
      if FOnTop then inc(PTN.Y, FNonClientHeight);
      if PtInRect(OKR, PTN) or PtInRect(CancelR, PTN) then
        Result := HTBORDER;
      //else
      // Result := HTCLIENT; //csreflector
     end
  end

end;

procedure TISDropDown.WMSetCursor(var Message: TWMSetCursor);
begin
  if (not (csDesigning in ComponentState)) and (ddsSized in FStyles)  then
  begin
  with Message do
   case Smallint(HitTest) of
     HTGROWBOX:
       if FOnTop then
        Windows.SetCursor(Screen.Cursors[crSizeNESW])
       else
        Windows.SetCursor(Screen.Cursors[crSizeNWSE]);
     else inherited;
   end;
   end
   else inherited;
end;

procedure TISDropDown.WMNCLMouseDown(var Message: TMessage);
var aPt : TSmallPoint;
    PTN : TPoint;
begin
  inherited;
  With Message do
  begin
   if ddsSized in FStyles then
   begin
      aPt := TSmallPoint(Lparam);

      if IsSizeGrip(aPt) then
      begin
        FSizing :=True;
        GetCursorPos(PTN);
        SizingStart.X := Ptn.X;
        SizingStart.Y := Ptn.Y;
        MouseCapture := True;
        exit;
      end;
   end;

   if (FStyles * [ddsOK, ddsOKCancel]) <> [] then
   begin
      OKDown :=false;
      CancelDown := False;
      aPt := TSmallPoint(Lparam);
      PTN := ScreenToClient(Point(aPt.X, aPt.Y));
      if FonTop then inc(PTN.Y, FNonClientHeight);
      if PtInRect(OKR, Ptn) then
      begin
         OKDown := true;
         OKUC := OKDown;
         Perform(WM_NCPAINT,0,0);//invalidateRect(Handle, @OKR, true);
      end
      else If PTInRect(CancelR, Ptn) then
      begin
         CancelDown :=True;
         CancelUC :=CancelDown;
         Perform(WM_NCPAINT,0,0);//InvalidateRect(Handle, @CancelR, true);
      end;
      MouseCapture := CancelDown or OKDown;
   end;  //OKDown, CancelDown
  end;
end;

procedure TISDropDown.WMLButtonUp(var Message: TWMLButtonUp);
var cHeight, CWidth : Integer;
begin
   if FSizing or OKDOwn or CancelDown then
   begin
      if OKUC and OKDown then
        FPlugEdit.CloseUp(True)
      else if CancelUC and CancelDown then
        FPlugEdit.CloseUp(False)
      else if FSizing then
      begin
       If FContainedControl <> nil then //
       begin
         cHeight := FContainedControl.Height;
         CWidth := FContainedControl.Width;
         FContainedControl.SetBounds(0,0, ClientWidth, ClientHeight);
         if (CHeight <> FContainedControl.Height) or
            (CWidth <> FContainedControl.Width) then
         begin
            FContainedControl.Perform(WM_SIZE, 0, MakeLong(FContainedControl.Width, FContainedControl.Height));
         end
         else
         begin
           ClientWidth := cWidth;
           clientHeight := cHeight;
         end;
       end;
       FSizing := False;
       OKDown := False;
       CancelDown := False;
       Perform(WM_NCPAINT, 0,0);
      end;
      MouseCapture := False;
   end
   else
     inherited;
end;

procedure TISDropDown.MouseMove(Shift: TShiftState; X, Y: Integer);
var UC : boolean;
    Pt : TPoint;
begin
   if FSizing then
   begin
     GetCursorPos(Pt);
     if (Abs(SizingStart.X - Pt.X) > Mouse.DragThreshold) or (Abs(SizingStart.Y - Pt.Y) > Mouse.DragThreshold) then
     begin
        if FOnTop then
        begin
         SetBounds(FLeftTop.X, FLeftTop.Y+(Pt.Y - SizingStart.Y), Width-(SizingStart.X - Pt.X), Height-(Pt.Y - SizingStart.Y));
         FLeftTop.Y := FLeftTop.Y+(Pt.Y - SizingStart.Y);
        end
        else
         SetBounds(FLeftTop.X, FLeftTop.Y, Self.Width + (Pt.X - SizingStart.X), Self.Height+(Pt.Y - SizingStart.Y));
         SizingStart :=Pt;//Point(Pt.X,Pt.Y);
         FContainedControl.Perform(WM_SIZE, 0, MakeLong(ClientWidth, ClientHeight));
{         If FContainedControl <> nil then //
         begin
           FContainedControl.SetBounds(0,0, ClientWidth, ClientHeight);
           FContainedControl.Perform(WM_SIZE,0,0);
         end;
}
     end;

   end
   else if OKDown or CancelDown then
   begin
      if FOnTop then inc(Y, FNonClientHeight);
      if OKDown then
      begin
         UC := PtInRect(OKR, Point(X,Y));
         if ((OKUC = False) and UC) or
            ((OKUC = True) and not UC) then
         begin
            OKUC := UC;
            Perform(WM_NCPAINT,0,0);
         end;
      end
      else if CancelDown then
      begin
         UC := PtInRect(CancelR, Point(X,Y));
         if ((CancelUC = False) and UC) or
            ((CancelUC = True) and not UC) then
         begin
          CancelUC := UC;
          Perform(WM_NCPAINT,0,0);
         end;
      end;
   end
   else
     Inherited;
end;

procedure TISDropDown.CMCloseUp(var Message: TMessage);
begin
  PostMessage(FPlugEdit.Handle, CM_CloseUp, Message.WParam, 0);
end;

procedure TISDropDown.WMActivate(var Message: TWMActivate);
begin
  inherited;
  if (Message.Active = WA_INACTIVE) and (FPlugEdit <> nil) {and (not CloseUpStarted)} then
  begin
    //if (Message.ActiveWindow <> FPlugEdit.Handle) then
       PostMessage(FPlugEdit.Handle, CM_CancelMode, 0, 0);
     Message.Result := 0;
  end;
end;




initialization
   {if GetClass('TIsPlugSection') = nil then
   begin
    Classes.Registerclass(TISPlugSections);
    Classes.Registerclass(TISPlugSection);
   end;
   }
   {$IFNDEF Registered}
   {$IFNDEF ACTIVEX}
   ShowURToolTip('',false);
   {$ENDIF}
   {$ENDIF}
   InstallHook;
finalization
   UninstallHook;
end.

